#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "y.tab.h"

node *make_temp();
node *leavecond();
node *addleaves();

/*
 * leavescan works on a single level in the parse tree.  It recursively
 * calls itself to deal with lower levels.  It has two passes.
 *
 * Since this routine is done at the ideal time, and has the correct
 * structure, I have also thrown in the creation of temporary
 * variables to replace IF and SELECT conditions.  (Even though it
 * has very little to do with LEAVE.)  -- The reason for this
 * substitution, is that these conditions are frequently dup'ed.
 * The is significant savings in dup'ing and ml'ing only the temporary.
 *
 * Pass 1:  leavescan lower levels and obtain a list of LEAVEs to be
 *  added to this level.  When a LEAVE is passed up a level, its conds
 *  field is modified to reflect the more strict condition under
 *  which the LEAVE actually occurs.
 *  Temporaries are inserted.
 * Pass 2:  Turn each set of LEAVEs encountered into IF statements
 *  which actually implement the leave.  Also assemble a list of LEAVEs
 *  to pass to the next level up.  LEAVEs are not passed up beyond
 *  their corresponding label.
 */
node *leavescan( leavelabel, root )
char *leavelabel;		/* name of calling stmt, NULL if none */
node **root;			/* subtree to process */
{
    node *current;		/* where we are in the level */
    node *last;			/* last node we looked at */
    node *next;			/* next node to look at */
    node *leaves;		/* leave statements returned to this level */
    node *ifstmt;		/* the IF to process */
    node *currentcase;		/* case is select to process */
    node *temp, *temp2;
    node *elsecond;		/* condition of else clause */
    node *tempnode;		/* temporary node to get the condition */
    node *assign;		/* assignment to tempnode */
    node *lastleave;		/* end of leave list */
    node *newif;		/* if statement to replace leaves */
    node *nindex;		/* node index */

    /* pass 1 */
    last = NULL;
    current = *root;
    while ( current != NULL ) {
	switch ( current->type ) {

	    case T_IF:
		/* put in a temporary variable for all but simple
		 * if conditions.  Simple = one node */
		if ( (current->un.bdif.expr->type != T_IS_STATE) &&
		     (current->un.bdif.expr->type != T_IS_CONSTANT) ) {
		    assign = node_create_fi( T_EQ, current->un.bdif.expr );
		    assign->un.assign.lexpr = tempnode =
			    make_temp( current->un.bdif.expr );
		    assign->un.assign.rexpr = current->un.bdif.expr;
		    current->un.bdif.expr = node_dup( tempnode );
		    if ( last == NULL ) {
			*root = last = assign;
		    } else {
			last->next = assign;
			last = assign;
		    }
		    assign->next = current;
		}

		ifstmt = current;
		leaves = leavescan( (char *) NULL, &ifstmt->un.bdif.thenb );
		current = addleaves( current, leaves, current->un.bdif.expr );
		if ( ifstmt->un.bdif.elseb != NULL ) {
		    leaves = leavescan( (char *) NULL,
			    &ifstmt->un.bdif.elseb );
		    elsecond = node_create_fi( T_NOT, ifstmt->un.bdif.expr );
		    elsecond->un.unary.expr = node_dup( ifstmt->un.bdif.expr );
		    current = addleaves( current, leaves, elsecond );
		    node_dispose( elsecond );
		}
		last = current;
		current = current->next;
		break;

	    case T_SELECT:
		currentcase = current->un.select.bdcase;

		/* insert temporary nodes for conditions */
		for( currentcase = current->un.select.bdcase;
			currentcase != NULL;
			currentcase = currentcase->next ) {
		    if ( (currentcase->un.bdcase.expr->type != T_IS_STATE) &&
			    (currentcase->un.bdcase.expr->type
			    != T_IS_CONSTANT) ) {
			assign = node_create_fi( T_EQ,
				currentcase->un.bdcase.expr );
			assign->un.assign.lexpr = tempnode =
				make_temp( currentcase->un.bdcase.expr );
			assign->un.assign.rexpr = currentcase->un.bdcase.expr;
			currentcase->un.bdcase.expr = node_dup( tempnode );
			if ( last == NULL ) {
			    *root = last = assign;
			} else {
			    last->next = assign;
			    last = assign;
			}
			assign->next = current;
		    }
		}

		/* Do leave processing */
		for( currentcase = current->un.select.bdcase;
			currentcase != NULL;
			currentcase = currentcase->next ) {
		    leaves = leavescan( (char *) NULL,
			    &currentcase->un.bdcase.stmt );
		    current = addleaves( current, leaves,
			    currentcase->un.bdcase.expr );
		}

		last = current;
		current = current->next;
		break;

	    case T_LABEL:
		leaves = leavescan( current->un.label.name->name,
			&current->un.label.stmt );
		temp = current->next;
		current->next = NULL;
		/* throw out LABEL node */
		if ( last == NULL ) {
		    *root = temp2 = current->un.label.stmt;
		} else {
		    last->next = temp2 = current->un.label.stmt;
		}
		current->un.label.stmt = NULL;
		NODE_DISPOSE( current );
		if ( temp2 == NULL ) {
		    current = last;
		} else {
		    for ( current = temp2; current->next != NULL;
			    current = current->next ) {
		    }
		}
		/* Hook on LEAVEs */
		if ( current == NULL ) {
		    *root = current = leaves;
		} else {
		    current->next = leaves;
		}
		if ( current != NULL ) {
		    while ( current->next != NULL ) {
			current = current->next;
		    }
		}
		/* Move on in level */
		if ( current == NULL ) {
		    *root = current = temp;
		} else {
		    last = current;
		    current->next = temp;
		    current = current->next;
		}
		break;

	    case T_LEAVE:
		if ( current->un.leave.conds != NULL ) {
		    ierror( "Conditional leave", current );
		}
		if ( current->next != NULL ) {
		    warning( "Statement unreachable", current->next );
		    NODE_DISPOSE( current->next );
		    current->next = NULL;
		}
		last = current;
		current = current->next;
		break;

	    default:
		last = current;
		current = current->next;
		break;
	}
    }

    /* pass 2 */
    leaves = NULL;		/* will contain leaves to be returned */
    last = NULL;
    current = *root;
    while ( current != NULL ) {
	next = current->next;
	if ( current->type == T_LEAVE ) {
	    /* grab contiguous leaves */
	    lastleave = current;
	    while ( (lastleave->next != NULL) &&
		    (lastleave->next->type == T_LEAVE) ) {
		lastleave = lastleave->next;
	    }
	    next = lastleave->next;		/* disconnect leaves */
	    lastleave->next = NULL;

	    /* make IF statement for the LEAVEs if it is needed */
	    newif = NULL;
	    if ( next != NULL ) {
		newif = node_create_fi( T_IF, current );
		newif->un.bdif.elseb = next;
		if ( current == lastleave ) {
		    newif->un.bdif.expr = leavecond( current->un.leave.conds );
		} else {
		    newif->un.bdif.expr = node_create_fi( T_UNARY_OR, current );
		    newif->un.bdif.expr->un.unary.expr =
			    leavecond( current->un.leave.conds );
		    for ( nindex = current->next; nindex != NULL;
			    nindex = nindex->next ) {
			temp = leavecond( nindex->un.leave.conds );
			temp->conc = newif->un.bdif.expr->un.unary.expr;
			newif->un.bdif.expr->un.unary.expr = temp;
		    }
		}
	    }


	    /* make a leave (result) list */
	    lastleave = NULL;		/* previous leave in list */
	    while ( current != NULL ) {
		temp = current->next;
		current->next = NULL;
		if ( (leavelabel != NULL) &&
			(strcmp( current->un.leave.name->name, leavelabel )
			== 0) ) {
		    node_dispose( current );
		} else {
		    /* put leave on leave list */
		    if ( lastleave == NULL ) {
			leaves = lastleave = current;
		    } else {
			lastleave->next = current;
			lastleave = lastleave->next;
		    }
		}
		current = temp;
	    }

	    /* hook on newif in place of leaves (newif may be NULL) */
	    if ( last == NULL ) {
		*root = newif;
	    } else {
		last->next = newif;
	    }
	    last = NULL;	/* this value of last should not matter */
	    /* continue on what's left of this level (if any) */
	    if ( newif == NULL ) {
		current = NULL;
	    } else {
		current = newif->un.bdif.elseb;
	    }

	} else {
	    last = current;
	    current = current->next;
	}
    }
    return leaves;
}


/*
 * Makes up the condition for one leave statement.  It makes the code above
 * look cleaner.
 */
node *leavecond( cond )
node *cond;		/* condition of leave */
{
    node *result;

    if ( cond != NULL && cond->conc == NULL ) {		/* only one condition */
	result = node_dup( cond );
    } else {
	/* NOTE: ANDR( nil ) = 1 */
	result = node_create_fi( T_UNARY_AND, cond );
	result->un.unary.expr = node_dup( cond );
    }
    return result;
}


/*
 * Inserts leaves the current level, and updates the LEAVE condition
 * list to include the condition of this branching statement.
 */
node *addleaves( current, leaves, cond )
node *current;		/* the branching stucture we will add after */
node *leaves;		/* LEAVE nodes to add */
node *cond;		/* condition to put on the LEAVEs. */
{
    node *temp;
    node *nindex;
    node *newcond;

    if ( leaves == NULL ) return current;
    temp = current->next;
    current->next = leaves;
    nindex = current;
    do {
	nindex = nindex->next;
	newcond = node_dup( cond );
	newcond->conc = nindex->un.leave.conds;
	nindex->un.leave.conds = newcond;
    } while ( nindex->next != NULL );
    nindex->next = temp;
    return nindex;
}


/*
 * will create a new state var -- internal condition
 */
node *make_temp( place )
node *place;		/* where in the file this all happens */
{
    char tname[12];		/* name of temp var for a condition */
    static int tnum = 0;	/* unique num in tname */
    node *condtemp;		/* temp var for a condition */
    symbol *symbolent;		/* symbol table entry for temp */

    condtemp = node_create_fi( T_IS_STATE, place );
    condtemp->un.bdvar.hbit = condtemp->un.bdvar.lbit = 0;
    condtemp->un.bdvar.rev = NULL;
    /* put temp in symtab */
    (void) sprintf( tname, "$$COND%d", tnum++ );
    symbolent = symtab_lookup( global_syms, tname, 1 );
    condtemp->un.bdvar.name = symbolent;
    symbolent->type = T_IS_STATE;
    symbolent->u.state.nobit = symbolent->u.state.lobit = symbolent->u.state.hibit = 0;
    symbolent->u.state.is_local = 0;
    symbolent->u.state.versions = NULL;
    symbolent->u.state.is_temp = 1;
    return condtemp;
}
