#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yylex.h"
#include "eval.h"
#include "mult.h"
#include "vov.h"

#include <sys/times.h>
#include <sys/wait.h>

extern int optind;
extern char *optarg;
extern node *input_ports, *output_ports;
extern char *model_name;

static int munch_routine(), cleanup_routine();
static node *process();
static int print_unspecs(), use_warn();

extern void trace_call_tree();

FILE *mltemp;
FILE *bdsyntemp;
FILE *outfile;
int eval_subst;			/* Eval forward substs if true */
int saw_complex;		/* 1 if complex op has been inserted */
int got_a_model;		/* 1 if yyparse() returns a model */

int watch_espresso = 0;
int make_unspec_set = 0;
int select_unique = 0;
int yylex_dump = 0;
int update = 0;			/* Give -u info? */
int exec_level = 50;		/* How far to run */
int bdsyn_minimize = 1;
int collapse_level = 2;		/* full blown collapse */
int dont_care_is_zero = 0;

int number_of_assigns = 0;
int assigns_to_ML = 0;
int nodes_alloced;

int short_1_bit = 0;	/* omit the <0> part for 1-bit variables */

#define EXECUTE(name, func, q, tree) \
    {long t = util_cpu_time(); func; \
    if (update) {\
	(void) printf("# done with %s\tn=%4d time=%s\n", name, nodes_alloced,\
	    util_print_time(util_cpu_time() - t));\
	(void) fflush(stdout);\
    }\
    if (tree != NIL(node) && exec_level == q){ (void) printf("\n\n");\
	if (exec_level < 9) {node_print(tree, 0, name);} \
	else {node_print_expr(tree);}}}

usage_message() 
{
    (void) fprintf(stderr, "bdsyn [-bcdenosuz] [filename]\n");
    (void) fprintf(stderr, "\t-b\tTurn off internal minimizing\n");
    (void) fprintf(stderr, "\t-c [n]\tSpecify less collapsing of logic\n");
    (void) fprintf(stderr, "\t  0 No collapsing\n");
    (void) fprintf(stderr, "\t  1 Local collapsing only\n");
    (void) fprintf(stderr, "\t-d\tGive a dump of tokens during the parsing\n");
    (void) fprintf(stderr, "\t-e [n]\tSpecify execution level (0 ,..., 13)\n");
    (void) fprintf(stderr, "\t  0 PARSE only\t\t 1 then dump\n");
    (void) fprintf(stderr, "\t  2 through FOR\t\t 3 then dump\n");
    (void) fprintf(stderr, "\t  4 through EVAL\t 5 then dump\n");
    (void) fprintf(stderr, "\t  6 through LEAVE\t 7 then dump\n");
    (void) fprintf(stderr, "\t  8 through VERSION\t 9 then dump\n");
    (void) fprintf(stderr, "\t 10 through CLEANUP\t11 then dump\n");
    (void) fprintf(stderr, "\t 12 through LIF\t\t13 then dump\n");
    (void) fprintf(stderr, "\t-n\tGive informaton about unspecifed variables\n");
    (void) fprintf(stderr, "\t-o\tOmit the trailing <0> for 1-bit variables\n");
    (void) fprintf(stderr, "\t-s\tMap all SELECTALL's to SELECT's\n");
    (void) fprintf(stderr, "\t-u\tGive updates as to progress of execution\n");
    (void) fprintf(stderr, "\t-z\tSet DONT_CARE's equal to zero\n");
    exit(1);
}


main(argc, argv)
int argc;
char **argv;
{
    int c;
    FILE *fp;
    char macro_filename[4096];

    node_init();
    yylex_init();

    outfile = stdout;
    while ((c = getopt(argc, argv, "bc:de:nosuw:z")) != EOF) {
	switch(c) {
	    case 'b':
		bdsyn_minimize = 0;
		break;
	    case 'c':
		if ((collapse_level = atoi(optarg)) > 1) {
		    usage_message();
		}
		break;
	    case 'd':
		yylex_dump = 1;
		break;
	    case 'e':
		if (!isdigit(*optarg)) {
		    usage_message();
		}
		exec_level = atoi(optarg);
		break;
	    case 's':
		select_unique = 1;
		break;
	    case 'u':
		update = 1;
		break;
	    case 'w':
		if (! isdigit(*optarg)) {
		    usage_message();
		}
		watch_espresso = atoi(optarg);
		break;
	    case 'n':
		make_unspec_set = 1;
		break;
	    case 'o':
		short_1_bit = 1;
		break;
	    case 'z':
		dont_care_is_zero = 1;
		break;
	    default:
		usage_message();
		break;
	}
    }

    VOVbegin( argc, argv );
    /* read the system standard library macro file first */
    (void) strcpy(macro_filename, "bdsyn.lib");
    fp = fopen(macro_filename, "r");
    if (fp == NULL) {
	char buffer[1024];

	(void) sprintf(buffer, "%s/lib/bdsyn/bdsyn.lib",
			       util_file_expand(CADROOT));
	(void) strcpy(macro_filename, buffer);
	fp = fopen(macro_filename, "r");
    }
    if (fp != NULL) {
	VOVinput( VOV_UNIX_FILE, macro_filename );
        (void) fclose(fp);
        if (yyopenfile(macro_filename, 0, 0) < 0) {
	   (void) fprintf(stderr, "strange error, file disappeared ..., sorry\n");
	} else {
	    (void) yyparse();
	    if (got_a_model) {
		(void) fprintf(stderr, "Expected no model in %s\n", macro_filename);
	    }
	}
    } else {
	(void) printf("Unable to open either bdsyn.lib or %s/lib/bdsyn/bdsyn.lib\n", CADROOT);
    }

    /* Parse the BDS input file */
    if (optind >= argc) {
	if (yyopenfile("(stdin)", 1, 0) < 0) {
	    exit(1);
	}
    } else if (optind == argc-1) {
	if (yyopenfile(argv[optind], 0, 0) < 0) {
	    exit(1);
	}
    } else {
	(void) fprintf(stderr, "bdsyn: too many filenames on command line\n");
	exit(1);
    }
    yy_reset_buffer();
    EXECUTE("parse", (void) yyparse(), -1, NIL(node));

    /* Verify that yyparse() matched a model (and not just a routine) */
    if (! got_a_model) {
	(void) fprintf(stderr, "No model found in the input file\n");
	exit(1);
    }

    /* Dump some debugging information depending on the command line options */
    if (exec_level == 1) {
	printf("\n\nGLOBAL SYMBOL TABLE:\n");
	symtab_info(global_syms);
	yacc_dump_symbol_table(global_syms);
    }

    /* Start producing the real output */
    if (exec_level >= 12) {
	if (collapse_level > 0) {
	    /* get mis on line and ready */
	    open_pipes_to_mlespresso(&mltemp, &bdsyntemp);
	}
	if (collapse_level == 0) {
	    /* talk to output, not to mis */
	    bdsyntemp = outfile;
	}
	/* print the model header */
	(void) fprintf(outfile, ".model %s\n", model_name);
	/* print port declarations (no inputs if making unspecs) */
	if (!make_unspec_set) {
	    print_ports( input_ports, 1 /* type = input */);
	}
	print_ports( output_ports, 2 /* type = output */);
	if (dont_care_is_zero) {
	    (void) fprintf(outfile, ".names DONT_CARE\n0\n");
	}
    }

    /* start tree crunching */
    EXECUTE("trace", trace_call_tree(), -1, NIL(node));
    symtab_foreach(global_syms, munch_routine, T_IS_ROUTINE);
    symtab_foreach(global_syms, cleanup_routine, T_IS_ROUTINE);

    /* Clean-up and take care of final outputs */
    if (exec_level >= 12) {
	if (collapse_level > 0) {
	    (void) fclose(bdsyntemp);
	    (void) fclose(mltemp);
	}
	(void) fflush( stdout );
	symtab_foreach( global_syms, use_warn, T_IS_OUTPUT_PORT ); 
        (void) fflush( stderr );
    }

    /* cleanup memory so we can check when all is said and done */
    symtab_discard(global_syms);
    symtab_discard(reserved_syms);
    if (input_ports != NIL(node)) node_dispose(input_ports);
    if (output_ports != NIL(node)) node_dispose(output_ports);
    if (model_name != NIL(char)) FREE(model_name);

    { 
	struct tms buffer;
	double parent, child, parent_s, child_s;

	if ( update ) {
	    (void) wait(0);
	    (void) times(&buffer);
	    parent = buffer.tms_utime / 60.0;
	    child = buffer.tms_cutime / 60.0;
	    parent_s = buffer.tms_stime / 60.0;
	    child_s = buffer.tms_cstime / 60.0;
	    (void) printf( "# Bdsyn: %3.2fu %3.2fs  MIS: %3.2fu %3.2fs\n",
		parent, parent_s, child, child_s);
	}
    }
    if (exec_level >= 12) (void) fprintf( outfile, ".end\n");
    VOVend(0);
}

static int munch_routine(psym)
symbol *psym;
{
    int width, constant;
    node *temp, *tree;
    revinfo *revdummy;
    extern node *inline_func();
    extern node *expand_name();
    extern node *leavescan();
    extern node *complex();

    if (exec_level == 1) {
	(void) printf("\n\n");
    }
    if (update) {
	(void) printf("# Routine is %s\n", psym->name);
    }

    /* check for an empty routine */
    if (psym->u.routine.body == NIL(node)) {
	return 1;
    }

    /* debugging information on this routine */
    if (exec_level == 1) {
	(void) printf("LOCAL SYMBOL TABLE:\n");
	symtab_info(psym->u.routine.locals);
	yacc_dump_symbol_table(psym->u.routine.locals);
	(void) printf("ROUTINE DEFINITION:\n");
	node_print(psym->u.routine.body, 0, "");
    }

    /* if this routine should be expanded ... */
    if (psym->u.routine.active) {
	EXECUTE("inline", tree = inline_func(psym), 1, tree);
	eval_subst = 0;
	do {
	    saw_complex = 0;
	    if ( exec_level >= 2 ) {
		/* perform META expansion */
		EXECUTE("meta", meta(tree), 3, tree);

		/* perform complex substitutions */
		EXECUTE("complex", tree = complex(tree,&constant,&width), 3, tree);
	    }

	    if (saw_complex) {
		EXECUTE("inline1", tree = expand_name(tree, psym->name), 3, tree);
	    }
	} while (saw_complex);

	/* now we need to do NODE_EVAL */
	if (exec_level >= 4) {
	    EXECUTE("eval", tree = node_eval(tree), 5, tree);
	}

	/* now we need to do LEAVE processing */
	if (exec_level >= 6) {
	    EXECUTE("leave", temp = leavescan((char *) NULL, &tree), 7, tree);
	    if (temp != NULL) {
		error("Leave does not exit to matching label", temp);
	    }
	}

	/* now we need to do SUBST_VERSION */
	if (exec_level >= 8) {
	    EXECUTE("premult", premult(tree, NIL(block)), 0, tree);
	    EXECUTE("version", tree =
		    subst_version(tree,rootblock,0,&revdummy), 9, tree);
	}

	/* now we need to do CLEANUP */
	if (bdsyn_minimize) {
	    eval_subst = 1;
	}
	if (exec_level >= 10) {
	    EXECUTE("cleanup", tree = clean_up(tree), 11, tree);
	    if (update) {
		(void) printf("# Number of assigns = %d\n", number_of_assigns);
	    }
	}

	/* now we need to do LIFASSIGN */
	if (exec_level >= 12) {
	    EXECUTE("lif ", lifassign(tree), 13, tree);
	    if (update) {
		(void) printf("# Assigns to ML = %d\n", assigns_to_ML);
	    }
	}

	/* we can discard this routine now, because nobody calls him */
	node_dispose(tree);
	symtab_discard(psym->u.routine.locals);
	node_dispose(psym->u.routine.formal_args);
	psym->u.routine.body = NIL(node);
	psym->u.routine.locals = NIL(symbol_table);
	psym->u.routine.formal_args = NIL(node);
    }

    return 1;
}


static int cleanup_routine(psym)
symbol *psym;
{
    if (psym->u.routine.body != NIL(node)) {
	node_dispose(psym->u.routine.body);
    }
    if (psym->u.routine.locals != NIL(symbol_table)) {
	symtab_discard(psym->u.routine.locals);
    }
    if (psym->u.routine.formal_args != NIL(node)) {
	node_dispose(psym->u.routine.formal_args);
    }
    return 1;
}

/*
 * Will print the name of a port from the symbol table
 */
print_ports( port_list, port_type )
node *port_list;
int port_type;			/* print_ports arg -- 1=inp 2=out */
{
    int i;
    char buf[512];
    int name_length;
    int line_length;
    node *current;		/* node for the port */
    symbol *p;			/* symtab entry of port */

    line_length = 0;
    if ( port_list != NIL(node) ) {
	for ( current = port_list; current != NIL(node);
		current = current->next ) {
	    p = current->un.bdvar.name;
	    for ( i = p->u.state.hibit; i>= p->u.state.lobit ; i-- ) {
		if (upper_strcmp(p->name, "DONT_CARE") == 0) {
		    (void) sprintf( buf, " %s", p->name);
		} else {
        	if (short_1_bit && p->u.state.nobit) {
           		(void) sprintf( buf, " %s", p->name );
        	}
			else { 
		    	(void) sprintf( buf, " %s<%d>", p->name, i );
			}
		}
		name_length = strlen( buf );
		if ( line_length + name_length > 80 ) {
		    (void) fprintf( outfile, "\n" );
		    line_length = 0;
		}
		if ( line_length == 0 ) {
		    if ( port_type == 1 ) {
			(void) fprintf( outfile, ".inputs");
			line_length = 7;
		    } else {
			(void) fprintf( outfile, ".outputs");
			line_length = 8;
		    }
		}
		line_length += name_length;
		(void) fprintf( outfile, "%s", buf );
	    }
	}
	(void) fprintf( outfile, "\n" );
    }
}


/*
 * Will print warnings about unset output ports.
 */
static int use_warn( p )
symbol *p;
{
    int i;
    char name[512];		/* name plus bit */
    char warn[512];

    for ( i = p->u.state.hibit; i>= p->u.state.lobit ; i-- ) {
	if ( (p->u.state.versions == NULL) ||
	    (p->u.state.versions[i - p->u.state.lobit].assigned.previous == 0) ) {
	    (void) sprintf(name, "%s<%d>", p->name, i );
	    (void) sprintf( warn,
		    "Output port %s is never set (zero assumed)", name );
	    warning( warn, (node *) NULL );
	    (void) fflush( stderr );
	    (void) fprintf( outfile, ".names %s<%d>\n", p->name, i );
	    (void) fflush( outfile );
	}
    }
    return(1);                  /* tell symtab_foreach to continue */
}
