#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "y.tab.h"


#define upper(c)	(islower(c) ? toupper(c) : (c))


/* 
 *  symtab_lookup: given a symbol table and a name, return the symbol
 *  table entry for the name.  Uses a simple hashing function.
 */
symbol *symtab_lookup(psym_table, name, create)
symbol_table *psym_table;	/* symbol table to look in */
char *name;			/* name to find in the table */
int create;			/* symbol should be defined if not found */
{
    register int index, i;
    register unsigned word;
    register symbol *psym;

    /* Compute a hash function over the name */
    word = 0;
    for(i = 0; name[i] != '\0'; i++)
	word = (word << 1) ^ upper(name[i]);
    index = word % psym_table->size;

    /* walk along the list looking for our entry */
    for(psym=psym_table->data[index]; psym != NIL(symbol); psym=psym->next) {
        if (upper_strcmp(psym->name, name) == 0) {
	    break;
	}
    }

    /* if entry not found (and we should create it) create a new entry */
    if (psym == NIL(symbol) && create) {
        psym = ALLOC(symbol, 1);
        psym->name = util_strsav(name);
	psym->type = T_IS_UNDEFINED;
        psym->next = psym_table->data[index];
        psym_table->data[index] = psym;
    }
    return psym;
}



/* 
 *  symtab_init: allocate and return a new symbol table
 */
symbol_table *symtab_init(size)
int size;
{
    register int i;
    register symbol_table *psym_table;

    psym_table = ALLOC(symbol_table, 1);
    psym_table->size = size;
    psym_table->data = ALLOC(symbol *, size);
    for(i = 0; i < size; i++) {
        psym_table->data[i] = NIL(symbol);
    }
    return psym_table;
}



/*  
 *  symtab_discard: free all of the memory used by a symbol table
 */
void
symtab_discard(psym_table)
symbol_table *psym_table;	/* the symbol table to discard */
{
    register int i;
    register symbol *p, *pnext;

    if (psym_table == NIL(symbol_table)) {
	return;
    }

    for(i = 0; i < psym_table->size; i++) {
	for(p = psym_table->data[i]; p != NIL(symbol); p = pnext) {
	    pnext = p->next;
	    if (p->name[0] != '*') { /* check for keyword in symbol table */
		if (p->type == T_IS_STATE || p->type == T_IS_INPUT_PORT 
		    || p->type == T_IS_OUTPUT_PORT) {
		    free_revision_struct(p);
		} else if (p->type == T_IS_MACRO) {
		    FREE(p->u.macro.value);
		}
	    }
	    FREE(p->name);
	    FREE(p);
	}
    }
    FREE(psym_table->data);
    FREE(psym_table);
}



/* 
 *  symtab_info -- print out report of efficiency of the hashing fct 
 */
void
symtab_info(psym_table)
symbol_table *psym_table;
{
    register symbol *p;
    register int i, length, occupied=0, max_length=0, collisions=0, hashed=0;
    double avg_length;

    if (psym_table == NIL(symbol_table)) {
	return;
    }

    for(i = 0; i < psym_table->size; i++) {

	length = 0;
	for(p = psym_table->data[i]; p != NIL(symbol); p = p->next) {
	    length++;
	}

	if (length > max_length) max_length = length;
	collisions += (length > 1);
	occupied += (length > 0);
	hashed += length;
    }

    avg_length = hashed / (double) occupied;
    (void) printf("Table size=%d; hashed count=%d; occupied=%d; collisions=%d\n",
	psym_table->size, hashed, occupied, collisions); 
    (void) printf("avg chain=%3.2f; max chain=%d\n", avg_length, max_length);
}


/*  
 *  symtab_foreach: execute user's function for each symbol in table
 */
symtab_foreach(psym_table, func, fetch_type)
symbol_table *psym_table;	/* the symbol table */
int (*func)();			/* the function to be executed */
int fetch_type;			/* -1 for no selection; otherwise type */
{
    register int i;
    register symbol *p;

    if (psym_table == NIL(symbol_table)) {
	return;
    }

    for(i = 0; i < psym_table->size; i++) {
	for(p = psym_table->data[i]; p != NIL(symbol); p = p->next) {
	    if (fetch_type == -1 || fetch_type == p->type) {
		if (func(p) == 0) {
		    return;
		}
	    }
	}
    }
}


int upper_strcmp(s1, s2)
register char *s1, *s2;
{
    while (upper(*s1) == upper(*s2)) {
	if (*s1 == '\0') {
	    return 0;
	}
	s1++;
	s2++;
    }
    return upper(*s1) - upper(*s2);
}
