#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "y.tab.h"
#include "yylex.h"

extern node *input_ports;
extern int dont_care_is_zero;

/*
 *  yacc_lookup -- return a symbol table entry for a symbol by checking first
 *  the local symbol table (if it exists), and then checking the global
 *  symbol table.
 *
 *  If the symbol does not already exist, it is created as T_IS_UNDEFINED.
 */
symbol *yacc_lookup(name)
char *name;
{
    symbol *psym;

    if (local_syms == NIL(symbol_table)) {
	psym = symtab_lookup(global_syms, name, /* create */ 1);
	psym->u.state.is_local = 0;
    } else { 
	psym = symtab_lookup(local_syms, name, /* create */ 1);
	psym->u.state.is_local = 1;
    }
    psym->u.state.is_temp = 0;
    psym->u.state.versions = NIL(struct revision_struct);
    return psym;
}



/*
 *  yacc_check_defined -- print an error if a symbol is not defined
 */
void
yacc_check_defined(psym)
symbol *psym;
{
    char errmsg[256];

    if (psym->type != T_IS_UNDEFINED) {
	(void) sprintf(errmsg, "%s already defined as %s", 
	    psym->name, map_token_to_name(psym->type));
	yyerror(errmsg);
    }
}



/*
 *  yacc_define_symbol -- define a variable as being either a T_IS_STATE, 
 *  T_IS_INPUT_PORT, T_IS_OUTPUT_PORT, or T_IS_META.  
 */
void
yacc_define_symbol(p, vartype)
node *p;			/* identifier to be defined */
int vartype;			/* type of definition */
{
    symbol *psym = p->un.bdvar.name;

    yacc_check_defined(psym);
    psym->type = vartype;
    switch(vartype) {
	case T_IS_INPUT_PORT:
	case T_IS_OUTPUT_PORT:
	case T_IS_STATE:
	case T_IS_META:
	    psym->u.state.nobit = p->un.bdvar.nbit;
	    psym->u.state.lobit = p->un.bdvar.lbit;
	    psym->u.state.hibit = p->un.bdvar.hbit;
	    psym->u.state.versions = NIL(struct revision_struct);
	    psym->u.state.local_def = NIL(symbol);
	    break;

	default:
	    (void) fprintf(stderr, "Unexpected vartype in yacc_define_symbol\n");
	    exit(1);
    }
    p->type = T_IS_STATE;	/* in case type was META */
}


/*
 *  yacc_define_synonym -- create synonym entry in symbol table
 */
void
yacc_define_synonym(p, def)
node *p;			/* node of type T_IS_STATE */
node *def;			/* node of type T_IS_STATE */
{
    symbol *psym = p->un.bdvar.name;
    symbol *defsym = def->un.bdvar.name;

    yacc_check_defined(psym);

    switch(defsym->type) {
	case T_IS_INPUT_PORT:
	case T_IS_OUTPUT_PORT:
	case T_IS_STATE:
	    psym->type = T_IS_SYNONYM;
	    psym->u.synonym.syn = defsym;
	    psym->u.synonym.lobit = def->un.bdvar.lbit - p->un.bdvar.lbit;
	    psym->u.synonym.hibit = def->un.bdvar.hbit - p->un.bdvar.lbit;
	    break;

	default:
	    yyerror("SYNONYM only allowed for previously defined STATE or PORT"); 
    }
}


/*
 *  yacc_define_constant -- create a constant entry in the symbol table 
 */
void
yacc_define_constant(p, def)
node *p;			/* node of type T_IS_STATE */
node *def;			/* node of type T_CONSTANT */
{
    symbol *psym = p->un.bdvar.name;

    yacc_check_defined(psym);
    psym->type = T_IS_CONSTANT;
    psym->u.constant.const_hbit = def->un.constant.hbit;
    psym->u.constant.const_num = def->un.constant.num;
}



/*
 *  yacc_define_routine -- create a routine entry in the symbol table
 */
void
yacc_define_routine(p, formals, forward_decl)
node *p;			/* node of type T_IS_STATE */
node *formals;
int forward_decl;		/* this is a forward declar of routine */
{
    symbol *psym = p->un.bdvar.name;
    node *q, *retval;
    char errmsg[256];

    if (psym->type == T_IS_ROUTINE) {
	if (psym->u.routine.body != NIL(node)) {
	    (void) sprintf(errmsg, "Attempt to redefine routine %s\n", 
		psym->name);
	    yyerror(errmsg);
	}
    } else {
	yacc_check_defined(psym);
    }

    /* define the formals arguments as local 'states' of type
     * T_IS_STATE, or T_IS_META */
    if (! forward_decl) {
	for(q = formals; q != NIL(node); q = q->next) {
	    yacc_define_symbol(q, q->type);
	}
    }

    /* create a local STATE to hold the return value */
    (void) strcpy(errmsg, "$");		/* guarantee? no clash */
    (void) strcat(errmsg, psym->name);
    retval = node_create(T_IS_STATE, yyfile, *yyline);
    retval->un.bdvar.name = yacc_lookup(errmsg);
    retval->un.bdvar.hbit = p->un.bdvar.hbit;
    retval->un.bdvar.lbit = p->un.bdvar.lbit;
    yacc_define_symbol(retval, p->type);

    /* Now define an entry for the routine */
    psym->type = T_IS_ROUTINE;
    psym->u.routine.body = NIL(node);
    psym->u.routine.locals = NIL(symbol_table);
    /* a little bad, the first formal argument is the return value ! */
    psym->u.routine.formal_args = retval;
    retval->next = formals;
    psym->u.routine.active = 0;
    psym->u.routine.hbit = p->un.bdvar.hbit;
    psym->u.routine.lbit = p->un.bdvar.lbit;
    psym->u.routine.call_count = 0;
}


node *yacc_make_variable(name, ret_type, left, right)
char *name;
int ret_type;
node *left;
node *right;
{
    node *ret;

    ret = node_create(ret_type, yyfile, *yyline);
    ret->un.bdvar.name = yacc_lookup(name);
	ret->un.bdvar.nbit = 0;

    left = node_eval(left);
    if (left->type == T_IS_CONSTANT) {
	ret->un.bdvar.hbit = left->un.constant.num; 
    } else {
	yyerror("Attempted bit-width definition with non-constant");
    }

    if (right == NULL) {
	ret->un.bdvar.lbit = left->un.constant.num; 
    } else {
	right = node_eval(right);
	if (right->type == T_IS_CONSTANT) {
	    ret->un.bdvar.lbit = right->un.constant.num; 
	} else {
	    yyerror("Attempted bit-width definition with non-constant");
	}
	node_dispose(right);
    }

    node_dispose(left);
    FREE(name);

    if (ret->un.bdvar.lbit > ret->un.bdvar.hbit) {
	yyerror("Attempt to use bit-range <hi:lo> with hi < lo");
    }
    return ret;
}

node *yacc_find_symbol(name)
char *name;
{
    symbol *psym;
    node *n;
    char errmsg[256];

    psym = NIL(symbol);
    if (local_syms != NIL(symbol_table))
	psym = symtab_lookup(local_syms, name, /* create */ 0);
    if (psym == NIL(symbol)) {
        psym = symtab_lookup(global_syms, name, /* create */ 0);
    }

    /* DONT_CARE's need not be declared by user -- make an input_port */
    if ((psym == NIL(symbol) && upper_strcmp(name, "DONT_CARE") == 0)) {
	psym = symtab_lookup(global_syms, "DONT_CARE", /* create */ 1);
	psym->type = T_IS_INPUT_PORT;
	psym->u.state.is_local = 0;
	psym->u.state.is_temp = 0;
	psym->u.state.versions = NIL(struct revision_struct);
	psym->u.state.nobit = psym->u.state.lobit = psym->u.state.hibit = 0;
	psym->u.state.local_def = NIL(symbol);
	if (!dont_care_is_zero) {
	    n = node_create(T_IS_STATE, yyfile, *yyline);
	    n->un.bdvar.name = psym;
	    n->un.bdvar.lbit = n->un.bdvar.hbit = 0;
	    input_ports = cons(input_ports, n);
	}
    }

    if (psym == NIL(symbol)) {
	(void) sprintf(errmsg, "Use of undefined symbol %s", name);
	yyerror(errmsg);
    }

    switch(psym->type) {
	case T_IS_STATE:
	case T_IS_INPUT_PORT:
	case T_IS_OUTPUT_PORT:
	case T_IS_META:
	    n = node_create(T_IS_STATE, yyfile, *yyline);
	    n->un.bdvar.name = psym;
	    n->un.bdvar.hbit = psym->u.state.hibit;
	    n->un.bdvar.lbit = psym->u.state.lobit;
	    break;

	case T_IS_SYNONYM:
	    n = node_create(T_IS_STATE, yyfile, *yyline);
	    n->un.bdvar.name = psym->u.synonym.syn;
	    n->un.bdvar.hbit = psym->u.synonym.hibit;
	    n->un.bdvar.lbit = psym->u.synonym.lobit;
	    break;
	
	case T_IS_CONSTANT:
	    n = node_create(T_IS_CONSTANT, yyfile, *yyline);
	    n->un.constant.num = psym->u.constant.const_num;
	    n->un.constant.hbit = psym->u.constant.const_hbit;
	    break;

	default:
	    (void) sprintf(errmsg, "Improper use of %s %s", 
		map_token_to_name(psym->type), psym->name);
	    yyerror(errmsg);
    }
    return n;
}


node *yacc_find_routine(name, args)
char *name;
node *args;
{
    symbol *psym;
    node *n;
    char errmsg[256];

    psym = symtab_lookup(global_syms, name, /* create */ 0);
    if (psym == NIL(symbol)) {
	(void) sprintf(errmsg, "Attempt to call undefined routine %s", name);
	yyerror(errmsg);
    }

    switch(psym->type) {
	case T_IS_ROUTINE:
	    n = node_create(T_FCALL, yyfile, *yyline);
	    n->un.fcall.def = psym;
	    n->un.fcall.actual_args = args;
	    break;

	default:
	    (void) sprintf(errmsg, "%s %s is not a routine", 
		map_token_to_name(psym->type), psym->name);
	    yyerror(errmsg);
    }
    return n;
}

/*
 *  modify A<hi:lo> into (A)<hi-x:lo-x> where x is the least bit of A
 */
node *yacc_hackup_tbit(p, n)
node *p, *n;
{
    node *p1, *p2;

    p->un.ternary.lexpr = n;
    if (! (n->type == T_IS_STATE && n->un.bdvar.name->u.state.lobit != 0)) {
	return p;
    }

    /* Fix the lexpr */
    if (p->un.ternary.mexpr != NIL(node)) {
	p1 = node_create(T_MINUS, yyfile, *yyline);
	p1->un.binary.lexpr = p->un.ternary.mexpr;
	p2 = node_create(T_IS_CONSTANT, yyfile, *yyline);
	p1->un.binary.rexpr = p2;
	p2->un.constant.num = n->un.bdvar.name->u.state.lobit;
	p2->un.constant.hbit = const_width((unsigned) p2->un.constant.num) - 1;
	p->un.ternary.mexpr = p1;
    }

    /* Fix the rexpr */
    if (p->un.ternary.rexpr != NIL(node)) {
	p1 = node_create(T_MINUS, yyfile, *yyline);
	p1->un.binary.lexpr = p->un.ternary.rexpr;
	p2 = node_create(T_IS_CONSTANT, yyfile, *yyline);
	p1->un.binary.rexpr = p2;
	p2->un.constant.num = n->un.bdvar.name->u.state.lobit;
	p2->un.constant.hbit = const_width((unsigned) p2->un.constant.num) - 1;
	p->un.ternary.rexpr = p1;
    }

    return p;
}


node *select_case_colon(nl, nr)
node *nl, *nr;
{
    int i;
    node *list, *x;

    if (nl->type == T_IS_CONSTANT && nr->type==T_IS_CONSTANT) {
	list = NIL(node);
	for(i = nl->un.constant.num; i <= nr->un.constant.num; i++) {
	    x = node_create(T_IS_CONSTANT, yyfile, *yyline);
	    x->un.constant.num = i;
	    x->un.constant.hbit = const_width((unsigned) i) - 1;
	    x->next = list;
	    list = x;
	}
    } else {
	yyerror("Colon not supported in SELECT's for nonconstant values");
/*			{MAKE_BINARY($$, $1, T_COLON, $3);}*/
    }
    return list;
}

read_constant(text, base_string, value, width)
register char *text;
char *base_string;
unsigned *value;
int *width;
{
    register unsigned int len = 0, c, base;

    base = atoi(base_string);

    *value = 0;
    while ((c = *text++) != '\0') {
	if (c >= '0' && c <= '9') c -= '0'; 
	else if (c >= 'A' && c <= 'F') c = c - 'A' + 10;
	else if (c >= 'a' && c <= 'f') c = c - 'a' + 10;
	if (c >= base) {
	    yyerror("badly formed constant");
	}
	*value = *value * base + c;
	len++;
    }

    switch(base) {
	case 2: 
	    *width = len;
	    break;
	case 8: 
	    *width = len * 3; 
	    break;
	case 10: 
	    *width = const_width(*value);
	    break;
	case 16: 
	    *width = len * 4; 
	    break;
	default:
	    yyerror("bad base for constant");
    }
}

int const_width(x)
unsigned x;
{
    if (x == 0) {
	return 1;
    } else {
	return log((double) x) / log(2.0) + 1.0001;
    }
}

node *slow_node_append(head, new)
node *head;		/* head of a linked list of nodes linked via "next" */
node *new;		/* another linked list of nodes linked via "next" */
{
    node *p;

    if (head == NIL(node)) {
	return new;
    } else {
	/* find end of the head chain (SLOW !) */
	for(p = head; p->next != NIL(node); p = p->next)
	    ;
	p->next = new;
	return head;
    }
}


/* a fast cons, assumes that new->next is NIL(node) */
node *cons(head, new)
node *head;
node *new;
{
    if (new == NIL(node)) {
	return head;
    } else {
	new->next = head; 
	return new;
    }
}


node *node_reverse(p) 
register node *p;
{
    register node *plast = NIL(node), *pnext;

    for( ; p != NIL(node); p = pnext) {
	pnext = p->next;
	p->next = plast;		/* point overselves backwards */
	plast = p;
    }
    return plast;
}

static int yacc_dump_symbol(psym)
symbol *psym;
{
    symbol *def;
    node *p;

    switch(psym->type) {
	case T_IS_INPUT_PORT:
	case T_IS_OUTPUT_PORT:
	case T_IS_STATE:
 	    (void) printf("name=%s range=<%d:%d> type=%s is_local=%d is_temp=%d\n",
		psym->name, psym->u.state.hibit, psym->u.state.lobit,
		map_token_to_name(psym->type), psym->u.state.is_local,
		psym->u.state.is_temp);
	    break;

	case T_IS_META:
 	    (void) printf("name=%s value=%d type=%s is_local=%d\n",
		psym->name, psym->VALUE, map_token_to_name(psym->type), 
		psym->u.state.is_local);
	    break;

	case T_IS_SYNONYM:
	    def = psym->u.synonym.syn;
 	    (void) printf("name=%s range=<%d:%d> type=%s\n",
		psym->name, psym->u.synonym.hibit, psym->u.synonym.lobit,
		map_token_to_name(psym->type));
 	    (void) printf("   synonym of: name=%s range=<%d:%d> type=%s is_local=%d\n",
		def->name, def->u.state.hibit, def->u.state.lobit,
		map_token_to_name(def->type), def->u.state.is_local);
	    break;

	case T_IS_CONSTANT:
 	    (void) printf("name=%s range=<%d:%d> type=%s value=%d\n",
		psym->name, psym->u.constant.const_hbit, 0,
		map_token_to_name(psym->type), psym->u.constant.const_num);
	    break;

	case T_IS_MACRO:
 	    (void) printf("name=%s type=%s line=%d value=%s\n",
		psym->name, map_token_to_name(psym->type), 
		psym->u.macro.starting_line, psym->u.macro.value);
	    break;

	case T_IS_ROUTINE:
 	    (void) printf("name=%s range=<%d:%d> type=%s active=%d calls=%d\n",
		psym->name, psym->u.routine.hbit, psym->u.routine.lbit,
		map_token_to_name(psym->type), psym->u.routine.active,
		psym->u.routine.call_count);
	    for(p = psym->u.routine.formal_args; p != NIL(node); p = p->next) {
		def = p->un.bdvar.name;
		(void) printf("    formal arg: name=%s range=<%d:%d> type=%s (%d)\n",
		    def->name, def->u.state.hibit, def->u.state.lobit,
		    map_token_to_name(def->type), def->u.state.is_local);
	    }
	    break;

	case T_LABEL:
	    (void) printf("name=%s type=%s\n", psym->name, 
		map_token_to_name(psym->type));
	    break;

	default:
	    (void) printf("Bad type in yacc_dump_symbol %s\n",
		map_token_to_name(psym->type));
	    break;
    }
    return 1;		/* continue with next symbol */
}

	    
yacc_dump_symbol_table(table)
symbol_table *table;
{
    symtab_foreach(table, yacc_dump_symbol, -1);
}
