%{
#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yylex.h"

/*
 *  Deviations allowed from BDS syntax:
 *	1) BEHAVIOR / ENDBEHAVIOR keywords are optional;
 *	2) name following ENDMODEL is optional;
 *	3) name following ENDROUTINE is optional;
 *	4) Declaring a variable with no size (disallowed by BDS) defines
 *	   its size as a single bit
 *	5) Syntax rule for ZXT/OXT/SXT is "SXT {width=expr} expr"
 *	6) Constant-valued expressions are allowed wherever constants are
 *	     (e.g., declaring variable sizes, and defining other constants)
 *
 *  We use operator priorities to resolve ambiguities in many cases; as in any
 *  good grammar, there is a single shift/reduce conflict (IF...THEN...ELSE).
 */


#define MAKE_ASSIGNMENT(ret, var, e1) \
    ret = node_create(T_EQ, yyfile, *yyline); ret->un.assign.lexpr = var;\
    ret->un.assign.rexpr = e1;\

#define MAKE_TERNARY(ret, op, e1, e2, e3) \
    ret = node_create(op, yyfile, *yyline); ret->un.ternary.lexpr = e1;\
    ret->un.ternary.mexpr = e2; ret->un.ternary.rexpr = e3;\

#define MAKE_BINARY(ret, e1, op, e2)\
    ret = node_create(op, yyfile, *yyline); ret->un.binary.lexpr = e1;\
    ret->un.binary.rexpr = e2;

#define MAKE_UNARY(ret, op, e1)\
    ret = node_create(op, yyfile, *yyline); ret->un.unary.expr = e1;

#define MAKE_IF(ret, e1, stmt1, stmt2)\
    ret = node_create(T_IF, yyfile, *yyline); ret->un.bdif.expr = e1; \
    ret->un.bdif.thenb = stmt1; ret->un.bdif.elseb = stmt2;

#define MAKE_SELECT(ret, op, e1, stmts)\
    ret = node_create(op, yyfile, *yyline); ret->un.select.expr = e1;\
    ret->un.select.bdcase = node_reverse(stmts);

#define MAKE_FOR(ret, op, v, from_e, to_e, by_e, st) \
    ret = node_create(op, yyfile, *yyline); ret->un.bdfor.var = v; \
    ret->un.bdfor.from = from_e; ret->un.bdfor.to = to_e; \
    ret->un.bdfor.by = by_e; ret->un.bdfor.stmt = st;

#define MAKE_CONST(ret, number, base) \
    ret = node_create(T_IS_CONSTANT, yyfile, *yyline);\
    read_constant(number, base, &dummy, &(ret->un.constant.hbit));\
    ret->un.constant.num = (int) dummy; \
    ret->un.constant.hbit -= 1;

#define MAKE_VAR(ret, op, varname, nobit, highbit, lowbit)\
    ret = node_create(op, yyfile, *yyline);\
    ret->un.bdvar.name = yacc_lookup(varname);\
    ret->un.bdvar.nbit = nobit; ret->un.bdvar.lbit = lowbit;\
	ret->un.bdvar.hbit = highbit;

#define MAKE_LEAVE(ret, label_name)\
    ret = node_create(T_LEAVE, yyfile, *yyline);\
    ret->un.leave.name = yacc_lookup(label_name);

#define MAKE_LABEL(ret, label_name, body)\
    ret = node_create(T_LABEL, yyfile, *yyline);\
    ret->un.label.name = yacc_lookup(label_name);\
    yacc_check_defined(ret->un.label.name);\
    ret->un.label.name->type = T_LABEL;\
    ret->un.label.stmt = body;\
    ret->un.label.name->u.label.def = ret;

#define SAVE_YYTEXT		strcpy(ALLOC(char, yylength + 1), yytext)

extern int select_unique;
extern int got_a_model;

node *input_ports, *output_ports;
char *model_name;

static char *routine_name;
%}


%union {
    int intval;
    char *strval;
    node *expr;
}

%start model_or_routine

%token <intval>
T_AND
T_BASE
T_BEGIN
T_BEHAVIOR
T_BUF
T_BY
T_CLOSE_ANGLE
T_CLOSE_BRACKET
T_CLOSE_CURLY
T_CLOSE_PAREN
T_COLON
T_COMMA
T_CONCAT
T_CONSTANT
T_DIV
T_DO
T_DOWNTO
T_ELSE
T_END
T_ENDBEHAVIOR
T_ENDMACRO
T_ENDMODEL
T_ENDMODULE
T_ENDROUTINE
T_ENDSELECT
T_ENDSELECTALL
T_ENDSELECTONE
T_ENTRY
T_EQ
T_EQL
T_EQV
T_EXIT
T_EXTERNAL
T_FOR
T_FORWARD
T_FOURSTATE
T_FROM
T_GEQ
T_GTR
T_HALT
T_IF
T_INFORM
T_INPUT
T_INVALID
T_LEAVE
T_LEQ
T_LSS
T_MACRO
T_MINUS
T_MOD
T_MODEL
T_MODULE
T_NAME
T_NAND
T_NEQ
T_NEXT
T_NOENTRY
T_NOR
T_NOT
T_NUMBER
T_OPEN_ANGLE
T_OPEN_BRACKET
T_OPEN_CURLY
T_OPEN_PAREN
T_OR
T_OTHERWISE
T_OUTPUT
T_OXT
T_PLUS
T_PORT
T_PRINT
T_REPEAT
T_RETURN
T_REVISION
T_ROUTINE
T_SELECT
T_SELECTALL
T_SELECTONE
T_SL0
T_SL1
T_SLR
T_SR0
T_SR1
T_SRR
T_STATE
T_STATIC
T_STRING
T_SXT
T_SYNONYM
T_THEN
T_TIMES
T_TO
T_TWOSTATE
T_UNTIL
T_VAR
T_WHILE
T_WIDTH
T_XOR
T_ZXT

%type <expr>
routine_body
routine_stmts
local_declaration
insertion
const_identifier
expression_list
identifier
routine_header
routine_head1
opt_routine_args
opt_arg_list
arg_list
routine_call
routine
constant
bit_subscripts
lstmt
stmt
stmt_list
if_statement
select_statement
for_statement
while_statement
until_statement
do_until_statement
do_while_statement
repeat_statement
return_statement
leave_statement
print_statement
expression
select_list
select_body
select_case_list
select_case

%type <strval>
string
number
name
opt_name

%left T_SXT T_ZXT T_OXT T_WIDTH
%left T_EQV T_XOR
%left T_OR T_NOR
%left T_AND T_NAND
%left T_NOT T_BUF
%left T_GEQ T_GTR T_LEQ T_LSS T_NEQ T_EQL
%left T_PLUS T_MINUS
%left T_TIMES T_DIV T_MOD
%left T_SL0 T_SR0 T_SR1 T_SL1 T_SLR T_SRR
%left T_CONCAT
%%

/*
 *   For now, syntax error for global modifier on behavior header
 *            syntax error for behavior subblock directives
 *  	
 *	      T_EXTERNAL ignored because MODULES are not implemented
 *	      PORT parsed and ignored (including directives)
 */

model_or_routine:   model
			{got_a_model = 1;}
		|   behavior_stmts
			{got_a_model = 0;}
		;

model		:   model_header behavior model_trailer
		;

model_header	:   T_MODEL name opt_arg_list T_EQ opt_arg_list T_NEXT
			{node *p;
			 model_name = $2;
			 for(p = $3; p != NIL(node); p = p->next) {
			    yacc_define_symbol(p, T_IS_OUTPUT_PORT);
			 }
			 for(p = $5; p != NIL(node); p = p->next) {
			    yacc_define_symbol(p, T_IS_INPUT_PORT);
			 }
			 input_ports = node_reverse($5);
			 output_ports = node_reverse($3);
			}
		;

model_trailer	:   T_ENDMODEL opt_name T_NEXT
			{if ($2 != NIL(char)) {
			    if (upper_strcmp(model_name, $2) != 0) {
			        yyerror("MODEL name and ENDMODEL name should match");
			    }
			    FREE($2);
			 }
			}
		;
			     
behavior	:   behavior_header behavior_stmts behavior_trailer
	 	;

behavior_header :   /* empty -- warning, not allowed by DECSIM ! */
		|   T_BEHAVIOR opt_name T_NEXT
			{if ($2 != NIL(char)) FREE($2);}
		;

behavior_trailer:   /* empty -- warning, not allowed by DECSIM ! */
		|   T_ENDBEHAVIOR opt_name T_NEXT
			{if ($2 != NIL(char)) FREE($2);}
		;

behavior_stmts	:   /* empty */
		|   behavior_stmts behavior_stmt 
		;

behavior_stmt   :   global_decl T_NEXT
		|   insertion T_NEXT
		;

global_decl	:   T_FORWARD T_ROUTINE const_identifier opt_routine_args
			{yacc_define_routine($3, $4, /*forward_decl*/ 1);
			 FREE($3); if ($4 != NIL(node)) node_dispose($4);
			}
		|   port_decl
		|   T_REVISION T_NUMBER
		|   routine
		|   local_declaration  /* poor name, really global decl */
		;

insertion	:   T_INFORM string
			{(void) fprintf(stderr, "%s\n", $2); FREE($2); $$ = NIL(node);}
		;

/*
 *  The basic unit we are interested in is the ROUTINE
 */

routine		:   routine_header T_NEXT routine_body T_ENDROUTINE opt_name
			{char buffer[256];
			 if ($5 != NIL(char)) {
			     if (upper_strcmp($5, $1->un.bdvar.name->name) != 0) {
			         yyerror("ROUTINE name and ENDROUTINE name should match");
			     }
			     FREE($5); 
			 }
			 /* wrap a LABEL around the routine */
			 (void) strcpy(buffer, "_LABEL:");
			 (void) strcat(buffer, routine_name);
			 MAKE_LABEL($$, buffer, $3);
			 $1->un.bdvar.name->u.routine.body = $$;

			 /* save the local symbol table and reset it to NIL */
			 $1->un.bdvar.name->u.routine.locals = local_syms;
			 local_syms = NIL(symbol_table);
			 FREE($1); 
			 $$ = NIL(node);
			}
  		;

routine_header	:   routine_head1 opt_routine_args
			{yacc_define_routine($1, $2, /*forward_decl*/ 0);}
		;			
routine_head1   :   T_ROUTINE const_identifier
		 	{local_syms = symtab_init(LOCAL_SYMTAB_SIZE); 
			 routine_name = $2->un.bdvar.name->name;
			 $$ = $2;
			}
		;

opt_routine_args:   /* empty */
			{$$ = NIL(node);}
		|   T_OPEN_PAREN opt_arg_list T_CLOSE_PAREN
			{$$ = $2;}
		;

opt_arg_list	:   /* empty */
			{$$ = NIL(node);}
		|   arg_list
		;

arg_list	:   const_identifier
                |   T_VAR const_identifier
                        {$$ = $2; $$->un.bdvar.var = 1;}
		|   arg_list T_COMMA const_identifier
			{$$ = cons($1, $3);}
		|   arg_list T_COMMA T_VAR const_identifier
			{$$ = cons($1, $4); $4->un.bdvar.var = 1;}
		;

routine_body    :   /* empty */
			{$$ = NIL(node);}
		|   routine_body routine_stmts
			{$$ = slow_node_append($1, $2);}

routine_stmts	:   local_declaration T_NEXT	/* should return $$ = NIL(node) */
		|   insertion T_NEXT		/* should return $$ = NIL(node) */
		|   lstmt T_NEXT		/* statements linked via next */
		;

opt_name 	:   /* empty */
			{$$ = NIL(char);}
		|   name
		;


/*
 *  Declarations (PORT, STATE, SYNONYM, CONSTANT)
 */

local_declaration:  state_decl 
			{$$ = NIL(node);} 
		|   const_decl 
			{$$ = NIL(node);} 
		|   synonym_decl 
			{$$ = NIL(node);} 
		;

/* PORT declarations */
port_decl       :   T_PORT port_list 
		;
port_list	:   port 
		|   port_list T_COMMA port 
		;
port		:   const_identifier
			{FREE($1);}
		|   const_identifier port_modifiers
			{FREE($1);}
		;

/* STATE declarations */
state_decl      :   T_STATE state_list
		;
state_list	:   state 
		|   state_list T_COMMA state 
		;
state		:   const_identifier
			{yacc_define_symbol($1, $1->type); FREE($1);}
		|   const_identifier port_modifiers
			{yacc_define_symbol($1, $1->type); FREE($1);}
		;

/* SYNONYM declaration */
synonym_decl    :   T_SYNONYM synonym_list
		;
synonym_list	:   synonym 
		|   synonym_list T_COMMA synonym 
		;
synonym	        :   const_identifier T_EQ const_identifier 
			{yacc_define_synonym($1, $3); FREE($1); FREE($3);}
		;

/* CONSTANT declaration */
const_decl      :   T_CONSTANT const_list 
		;
const_list	:   dconstant 
		|   const_list T_COMMA dconstant 
		;
dconstant	:   const_identifier T_EQ constant
			{yacc_define_constant($1, $3); FREE($1); FREE($3);}

		;

/*  an identifier with (optional) constant subscripts  */
const_identifier:   name
			{MAKE_VAR($$, T_IS_STATE, $1, 1, 0, 0); FREE($1);}
		|   name T_OPEN_ANGLE T_CLOSE_ANGLE
			{MAKE_VAR($$, T_IS_META, $1, 0, 31, 0); FREE($1);}
		|   name T_OPEN_ANGLE expression T_CLOSE_ANGLE
			{$$ = yacc_make_variable($1, T_IS_STATE, $3, NIL(node));}
		|   name T_OPEN_ANGLE expression T_COLON expression T_CLOSE_ANGLE
			{$$ = yacc_make_variable($1, T_IS_STATE, $3, $5);}
		;

/*
 *  YACC Grammar for BDS (BDSYN) Statements
 */
lstmt		:   stmt
		|   name T_COLON stmt
			{MAKE_LABEL($$, $1, $3);}
		;

stmt		:   T_BEGIN stmt_list T_END
			{$$ = node_reverse($2);}
		|   expression
			{$$ = $1;}
		|   identifier T_EQ expression
			{MAKE_ASSIGNMENT($$, $1, $3);}
		|   identifier T_EQ T_OPEN_CURLY T_NUMBER T_CLOSE_CURLY expression
			{MAKE_ASSIGNMENT($$, $1, $6);}
		|   if_statement
		|   select_statement
  		|   for_statement
		|   leave_statement
		|   return_statement
		|   print_statement
		|   while_statement
		|   until_statement
		|   do_until_statement
		|   do_while_statement
		|   repeat_statement
		|   T_HALT
			{$$ = NIL(node);}
		|   error T_NEXT
			{$$ = NIL(node);}
		;

stmt_list	:   lstmt T_NEXT
		|   stmt_list lstmt T_NEXT
			{$$ = cons($1, $2);}
		;


for_statement	:   T_FOR identifier T_FROM expression T_TO expression 
		    T_DO lstmt
			{MAKE_FOR($$, T_FOR, $2, $4, $6, NIL(node), $8);}
		|   T_FOR identifier T_FROM expression T_TO expression 
		    T_BY expression T_DO lstmt
			{MAKE_FOR($$, T_FOR, $2, $4, $6, $8, $10);}
		|   T_FOR identifier T_FROM expression T_DOWNTO expression 
		    T_DO lstmt
			{MAKE_FOR($$, T_FOR_DOWN, $2, $4, $6, NIL(node), $8);}
		|   T_FOR identifier T_FROM expression T_DOWNTO expression 
		    T_BY expression T_DO lstmt
			{MAKE_FOR($$, T_FOR_DOWN, $2, $4, $6, $8, $10);}
		;


if_statement	:   T_IF expression T_THEN lstmt
			{MAKE_IF($$, $2, $4, NIL(node));}
		|   T_IF expression T_THEN lstmt T_ELSE lstmt
			{MAKE_IF($$, $2, $4, $6);}
		;


select_statement:   T_SELECT expression T_FROM select_list T_ENDSELECT
			{MAKE_SELECT($$, T_SELECT, $2, $4);}
		|   T_SELECTONE expression T_FROM select_list T_ENDSELECTONE
			{MAKE_SELECT($$, T_SELECTONE, $2, $4);}
		|   T_SELECTALL expression T_FROM select_list T_ENDSELECTALL
			{if (select_unique) 
			    {MAKE_SELECT($$, T_SELECT, $2, $4);}
			 else 
			    {MAKE_SELECT($$, T_SELECTALL, $2, $4);}
			}
		;


leave_statement	:   T_LEAVE name 
			{MAKE_LEAVE($$, $2);}
		;


return_statement:   T_RETURN
			{char buffer[256]; 
			 (void) strcpy(buffer, "_LABEL:");
			 (void) strcat(buffer, routine_name);
			 MAKE_LEAVE($$, buffer);
			}
		|   T_RETURN expression
			{char buffer[256]; node *ret;
			 (void) strcpy(buffer, "$"); 
			 (void) strcat(buffer, routine_name);
			 MAKE_ASSIGNMENT(ret, yacc_find_symbol(buffer), $2);
			 (void) strcpy(buffer, "_LABEL:");
			 (void) strcat(buffer, routine_name);
			 MAKE_LEAVE(ret->next, buffer);
			 $$ = node_create(T_LABEL, yyfile, *yyline);
			 $$->un.label.name = yacc_lookup("");
			 $$->un.label.name->type = T_LABEL;
			 $$->un.label.stmt = ret;
			 $$->un.label.name->u.label.def = $$;
			}
		;



/*
 *  Details for the select statement
 */


select_list	:   select_body
		|   select_list select_body
			{$$ = cons($1, $2);}
		;

select_body	:   T_OPEN_BRACKET select_case_list T_CLOSE_BRACKET T_COLON 
		    lstmt T_NEXT
			{$$ = node_create(T_CASE, yyfile, *yyline);
			$$->un.select.expr = node_reverse($2); 
			$$->un.select.bdcase = $5;}
		|   T_OPEN_BRACKET select_case_list T_CLOSE_BRACKET T_COLON T_NEXT
			{$$ = node_create(T_CASE, yyfile, *yyline);
			$$->un.select.expr = node_reverse($2); 
			$$->un.select.bdcase = NIL(node);}
  		;

select_case_list:   select_case
		|   select_case_list T_COMMA select_case
			{$$ = cons($1, $3);}
		;

select_case	:   expression
		|   expression T_COLON expression
			{$$ = select_case_colon($1, $3);}
		|   T_OTHERWISE
			{$$ = node_create(T_OTHERWISE, yyfile, *yyline);}
		;		


/*
 *  Unsupported and ignored constructs
 */

while_statement	:   T_WHILE expression T_DO lstmt 
			{(void) fprintf(stderr, "WHILE ... DO ignored\n"); $$=NIL(node);}
		;

until_statement	:   T_UNTIL expression T_DO lstmt 
			{(void) fprintf(stderr, "UNTIL ... DO ignored\n"); $$=NIL(node);}
		;

do_until_statement: T_DO lstmt T_UNTIL expression 
			{(void) fprintf(stderr, "DO ... UNTIL ignored\n"); $$=NIL(node);}
		;

do_while_statement: T_DO lstmt T_WHILE expression 
			{(void) fprintf(stderr, "DO ... WHILE ignored\n"); $$=NIL(node);}
		;

repeat_statement:   T_REPEAT lstmt 
			{(void) fprintf(stderr, "REPEAT ignored\n"); $$=NIL(node);}
		;

print_statement	:   T_PRINT T_OPEN_PAREN p_list T_CLOSE_PAREN
		        {$$ = NIL(node);}
		;
p_list		:   p_expr 
		|   p_list T_COMMA p_expr 
		;
p_expr		:   /* empty */ 
		|   T_STRING 
		|   expression 
		;

/*
 *  expressions
 */
expression	:   identifier

		|   constant

		|   T_OPEN_PAREN expression T_CLOSE_PAREN
			{$$ = $2;}
		|   T_OPEN_PAREN expression T_CLOSE_PAREN bit_subscripts
			{$$ = $4; $4->un.ternary.lexpr = $2;}

		|   routine_call

	/* unary logical operators */
		|   T_NOT expression 
			{MAKE_UNARY($$, T_NOT, $2);}
		|   T_BUF expression 
			{MAKE_UNARY($$, T_BUF, $2);}
		|   T_WIDTH expression %prec T_WIDTH
			{MAKE_UNARY($$, T_WIDTH, $2);}

	/* unary arthimetic operators */
		|   T_PLUS expression %prec T_SL0
			{MAKE_UNARY($$, T_UNARY_PLUS, $2);}
		|   T_MINUS expression %prec T_SL0
			{MAKE_UNARY($$, T_UNARY_MINUS, $2);}

	/* binary logical operators */
		|   expression T_AND expression
			{MAKE_BINARY($$, $1, T_AND, $3);}
		|   expression T_NAND expression
			{MAKE_BINARY($$, $1, T_NAND, $3);}
		|   expression T_OR expression
			{MAKE_BINARY($$, $1, T_OR, $3);}
		|   expression T_NOR expression
			{MAKE_BINARY($$, $1, T_NOR, $3);}
		|   expression T_XOR expression
			{MAKE_BINARY($$, $1, T_XOR, $3);}
		|   expression T_EQV expression
			{MAKE_BINARY($$, $1, T_EQV, $3);}

	/* binary shift operators */ 
		|   expression T_SL0 expression
			{MAKE_BINARY($$, $1, T_SL0, $3);}
		|   expression T_SR0 expression
			{MAKE_BINARY($$, $1, T_SR0, $3);}
		|   expression T_SL1 expression
			{MAKE_BINARY($$, $1, T_SL1, $3);}
		|   expression T_SR1 expression
			{MAKE_BINARY($$, $1, T_SR1, $3);}
		|   expression T_SLR expression
			{MAKE_BINARY($$, $1, T_SLR, $3);}
		|   expression T_SRR expression
			{MAKE_BINARY($$, $1, T_SRR, $3);}

	/* binary (well, actually unary) sign-extend operators */
		|   T_SXT T_OPEN_CURLY T_WIDTH T_EQ expression 
		    T_CLOSE_CURLY expression 	%prec T_SXT
			{MAKE_BINARY($$, $7, T_SXT, $5);}
		|   T_ZXT T_OPEN_CURLY T_WIDTH T_EQ expression 
		    T_CLOSE_CURLY expression    %prec T_ZXT
			{MAKE_BINARY($$, $7, T_ZXT, $5);}
		|   T_OXT T_OPEN_CURLY T_WIDTH T_EQ expression 
		    T_CLOSE_CURLY expression    %prec T_OXT
			{MAKE_BINARY($$, $7, T_OXT, $5);}

	/*  
         * Concatenate bits.  We see high order bits first; the high
         * order bits end up at the end of the chain (no node_reverse
         * is needed)
	 */
		|   expression T_CONCAT expression
			{$3->conc = $1; $$ = $3;}

	/* binary arithmetic operators */
		|   expression T_PLUS expression
			{MAKE_BINARY($$, $1, T_PLUS, $3);}
		|   expression T_MINUS expression
			{MAKE_BINARY($$, $1, T_MINUS, $3);}
		|   expression T_TIMES expression
			{MAKE_BINARY($$, $1, T_TIMES, $3);}
		|   expression T_DIV expression
			{MAKE_BINARY($$, $1, T_DIV, $3);}
		|   expression T_MOD expression
			{MAKE_BINARY($$, $1, T_MOD, $3);}

	/* binary relational operators */
		|   expression T_EQL expression
			{MAKE_BINARY($$, $1, T_EQL, $3);}
		|   expression T_NEQ expression
			{MAKE_BINARY($$, $1, T_NEQ, $3);}
		|   expression T_GEQ expression
			{MAKE_BINARY($$, $1, T_GEQ, $3);}
		|   expression T_LEQ expression
			{MAKE_BINARY($$, $1, T_LEQ, $3);}
		|   expression T_GTR expression
			{MAKE_BINARY($$, $1, T_GTR, $3);}
		|   expression T_LSS expression
			{MAKE_BINARY($$, $1, T_LSS, $3);}
		;

routine_call	:   name T_OPEN_PAREN T_CLOSE_PAREN
			{$$ = yacc_find_routine($1, NIL(node));}
		|   name T_OPEN_PAREN expression_list T_CLOSE_PAREN
			{$$ = yacc_find_routine($1, $3); FREE($1);}
		;

expression_list :   expression
		|   expression_list T_COMMA expression
			{$$ = cons($1, $3);}
		;


/*  identifier -- return either a T_IS_STATE node, or update a T_BIT node */
identifier	:   name 
			{$$ = yacc_find_symbol($1); FREE($1);}
		|   name bit_subscripts
			{if ($2 == NIL(node)) {
			    $$ = yacc_find_symbol($1);
			 } else {
			    $$ = yacc_hackup_tbit($2, yacc_find_symbol($1));
			 }
			 FREE($1);
			}
		;


/*  bit_subscripts -- return a T_BIT operator node */
bit_subscripts	:   T_OPEN_ANGLE T_CLOSE_ANGLE
			{$$ = NIL(node);}
		|   T_OPEN_ANGLE expression T_CLOSE_ANGLE
			{MAKE_TERNARY($$, T_BIT, NIL(node), $2, NIL(node));}
/*			{MAKE_TERNARY($$, T_BIT, NIL(node), $2, node_dup($2));}*/
		|   T_OPEN_ANGLE expression T_COLON expression T_CLOSE_ANGLE
			{MAKE_TERNARY($$, T_BIT, NIL(node), $2, $4);}
  		;


/* a constant node */
constant	:   number
			{int dummy; MAKE_CONST($$, $1, "10"); FREE($1);}
		|   number T_BASE number
			{int dummy; MAKE_CONST($$, $1, $3); FREE($1); FREE($3);} 
		|   name T_BASE number
			{int dummy; MAKE_CONST($$, $1, $3); FREE($1); FREE($3);} 
		;


name		:   T_NAME
			{$$ = SAVE_YYTEXT;}
		;

number		:   T_NUMBER
			{$$ = SAVE_YYTEXT;}
		;

string	        :   T_STRING
			{$$ = SAVE_YYTEXT;}
		;

port_modifiers  :   port_modifier
		|   port_modifiers port_modifier
		;

port_modifier	:   T_INPUT
		|   T_OUTPUT
		|   T_ENTRY T_OPEN_PAREN T_NAME T_CLOSE_PAREN
		|   T_NOENTRY
		|   T_TWOSTATE 
		|   T_FOURSTATE 
		;
