#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "vov.h"
#include "yylex.h"
#include "y.tab.h"

#define GETC()	    (*npushback != 0 ? getc_pushback() : \
			((*is_macro) ? getc_macro() : mygetc(lexin)))
#define UNGETC(c)   (input_stream[n_input_stream-1].pushback[(*npushback)++] = c);


/* 
 *  Define the stack of input streams (macros and files)
 */
#define MAXSTACK 24

struct input_stream_stack_struct {
    char *filename;		/* name of file (or name of macro) */
    int line;			/* current line number */
    int is_macro;		/* indicates entry is a macro */
    int npushback;		/* number of characters pushed back */
    char pushback[10];		/* characters pushed back on this stream */
    union input_type_union {
	FILE *fp;		/* file pointer */
	char *macro;		/* string defining the macro */
    } input_type;
} input_stream[MAXSTACK];

int n_input_stream;		/* number of input streams */


/* 
 *  Fast access to the current information on the input_stream stack
 *  yyline and yyfile are available for use by the parser
 */
static FILE *lexin;		/* current file pointer */
static int *npushback;		/* current pushback counter */
static int *is_macro;		/* current is_macro indicator */
int *yyline;			/* current line number in that file */
char *yyfile;			/* current file name */


/* We pass information in yytext/yylength for use by yyparse routines */
char yytext[1024];			/* the token read from the file */
int yylength;


/* Define the symbol tables */
symbol_table *local_syms, *global_syms, *reserved_syms;

/* magic character list (initialized in yylex_init()) */
static int magic_char_lookup[256];

/* Reserved word list */
static struct reserved_words_struct {
    char *keyword;
    int token_value;
} reserved_words[] = {
    "AND", T_AND,
    "BEGIN", T_BEGIN,
    "BEHAVIOR", T_BEHAVIOR,
    "BUF", T_BUF,
    "BY", T_BY,
    "CONSTANT", T_CONSTANT,
    "DO", T_DO,
    "DOWNTO", T_DOWNTO,
    "ELSE", T_ELSE,
    "END", T_END,
    "ENDBEHAVIOR", T_ENDBEHAVIOR,
    "ENDMACRO", T_ENDMACRO,
    "ENDMODEL", T_ENDMODEL,
    "ENDMODULE", T_ENDMODULE,
    "ENDROUTINE", T_ENDROUTINE,
    "ENDSELECT", T_ENDSELECT,
    "ENDSELECTALL", T_ENDSELECTALL,
    "ENDSELECTONE", T_ENDSELECTONE,
    "ENTRY", T_ENTRY,
    "EQL", T_EQL,
    "EQV", T_EQV,
    "EXIT", T_EXIT,
    "EXTERNAL", T_EXTERNAL,
    "FOR", T_FOR,
    "FORWARD", T_FORWARD,
    "FOURSTATE", T_FOURSTATE,
    "FROM", T_FROM,
    "GEQ", T_GEQ,
    "GTR", T_GTR,
    "HALT", T_HALT,
    "IF", T_IF,
    "INFORM", T_INFORM,
    "INPUT", T_INPUT,
    "LEAVE", T_LEAVE,
    "LEQ", T_LEQ,
    "LSS", T_LSS,
    "MACRO", T_MACRO,
    "MOD", T_MOD,
    "MODEL", T_MODEL,
    "MODULE", T_MODULE,
    "NAND", T_NAND,
    "NEQ", T_NEQ,
    "NOENTRY", T_NOENTRY,
    "NOR", T_NOR,
    "NOT", T_NOT,
    "OR", T_OR,
    "OTHERWISE", T_OTHERWISE,
    "OUTPUT", T_OUTPUT,
    "OXT", T_OXT,
    "PORT", T_PORT,
    "%PRINT", T_PRINT,
    "REPEAT", T_REPEAT,
    "REQUIRE", T_REQUIRE,
    "RETURN", T_RETURN,
    "REVISION", T_REVISION,
    "ROUTINE", T_ROUTINE,
    "SELECT", T_SELECT,
    "SELECTALL", T_SELECTALL,
    "SELECTONE", T_SELECTONE,
    "SL0", T_SL0,
    "SL1", T_SL1,
    "SLR", T_SLR,
    "SR0", T_SR0,
    "SR1", T_SR1,
    "SRR", T_SRR,
    "STATE", T_STATE,
    "STATIC", T_STATIC,
    "SXT", T_SXT,
    "SYNONYM", T_SYNONYM,
    "THEN", T_THEN,
    "TO", T_TO,
    "TWOSTATE", T_TWOSTATE,
    "UNTIL", T_UNTIL,
    "VAR",   T_VAR,
    "WHILE", T_WHILE,
    "WIDTH", T_WIDTH,
    "XOR", T_XOR,
    "ZXT", T_ZXT,

/* The next symbols are NOT allowed to appear in the input, but are provided
 * so that "map_token_to-name" may return a name for them.  The leading asterisk
 * guarantees that they will NEVER be seen as valid tokens (a leading asterisk
 * is always eaten as a magic character)
 */
    "*LIF", T_LIF,
    "*STATE", T_IS_STATE,
    "*META", T_IS_META,
    "*CONSTANT", T_IS_CONSTANT,
    "*SYNONYM", T_IS_SYNONYM,
    "*ROUTINE", T_IS_ROUTINE,
    "*INPUT_PORT", T_IS_INPUT_PORT,
    "*OUTPUT_PORT", T_IS_OUTPUT_PORT,
    "*MACRO", T_IS_MACRO,
    "*UNDEFINED", T_IS_UNDEFINED,
    "*DUP", T_DUP,
    "*ANDR", T_UNARY_AND,
    "*ORR", T_UNARY_OR,
    "*NANDR", T_UNARY_NAND,
    "*NORR", T_UNARY_NOR,
    "*XORR", T_UNARY_XOR,
    "*EQVR", T_UNARY_EQV,
    "*UNARY_PLUS", T_UNARY_PLUS,
    "*UNARY_MINUS", T_UNARY_MINUS,
    "*<>", T_BIT,
    "*FCALL", T_FCALL,
    "*LABEL", T_LABEL,
};
#define NUM_RESERVED_WORDS	\
	(sizeof(reserved_words) / sizeof(struct reserved_words_struct))


static struct magic_chars_struct {
    char c;
    int token_value;
} magic_chars[] = {
    ':', T_COLON,
    '=', T_EQ,
    ',', T_COMMA,
    ';', T_NEXT,
    '&', T_CONCAT,
    '(', T_OPEN_PAREN,
    ')', T_CLOSE_PAREN,
    '[', T_OPEN_BRACKET,
    ']', T_CLOSE_BRACKET,
    '{', T_OPEN_CURLY,
    '}', T_CLOSE_CURLY,
    '<', T_OPEN_ANGLE,
    '>', T_CLOSE_ANGLE,
    '#', T_BASE,
    '+', T_PLUS,
    '-', T_MINUS,
    '*', T_TIMES,
    '/', T_DIV,
};
#define NUM_MAGIC_CHARS		\
	sizeof(magic_chars) / sizeof(struct magic_chars_struct)


#define LINELENGTH 1025

static char line[LINELENGTH] = "";
static int linepos = 0, linenum = 0;
static int errorcount = 0, list = 0;

yy_reset_buffer()
{
    *line = '\0';
    linepos = 0;
}

where ()

{
	int i;

	if (*is_macro) {
		(void) fprintf(stderr, "syntax error in macro %s at line %d\n", 
  			yyfile, *yyline);
	} else {
		(void) fprintf(stderr, "syntax error in file %s at line %d\n", 
    		yyfile, *yyline);
	}
   	for(i = n_input_stream-2; i >= 0; i--) {
		(void) fprintf(stderr, "called from ");
		if (input_stream[i].is_macro) {
   			(void) fprintf(stderr, "macro %s at line %d\n", 
			input_stream[i].filename, input_stream[i].line);
		} else {
   			(void) fprintf(stderr, "file %s at line %d\n", 
			input_stream[i].filename, input_stream[i].line);
		}
   	}
}

yyerror (s)
register char *s;
{
	yyshowerror (s, 0, 1 /* exit */);
}

yysoft (s)
register char *s;
{
	yyshowerror (s, 0, 0 /* exit */);
}

yyshowerror (s, t, fatal)
register char *s;
int t, fatal;
{
	int lp;

	if (s || ! list) {
	/* print position */
		if (linepos < 0) {
			(void) fprintf (stderr, "unexpected end of file at line %d\n", linenum);
		}
		else {
			(void) fputs(line, stderr);
			for (lp = 0; lp < linepos - 1; lp++)
				if (line[lp] == '\t')
					putc('\t', stderr);
				else	putc(' ', stderr);
			(void) fputs ("^\n", stderr);
		}
	/* print message */
		(void) fprintf(stderr, "[%d] ", errorcount+1);
		if (s) {
			(void) fputs(s, stderr);
			putc('\n', stderr);
			where ();
			if (fatal) exit(-1);
			else return;
		}
		where ();
		(void) fputs("expecting one of: ", stderr);
		(void) fputs(map_token_to_name (t), stderr);
		list = 1;
		return;
	}
	if (t) { /* other tokens */
		putc (' ', stderr);
		(void) fputs(map_token_to_name (t), stderr);
		return;
	}
	/* end of message */
	putc('\n', stderr);
	list = 0;
	if (fatal) exit(-1);
	else return;
}

int mygetc (inputfile)
FILE *inputfile;
{
	if ((linepos == -1) || (line[linepos] == '\0')) {
		if (fgets (line, (LINELENGTH - 1), inputfile)) {
			linenum++;
			linepos = 0;
	/* insert here listing generation */
		}
		else {
			linepos = - 1;
			return (EOF);
		}
	}
	return ((int) line[linepos++]);
}


/*
 *  map_token_to_name
 */
char *map_token_to_name(t)
int t;
{
    register int i;
    static char s[2];

    for(i = 0; i < NUM_RESERVED_WORDS; i++) {
	if (t == reserved_words[i].token_value) {
	    if (reserved_words[i].keyword[0] == '*')
		return reserved_words[i].keyword + 1;
	    else
		return reserved_words[i].keyword;
	}
    }
    for(i = 0; i < NUM_MAGIC_CHARS; i++) {
	if (t == magic_chars[i].token_value) {
	    s[0] = magic_chars[i].c;
	    s[1] = '\0';
	    return s;
	}
    }
    return "??WHAT??";
}



/*
 *  yylex_init -- called first to initialize some tables to speed the
 *  processing of individual characters
 */
yylex_init()
{
    register int i;
    register symbol *psym;

    /* Define a lookup table for the magic characters */
    for(i = 0; i < 256; i++)
	magic_char_lookup[i] = 0;
    for(i = 0; i < NUM_MAGIC_CHARS; i++)
	magic_char_lookup[magic_chars[i].c] = magic_chars[i].token_value;

    /* Hash the reserved words into a reserved words symbol table */
    reserved_syms = symtab_init(117);
    for(i = 0; i < NUM_RESERVED_WORDS; i++) {
	psym = symtab_lookup(reserved_syms, reserved_words[i].keyword, 1);
	psym->type = reserved_words[i].token_value;
    }

    global_syms = symtab_init(GLOBAL_SYMTAB_SIZE);
    local_syms = NIL(symbol_table);
}


/*  getc_pushback -- fetch character from the pushback buffer */
int getc_pushback()
{
    return input_stream[n_input_stream-1].pushback[--(*npushback)];
}


/*  getc_macro -- fetch a character from an active macro  */
int getc_macro()
{
    register int c;
    char **macro = &(input_stream[n_input_stream-1].input_type.macro);

    if ((c = *(*macro)++) == '\0') {
	return EOF;
    } else {
	return c;
    }
}



/* read_string -- read text characters up to a specified terminator */
int read_string(word, terminator)
char *word;
int terminator;		/* character to terminate on */
{
    register int i = 0, c;

    while ((c = GETC()) != EOF && (c != terminator))
	word[i++] = c;
    word[i] = '\0';
    return i;
}


/* read_number -- read digit characters up to the first nondigit */
int read_number(word, first_char)
char *word;
int first_char;
{
    register int i = 0, c;

    word[i++] = first_char;
    while ((c = GETC()) != EOF && isxdigit(c))
	word[i++] = c;
    word[i] = '\0';
    UNGETC(c);
    return i;
}


/* read_token -- read alphabetics until first nonalphabetic */
int read_token(word, first_char)
register char *word;
register int first_char;
{
    register int i = 0, c;

    word[i++] = first_char;
    while ((c = GETC()) != EOF && 
	(isalpha(c) || isdigit(c) || c == '_' || c == '%' || c == '$'))
	    word[i++] = c;
    word[i] = '\0';
    UNGETC(c);
    return i;
}


/* parse_require -- parse the REQUIRE mechanism for including other files */
parse_require()
{
    int c;
    char filename[256], *pfilename;

    if ((c = yylex()) != T_NAME && (c != T_STRING)) {
	yyerror("REQUIRE file name missing");
    }

    (void) strcpy(filename, yytext);
    if (strchr(filename, '.') == NIL(char)) {
	(void) strcat(filename, ".nrq");
    }
    if ((pfilename = strrchr(filename, ']')) == NIL(char)) {
	pfilename = filename;
    } else {
	pfilename += 1;		/* skip past the bracket */
    }
    if ((c = yylex()) != T_NEXT) {
	yyerror("REQUIRE syntax error (missing semicolon)");
    }
    if (yyopenfile(pfilename, /*use_stdin*/ 0, /*is_a_macro*/ 0) < 0) {
	yyerror("REQUIRE file not found");
    }
}


/* parse_macro -- parse macro definitions */
parse_macro()
{
    int i, c, macro_len, nbuf;
    char buffer[32], *macro_name;
    symbol *psym;

    /* read the macro name */
    c = yylex();
    macro_name = util_strsav(yytext);
    if ((c = yylex()) != T_EQ) {
	yyerror("MACRO Syntax (warning: parameters not supported)");
    }

    /* find space in the symbol table for the macro */
    psym = symtab_lookup(reserved_syms, macro_name, /* create */ 1);
    FREE(macro_name);

    /* allow redefinition of a macro */
    if (psym->type == T_IS_MACRO) {
	FREE(psym->u.macro.value);
    }

    psym->type = T_IS_MACRO;
    psym->u.macro.starting_line = *yyline;

    /* read the macro text */
    psym->u.macro.value = ALLOC(char, 32);
    psym->u.macro.value[0] = '\0';
    macro_len = 0;
    nbuf = 0;

    while((c = GETC()) != EOF) {
	if (c == '\n') {
	    (*yyline)++;
	}
	/* Check for the end of the macro */
	if (c == '$') {
	    yylength = read_token(yytext, c);
	    if (upper_strcmp(yytext, "$ENDMACRO") == 0) {
		break;
	    } else {
		for(i = 0; i < yylength; i++) {
		    save_char(buffer, &nbuf, &(psym->u.macro.value), &macro_len, 
			yytext[i]);
		}
	    }
	} else {
	    save_char(buffer, &nbuf, &(psym->u.macro.value), &macro_len, c);
	}
    }

    save_char(buffer, &nbuf, &(psym->u.macro.value), &macro_len, EOF);

    if ((c == EOF) || (c = yylex()) != T_NEXT) {
	yyerror("MACRO Syntax (missing semicolon)");
    }
}


save_char(buffer, nbuf, macro_def, macro_len, c)
char *buffer;
int *nbuf;		/* will be changed */
char **macro_def;
int *macro_len;
int c;
{
    if (c != EOF) {
	buffer[(*nbuf)++] = c;
    }
    if (c == EOF || *nbuf >= 31) {
	*macro_len += *nbuf;
	*macro_def = REALLOC(char, *macro_def, *macro_len + *nbuf + 1);
	buffer[*nbuf] = '\0';
	(void) strcat(*macro_def, buffer);
	*nbuf = 0;
    }
}

extern int yylex_dump;

yylex()
{
    int c = yylex_stub();
    if (yylex_dump)
	(void) printf("YYLEX returns %s\n", yytext);
    return c;
}


yylex_stub()
{
    register int c;
    register symbol *psym;

start:
    c = GETC();

    /* Check for end of file, keep returning EOF until yacc satisfied */
    if (c == EOF) {
	yyclosefile();
	if (n_input_stream == 0) {
	    return EOF;
	} else {
	    goto start;
	}

    /* ignore whitespace for the most part */
    } else if (isspace(c)) {
	if (c == '\n') (*yyline)++;
	goto start;

    /* Check for a comment (comments are until end of line) */
    } else if (c == '!') {
	while ((c = GETC()) != EOF && (c != '\n'))
	    ;
	(*yyline)++;
	goto start;

    /* check for quoted name */
    } else if (c == '"') {
	yylength = read_string(yytext, '"');
	return T_NAME;

    /* check for string */
    } else if (c == '\'') {
	yylength = read_string(yytext, '\'');
	return T_STRING;

    /* check for special character */
    } else if (magic_char_lookup[c] != 0) {
	yytext[0] = c;
	yytext[1] = '\0';
	yylength = 1;
	return magic_char_lookup[c];

    /* check for number string */
    } else if (isdigit(c)) {
	yylength = read_number(yytext, c);
	return T_NUMBER;

    /* otherwise must be a reserved word or identifier (includes %PRINT) */
    } else if (isalpha(c) || (c == '%') || (c == '$')) {
	yylength = read_token(yytext, c);

	/* Check the reserved words symbol table for keywords */
	psym = symtab_lookup(reserved_syms, yytext, 0);
	if (psym == NIL(symbol)) {
	    return T_NAME;
	} else if (psym->type == T_IS_MACRO) {
	    (void) yyopenfile(yytext, /*use_stdin*/0, /* is_a_macro */ 1);
	    input_stream[n_input_stream-1].input_type.macro = psym->u.macro.value;
	    input_stream[n_input_stream-1].line = psym->u.macro.starting_line;
	    goto start;
	} else if (psym->type == T_MACRO) {
	    parse_macro();
	    goto start;
	} else if (psym->type == T_REQUIRE) {
	    parse_require();
	    goto start;
	} else {
	    return psym->type;
	}

    /* else its something we don't quite understand: let YACC complain */
    } else {
	return T_INVALID;
    }
}



int yyopenfile(s, use_stdin, is_a_macro)
char *s;			/* name of the file */
int use_stdin;			/* if 1, use standard input */
int is_a_macro;			/* if 1, we are opening a macro */
{
    FILE *fp;

    if (n_input_stream >= MAXSTACK) {
	yyerror("Input file/Macro Nesting too deep");
    }

    input_stream[n_input_stream].filename = util_strsav(s);
    input_stream[n_input_stream].line = 0;
    input_stream[n_input_stream].is_macro = is_a_macro;
    input_stream[n_input_stream].npushback = 0;

    /* actually open the file if its a file */
    if (! input_stream[n_input_stream].is_macro) {
	if (use_stdin) {
	   fp = stdin;
	} else {
	    if ((fp = fopen(s, "r")) == NULL) {
		(void) fprintf(stderr, "bdsyn: Unable to open \"%s\"\n", s);
		return -1;
	    }
	    VOVinput( VOV_UNIX_FILE, s );
	}
	input_stream[n_input_stream].input_type.fp = fp;
    }

    /* Setup the pointers to the current input stream */
    if (! input_stream[n_input_stream].is_macro) {
	lexin = input_stream[n_input_stream].input_type.fp;
    }
    yyfile = input_stream[n_input_stream].filename;
    yyline = &(input_stream[n_input_stream].line);
    is_macro = &(input_stream[n_input_stream].is_macro);
    npushback = &(input_stream[n_input_stream].npushback);

    n_input_stream++;
    *yyline = 1;
    return 0;
}


yyclosefile()
{
    /* pop one level on the stack */
    if (n_input_stream == 0)
	return;
    n_input_stream--;

    /* close the file it was a file, otherwise do nothing */
    if (! input_stream[n_input_stream].is_macro)
	if (input_stream[n_input_stream].input_type.fp != stdin)
	    (void) fclose(input_stream[n_input_stream].input_type.fp);

    /* Reinitialize the globals which point to the current entry */
    if (n_input_stream > 0) {
	lexin = input_stream[n_input_stream-1].input_type.fp;
	yyfile = input_stream[n_input_stream-1].filename;
	yyline = &(input_stream[n_input_stream-1].line);
	is_macro = &(input_stream[n_input_stream-1].is_macro);
	npushback = &(input_stream[n_input_stream-1].npushback);
    }

    /* note that yyfile, yyline, and is_macro point to the (still valid)
    last filename and line number if n_input_stream == 0
    (in case an error is detected past eof) */
}


yyreparse(string)
char *string;
{
    (void) yyopenfile("<re-parse>", /*use_stdin*/0, /* is_a_macro */ 1);
    input_stream[n_input_stream-1].input_type.macro = string;
    (void) yyparse();
}
