#include "mis.h"
#include "atpg_int.h"

static void print_pattern();
static void report_error();
static void simulate_network_32();
static int  simulate_network_32_rec();
static int  simulate_node_32();

/* minor hack: random generates numbers from 0 to 2^31 only
 */
void gen_random_pattern(n, pattern)
int n;
int *pattern;
{
    int i;

    for (i = 0; i < n; i++) {
        pattern[i] = (random() << 16) | (random() & 0xffff);
    }
}

static int 
simulate_network_32_rec(network, table, node)
network_t *network;
st_table *table;
node_t *node;
{
    int i;
    char *value;
    int result;
    int n_inputs = node_num_fanin(node);
    int **inputs;

    if (st_lookup(table, (char *) node, &value))
        return (int) value;
    assert(node->type != PRIMARY_INPUT);
    inputs = ALLOC(int *, 4);
    for (i = 0; i < 4; i++) {
        inputs[i] = ALLOC(int, n_inputs);
    }
    assert(ONE < 4 && ZERO < 4 && TWO < 4);
    for (i = 0; i < n_inputs; i++) {
        node_t *input = node_get_fanin(node, i);
        inputs[ONE][i] = simulate_network_32_rec(network, table, input);
        inputs[ZERO][i] = ~ inputs[ONE][i];
        inputs[TWO][i] = 0xffffffff;
    }
    result = simulate_node_32(node, inputs);
    st_insert(table, (char *) node, (char *) result);
    for (i = 0; i < 3; i++)
        FREE(inputs[i]);
    return result;
}

static int 
simulate_node_32(node, inputs)
node_t *node;
int **inputs;
{
    register int i, j;
    node_cube_t cube;
    int num_cubes = node_num_cube(node);
    int and_result;
    int  result = 0;

    for (i = 0; i < num_cubes; i++) {
        and_result = 0xffffffff;
        cube = node_get_cube(node, i);
	for (j = 0; j < node_num_fanin(node); j ++) {
            int literal = node_get_literal(cube, j);
            and_result &= inputs[literal][j];
        }
        result |= and_result;
    }
    return result;
}


void 
fsim_network(network, table, nodes, pi_values, po_values)
network_t  *network;
st_table   *table;
sim_node_t *nodes;
int        *pi_values;
int        *po_values;
{
    node_t *pi, *po;
    lsGen  gen;
    int    i, uid;
  
    i = 0;
    foreach_primary_input(network, gen, pi) {
        if (st_lookup(table, (char *) pi, (char **) &uid)) {
            nodes[uid].value = pi_values[i];
            i++;
        }
    }

    for (i = 0; i < st_count(table); i++) {
        (*(nodes[i].eval))(nodes, i);
    }

    i = 0;
    foreach_primary_output(network, gen, po) {
        assert(st_lookup(table, (char *) po, (char **) &uid));
        po_values[i] = nodes[uid].value;
        i++;
    }
}

/* This routine does parallel fault simulation based on Waicukauski VLSI 1985
 * paper. It only recomputes the values of the transitive fanout of the fault 
 * site.
 * tfo is assumed to contain all the index numbers of the 
 * nodes in the transitive fanout */
void 
fast_fsim_network(network, table, nodes, pi_values, po_values, tfo, refresh_pi)
network_t  *network;
st_table   *table;
sim_node_t *nodes;
int        *pi_values;
int        *po_values;
array_t    *tfo;
int        refresh_pi;
{
    node_t *pi, *po;
    lsGen  gen;
    int    i, x, uid;
  
    if (refresh_pi == 1) {
        i = 0;
        foreach_primary_input(network, gen, pi) {
            assert(st_lookup(table, (char *) pi, (char **) &uid));
            nodes[uid].value = pi_values[i];
            i++;
        }
    }

    for (i = 0; i < array_n(tfo); i++) {
        x = array_fetch(int, tfo, i);
        if (x == 1) {
            (*(nodes[i].eval))(nodes, i);
        }
    }

    i = 0;
    foreach_primary_output(network, gen, po) {
        assert(st_lookup(table, (char *) po, (char **) &uid));
        po_values[i] = nodes[uid].value;
        i++;
    }
}

void 
recursive_tfo(current_node, fanout_list, tfo)
int     current_node;
array_t *fanout_list;
array_t *tfo;
{
    int i, j;
    array_t *list;

    array_insert(int, tfo, current_node, 1);
    list = array_fetch(array_t *, fanout_list, current_node);
    for ( i = 0; i < array_n(list); i++) {
        j = array_fetch(int, list, i);
        if ((array_fetch(int, tfo, j)) == 0) {
            recursive_tfo(j, fanout_list, tfo);
        }
    }
}

void print_array2(array)
array_t *array;
{
    int i;
    for (i=0; i<array_n(array); i++)
        fprintf(misout, "%d ", array_fetch(int, array, i));
    fprintf(misout, "\n");
}

void print_fanout_list(fanout_list, nodes)
array_t *fanout_list;
sim_node_t *nodes;
{
    int i, j;
    array_t *list;

    for (i=0; i<array_n(fanout_list); i++) {
        fprintf(misout, "%s: ", node_long_name(nodes[i].node));
        fprintf(misout, "%d: ", i);
        list = array_fetch(array_t *, fanout_list, i);
        for (j=0; j < array_n(list); j++) {
            fprintf(misout, "%d ", array_fetch(int, list, j));
        }
        fprintf(misout, "\n");
    }
}
  
void 
clear_tfo(tfo, table)
array_t *tfo;
st_table *table;
{
    int i;

    for (i=0; i < st_count(table); i++) {
        array_insert(int, tfo, i, 0);
    }
}
  
/* This routine takes in the node address of the fault site and then 
 * traverses the transitive fanout of the fault site. */
/* ARGUSED */
void 
create_tfo(nodes, fanout_list, fault_site, tfo)
sim_node_t *nodes;
array_t    *fanout_list;
int        fault_site;
array_t    *tfo;
{
  recursive_tfo(fault_site, fanout_list, tfo);
}

array_t *
create_fanout_list(nodes, table)
sim_node_t *nodes;
st_table   *table;
{
    array_t *fanout_list, *list;
    node_t  *fanout;
    lsGen   gen;
    int     flag, i, j;

    fanout_list = array_alloc(array_t *, st_count(table));
    for (i=0; i < st_count(table); i++) {
        list = array_alloc(int, 0);
        array_insert(array_t *, fanout_list, i, list);
        foreach_fanout(nodes[i].node, gen, fanout) {
            j = 0;
            flag = 0;
            while ((flag == 0) && (j < st_count(table))) {
                if (nodes[j].node == fanout) {
                    array_insert_last(int, list, j);
                    flag =1;
                }
                j++;
            }
        }
    }
    return fanout_list;
}
