#include "mis.h"
#include "atpg_int.h"
#include "satisfy.h"

static void extract_buffer();
static void extract_inverter();
static void extract_complex_node();
static void add_output_constraints();
static int  extract_node();
static int  extract_and();
static int  get_literal_value();
static int  extract_or();

int atpg_variable_count = 1;

static st_table *node_table;

int 
get_variable(node)
node_t *node;
{
  int var;

  if (node_table == NIL(st_table)) {
    node_table = st_init_table(st_numcmp, st_numhash);
    atpg_variable_count = 1;
  }
  if (! st_lookup(node_table, (char *) node, (char **) &var)) {
    var = atpg_variable_count ++;
    st_insert(node_table, (char *) node, (char*) var);
  }
  return var;
}

void 
clear_node_table()
{
    if (node_table != NIL(st_table)) {
	st_free_table(node_table);
	node_table = NIL(st_table);
    }
}

int 
*copy_array(a)
array_t *a;
{
  int i, n = array_n(a);
  int *result = ALLOC(int, n);

  for (i = 0; i < n; i++) {
    result[i] = array_fetch(int, a, i);
  }
  return result;
}

int 
extract(array, network)
array_t *array;
network_t *network;
{
    int i, break_flag = 0;
    array_t *nodevec;
    node_t *node;

    nodevec = network_dfs_complete(network);
    for(i = 0; i < array_n(nodevec); i++) {
        node = array_fetch(node_t *, nodevec, i);
        switch (extract_node(array,node)) {
            /* the return codes for extract_node are described in the comments
             * preceding the procedure */
            case (-1):
                return 0;
            case 0:
                /* a complex node was encountered */
                decomp_tech_network(network, 100, 100);
                extract(array, network); /* start all over */
                break_flag = 1;
                break;
            default:
                break;
        }
        if (break_flag)
            break;
    }
    array_free(nodevec);
    return 1;
}

/* I have set up three different kinds of return codes for this procedure
 * a -1 indicates that a constant node has been encountered and control should
 * be returned to the command line, a 0 indicates that a complex node has been
 * encountered (which means that the calling procedure should tech_decomp to
 * break it down into and/or gates), finally a 1 indicates that extract completed
 * successfully */
static 
int extract_node(array, node)
array_t *array;
node_t *node;
{

    switch (node_function(node)) {
        case NODE_UNDEFINED:
            fail("Undefined nodes should have been removed");
            break;
        case NODE_0:
        case NODE_1:
            fprintf(miserr,"Constant nodes should have been removed from the network.  Aborting ATPG.\n");
            return (-1);
        case NODE_PI:
            break;
        case NODE_PO:
        case NODE_BUF:
            extract_buffer(array, node);
            break;
        case NODE_INV:
            extract_inverter(array, node);
            break;
        case NODE_AND:
            extract_and2(array, node);
            break;
        case NODE_OR:
            extract_or2(array,node);
            break;
        case NODE_COMPLEX:
            fprintf(misout,"Complex node encountered.  Breaking down and proceeding with atpg.\n\n");
            return 0;
            break;
    }
    return 1;
}

static void 
extract_buffer(array, node)
array_t *array;
node_t *node;
{
  int var_in = get_variable(node_get_fanin(node, 0));
  int var_out = get_variable(node);
  add_clause_2(array, var_in, - var_out);
  add_clause_2(array, - var_in, var_out);
}

static void 
extract_inverter(array, node)
array_t *array;
node_t *node;
{
  int var_in = get_variable(node_get_fanin(node, 0));
  int var_out = get_variable(node);
  add_clause_2(array, var_in, var_out);
  add_clause_2(array, - var_in, - var_out);
}


void 
extract_or2(array, node)
array_t *array;
node_t *node;
{
    int                n_cubes        = node_num_cube(node);
    int                *input_set    = ALLOC(int, n_cubes);
    node_literal_t    literal;
    node_t            *fanin;
    node_cube_t        cube;
    int                var_out, var_in, j, i;
    
    cube = node_get_cube(node, 0);
    var_out = get_variable(node);

    for (i=0; i < n_cubes; i++) {
        cube = node_get_cube(node,i);
        j = 0;
        while (node_get_literal(cube,j) == TWO)
            j++;
        literal = node_get_literal(cube,j);
        fanin = node_get_fanin(node,j);
        switch(literal) {
            case ZERO:
                var_in =  get_variable(fanin);
                add_clause_2(array,  var_out,  var_in);
                input_set[j] = - var_in;
                break;
            case ONE: 
                var_in =  get_variable(fanin);
                add_clause_2(array,  var_out, - var_in);
                input_set[j] =  var_in;
                break;
            case TWO:
                input_set[j] = 0;
                break;
        }
    }
    array_insert_last(int, array, - var_out);
    for (i=0; i < n_cubes; i++)
        array_insert_last(int, array, input_set[i]);
    array_insert_last(int, array, 0);
    FREE(input_set); /*change*/
}

void 
extract_and2(array, node)
array_t *array;
node_t *node;
{
    int n_inputs = node_num_fanin(node);
    int *input_set = ALLOC(int, n_inputs);
    node_literal_t literal;
    node_t *fanin;
    node_cube_t cube;
    int var_out, var_in,j,i;
    
    
    cube = node_get_cube(node, 0);
    var_out = get_variable(node);
    foreach_fanin(node,j,fanin) {
    literal = node_get_literal(cube,j);
        switch(literal) {
        case ZERO:
        var_in =  get_variable(fanin);
        add_clause_2(array, - var_out, - var_in);
        input_set[j] = var_in;
        break;
        case ONE: 
        var_in =  get_variable(fanin);
        add_clause_2(array, - var_out, var_in);
        input_set[j] = - var_in;
        break;
        case TWO:
        input_set[j] = 0;
         break;
    }
    }
    array_insert_last(int, array, var_out);
    for (i=0; i < n_inputs; i++) {
    array_insert_last(int, array, input_set[i]);
    }
    array_insert_last(int, array, 0);
    FREE(input_set);
}

void 
add_clause_2(array, var1, var2)
array_t *array;
int var1;
int var2;
{
  array_insert_last(int, array, var1);
  array_insert_last(int, array, var2);
  array_insert_last(int, array, 0);
}

void 
add_clause_3(array, var1, var2, var3)
array_t *array;
int var1;
int var2;
int var3;
{
  array_insert_last(int, array, var1);
  array_insert_last(int, array, var2);
  array_insert_last(int, array, var3);
  array_insert_last(int, array, 0);
}

void 
dump_node_bindings(network)
network_t *network;
{
  int i, var;
  array_t *nodevec;
  node_t *node;

  nodevec = network_dfs_complete(network);
  for(i = 0; i < array_n(nodevec); i++) {
    node = array_fetch(node_t *, nodevec, i);
    var = get_variable(node);
    fprintf(misout, "%8 s -> %4 d\n", node->name, var);
  }
  array_free(nodevec);
}

void 
dump_clauses(array)
array_t *array;
{
    int i, c = 0, start = 1, n = array_n(array);

    for (i = 0; i < n; i++) {
        int value = array_fetch(int, array, i);
        if (start) {
            fprintf(misout, "%4 d: ", c);
            start = 0;
        }
        if (value == 0) {
            c++;
            start = 1;
            fprintf(misout, "\n");
        }
        else
            fprintf(misout, "%4 d ", value);
    }
}

/* we isolate here the primary inputs, in their network order */
int 
*extract_imp_vars(network, n_vars)
network_t *network;
unsigned  *n_vars;
{
    int        i;
    lsGen    gen;
    node_t    *input;
    int        n = network_num_pi(network);
    int        *result= ALLOC(int, n);

    *n_vars = (unsigned) n;
    i = 0;
    foreach_primary_input(network, gen, input)
        result[i++] = get_variable(input);
    return result;
}
