#include "mis.h"
#include "util.h"
#include "atpg.h"
#include "atpg_int.h"
#include "satisfy.h"

extern int  atpg_variable_count;
static void print_values();

init_atpg()
{
    com_add_command("atpg", com_atpg, /* changes_network */ 0);
    com_add_command("red_removal", com_red_removal, /* changes_network */ 1);
}

end_atpg()
{
}

static void
print_usage()
{
    fprintf(misout, "usage: atpg [-acdfilnrtT]\n");
    fprintf(misout, "-a\tdon't use active clauses\n");
    fprintf(misout, "-c\tdon't use critical clauses\n");
    fprintf(misout, "-d\tuse dominators\n");
    fprintf(misout, "-f\tdon't use fault simulation\n");
    fprintf(misout, "-i\tdon't use additional critical clauses\n");
    fprintf(misout, "-l\tRTG give up level\n");
    fprintf(misout, "-n\tdon't use non-local implications\n");
    fprintf(misout, "-r\tdon't use random test generation\n");
    fprintf(misout, "-t\ttrace through test generation\n");
    fprintf(misout, "-T\ttrace through the satisfier\n");
}

/* main routine for doing automatic test pattern generation */
int
com_atpg(network_1, argc, argv)
network_t **network_1;
int       argc;
char      **argv;
{
    network_t  *network;
    sim_node_t *sim_nodes;
    st_table   *table, *dom_table;
    lsGeneric  userData;
    lsHandle   itemHandle;
    lsList     flist, redundant_faults, aborted_faults, tested_faults,
               incr_tested;
    array_t    *master_constraint_clauses = array_alloc(int, 0), 
               *constraint_clauses, *non_local_clauses;
    saf_t      *fnode;
    unsigned   n, n_vars;
    int        level = 10, sat_debug=0, tpg_debug = 0,
	       active_clauses=1, fs = 1, ct, result, rtg=1, 
	       nl = 1, critical=1, imp1=1, dom=0, total_nodes, 
	       c, *values, *marker, *vars, *clauses;

    /* get rid of fanout-free nodes which are not primary output nodes */
    network = network_dup(*network_1);
    network_sweep(network); 
    decomp_tech_network(network, INFINITY, INFINITY); 
    total_nodes = network_num_pi(network) + network_num_internal(network) + 
		  network_num_po(network);
    values = ALLOC(int, network_num_pi(network));

    getopt_reset();
    while ((c = getopt(argc, argv, "l:acdfinrtT")) != EOF) {
        switch (c) {
            case 'a':
                active_clauses = 0;
                break;
            case 'c':
                critical = 0;
                break;
            case 'd':
                dom=1;    
                break;
            case 'f':
                fs = 0;
                break;
            case 'i':
                imp1 = 0;
                break;
            case 'l':
		level = atoi(optarg);
		break;
            case 'n':
                nl = 0;
                break;
            case 'r':
                rtg = 0;
                break;
            case 't':
                tpg_debug = 1;
                break;
            case 'T':
                sat_debug = 1;
                break;
            default:
                print_usage();
                return 1;
        }
    }

    redundant_faults = lsCreate();
    aborted_faults = lsCreate();
    table = st_init_table(st_ptrcmp, st_ptrhash);
    sim_nodes = sim_extract_network(network, table);
    flist = gen_fault_list(network, table);
    /*
    if (tpg_debug) {
	print_fault_list(flist);
    }
    */
        
    if (rtg == 1) {
        tested_faults = rtp_generate(network, table, sim_nodes, flist, 
				     level, tpg_debug );
    }
    else {
	tested_faults = lsCreate();
    }

    atpg_variable_count = 1;
    if (extract(master_constraint_clauses, network) != 1) {
        array_free(master_constraint_clauses);
        return 1;
    }

    /* Do Non-local implications */
    if (nl == 1) {
        non_local_clauses = array_alloc(int, 0);
        nl_impl(network, non_local_clauses);
    }

    /* Add dominators */
    if (dom == 1) {
        dom_table = st_init_table(st_ptrcmp, st_ptrhash);
        dominators(network, dom_table);
    }
    /* Dump clauses for debugging purposes */
    /*
    if (tpg_debug) {
        dump_node_bindings(network);
        dump_clauses(master_constraint_clauses);
    }
    */

    /* test pattern generation loop */
    while (lsFirstItem(flist, (lsGeneric*)&fnode, &itemHandle) == LS_OK) {
	if (tpg_debug) {
	    print_fault(fnode);
	}
        constraint_clauses = array_dup(master_constraint_clauses);
        fault_extract(network, constraint_clauses, fnode, active_clauses,
		      &marker, critical, imp1);
        if (nl == 1) {
            array_append(constraint_clauses, non_local_clauses);
            nl_bad_clauses(constraint_clauses, non_local_clauses, marker);
        }
        if (dom == 1)
            app_all_doms(constraint_clauses, dom_table, marker, total_nodes);
	/*
        if (tpg_debug) {
            dump_node_bindings(network);
            dump_clauses(constraint_clauses);
        }
	*/

        FREE(marker);

        vars = extract_imp_vars(network, &n_vars);
        n = array_n(constraint_clauses);
        clauses = copy_array(constraint_clauses);
	/*
	if (sat_debug) {
	    dump_clauses(constraint_clauses);
	}
	*/
        array_free(constraint_clauses);
        result = satisfy(clauses, n, vars, values, n_vars, sat_debug);
        FREE(clauses);
        FREE(vars);
        switch (result) {
            case 0: /* redundant fault */
            case 2: /* aborted fault */
		if (result == 0) {
		    if (tpg_debug) {
			fprintf(misout, "Redundant fault encountered\n");
		    }
		    lsNewEnd(redundant_faults, (lsGeneric)fnode, 0);
		}
		else {
		    if (tpg_debug) {
			fprintf(misout, "Aborted fault\n");
		    }
		    lsNewEnd(aborted_faults, (lsGeneric)fnode, 0);
		}
		lsRemoveItem(itemHandle, &userData); 
		break;

            case 1:
		/* print the necessary input values
		 * for the stuck-at fault currently being tested */
		if (tpg_debug) {
		    fprintf(misout, "Test vector: ");
		    print_values(network, values);
		}
		if (fs == 1) {
		    for (ct = 0; ct < network_num_pi(network); ct++) {
			if (values[ct] == 1)
			values[ct] = ~0;
		    }
		    incr_tested = fsim_vector(network, table, sim_nodes, 
					      flist, values);
		    /* add these detected faults to tested list */
		    concat_lists(tested_faults, incr_tested);
		}
		else {
		    lsNewEnd(tested_faults, (lsGeneric)fnode, 0);
		    lsRemoveItem(itemHandle, &userData); 
		}
		break;
        }
    }
    
    fprintf(misout, "%d total faults\n", lsLength(tested_faults) + 
	    lsLength(redundant_faults) + lsLength(aborted_faults));
    fprintf(misout, "%d redundant faults\n", lsLength(redundant_faults));
    fprintf(misout, "%d aborted faults\n", lsLength(aborted_faults));

    sim_free_nodes(sim_nodes, table);
    st_free_table(table);
    array_free(master_constraint_clauses);    
    if (nl == 1) {
	array_free(non_local_clauses);
    }
    lsDestroy(redundant_faults, free);
    lsDestroy(aborted_faults, free);
    lsDestroy(tested_faults, free_pattern);
    lsDestroy(flist, free);
    clear_node_table();
    FREE(values);

    return 0; 
}

void
free_pattern(f)
saf_t *f;
{
    FREE(f->pattern);
    FREE(f);
}

static void 
print_values(network, values)
network_t *network;
int *values;
{
    int ct;

    for (ct = 0; ct < network_num_pi(network); ct++)
        fprintf(misout, "%d ", values[ct]);
    fprintf(misout, "\n");
}

/* append elements of l2 to l1; l2 is freed */
void
concat_lists(l1, l2)
lsList l1;
lsList l2;
{
    lsGeneric data;

    while (lsDelBegin(l2, &data) != LS_NOMORE) {
	lsNewEnd(l1, data, 0);
    }
    lsDestroy(l2, free);
}
