#include "mis.h"
#include "atpg_int.h"

static lsList order_flist();
static saf_t  *new_fault();
static void   fanout_free_collapse();
static bool   fanout_free_fanin();
static bool   input_collapse();
static int    collapse_fault_list();
static int    gate_collapse();

/* Given a network, gen_fault_list assumes a worst case situation and generates
 * two s-a-faults for each fanin and fanout of every node within the network 
 * (one for the sa-0 case and one for the sa-1 case) it then calls 
 * collapse_fault_list to attempt to minimize the number of generated faults, 
 * finally returning all the faults in a generic list 
 */


/* The fault collapsing techniques used in this project are the results of 
 * Breuer and Friedman. First, only n+1 faults need to be considered for an 
 * n-input gate. Second, any setof tests which detects all single stuck faults
 * on all single stuck faults on all primary inputs and all branches of 
 * fanout points detects all single faults.
 */
lsList 
gen_fault_list(network, table)
network_t *network;
st_table  *table;
{
    node_t  *node, *fanin;
    saf_t   *fnode;
    lsList  fault_list, fc;
    lsGen   gen;
    int     i;

    fault_list = lsCreate();
    /* First, all wires into the fault list */
    foreach_node(network, gen, node) {
	/* faults on node inputs */
        foreach_fanin(node, i, fanin) {
	    fnode = new_fault(table, node, fanin, 0);
            lsNewEnd(fault_list, (lsGeneric)fnode, 0);
	    fnode = new_fault(table, node, fanin, 1);
            lsNewEnd(fault_list, (lsGeneric)fnode, 0);
        }
	/* faults on node output */
	fnode = new_fault(table, node, NULL, 0);
	lsNewEnd(fault_list, (lsGeneric)fnode, 0);
	fnode = new_fault(table, node, NULL, 1);
	lsNewEnd(fault_list, (lsGeneric)fnode, 0);
    }
    fc = order_flist(fault_list);
    collapse_fault_list(fc);
    return fc;
}

static saf_t *
new_fault(table, node, fanin, value)
st_table *table;
node_t   *node;
node_t   *fanin;
int      value;
{
    saf_t *f;
    int   uid;

    f = ALLOC(saf_t, 1);
    f->pattern = 0;
    f->redundant = 0;
    f->covered = 0;
    f->bit = -1;
    f->node = node;
    st_lookup(table, (char *) node, (char **) &uid);
    f->node_id = uid;
    f->value = value;
    f->fanin = fanin;
    if (fanin == NULL) {
	f->input = -1;
    }
    else {
	f->input = node_get_fanin_index(node, fanin);
    }
    return f;
}

/* creates an ordered fault list such that the faults nearest
 * to the primary outputs of the circuits are in the beginning of the list
 * and the faults near the primary inputs are in the end of the list.
 * faults structures are unchanged 
 * f is destroyed */
static lsList 
order_flist(f)
lsList f;
{
    lsHandle  handle1, handle2;
    lsList    flist;
    lsGen     gen1, gen2;
    saf_t     *crec, *key;
    int       flag;

    flist = lsCreate();
    gen1 = lsStart(f);
    while (lsNext(gen1, (lsGeneric *) &key, &handle1) == LS_OK) {
        gen2 = lsStart(flist);
        flag = 0;
        while ((flag == 0) && 
	    (lsNext(gen2, (lsGeneric *) &crec, &handle2) == LS_OK)){
            if (crec->node_id >= key->node_id) {
                flag = 1;
                lsInBefore(gen2, (lsGeneric) key, &handle2);
            }
        }
        lsFinish(gen2);
        if (flag == 0) {
            lsNewEnd(flist, (lsGeneric)key, &handle2);
        }
    }
    lsFinish(gen1);
    lsDestroy(f, 0);
    return flist;
}

void 
print_fault_list(fault_list)
lsList fault_list;
{
    lsGen gen;
    saf_t *fnode;

    gen = lsStart(fault_list);
    while (lsNext(gen, (lsGeneric*)&fnode, 0) == LS_OK) {
        printf("Node name: %s\n", node_long_name(fnode->node));
        if (fnode->fanin != NULL) {
            printf("Source node: %s\n", node_long_name(fnode->fanin));
        } 
        else {
            printf("Source node: NULL\n");  
        }
        printf("Stuck Type: %d\n", fnode->value);
        printf("\n");
    }
    lsFinish(gen);
}

/* taking the fault_list generated in gen_fault_list attempt to minimize the
 * number of faults.  
 */
static int 
collapse_fault_list(fault_list)
lsList fault_list;
{
    fprintf(misout, "%d uncollapsed faults enumerated\n", lsLength(fault_list));

    /* Remove all faults that are not part of signals with fanout greater than
     * one.
     */
    fanout_free_collapse(fault_list);
    return gate_collapse(fault_list); 
}

static void 
fanout_free_collapse(fault_list)
lsList fault_list;
{
    lsGeneric userData;
    lsHandle  itemHandle;
    lsGen     gen;
    saf_t     *fnode;

    gen = lsStart(fault_list);
    while (lsNext(gen, (lsGeneric*)&fnode, &itemHandle) == LS_OK) {
        if (node_function(fnode->node) != NODE_PI) {
            if (fnode->fanin == NULL) {
		/* If a fault site is an output of a node and the node is 
		 * not a primary input, remove it from the fault list.
		 */
                FREE(fnode);
                lsRemoveItem(itemHandle, &userData);
            }
            else {
		/* If the input of a node comes from a node with a fanout 
		 * equal to one, remove it from the fault list.
		 */
                if (node_num_fanout(fnode->fanin) == 1) {
                    FREE(fnode);
                    lsRemoveItem(itemHandle, &userData);
                }
	    }
	}
    }
    lsFinish(gen);
}

/* return true if all the inputs of the node have source nodes with fanout 
 * equal to one 
 */
static bool
fanout_free_fanin(node)
node_t *node;
{
    node_t *fanin;
    int    i;

    foreach_fanin(node, i, fanin) {
        if (node_num_fanout(fanin) > 1) {
	    return FALSE;
        }
    }
    return TRUE;
}

/* Return true if the fanout node has inputs with a lower index compared to 
 * 'node' and has a source node with fanout  equal to one.
 * Also, return true if the fanout node has inputs with source nodes with inputs
 * equal to one and it is not a primary input.
 */
static int 
input_collapse(node, fanout)
node_t *node, *fanout;
{
    node_t *fanin;
    int    i;

    foreach_fanin(fanout, i, fanin) {
        if (node_num_fanout(fanin) == 1) {
            if ((node_get_fanin_index(fanout, node) > i) || 
		(node_function(fanin) != NODE_PI)) {
		return TRUE;
            }
        } 
    }
    return FALSE;
}

static int 
gate_collapse(fault_list)
lsList fault_list;
{
    lsGeneric userData;
    lsHandle  itemHandle;
    lsGen     gen;
    node_t    *fanout;
    saf_t     *fnode;

    gen = lsStart(fault_list);
    while (lsNext(gen, (lsGeneric*)&fnode, &itemHandle) == LS_OK) {
        switch(node_function(fnode->node)) {
        case NODE_AND:
            if (fnode->fanin != NULL) {
                if ((fnode->value == 0) && 
		    (node_input_phase(fnode->node, fnode->fanin) == POS_UNATE)) {
		    /* remove fault site if it is an input to an AND gate, 
		     * stuck-at-0, and is not the first input of the AND 
		     * gate */
                    if (node_get_fanin_index(fnode->node, fnode->fanin) != 0) {
                        FREE(fnode);
                        lsRemoveItem(itemHandle, &userData);
                    } 
                    else {
			/* If it is the first input, and all the inputs 
			 * have source nodes with fanout equal to 
			 * one, also remove the fault site */  
                        if (fanout_free_fanin(fnode->node)) {
                            FREE(fnode);
                            lsRemoveItem(itemHandle, &userData);
                        }
                    }
                } 
                else {
		    /* Take into account negative unate inputs */
                    if ((fnode->value == 1) && 
			(node_input_phase(fnode->node, fnode->fanin) == NEG_UNATE)) {
                        if (node_get_fanin_index(fnode->node, fnode->fanin) != 0) {
                            FREE(fnode);
                            lsRemoveItem(itemHandle, &userData);
                        } 
                        else {
                            if (fanout_free_fanin(fnode->node)) {
                                FREE(fnode);
                                lsRemoveItem(itemHandle, &userData);
                            }
                        }
                    }
                }
            }
            break;
        case NODE_OR:
	    /* Basically the same thing as AND gates with the stuck-at 
	     * values inverted */
            if (fnode->fanin != NULL) {
                if ((fnode->value == 1) && 
		    (node_input_phase(fnode->node, fnode->fanin) == POS_UNATE)) {
                    if (node_get_fanin_index(fnode->node, fnode->fanin) != 0) {
                        FREE(fnode);    
                        lsRemoveItem(itemHandle, &userData);
                    } 
                    else {
                        if (fanout_free_fanin(fnode->node)) {
                            FREE(fnode);
                            lsRemoveItem(itemHandle, &userData);
                        }
                    }
                } 
                else {
                    if ((fnode->value == 0) && 
			(node_input_phase(fnode->node, fnode->fanin) == NEG_UNATE)) {
                        if (node_get_fanin_index(fnode->node, fnode->fanin) != 0) {
                            FREE(fnode);
                            lsRemoveItem(itemHandle, &userData);
                        } 
                        else {
                            if (fanout_free_fanin(fnode->node)) {
                                FREE(fnode);
                                lsRemoveItem(itemHandle, &userData);
                            }
                        }
                    }
                }
            }
            break;
        case NODE_PI:
            if (node_num_fanout(fnode->node) == 1) {
                fanout = node_get_fanout(fnode->node, 0);
                if (node_function(fanout) == NODE_AND) {
                    if ((fnode->value == 0) && 
			(node_input_phase(fanout, fnode->node) == POS_UNATE)) {

			/* Remove fault site if it is a primary input, 
			 * fanout=1, fanout node = AND gate, stuck-at-0, 
			 * pos_unate destination, and input collapse returns 
			 * true */
                        if (input_collapse(fnode->node, fanout)) {
                            FREE(fnode);
                            lsRemoveItem(itemHandle, &userData);
                        }
                    } 
                    else {

			/* Do the same thing for negative unate destinations 
			 * with 1 */
                        if ((fnode->value == 1) && 
			    (node_input_phase(fanout, fnode->node) == NEG_UNATE)) {
                            if (input_collapse(fnode->node, fanout)) {
                                FREE(fnode);
                                lsRemoveItem(itemHandle, &userData);
                            }
                        }
                    }
                } 

		/* Now, handle OR gates */
                if (node_function(fanout) == NODE_OR) {
                    if ((fnode->value == 1) && 
			(node_input_phase(fanout, fnode->node) == POS_UNATE)) {
                        if (input_collapse(fnode->node, fanout)) {
                            FREE(fnode);
                            lsRemoveItem(itemHandle, &userData);
                        }
                    } 
                    else {
                        if ((fnode->value == 0) && 
			    (node_input_phase(fanout, fnode->node) == NEG_UNATE)) {
                            if (input_collapse(fnode->node, fanout)) {
                                FREE(fnode);
                                lsRemoveItem(itemHandle, &userData);
                            }
                        }
                    }
                } 
            }
            break;
        case NODE_PO:
        case NODE_INV:
        case NODE_BUF:
            break;
        case NODE_COMPLEX:
	    fail("can't handle complex nodes as yet");
            break;
        case NODE_UNDEFINED:
	    fail("Undefined node");
            break;
        case NODE_0:
            fail("NODE_0s are not allowed");
            break;
        case NODE_1:
            fail("NODE_1s are not allowed");
            break;
        default:
            fail("Node does not have a function type");
            break;
        }
    }
    lsFinish(gen);
    fprintf(misout, "%d faults after collapse\n", lsLength(fault_list));
    return TRUE;
} 
