#include "mis.h"
#include "atpg_int.h"

static array_t *sim_extract_faults();
static lsList  fast_fault_sim();
static lsList  random_cover();
static void    get_uniq_pattern();
static void    sim_reset_masks();
static void    sim_set_masks();
static int     rec_fast_fsim();
static void    rec_par_fsim();

static int     g_level;
static int     g_verbosity;

lsList
rtp_generate(network, table, sim_nodes, flist, level, rtg_debug)
network_t  *network;
st_table   *table;
sim_node_t *sim_nodes;
lsList     flist;
int        level;
int        rtg_debug;
{
    lsList f_covered;
    double ratio;

    if (lsLength(flist) == 0)
	return lsCreate();

    g_level = level;
    g_verbosity = rtg_debug;
    f_covered = random_cover(network, sim_nodes, table, flist);
    ratio = ((double) lsLength(f_covered)) / 
	    ((double) (lsLength(f_covered) + lsLength(flist)));
    if (g_verbosity) { 
        fprintf(misout, "coverage with random patterns: %2.4f\n", ratio);
        fprintf(misout, "%d faults remain\n", lsLength(flist));
    }
    return f_covered;
}

static lsList 
random_cover(network, sim_nodes, table, flist)
network_t  *network;
sim_node_t *sim_nodes;
st_table   *table;
lsList     flist;
{
    array_t   *fanout_list;
    lsList    fc, lc;
    int       limit_count = 0, ct;

    fanout_list = create_fanout_list(sim_nodes, table);

    fc = lsCreate();
    while (lsLength(flist) != 0 && limit_count < g_level) {
        lc = fast_fault_sim(sim_nodes, table, network, flist, fanout_list); 
        if (lsLength(lc) > 0) {
            limit_count = 0;
        }
        else {
            limit_count ++;
        }
	concat_lists(fc, lc);
        if (g_verbosity) {
            fprintf(misout, "covered: %d\tuncovered: %d\n", lsLength(fc), 
		    lsLength(flist));
	}
    }

    for (ct = 0; ct < array_n(fanout_list); ct++) {
        array_free(array_fetch(array_t *, fanout_list, ct));
    }
    array_free(fanout_list);
    return fc;
}

/* return list of covered faults */
static lsList
fast_fault_sim(sim_nodes, table, network, flist, fanout_list)
sim_node_t *sim_nodes;
st_table   *table;
network_t  *network;
lsList     flist;
array_t    *fanout_list;
{
    lsGeneric userData;
    lsHandle  handle1;
    lsList    local_cov;
    lsGen     gen1, gen2;
    array_t   *tfo;
    saf_t     **faults = ALLOC(saf_t *, WORD_LENGTH);
    saf_t     *prev_fsite;
    char      *key;
    int       refresh_pi = 0, i;
    int       n_pi = network_num_pi(network);
    int       n_po = network_num_po(network);
    int       *pattern = ALLOC(int, n_pi);
    int       *correct_po_val = ALLOC(int, n_po);
    int       *faulty_po_val  = ALLOC(int, n_po);

    prev_fsite = NULL;
    tfo = array_alloc(int, st_count(table));
    gen_random_pattern(n_pi, pattern);
    fsim_network(network, table, sim_nodes, pattern, correct_po_val);

    gen1 = lsStart(flist);
    local_cov = lsCreate();
    gen2 = lsStart(local_cov);
    while (lsNext(gen1, &key, &handle1) == LS_OK) {
        for (i = 0; i < WORD_LENGTH; i++) {
            faults[i] = (saf_t *) key;
        }
        sim_set_masks(sim_nodes, faults);
        clear_tfo(tfo, table);
        if (prev_fsite != NULL) {
            create_tfo(sim_nodes, fanout_list, prev_fsite->node_id, tfo);
        }
        create_tfo(sim_nodes, fanout_list, faults[0]->node_id, tfo);
        if (prev_fsite != NULL) {
            if (node_function(prev_fsite->node) == NODE_PI) {
                refresh_pi = 1;
            }    
            else {
                refresh_pi = 0;
            }
        }
        fast_fsim_network(network, table, sim_nodes, pattern, 
			    faulty_po_val, tfo, refresh_pi);
        prev_fsite = faults[0];
        sim_reset_masks(sim_nodes, faults);
        if (rec_fast_fsim(faults[0], n_po, correct_po_val, faulty_po_val)
	    == 1) {
            lsRemoveItem(handle1, &userData);
	    lsNewEnd(local_cov, userData, 0);
        }
    }
    lsFinish(gen1);
    lsFinish(gen2);

    if (lsLength(local_cov) > 0) {
        get_uniq_pattern(n_pi, pattern, local_cov);
    }

    FREE(pattern);
    FREE(correct_po_val);
    FREE(faulty_po_val);
    FREE(faults);
    array_free(tfo);

    return local_cov;
}

static int 
rec_fast_fsim(fault, n_po, po_val_0, po_val_1)
saf_t *fault;
int n_po;
int *po_val_0;
int *po_val_1;
{
    int i, j;

    for (i = 0; i < n_po; i++) {
        if (po_val_0[i] != po_val_1[i]) {
            int val0 = po_val_0[i];
            int val1 = po_val_1[i];

            for (j = 0; j < WORD_LENGTH; j++) {
                if (EXTRACT_BIT(val0, j) != EXTRACT_BIT(val1, j))
                    break;
            }
	    fault->bit = j;
            return 1;
        }
    }
    return 0;
}

 /* CONVENTION: a NIL pointer to a fault means no fault */
static void sim_set_masks(sim_nodes, flist)
sim_node_t *sim_nodes;
saf_t **flist;
{
    int i;

    for (i = 0; i < WORD_LENGTH - 1; i++) {
        saf_t *fault = flist[i];
        sim_node_t *node;

        if (fault == NIL(saf_t)) continue;
        node = &sim_nodes[fault->node_id];
        assert(node->id == fault->node_id);
        if (fault->input == -1) {
            if (fault->value == 1) {         /* stuck at 1 at output */
                node->or_output_mask |= (1 << i);
            }
            else {                 /* stuck at 0 at output */
                node->and_output_mask &= ~ (1 << i);
            }
        }
        else {
            if (fault->value == 1) {     /* stuck at 1 at some input */
                node->or_input_masks[fault->input] |= (1 << i);
            }
            else {                 /* stuck at 0 at some input */
                node->and_input_masks[fault->input] &= ~ (1 << i);
            }
        }
    }
}

static void sim_reset_masks(sim_nodes, flist)
sim_node_t *sim_nodes;
saf_t **flist;
{
    int i;

    for (i = 0; i < WORD_LENGTH - 1; i++) {
        saf_t *fault = flist[i];
        sim_node_t *node;

        if (fault == NIL(saf_t))
            continue;
        node = &sim_nodes[fault->node_id];
        if (fault->input == -1) {
            node->or_output_mask = ALL_ZEROS;
            node->and_output_mask = ALL_ONES;
        }
        else {
            node->or_input_masks[fault->input] = ALL_ZEROS;
            node->and_input_masks[fault->input] = ALL_ONES;
        }
    }
}

static void 
get_uniq_pattern(n_pi, pattern, local_cov)
int    n_pi;
int    *pattern;
lsList local_cov;
{
    lsHandle  handle;
    lsGen     gen;
    saf_t     *fault;
    int       *used_pattern = ALLOC(int, WORD_LENGTH);
    int       **alloc_pattern = ALLOC(int *, WORD_LENGTH);
    int       i, j;

    for (i = 0; i < WORD_LENGTH; i++) {
        used_pattern[i] = 0;
    }
    gen = lsStart(local_cov);
    while (lsNext(gen, (lsGeneric*)&fault, &handle) == LS_OK) {
	assert((fault->bit != -1));
        used_pattern[fault->bit] = 1;
    }

    for (i = 0; i < WORD_LENGTH; i++) {
        if (used_pattern[i]) {
            alloc_pattern[i] = ALLOC(int, n_pi);
            for (j = 0; j < n_pi; j++) {
                alloc_pattern[i][j] = EXTRACT_BIT(pattern[j], i);
            }
        }
    }
    lsFinish(gen);
    gen = lsStart(local_cov);
    while (lsNext(gen, (lsGeneric *) &fault, &handle) == LS_OK) {
        assert(used_pattern[fault->bit]);
	fault->pattern = ALLOC(int, n_pi);
	i = fault->bit;
	for (j = 0; j < n_pi; j++) {
	    fault->pattern[j] = alloc_pattern[i][j];
	}
    }
    lsFinish(gen);
    for (i = 0; i < WORD_LENGTH; i++) {
        if (used_pattern[i]) {
	    FREE(alloc_pattern[i]);
	}
    }
    FREE(alloc_pattern);
    FREE(used_pattern);
}

void 
print_fault(fault)
saf_t *fault;
{
    if (fault->input == -1) {
        fprintf(misout, "%s[%d] o(%d) c(%d)\n", fault->node->name, 
        fault->node_id, fault->value, fault->covered);
    }
    else {
        fprintf(misout, "%s[%d] i%d(%d) c(%d)\n", fault->node->name, 
        fault->node_id, fault->input, fault->value, fault->covered);
    }
}

/* Do fault simulation, remove all faults from the fault list that are
 * satisfied by the vector "pattern", return the list of faults that were
 * covered by the particular pattern */
lsList
fsim_vector(network, table, sim_nodes, flist, pattern)
network_t *network;
st_table *table;
sim_node_t *sim_nodes;
lsList flist;
int *pattern;
{
    lsGeneric userData;
    lsHandle  handle;
    lsList    fc;
    lsGen     gen;
    saf_t     **f_ptr;
    char      *key;
    int       i, n_sim, n_entries, from;
    int       n_pi = network_num_pi(network);
    int       n_po = network_num_po(network);
    int       *po_val = ALLOC(int, n_po);

    n_entries = lsLength(flist);
    n_entries = (n_entries % (WORD_LENGTH - 1)) ? 
		 (n_entries / (WORD_LENGTH - 1)) + 1 : 
		 (n_entries / (WORD_LENGTH - 1));
    n_entries *= WORD_LENGTH;
    f_ptr = ALLOC(saf_t *, n_entries);
  
    /* copy uncovered faults into f_ptr */
    from = 0;
    gen = lsStart(flist);
    while (lsNext(gen, &key, &handle) == LS_OK) {
        if (from % WORD_LENGTH == (WORD_LENGTH - 1)) {
            f_ptr[from ++] = NIL(saf_t);
        }
        f_ptr[from] = (saf_t *) key;
        assert(f_ptr[from]->covered == 0); 
        from ++;
    }
    lsFinish(gen);
    for (i = from; i < n_entries; i++) {
        f_ptr[i] = NIL(saf_t);
    }
    n_sim = n_entries / WORD_LENGTH;
    for (i = 0; i < n_sim; i++) {
        sim_set_masks(sim_nodes, &f_ptr[i * WORD_LENGTH]);
        fsim_network(network, table, sim_nodes, pattern, po_val);
        sim_reset_masks(sim_nodes, &f_ptr[i * WORD_LENGTH]);
        rec_par_fsim(&f_ptr[i * WORD_LENGTH], pattern, n_pi, n_po, po_val);
    }

    /* delete covered faults */
    fc = lsCreate();
    gen = lsStart(flist);
    while (lsNext(gen, &key, &handle) == LS_OK) {
	if (((saf_t *)key)->covered) {
	    lsRemoveItem(handle, &userData);
	    lsNewEnd(fc, userData, 0);
	}
    }
    lsFinish(gen);

    FREE(f_ptr);
    FREE(po_val);

    return fc;
}

static void 
rec_par_fsim(f_ptr, pattern, n_pi, n_po, po_val)
saf_t  **f_ptr;
int    *pattern;
int    n_pi;
int    n_po;
int    *po_val;
{
    int i, j, k, output, correct_value;

    for (i = 0; i < n_po; i++) {
        output = po_val[i];
        correct_value = EXTRACT_BIT(output, WORD_LENGTH - 1);
        for (j = 0; j < WORD_LENGTH - 1; j++) {
            if (EXTRACT_BIT(output, j) != correct_value) {
		f_ptr[j]->covered = 1;
		f_ptr[j]->pattern = ALLOC(int, n_pi);
		for (k = 0; k < n_pi; k ++) {
		    f_ptr[j]->pattern[k] = pattern[k];
		}
            }
        }
    }
}
