/* SAT package -------------------------------------------------------- *\

  This module is a generic SAT solver.  Implemented fairly efficiently
  but remember there is no way to fool mother nature (i.e. it is still
  NP-complete).

  variable: binary value, either true, false or unassigned.  For
            this package, variables are numbered 1 to N, where
            N is the number of variables.
  literal:  specific assignment to a variable, positive means the
            variable is true, negative means it is false.  For
            example 6 and -6 are the two literals for variable 6.
  clause:   disjunction of 1 or more literals.  A clause with n
            literals is called an n-clause.  For example (6 + -7 + 9)
            is a 3-clause.

  INPUT:
    A CNF expression given in an integer array.  The literals for a
    clause are terminated by a 0.  For example: [-2,3,4,0].  Variables
    are any integer from 1 to MaxInt, and should be consecutive.  The
    vars array is a list of variables to return the values for.

  OUTPUT:
    This routine returns 1 if the form is satisfiable, 0 if not. If
    trace flag is greater than 0, output is also printed to stdout.
    If the result is 1, then the values array holds the assigned values
    which satisfied the equation, 0=false, 1=true, 2=don't care.
    Note the assignment to satisfy the expression need not be unique.

  EXAMPLE:
    int expr[] = {1,2,0, -1,-2,0, -1,-2,3,0};
    int values[2], vars[2] = {1,2};
    int result;
    result = satisfy (expr,10, vars,values,2, 0);
    
    For this example, the result is true, and one possible assignment
    to values is {1,0}.

  ALGORITHM:
    Based on backtracking with heuristic for ordering the variables
    during the backtracking.  Linear space complexity, worst case
    exponential time complexity.

  REFERENCES:
    Cook, S.A., 1971, "The Complexity of Theorem-Proving Procedures,"
        Proc. 3rd Ann. ACM Symp. on Theory of Computing, Association
        for Computing Machinery, NY, 151-158.
    Garey, M., D. Johnson, 1979, "Computers and Intractability,"
        W.H. Freeman & Co., NY.
    Larrabee, T., 1989, "Efficient Generation of Test Patterns Using
        Boolean Difference," IEEE International Test Conference.
    Larrabee, T., 1989, "A Framework for Evaluating Test Pattern
        Generation Strategies," IEEE ICCAD '89.

  IDEAS:
    * ~octtools/demo/misII/amd files for examples
    * only use "what" for debugging?
    * use var_order more often instead of full loops

  HISTORY:
\* -------------------------------------------------------------------- */

#include "port.h"
#include "satisfy_int.h"            /* this includes some other files   */

static int      trace_flag;         /* enables debugging code/messages  */
static unsigned     n_backtrack;
static unsigned     bktrack_lim;
static Boolean      give_up;
static VarPtr          *high_tide;
static VarPtr          *sea_level;

#ifndef NDEBUG
static long     n_assign_bb;        /* total variable assignments made  */
static long     n_assign_nli;
#endif /* NDEBUG */

static SatRec   *g_sat;

static VarPtr    *var_trail;        /* stack: variables to reset value  */
static unsigned   n_vt;             /* max depth of var_trail stack     */
static VarPtr    *VT;               /* one past top processed variable  */
static VarPtr    *IM;               /* one past top assigned variable   */


static VarPtr    *equal_trail;
static unsigned   n_et;
static VarPtr    *ET;

static StackStatus stbase;

static SaveCount *incr_trail;       /* stack: clauses to undo var_count */
static unsigned   n_it;             /* max depth of incr_trail stack    */
static SaveCount *IT;               /* one past top stack element       */


static TempImpl  *impl_trail;
static TempImpl  *free_impl = NULL;


int
sat_solve (clauses,n, vars,values,n_vars, options)
int *clauses;       /*i binary cnf in format described above    */
unsigned n;         /*i length of array clauses                 */
int *vars;          /*i variables to return values for          */
int *values;        /*o satisfying assignment if result is 1    */
unsigned n_vars;    /*i number of entries in vars & values      */
sat_options *options;   /*i control options for sat         */
{
    Boolean    result = false;
    SatRec     theSat;
    unsigned   est_n_var, est_n_lit, est_n_ncl; /* est. problem size */
    int        pass_n;

    trace_flag  = options->trace_flag;
    debug (1, sat_write (clauses,n,options->out_file,options->append_out));
    debug (0, n_assign_bb=n_assign_nli=0);
    estimate_sizes (clauses,n, &est_n_var,&est_n_lit,&est_n_ncl);
    alloc_sat (&theSat, est_n_var,est_n_ncl,est_n_lit);

    n_vt = 2*est_n_var;  VT = IM = var_trail   = Malloc (n_vt, VarPtr);
    n_it = est_n_lit;         IT = incr_trail  = Malloc (n_it, SaveCount);
    n_et = 2*est_n_var;       ET = equal_trail = Malloc (n_et, VarPtr);
    impl_trail = NULL;
    if (var_trail == NULL || incr_trail == NULL) Fail("Out of memory");
    stbase.save_vt = var_trail; stbase.save_et = equal_trail;
    stbase.save_it = incr_trail;    stbase.save_im = impl_trail;

    if (trace_flag >= 1) print_summary (clauses,n);

    if ( initialize (clauses,n, &theSat,options->n_init_nli) ) {
        for (pass_n=0; pass_n < options->n_pass; pass_n++) {
            good_var_order (&theSat,options->passes+pass_n);
            if (trace_flag)
                (void) printf("Strategy %d: %d var\n",
                pass_n+1,theSat.n_free_var);
            result = branch_and_bound (&theSat,options->passes+pass_n);
            if (result != undecided) break;
        }
    }

    if (result == true) {
        check_assignment (clauses,n, &theSat);
        copy_assignment (vars,values,n_vars, &theSat);
    }

    Free (equal_trail);
    Free (incr_trail); 
    Free (var_trail);
    dump_impl ();
    free_sat (&theSat);

    return (int) result;
}


/* ======================== Branch and Bound =========================== */

static Boolean use_dyn_eq;

static Boolean branch_and_bound (sat,pass)
SatRec *sat;
sat_pass *pass;
{
    Boolean result;
    int tide_mark, n;
    int branch_depth;
    VarPtr *vp;
    StackStatus init_status;

    g_sat = sat;
    n_backtrack = 0;
    give_up = false;
    sea_level = sat->var_order;
    high_tide = sat->var_order + sat->n_free_var;
    bktrack_lim = pass->backtrack_lim;
    use_dyn_eq = (Boolean) (pass->use_dyn_eq != 0);
    save_stacks (&init_status);

    result = branch (sat->var_order);

    if (give_up) {
        result = undecided;
        n_backtrack--;
        debug (1, printf("Gave up on solution.\n"));
        if (trace_flag) {
            /* High tide is the topmost var that has been undone. */
            tide_mark = (high_tide-1) - sat->var_order;
            branch_depth = 0;
            for (vp=sat->var_order,n=tide_mark; n--; vp++)
                if ((*vp)->primary) branch_depth++;
            (void) printf("high tide %d/%d (%d branch)\n",
                tide_mark,sat->n_free_var,branch_depth);
        }
        /* Clean up in case we want to retry. */
        undo_assign (&init_status);
    }

    debug (1, printf("b&b: %d, %lu+%lu assigns, %lu backtracks.\n",
            result,n_assign_nli,n_assign_bb,n_backtrack));
    return result;
}

static Boolean branch (next_var)
VarPtr *next_var;
{
    StackStatus  stinfo;
    Variable    *v;

    do {
        v = *(next_var++);
        if (v == NULL) return true;
        assert ((v->next_equal==NULL) == (v->opposite->next_equal==NULL));
    } while (Bound(v));

    save_stacks (&stinfo);
    debug (2, printf("Try [%d] : %x\n",v->what,v->weight));
    v->primary = true;
    if (bound(v) && branch(next_var)) return true;
    if (++n_backtrack > bktrack_lim && bktrack_lim != 0) {
        give_up = true;
        return undecided;
    }
    undo_assign (&stinfo);

    /* This failed. Try the other value. */
    debug (2, printf("Retry [%d] : %x\n",v->opposite->what,v->opposite->weight));
    if (next_var < high_tide) {
        high_tide = next_var;
        if (trace_flag >= 2) printf("high tide=%d\n",
            (high_tide-1)-sea_level);
        if ((high_tide-1)-sea_level == 0)
            high_tide = sea_level + g_sat->n_free_var;
    }
    if (bound(v->opposite) && branch(next_var)) return true;

    return false;       /* Both assignments to v failed. */
}

static void save_stacks (stinfo)
StackStatus *stinfo;
{
    stinfo->save_vt = VT;
    stinfo->save_it = IT;
    stinfo->save_im = impl_trail;
    stinfo->save_et = ET;
}

static void undo_assign (stinfo)
StackStatus *stinfo;
{
    VarPtr    *var_p;
    SaveCount *cl_p;
    Clause     *clause_p;
    TempImpl  *impl_p;
    Variable  *v;

    for (var_p=ET; var_p > stinfo->save_et; ) {
        v = *(--var_p);
        v->next_equal->prev_equal = NULL;
        v->next_equal = NULL;
    }
    ET = var_p;

    assert (IM >= VT);
    for (var_p=IM; var_p > stinfo->save_vt; ) {
        (*--var_p)->value = false;
        (*var_p)->primary = false;
    }
    VT = IM = var_p;

    for (cl_p=IT; cl_p > stinfo->save_it; ) {
        clause_p = (--cl_p)->clause_p;
        assert (cl_p->prev_count > clause_p->var_count);
        clause_p->var_count = cl_p->prev_count;
    }
    IT = cl_p;

    while (impl_trail != stinfo->save_im) {
        impl_p = impl_trail;
        impl_trail = impl_p->next;
        impl_p->var->n_implies--;
        /* Maintain a free list. */
        impl_p->next = free_impl;
        free_impl = impl_p;
    }
}

static Boolean bound (v)
Variable *v;
{
    int        i, j;
    ClausePtr *L;
    Clause    *C;
    Variable  *v2, *var;
    VarPtr    *var_list;

    assert (IM == VT);
    assert (!v->value && !v->opposite->value);
    push_vt (v);

    while (more_impl()) {

        while (more_impl()) {

            var = next_impl();
            debug (3, printf("processing [%d]\n",var->what));
            assert (var->value && !var->opposite->value);
            debug (0,n_assign_bb++);

            for (var_list=var->implies,i=var->n_implies; i--; var_list++) {
                v2 = *var_list;
                debug (5, printf("2-implied [%d]\n", v2->what));
                if (v2->opposite->value) return false;
                if (v2->value) continue;
                push_vt (v2);
            }

            for (L=var->satisfies,j=var->n_satisfies; j--; L++) {
                C = *L;  i = C->var_count;
                if (i == 0) continue;
                debug (5, printf("satisfied %s\n",clause_name(C)));
                push_it (i,C);
                C->var_count = 0;
            } /* end for each satisifed clause */

            for (L=var->opposite->satisfies,j=var->opposite->n_satisfies; j--; L++) {
                C = *L;
                if (C->var_count == 0) continue;
                debug (5, printf("remove literal from %s\n", clause_name(C)));
                if (C->var_count == 1) return false;
                push_it (C->var_count,C);
                if (--C->var_count > 2) continue;
                if (drop_literal (C) == false) return false;
            } /* end for each orphaned clause */

        } /* end while */

        if (!check_equivalence (use_dyn_eq)) return false;

    } /* end while more implications */


    return true;    /* entire assignment was successful */

} /* end bound */

static int temp_equal (var1,var2)
Variable *var1, *var2;
{
    int j;
    int rc = 0;
    Variable *xv, *v1, *v2;
    Variable *xvar;
    Clause *cl;
    ClausePtr *cp1;

    v1 = var1; DEREF(v1);
    v2 = var2; DEREF(v2);
    if (v1 == v2) return 0;

    assert (!v1->value || !v1->opposite->value);
    assert (!v2->value || !v2->opposite->value);
    if (v1 == v2->opposite) return (-1);

    if (v1->value || v2->value || v1->opposite->value || v2->opposite->value) {
        /* An ugly amount of cases here. */
        if (v1->value && v2->opposite->value) return (-1);
        if (v1->opposite->value && v2->value) return (-1);
        if (v1->value && !v2->value) push_vt (v2);
        if (v2->value && !v1->value) push_vt (v1);
        v1 = v1->opposite;  v2 = v2->opposite;
        if (v1->value && !v2->value) push_vt (v2);
        if (v2->value && !v1->value) push_vt (v1);
        return 0;
    }

    rc = 1;
    if (comp_var (v1,v2) > 0) SWAP (Variable *,v1,v2);
    v2->next_equal = v1;
    debug (3, printf("temp equal [%d] = [%d]\n",v2->what,v1->what));

    /* Remove any duplicate references and deduce from them. */
    for (xvar=v1; xvar != NULL; xvar=xvar->prev_equal) {
        for (cp1=xvar->satisfies,j=xvar->n_satisfies; j--; cp1++) {
            (*cp1)->cflag = 1;
        }
    }
    for (xvar=v2; rc >= 0 && xvar != NULL; xvar=xvar->prev_equal) {
        for (cp1=xvar->satisfies,j=xvar->n_satisfies; j--; cp1++) {
            cl = *cp1;
            if (cl->cflag) {
                if (cl->var_count == 0) continue;
                if (cl->var_count == 1) return (-1);
                if (drop_literal (cl) == false) {
                    rc = -1;
                    break;
                }
                debug(3,printf ("new %d-cl\n",cl->var_count));
            }
        }
    }

    for (xvar=v1; xvar != NULL; xvar=xvar->prev_equal) {
        for (cp1=xvar->satisfies,j=xvar->n_satisfies; j--; cp1++) {
            (*cp1)->cflag = 0;
        }
    }

    /* set v2 temporarily equal to v1. */
    /* Sort of a DLL_CAT... */
    for (xv=v1; xv->prev_equal != NULL; ) xv = xv->prev_equal;
    xv->prev_equal = v2;
    v2->next_equal = xv;
    push_et (v2);
    return rc;
}

static Boolean check_equivalence (use_de)
Boolean use_de;
{
    Variable *v;
    int i,x, n_new_loops = 0;

    if (!use_de) return true;

    /* Have deduced some new implications, so search for more loops. */
    for (v=g_sat->variable_array,i=g_sat->n_variable; i--; v++) {
        v->processed = -2;
        if (v->next_equal != NULL) v->processed = -1;
    }

    for (v=g_sat->variable_array,i=g_sat->n_variable; i--; v++) {
        if (Bound(v)) continue;
        x = find_loops (v,temp_equal,false);
        if (x < 0)
            n_new_loops = x;
        else
            n_new_loops += x;
    }
    debug (2, printf("found %d temp loops\n",n_new_loops));
    clear_var_flags (g_sat);
    return (n_new_loops >= 0) ? true : false;
}

static Boolean drop_literal (C)
Clause *C;
{
    /* true=ok, false=contradiction */
    int  i;
    VarPtr *var_list;
    Variable *v1,*v2;

    for (i=C->type,var_list=C->var; i--; var_list++) {
        v1 = *var_list;
        DEREF(v1);
        /* A satisfied clause is of no interest. */
        if (v1->value) return true;
        if (!v1->opposite->value) break;
    }
    /* Found no uncommited literals. */
    if (i < 0) return false;

    for (var_list++; i--; var_list++) {
        v2 = *var_list;
        DEREF(v2);
        if (v2 == v1) continue;
        if (v2->value) return true;
        if (!v2->opposite->value) break;
    }
    if (i < 0) {
        /* Only found one uncommited literal. */
        assert (C->var_count >= 1);
        assert (!v1->value && !v1->opposite->value);
        push_vt (v1);
        return true;
    }

    for (var_list++; i--; var_list++) {
        Variable *vx = *var_list;
        DEREF(vx);
        if (vx == v1 || vx == v2) continue;
        if (vx->value) return true;
        if (!vx->opposite->value) break;
    }
    if (i < 0) {
        /* Only found two uncommited literals. */
        assert (C->var_count >= 2);
        /* Add the two new implications. */
        v1 = v1->opposite;
        v2 = v2->opposite;
        DEREF(v1); DEREF(v2);
        if (add_impl (v1, v2->opposite)) push_im (v1);
        if (add_impl (v2, v1->opposite)) push_im (v2);
        /* Essentially dump this clauses as satisfied. */
        push_it (C->var_count,C);
        C->var_count = 0;
    }
    return true;
}


/* ========================== What is this? ============================ */

static int global_n_nli;        /* Set in find_nli, used in initialize. */


static Boolean initialize (clauses,n,sat,n_nli)
int *clauses;
unsigned n;
SatRec *sat;
int n_nli;
{
    int pass_n = 0, n_new_loops;
    void check_valid ();
    Boolean result = true;

    /* Start up the variables array. */
    init_variables (sat);

    /* Build structures, check for immediate contradictions. */
    result = init_clauses (clauses,n,sat);
    if (!result) return false;
    init_var_links (sat);

    /* Propagate 1-clauses permanently, check for contradictions. */
    debug (3, printf("do %d 1-cl.\n",(int)(IM-var_trail)));
    if (do_1_clauses() == false) return false;
    set_useless_var (sat);
    n_new_loops = find_all_loops (sat);

    while (n_nli--) {

        if (trace_flag >= 2) summarize_sat (sat);
        debug (2, printf("start pass %d:\n",pass_n));
        if (++pass_n > 20) Fail ("Infinite loop?");

        result = find_all_nli (sat);
        if (!result) break;
        check_valid (sat);

        n_new_loops = find_all_loops (sat);
        if (n_new_loops == 0) break;

    }

    if (result) result = end_stuff (sat);
    debug (2, printf("exit initialize: %d, %d passes\n",result,pass_n));
    return result;
}

void check_valid (sat)
SatRec *sat;
{
    Variable *v;
    int i;
    int j, n;
    VarPtr *vp;

    n = 0;
    for (v=sat->variable_array,i=sat->n_variable; i--; v++) {
        if (Bound(v)) continue;
        for (vp=v->implies,j=v->n_implies; j--; vp++) {
            if ((*vp)->opposite->value) n++;
        }
    }
    if (n != 0) printf("%d contradictions found.\n",n);
}

static void set_useless_var (sat)
SatRec *sat;
{
    Variable *var;
    int i, n_new_values, n_not_used;

    debug (3, printf("enter set_useless_var\n"));
    n_new_values = n_not_used = 0;

    for (var=sat->variable_array,i=sat->n_variable; i--; var++) {
        assert ((var->next_equal==NULL) == (var->opposite->next_equal==NULL));
        if (Bound(var)) continue;
        assert (var->opposite->next_equal == NULL);
        if (var->n_satisfies == 0 && var->opposite->n_implies == 0) {
            debug (4, printf("[%d] useless, try [%d].\n",var->what,var->opposite->what));
            if (var->opposite->n_satisfies == 0 && var->n_implies == 0) {
                /* This variable never appears. */
                var->next_equal = sat->pivot_var;
                var->opposite->next_equal = sat->pivot_var->opposite;
                n_not_used++;
                continue;
            }
            if (perm_assign (var->opposite) == false) {
                debug (3, printf("      ~[%d] => [%d]\n",var->opposite->what,var->what));
                if (perm_assign (var) == false) Fail ("Bad Logic");
            }
            n_new_values++;
        }
    } /* end for each variable */

    debug (2, printf("var: %d useless, %d not used\n",n_new_values,n_not_used));
}


/* ========================== Initialization =========================== */

static int comp_var (v1,v2)
Variable *v1, *v2;
{
    int rc = 0;

    /* Get the positive phase of both variables. */
    v1 = (v1 > v1->opposite) ? v1 : v1->opposite;
    v2 = (v2 > v2->opposite) ? v2 : v2->opposite;
    if (v1 > v2)
        rc = 1;
    else if (v1 < v2)
        rc = -1;
    return rc;
}

static int set_equal (vp1,vp2)
Variable *vp1, *vp2;
{
    int   rc = 0;
    Variable *v1, *v2;

    v1 = vp1; while (v1->next_equal != NULL) v1 = v1->next_equal;
    v2 = vp2; while (v2->next_equal != NULL) v2 = v2->next_equal;

    if (v1 != v2) {
        /* Make v1 the "smaller" variable. */
        if (comp_var (v1,v2) > 0) {
            Variable *v_swap = v1; v1 = v2; v2 = v_swap;
        }
        debug (4, printf("set equal: [%d] == [%d]\n",v2->what,v1->what));
        /* Set v2 equal to v1. */
        v2->next_equal = v1;
        if (!v2->loop_flag) {
            v2->loop_flag = true;
            *(IM++) = v2;
        }
        rc = 1;
    }
    return rc;
}

static int find_loops (v,fix_loop,pushit)
Variable *v;
int (*fix_loop)();
Boolean pushit;
{
    /* set all processed flags to -2 for all variables. */
    /* returns -1 if contradictions, otherwise # of loops. */
    int impl_n, x, rc = 0;
    VarPtr *vp;
    Variable *v2;

    assert (v->processed >= -2);
    debug (6, printf("find loops [%d]\n",v->what));
    if (v->processed >= 0) {
        v2 = v;
        debug (5, printf("loop:"));
        do {
            debug (5, printf(" %d",v2->what));
            x = (*fix_loop) (v2,v);
            if (x < 0) return -1;
            if (x > 0) rc = 1;
            x = (*fix_loop) (v2->opposite,v->opposite);
            if (x < 0) return -1;
            if (x > 0) rc = 1;
            assert (v2->processed >= 0 && v2->processed < v2->n_implies);
            v2 = v2->implies[v2->processed];
            assert (v2->processed >= 0);        
        } while (v2 != v);
        debug (5, printf(" .\n"));
    }
    else if (v->processed == -2) {
        v->processed = 0;
        for (impl_n=v->n_implies,vp=v->implies; impl_n--; vp++) {
            v2 = *vp;
            x = find_loops (v2,fix_loop,pushit);
            if (x < 0) return -1;
            rc += x;
            if (pushit && v2->next_equal != NULL && !v->loop_flag) {
                v->loop_flag = true;
                *(IM++) = v;
            }
            v->processed++;
        }
        v->processed = -1;
    }
    debug (6, printf("exit loops [%d]\n",v->what));
    return rc;
}


static Boolean init_clauses (clauses,n,sat)
int *clauses;
unsigned n;
SatRec *sat;
{
    Boolean result = true;
    int *cl_p;
    register int i, j;
    Variable  *v, *vp1, *vp2;
    int clause_n;
    Clause *new_cl = sat->clause_array;  /* next free clause        */
    VarPtr *new_bp = sat->literal_array; /* next free back pointer  */

    enum {
        s_0_clause, s_1_clause, s_2_clause, s_N_clause
    }
    state;

    sat->n_clause = 0;
    clause_n = 0;

    for (cl_p=clauses,i=n; i--; cl_p++) {

        state = s_0_clause;
        clause_n++;

        while (*cl_p != 0) {
            v = sat->pivot_var + (*cl_p);
            i--; cl_p++;
            if (v->processed == clause_n) continue;
            v->processed = clause_n;

            if (state == s_0_clause) {
                vp1 = v;
                state = s_1_clause;
            }
            else if (state == s_1_clause) {
                vp2 = v;
                state = s_2_clause;
            }
            else if (state == s_2_clause) {
                /* Start up another n-clause. */
                new_cl->var = new_bp;  j = 0;
                new_bp[j++] = vp1;
                new_bp[j++] = vp2;
                new_bp[j++] = v;
                state = s_N_clause;
            }
            else if (state == s_N_clause) {
                new_bp[j++] = v;
                /* stay in state s_N_clause. */
            }
            else {
                Fail ("Unexpected state");
            }
        }

        if (state == s_0_clause) {
            /* A null clause is never satisfied. */
            result = false;
            break;
        }
        else if (state == s_1_clause) {
            if (vp1->opposite->value) {
                result = false;
                break;
            }
            /* Just drop this on the implication stack. */
            debug (3, printf("Adding 1-clause (%d)\n",vp1->what));
            if (!vp1->value) push_vt (vp1);
        }
        else if (state == s_2_clause) {
            /* Break 2-clause into two implications. */
            (void) add_impl (vp1->opposite, vp2);
            (void) add_impl (vp2->opposite, vp1);
        }
        else if (state == s_N_clause) {
            /* Complete the construction of this N-clause. */
            new_cl->var_count = new_cl->type = j;
            new_cl->cflag = 0;
            new_bp += j;
            new_cl++;  sat->n_clause++;
        }
        else {
            Fail ("Unexpected state");
        }

    } /* end for all clauses */

    debug (3, printf("exit init_clauses: %d cl, rc=%d\n",sat->n_clause,result));
    return result;
}

static void init_var_links (sat)
SatRec *sat;
{
    register int     i,j;
    register Clause *cl;
    register VarPtr *var_p;

    for (cl=sat->clause_array,i=sat->n_clause; i--; cl++) {
        /* What is this next line good for? */
        if (cl->var_count == 0) continue;
        for (var_p=cl->var,j=cl->type; j--; var_p++) {
            add_clause (*var_p,cl);
        }
    }
}

static void init_variables (sat)
SatRec *sat;
{
    int i,n_binary_var;
    Variable *var;

    n_binary_var = sat->n_binary_var;   /* to make this a signed int */
    for (i=(-n_binary_var); i <= n_binary_var; i++) {
        if (i == 0) continue;
        var = sat->pivot_var + i;
        var->opposite = sat->pivot_var - i;
        debug (0, var->what = i);
    }
}

static void cat_impl (v2,var)
Variable *v2, *var;
{
    unsigned n_alloc, n_needed;
    int n;
    VarPtr *vp, *var_p;
    Variable *v;

    assert (v2->next_equal == NULL);
    n_needed = (v2->n_implies+var->n_implies+3) & (~0x3);
    n_alloc  = (v2->n_implies+3) & (~0x3);
    if (n_needed > n_alloc) {
        v2->implies = Realloc (v2->implies,n_needed,VarPtr);
        if (v2->implies == NULL) Fail("Out of memory");
    }

    /* Copy all implications from var to v2. */
    vp = v2->implies + v2->n_implies;
    for (var_p=var->implies,n=var->n_implies; n--; var_p++) {
        v = *var_p;
        while (v->next_equal != NULL) v = v->next_equal;
        if (v == v2) continue;
        *(vp++) = v;  v2->n_implies++;
    }

    /* Clean out the withered husk of var. */
    var->n_implies = 0;
    if (var->implies != NULL) Free (var->implies);
    var->implies = NULL;
}

static void eliminate_var (var)
Variable *var;
{
    int i,j;
    Variable *v2;
    VarPtr *var_p;
    Clause *cl;
    ClausePtr *cl_p;

    v2 = var->next_equal;
    assert (v2 != NULL);
    while (v2->next_equal != NULL) v2 = v2->next_equal;

    /* Copy over all implications. */
    cat_impl (v2,var);

    /* Copy links to satisfied clauses. */
    for (cl_p=var->satisfies,j=var->n_satisfies; j--; cl_p++) {
        cl = *cl_p;
        for (var_p=cl->var,i=cl->type; i--; var_p++) {
            if (*var_p == var) {
                *var_p = v2;
                add_clause (v2, cl);
                break;
            }
        }
        /* sometimes no var is found.  why not? */
    }
    var->n_satisfies = 0;
    if (var->satisfies != NULL) Free (var->satisfies);
    var->satisfies = NULL;
}

static void fix_impl (var)
Variable *var;
{
    int j;
    VarPtr *var_p;
    Variable *v2;

    assert (var->next_equal == NULL);

    for (var_p=var->implies,j=0; j < var->n_implies; j++) {
        v2 = var_p[j];
        if (v2->next_equal != NULL || v2 == var) {
            while (v2->next_equal != NULL) v2 = v2->next_equal;
            if (v2 == var)
                /* delete this */
                var_p[j--] = var_p[--var->n_implies];
            else
                var_p[j] = v2;
        }
    }
}

static int find_all_loops (sat)
SatRec *sat;
{
    int i, n_new_loops;
    Variable *var;
    VarPtr *vp;

    debug (3, printf("enter find_all_loops\n"));
    vp = sat->var_order;  sat->n_free_var = 0;
    for (var=sat->variable_array,i=sat->n_variable; i--; var++) {
        var->processed = -2;
        var->loop_flag = false;
        if (Bound(var)) continue;
        *(vp++) = var;
        sat->n_free_var++;
    }

    assert (IM == var_trail);
    n_new_loops = 0;
    for (vp=sat->var_order,i=sat->n_free_var; i--; vp++) {
        var = *vp;
        if (var->next_equal != NULL) continue;
        assert (var->value == false);
        n_new_loops += find_loops (var,set_equal,true);
    }

    while (IM > var_trail) {
        /* This is a variable to process. */
        var = *(--IM);
        assert (var->loop_flag);
        var->loop_flag = false;
        if (var->next_equal == NULL) {
            debug (3, printf("fix impl of [%d]\n",var->what));
            fix_impl (var);
        }
        else {
            debug (3, printf("eliminate [%d]\n",var->what));
            eliminate_var (var);
        }
    }

    debug (2, printf("%d loops\n",n_new_loops));
    return n_new_loops;
}

static Boolean end_stuff (sat)
SatRec *sat;
{
    Variable *var, *v1, *v2;
    int i;

    /* Clean up the processed flag. */
    for (var=sat->variable_array,i=sat->n_variable; i--; var++) {
        if (var->next_equal != NULL) {
            /* Make sure that var is not equivalent to var' */
            v1 = var;  while (v1->next_equal != NULL) v1 = v1->next_equal;
            v2 = var->opposite;  while (v2->next_equal != NULL) v2 = v2->next_equal;
            if (v1 == v2) return false;
        }
        assert ((var->next_equal==NULL) == (var->opposite->next_equal==NULL));
        var->processed = -2;
        var->loop_flag = false;
    }

    return true;
}


/* ================== Selecting a variable ordering ==================== */


static void report_loop (v)
Variable *v;
{
    Variable *next_var, *var = v;
    int       n, timeout = 0;

    (void) printf("warning: loop ");
    do {
        (void) printf("[%d]-->",var->what);
        for (next_var=NULL,n=0; n < var->n_implies; n++) {
            if (var->implies[n]->processed == 0) {
                next_var = var->implies[n];
            }
        }
        assert (next_var != NULL);
        if (++timeout > 30) {
            printf("...infinite loop?");
            break;
        }
        var = next_var;
    } while (var != v);
    (void) printf("...\n");
}

static int levelize (v)
Variable *v;
{
    VarPtr *vp;
    int impl_n;
    int x;

    debug (0, if (v->processed >= 0) report_loop (v));

    if (v->processed == -2) {
        v->processed = 0;
        v->level = 1;
        v->n_ti = v->n_implies;
        for (vp=v->implies,impl_n=v->n_implies; impl_n--; vp++) {
            x = levelize (*vp) + 1;
            v->n_ti += (*vp)->n_ti;
            if (x > v->level) v->level = x;
        }
        v->processed = -1;
    }
    return v->level;
}

static int compare_v (vp1,vp2)
VarPtr *vp1, *vp2;
{
    /* >0: vp1 first, =0: equal, <0: vp2 first */
    Variable *v1 = *vp1, *v2 = *vp2;

    /* The implication graph is only a partial order (topological). */
    if (v1->level+v1->opposite->level > v2->level+v2->opposite->level) {
        return 1;
    }
    if (v2->level+v2->opposite->level > v1->level+v1->opposite->level) {
        return -1;
    }
    /* Implication graph does not order these two.  Use the weight. */
    return (v2->weight - v1->weight);
}

static Boolean is_descendant_of (v,target)
Variable *v;
Variable *target;
{
    Boolean rc = false;
    VarPtr *vp;
    int     i;

    if (v == target) return true;
    if (v->level <= target->level) return false;
    if (v->processed == -2) {
        v->processed = 0;
        for (vp=v->implies,i=v->n_implies; i--; vp++) {
            rc = is_descendant_of (*vp,target);
            if (rc) break;
        }
        v->processed = -2;
    }
    return rc;
}

static void clear_var_flags (sat)
SatRec *sat;
{
    Variable *v;
    int i;

    for (v=sat->variable_array,i=sat->n_variable; i--; v++) {
        v->processed = -2;
    }
}

static void clear_trace_flags (v)
Variable *v;
{
    VarPtr *vp;
    int     i;

    if (v->processed == -1) {
        v->processed = -2;
        for (vp=v->implies,i=v->n_implies; i--; vp++)
            clear_trace_flags (*vp);
    }
}

static int tn_implies (v,phase)
Variable *v;
int phase;
{
    int rc = 0;
    VarPtr *vp;
    int     i;

    if (v->processed == -2) {
        v->processed = -1;
        if (phase == 0 && v->what < 0) rc = 1;
        if (phase == 1 && v->what > 0) rc = 1;
        for (vp=v->implies,i=v->n_implies; i--; vp++)
            rc += tn_implies (*vp,phase);
    }
    return rc;
}

static void fill_lit_info (v,lit_info)
Variable *v;
sat_lit_info *lit_info;
{
    lit_info->level_n = v->level;
    lit_info->n_satisfies = v->n_satisfies;
    lit_info->n_trans_implies = v->n_ti;
    lit_info->n_implies_0 = tn_implies (v,0);
    clear_trace_flags (v);
    lit_info->n_implies_1 = tn_implies (v,1);
    clear_trace_flags (v);
}

static void good_var_order (sat,pass_info)
SatRec *sat;
sat_pass *pass_info;
{
    Variable *v;
    VarPtr *vp;
    int i, weight, which_phase;
    int n_phase_swap;
    sat_var_info var_info;

    clear_var_flags (sat);

    /* Initialize the var_order array for good_var_order. */
    vp = sat->var_order;  sat->n_free_var = 0;
    for (v=sat->pos_var_array,i=sat->n_binary_var; i--; v++) {
        v->processed = -2;
        v->opposite->processed = -2;
        if (Bound(v)) continue;
        *(vp++) = v;
        sat->n_free_var++;
        (void) levelize (v);
        (void) levelize (v->opposite);
    }
    *vp = NULL;     /* sentinel null */

    clear_var_flags (sat);
    n_phase_swap = 0;
    /* Get the variable ordering and variable phase to use. */

    for (vp=sat->var_order,i=sat->n_free_var; i--; vp++) {
        v = *vp;  assert (v->what > 0);
        fill_lit_info (v,           &var_info.pos);
        fill_lit_info (v->opposite, &var_info.neg);
        weight = (*pass_info->weight_fcn) (&var_info,&which_phase);
        if (which_phase == 0) {
            n_phase_swap++;
            *vp = v->opposite;
        }
        (*vp)->weight = weight;
    }

    qsort ((char*)sat->var_order, sat->n_free_var, sizeof(VarPtr), compare_v);
    debug (2, printf("Done sorting %d free var.\n",sat->n_free_var));
    debug (2, printf("Swapped phase of %d var.\n",n_phase_swap));
}


/* ====================== Find nonlocal Implications ======================= */


static Boolean find_all_nli (sat)
SatRec *sat;
{
    Variable *var;
    VarPtr   *vp;
    int i, n_nli, n_perm, n_skip;
    Boolean result;
    int max_level, l;
    unsigned n_singleton;

    debug (3, printf("enter find_all_nli\n"));
    n_nli  = 0;
    n_skip = 0;
    n_perm = 0;
    result = true;

    for (var=sat->variable_array,i=sat->n_variable; i--; var++) {
        var->processed = -2;
    }

    vp = sat->var_order;  sat->n_free_var = 0;
    max_level = 0;
    n_singleton = 0;
    for (var=sat->variable_array,i=sat->n_variable; i--; var++) {
        if (Bound(var)) continue;
        if (var->n_implies <= 1) continue;
        *(vp++) = var;
        sat->n_free_var++;
        l = levelize (var);
        if (l > max_level) max_level = l;
        if (var->n_implies <= 1) n_singleton++;
    }

    /* Now sort all var in topological order. */
    /* levelize leaves all processed flags at -1. */
    debug (2, printf("%d/%d var < 2 impl\n",n_singleton,sat->n_free_var));
    debug (2, printf("Max var level = %d.\n",max_level));
    dsort ((char **)sat->var_order, sat->n_free_var, 0,max_level, tellv);

    for (vp=sat->var_order,i=sat->n_free_var; i--; vp++) {

        var = *vp;
        if (Bound(var)) continue;
        /* if (var->processed == -2) continue; */
        debug (4, printf("trying nli of [%d]\n",var->what));
        global_n_nli = 0;
        result = find_nli (var,true);
        var->processed = -2;
        dump_impl ();
        undo_assign (&stbase);
        n_nli += global_n_nli;

        if (!result) {
            debug (4, printf("~[%d] => [%d].\n",var->what,var->opposite->what));
            n_perm++;
            result = perm_assign (var->opposite);
            debug (4, if (!result) printf("~[%d] ^ ~[%d] => f\n",var->what,var->opposite->what));
            if (!result) break;
        }

    } /* end for each variable */

    { Clause *cl; int n_cl;
      for (cl=sat->clause_array,n_cl=sat->n_clause; n_cl--; cl++) {
        assert (cl->var_count == cl->type);
      }
    }

    debug (2, printf("%d var skippable\n",n_skip));
    debug (2, printf("%d nli (%d states) %d perm (result %d)\n",
        n_nli,n_assign_nli,n_perm,result));
    return result;
}

static int tellv (vp)
Variable *vp;
{
    /* For sorting var by level. */
    return (vp)->level;
}

static Boolean find_nli (v,learn)
Variable *v;
Boolean learn;
{
    /* In order of preference for being register variables. */
    register int         i,j;
    register ClausePtr  *L;
    register Clause     *C;
    register Variable   *v2;
    Variable *var;
    VarPtr   *var_list;

    if (v != NULL) {
        assert (!v->value && !v->opposite->value);
        push_vt (v);
    }

    while (more_impl()) {

        var = next_impl();
        debug (4, printf("processing [%d]\n",var->what));
        assert (var->value && !var->opposite->value);
        debug (0,n_assign_nli++);

        for (i=var->n_implies; i--; ) {
            v2 = var->implies[i];
            debug (5, printf("2-cl => [%d]\n", var->implies[i]->what));
            if (v2->opposite->value) return false;
            if (v2->value) continue;
            push_vt (v2);
        }

        for (L=var->opposite->satisfies,j=var->opposite->n_satisfies; j--; L++) {
            C = *L; i = C->var_count;
            if (i == 0) continue;
            debug (5, printf("remove literal from %s\n", clause_name(C)));
            assert (i != 1);
            push_it (i,C);
            if (--C->var_count != 1) continue;
            /* The remaining uncommitted literal is now implied. */
            for (i=C->type,var_list=C->var; i--; var_list++) {
                v2 = *var_list;
                if (v2->opposite->value) continue;
                if (v2->value) break;
                debug (5, printf("n-cl => [%d]\n", v2->what));
                push_vt (v2);
                if (learn) {
                /* Since v -> v2, we can deduce that v2' -> v' */
                debug (4, printf("deduced [%d] => [%d]\n",v2->opposite->what,v->opposite->what));
                if (add_impl (v2->opposite,v->opposite)) {
                    push_im (v2->opposite);
                    global_n_nli++;
                }
                }
                break;
            }
            if (i < 0) return false;
        } /* end for each orphaned clause */
    
        for (L=var->satisfies,j=var->n_satisfies; j--; L++) {
            C = *L; i = C->var_count;
            if (i == 0) continue;
            debug (5, printf("satisfied %s\n",clause_name(C)));
            push_it (i,C);
            C->var_count = 0;
        } /* end for each satisifed clause */

    } /* end while more implications */

    return true;    /* entire assignment was successful */

} /* end find_nli */

/* ================ Preprocessing assignment functions ================= */

static Boolean do_1_clauses ()
{
    Boolean result = true;
    Variable *v;

    if (more_impl()) {
        /* Steal the top of the stack. */
        v = *(--IM);
        assert (v->value);
        v->value = false;
        result = perm_assign (v);
    }
    return result;
}

static Boolean perm_assign (v)
Variable *v;        /*u variable to assign. */
{
    int       i;
    VarPtr   *var_p;
    Variable *var;
    Clause   *cl;

    assert (v != NULL);
    if (find_nli(v,false) == false) {
        undo_assign (&stbase);
        return false;
    }

    /* Make an assignment permanent. */
    assert (VT == IM);
    IM = var_trail;
    while (VT > var_trail) {
        var = *(--VT);
        while (var->n_satisfies != 0) {
            drop_clause (*var->satisfies);
        }
    }

    while (IT > incr_trail) {
        cl = (--IT)->clause_p;
        if (cl->var_count == 0) continue;
        /* Remove any var's which are false. */
        for (var_p=cl->var,i=0; i < cl->type; i++) {
            var = var_p[i];
            assert (!var->value);
            if (var->opposite->value) {
                /* This literal has dropped out. */
                var_p[i--] = cl->var[--cl->type];
            }
        }
        assert (cl->type == cl->var_count);
        if (cl->type == 2) {
            debug (3, printf("new 2 clause\n"));
            (void) add_impl (cl->var[0]->opposite, cl->var[1]);
            (void) add_impl (cl->var[1]->opposite, cl->var[0]);
            drop_clause (cl);
        }
    }

    return true;    /* assignment succeeded */
}

static void drop_clause (target_cl)
Clause *target_cl;
{
    int i,j;
    VarPtr *var_p;
    Variable *var;
    ClausePtr *cl_p;

    target_cl->var_count = 0;

    debug (4, printf("dropping clause %s\n",clause_name(target_cl)));
    /* Remove this clause from the sat problem. */
    for (var_p=target_cl->var,i=target_cl->type; i--; var_p++) {
        var = *var_p;
        for (cl_p=var->satisfies,j=var->n_satisfies; j--; cl_p++) {
            if (*cl_p == target_cl) {
                *cl_p = var->satisfies[--var->n_satisfies];
                break;
            }
        }
        assert (j >= 0);
    }
    target_cl->type = target_cl->var_count = 0;
}


/* ========================== Miscellaneous ============================ */


static void estimate_sizes (clauses, n, n_var,n_lit,n_ncl)
int *clauses;       /* in format described above */
unsigned  n;        /* length of array clauses */
unsigned *n_var, *n_lit;
unsigned *n_ncl;
{
    register int i, var, w, *p;

    /* Rest of satisfy depends on the following. */
    if (n == 0) 
		Fail("Bad input");
    if (clauses[n-1] != 0) 
		Fail("Bad input");
    *n_lit = *n_var = *n_ncl = 0;
    w = 0;

    for (p=clauses,i=n; i--; p++) {
        if ((var=(*p)) == 0) {
            if (w > 2)
                { ++*n_ncl; *n_lit += w; }
            w = 0;
        }
        else {
            w++;
            if (var < 0) var = - var;
            if (var > *n_var) *n_var = var;
        }
    }
    debug (2, printf("%d total var.\n",*n_var));
}

static void alloc_sat (sat,n_var,n_ncl,n_lit)
SatRec *sat;
unsigned n_var, n_ncl, n_lit;
{
    Variable *v, *hidden_var;
    int i;

    sat->n_binary_var   = n_var;
    sat->a_variable     = 2 * (n_var+1);
    sat->a_list         = n_lit;
    sat->a_literal      = n_lit;
    sat->a_clause       = n_ncl;

    sat->list_array     = Malloc (sat->a_list, ClausePtr);
    sat->literal_array  = Malloc (sat->a_literal,  VarPtr);
    sat->clause_array   = Calloc (sat->a_clause, Clause);
    sat->variable_array = Calloc (sat->a_variable, Variable);
    sat->var_order      = Malloc (sat->a_variable, VarPtr);

    if (sat->list_array == NULL   || sat->literal_array == NULL  ||
            sat->clause_array == NULL || sat->variable_array == NULL ||
            sat->var_order == NULL) Fail("Out of memory");

    /* Set up the convenience pointers to the variable array. */
    sat->neg_var_array  = sat->variable_array;
    sat->pivot_var      = sat->neg_var_array + sat->n_binary_var;
    sat->pos_var_array  = sat->pivot_var + 1;

    sat->n_variable = 2 * sat->n_binary_var + 1;
    for (v=sat->variable_array,i=sat->n_variable; i--; v++) {
        v->implies = NULL;
        v->n_implies = 0;
    }

    /* Special "zero" variable pair. */
    hidden_var = sat->pos_var_array + sat->n_binary_var;
    sat->pivot_var->opposite = hidden_var;
    sat->pivot_var->next_equal   = NULL;
    sat->pivot_var->value    = true;
    hidden_var->opposite = sat->pivot_var;
    hidden_var->next_equal   = NULL;
    hidden_var->value    = false;
}

static int add_impl (v,lit)
Variable *v;
Variable *lit;
{
    register VarPtr *vp;
    register int     i;
    unsigned n_to_alloc;

    assert (v != NULL && lit != NULL);
    while (lit->next_equal != NULL) lit = lit->next_equal;
    if (v->next_equal != NULL) /* temporary change from the "assert"
		abort();				* so as to cause a core dump */
    if (v == lit) return 0;

    for (vp=v->implies,i=v->n_implies; i--; vp++) {
        if (*vp == lit) return 0;
    }

    if ((v->n_implies&0x3) == 0) {
        n_to_alloc = v->n_implies + 4;
        if (v->implies == NULL)
            v->implies = Malloc (n_to_alloc,VarPtr);
        else
            v->implies = Realloc (v->implies,n_to_alloc,VarPtr);
        if (v->implies == NULL) Fail("Out of memory");
    }

    v->implies[v->n_implies++] = lit;
    return 1;
}

static void add_clause (v,cl)
Variable *v;
Clause *cl;
{
    unsigned n_to_alloc;

    assert (v->next_equal == NULL);
    if ((v->n_satisfies&0x3) == 0) {
        n_to_alloc = v->n_satisfies + 4;
        if (v->satisfies == NULL)
            v->satisfies = Malloc (n_to_alloc,ClausePtr);
        else
            v->satisfies = Realloc (v->satisfies,n_to_alloc,ClausePtr);
        if (v->satisfies == NULL) Fail("Out of memory");
    }

    v->satisfies[v->n_satisfies++] = cl;
}

static void dump_impl ()
{
    /* Dump all impl stack elements to the free list. */
    while (impl_trail != NULL) {
        TempImpl *ti = impl_trail;  impl_trail = ti->next;
        ti->next = free_impl;       free_impl  = ti;
    }
}

static void free_sat (sat)
SatRec *sat;
{
    register Variable *v;
    register int i, n_imp;
    unsigned long imp_mem, lit_mem, ncl_mem, var_mem;

#define MALLOC_HDR  4

    if (trace_flag >= 1) {
        var_mem = sat->a_variable *
                    (sizeof(Variable)+2*sizeof(VarPtr));
        lit_mem = sat->a_literal *
                    (sizeof(SaveCount)+sizeof(VarPtr)+sizeof(ClausePtr));
        ncl_mem = sat->a_clause * (sizeof(Clause));
        imp_mem = 0;
        n_imp = 0;
        for (v=sat->variable_array,i=sat->n_variable; i--; v++) {
            n_imp += v->n_implies;
            imp_mem += MALLOC_HDR + sizeof(VarPtr) * (v->n_implies+3)/4;
            lit_mem += MALLOC_HDR + sizeof(ClausePtr) * (v->n_satisfies+3)/4;
        }
    }

    if (trace_flag >= 2) {
        /* Print estimate of memory usage for this problem. */
        (void) printf("SAT: %5d var (%6lu bytes)\n     %5d lit (%6lu bytes),\n",
                sat->a_variable, var_mem, sat->a_literal, lit_mem);
        (void) printf("     %5d ncl (%6lu bytes)\n     %5d imp (%6lu bytes).\n",
                sat->a_clause, ncl_mem, n_imp, imp_mem);
    }
    if (trace_flag >= 1) {
        (void) printf("Memory estimate: %lu bytes for %u var\n",
            var_mem+lit_mem+ncl_mem+imp_mem,sat->n_binary_var);
    }

    for (v=sat->variable_array,i=sat->n_variable; i--; v++) {
        if (v->implies  != NULL) Free (v->implies);
        if (v->satisfies != NULL) Free (v->satisfies);
    }

    Free (sat->variable_array);
    Free (sat->var_order);
    Free (sat->clause_array); 
    Free (sat->literal_array);
    Free (sat->list_array);
}


#define MAX_CCOUNT  7
#define MAX_ICOUNT  5

static void summarize_sat (sat)
SatRec *sat;
{
    int i, w, total_used = 0;
    Variable *var;
    int icount[MAX_ICOUNT], max_nimpl = 0;

    for (i=0; i < MAX_ICOUNT; i++) icount[i] = 0;

    for (i=0,var=sat->variable_array; i < sat->n_variable; i++,var++) {
        if (Bound(var)) continue;
        total_used  += var->n_implies;
        w = var->n_implies;
        if (w > max_nimpl) max_nimpl = w;
        if (w >= MAX_ICOUNT-1)
            icount[MAX_ICOUNT-1]++;
        else
            icount[w]++;
    }

    (void) printf("n_implies:");
    for (i=0; i < MAX_ICOUNT-1; i++)
        if (icount[i] != 0) (void) printf("  %d:%d",i,icount[i]);
    if (max_nimpl == MAX_ICOUNT-1)
        (void) printf("  %d:%d",max_nimpl,icount[MAX_ICOUNT-1]);
    else if (max_nimpl >= MAX_ICOUNT)
        (void) printf("  %d-%d:%d",MAX_ICOUNT-1,max_nimpl,icount[MAX_ICOUNT-1]);
    (void) printf("\nTotal: %d used.\n",total_used);
}

static void print_summary (clauses,n)
int *clauses;
unsigned n;
{
    int i,n_cl,w;
    int ccount[MAX_CCOUNT], max_clause = 0;
    register *p;
    float percent_2cl;

    n_cl = 0;
    for (i=0; i < MAX_CCOUNT; i++) ccount[i] = 0;

    for (i=0; i < n; i+=w+1) {

        for (p=clauses+i,w=0; *p != 0; p++) w++;
        if (w > max_clause) max_clause = w;
        n_cl++;
        if (w >= MAX_CCOUNT) {
            ccount[0]++;
        }
        else if (w != 0) {
            ccount[w]++;
        }
    }

    percent_2cl = 100.*((float)ccount[2]/(float)n_cl);
    (void) printf("%d clauses:",n_cl);
    for (i=1; i < MAX_CCOUNT; i++) {
        if (ccount[i] != 0) (void) printf("  %d:%d",i,ccount[i]);
        if (i == 2) (void) printf(" (%.0f%%)",percent_2cl);
    }
    if (ccount[0] != 0) {
        if (max_clause == MAX_CCOUNT)
            (void) printf("  %d:%d",MAX_CCOUNT,ccount[0]);
        else
            (void) printf("  %d-%d:%d",MAX_CCOUNT,max_clause,ccount[0]);
    }
    (void) printf("\n");

}


static void check_assignment (clauses,n,sat)
int      *clauses;
unsigned  n;
SatRec   *sat;
{
    /* Checks the assignment against the original clause list. */
    unsigned i;
    Variable *var;
    int c, *p, n_err = 0;
    Boolean cl_satisfied;

    for (var=sat->variable_array,i=sat->n_variable; i--; var++) {
        if (var->value && var->opposite->value) {
            (void) fprintf(stderr,"\nERROR ### [%d] and [%d] are both true!\n",
                var->what, var->opposite->what);
            dump_var (var);
            dump_var (var->opposite);
        }
    }

    for (cl_satisfied=false,p=clauses,i=n; i--; p++) {
        if (*p == 0) {
            if (!cl_satisfied) {
                n_err++;
                (void) fprintf(stderr,"ERROR ### Clause ");
                while (p > clauses)
                    if (*(--p) == 0) {p++; break;}
                for (c='('; *p != 0; p++) {
                    (void) fprintf(stderr,"%c%d",c,*p); c=' ';
                }
                (void) fprintf(stderr,") is not satisfied.\n");
            }
            cl_satisfied = false;
        }
        else if (!cl_satisfied) {
            if (var_value (*p,sat) == 1) cl_satisfied = true;
        }
    }

    if (trace_flag >= 2 && n_err != 0) {
        (void) fprintf(stderr,"\nVariable values:");
        for (i=0; i <= sat->n_binary_var; i++) {
            if (i%50 == 0) (void) fprintf(stderr,"\n",i);
            if (i%10 == 0) (void) fprintf(stderr," %4d) ",i);
            if (i == 0) {(void)fprintf(stderr," "); continue;}
            (void) fprintf(stderr,"%d",var_value((int)i,sat));
        }
        (void) fprintf(stderr,"\n\n");
    }
}


static char *clause_name (c)
Clause *c;
{
    static char buf[90];
    int i,k;

    for (k=0,i=0; i < c->type; i++) {
        (void) sprintf(buf+k,"%c%d",(i==0)?'(':' ',c->var[i]->what);
        if (c->var[i]->value) (void) strcat(buf,"t");
        if (c->var[i]->opposite->value) (void) strcat(buf,"f");
        k += strlen(buf+k);
        /* Watch out for the godzilla clause. */
        if (k >= sizeof(buf)-10) {
            (void) strcpy(buf+k," ...");
            k += 4; break;
        }
    }
    (void) sprintf(buf+k,")");
    return buf;
}


static int var_value (var_n,sat)
int     var_n;
SatRec *sat;
{
    Variable *v;
    int pos_var_n, x = 2;

    pos_var_n = ABS(var_n);
    if (pos_var_n > 0 && pos_var_n <= sat->n_binary_var) {
        v = sat->pivot_var + var_n;
        while (v->next_equal != NULL) v = v->next_equal;
        if (v->value)
            x = 1;
        else if (v->opposite->value)
            x = 0;
    }
    return x;
}


static void copy_assignment (vars,values,n_vars,sat)
int *vars;          /*i variables to return values for */
int *values;        /*o the values that were assigned */
unsigned n_vars;    /*i number of entries in vars & values */
SatRec *sat;        /*i the solved sat problem */
{
    register int n;

    for (n=n_vars; n--; ) {
        values[n] = var_value (vars[n],sat);
        debug (3, printf("[%d] = %d\n",vars[n],values[n]));
    }
}

void dump_var (v)
Variable *v;
{
    int i,j;
    (void) printf("[%d] opp [%d] equ [%d] val %d level %d n_ti %d\n",
        v->what,v->opposite->what,v->next_equal==NULL?0:v->next_equal->what,
        v->value,v->level,v->n_ti);
    (void) printf("       imp2 %d ncl %d weight %d proc %d loop %d\n",
        v->n_implies,v->n_satisfies,v->weight,v->processed,v->loop_flag);
    for (i=0; i < v->n_implies; ) {
        (void) printf("-->");
        for (j=8; j-- && i < v->n_implies; i++)
            (void) printf(" %d",v->implies[i]->what);
        (void) printf("\n");
    }
}

/*
    Appends the given set of clauses to file "saved_clauses" in a simple
    format so they can be easily read in again by a SAT test driver program.  */

void sat_write (clauses,n,fname,append)
int  *clauses;
unsigned n;
char *fname;
int   append;
{
    static int problem_n = 0;
    int column_n = 1;
    char temp[10];
    FILE *fout;
    int *cl, i, x;

    if (fname == NULL) fname = "saved_clauses";
    fout = fopen(fname,(append)?"a":"w");
    if (fout == NULL) return;

    (void) fprintf(fout,"\n");
    (void) fprintf(fout,"prob%d 2 0\n",++problem_n);

    for (cl=clauses,i=n; i--; cl++) {
        (void) sprintf(temp,"%d%c",*cl,(i==0)?'.':',');
        x = strlen(temp);
        column_n += x;
        if (column_n > 80) {
            (void) fprintf(fout,"\n");
            column_n = x+1;
        }
        (void) fprintf(fout,"%s",temp);
    }

    (void) fprintf(fout,"\n");
    if (fclose (fout) == EOF) Fail("fclose error");
}

static void dsort (array,n_el,minkey,maxkey,getkey)
char     **array;
unsigned   n_el;
int        minkey, maxkey;
int      (*getkey)();
{
    unsigned n_bin;
    char **p, **locn_for_p, *key_to_place, *temp;
    unsigned j, bin_n;
    int pkey;
    unsigned k;
    struct BinInfo {
        unsigned qty;
        unsigned bounds;
    } *counts;

    /* Allocate bins. */
    n_bin = maxkey - minkey + 2;
    counts = (struct BinInfo *) calloc (n_bin,sizeof(struct BinInfo));
    assert (counts != NULL);

    for (p=array,j=n_el; j--; p++) {
        k = 1 + maxkey - (*getkey) (*p);
        if (k >= n_bin) k = n_bin-1;
        counts[k].qty++;
    }

    counts[0].bounds = 0;
    for (j=1; j < n_bin; j++) {
        counts[j].qty += counts[j-1].qty;
        counts[j-1].bounds = counts[j].qty;
    }
    assert (counts[n_bin-2].bounds == n_el);

    for (bin_n=0; bin_n < n_bin-1; bin_n++) {
        while (counts[bin_n].qty < counts[bin_n].bounds) {
            p = array + counts[bin_n].qty;
            temp = *p;
            pkey = maxkey - (*getkey) (temp);
            if (pkey == bin_n) {
                counts[bin_n].qty++;
                continue;
            }

        do {
            key_to_place = temp;
            pkey = maxkey - (*getkey) (key_to_place);
            k = counts[pkey].qty++;
            locn_for_p = array + k;
            temp = *locn_for_p;
            /* This puts it in a valid position. */
            *locn_for_p = key_to_place;
        } while (locn_for_p != p);

        } /* end while more in this bin */

    } /* end for every element */

#ifndef NDEBUG
    /* Verify the array is in descending order. */
    {   int prev_key = maxkey, key;
        for (p=array,j=n_el; j--; p++) {
            key = (*getkey) (*p);
            assert (key <= prev_key);
            prev_key = key;
        }
    }
#endif
	free(counts);
} /* end dsort */
