/* satisfy.h 

   There are two callable functions, and a description of the user
   functions you must write to define the search strategies.

   int sat_solve (int *clauses, unsigned n,
                  int *outvar, int *values, unsigned n_out,
                  sat_options *options);

     Tries to solve the specified SAT problem using the user-defined
     strategies in the options structure (see below).
     Returns: 0 = not solvable
              1 = solved, values for outvar are in values array
              2 = gave up on all strategies
              3 = ran out of memory
              4 = internal error

     Any other return value should also be interpreted as some sort of
     internal error, with no information on whether the problem can be
     solved or not.

   void sat_write (int *clauses, unsigned n, char *fname, int append);

     Writes the clauses to the specified file, either appending if
     argument append is nonzero, or overwriting otherwise.

   int weight_fcn (sat_var_info *var_info, int *which_phase);

     Optionally using the information in var_info, this user-specified
     function must return a weight for the variable to indicate the
     ordering used in the branch and bound phase, a higher weight means
     the variable has higher priority.  It also must set *which_phase
     to 1 or 0 depending on which phase should be tried first.  It is
     okay to use the same weight function for more than one search
     strategy.
*/


typedef struct {
    int level_n;        /* minimum level # in impl. graph   */
    int n_satisfies;    /* # of 3+ clauses satisfied        */
    int n_implies_0;    /* # of var this forces to zero     */
    int n_implies_1;    /* # of var this forces to one      */
    int n_trans_implies;    /* transitive n_implies for this literal*/
}
sat_lit_info;

typedef struct {
    sat_lit_info    pos;    /* info for when variable is true   */
    sat_lit_info    neg;    /* info for when variable is false  */
}
sat_var_info;

typedef struct {
    int (*weight_fcn) ();   /* variable weight and phase ordering   */
    int backtrack_lim;  /* limit on backtracks for this phase   */
    int use_dyn_eq;     /* do dynamic equal variable checks?    */
    int use_dyn_nli;    /* do dynamic nonlocal impl checks? */
}
sat_pass;

typedef struct {
    int       trace_flag;   /* 0=no trace, positive=enable output   */
    char     *out_file; /* NULL, or file to write clauses to    */
    int       append_out;   /* 1=overwrite file, 0=append to file   */
    int       n_init_nli;   /* number of initial nli checks to make */
    int       n_pass;   /* number of passes for branch&bound    */
    sat_pass *passes;   /* description of the passes        */
}
sat_options;

int  sat_solve ();
void sat_write ();
