/* Compile with -DNDEBUG or add #define NDEBUG to compile out all
   debugging output and assertions. */

#include "utility.h"
#include "util.h"
#include "satisfy.h"

#define Fail(msg)\
    {(void)fprintf(stderr,"Error:%s: file %s, line %d\n",\
     msg,__FILE__,__LINE__);\
     exit(1);\
    }

#ifdef NDEBUG
#define debug(n,x)
#else
#define debug(n,x)  if (trace_flag >= n) x
#endif

typedef enum {
  false,
  true,
  undecided
} Boolean;

#define ushort unsigned short
typedef struct Variable Variable, *VarPtr;
typedef struct Clause   Clause,   *ClausePtr;


struct Variable {
  Variable  *opposite;      /* pointer to the opposite phase literal       */
  Variable  *next_equal;    /* equivalent variable to use instead          */
  Variable  *prev_equal;    /* for dynamic variable equivalence            */
  ClausePtr *satisfies;     /* clauses satisifed by this variable          */
  VarPtr    *implies;       /* list of implications (mostly 2-clauses)     */
  Clause    *work;      /* for detecting duplicate var in clauses      */
  int        weight;        /* for sorting the variables as input order    */
  short      what;          /* notation to use for printing this variable  */
  ushort     n_satisfies;   /* number of clauses it causes to be satisfied */
  ushort     n_implies;     /* number of elements in implies               */
  ushort     level;         /* level number based on implication graph     */
  ushort     n_ti;          /* number of transitive implications           */
  short      processed;     /* general recursive processing flag           */
  Boolean    value;     /* current value of variable                   */
  Boolean    loop_flag;     /* used to mark var while tracing loops        */
  Boolean    primary;       /* is variable assigned in branch phase?       */
};

struct Clause {
  VarPtr    *var;           /* pointers to literals in this clause         */
  short      type;      /* number of elements in var array             */
  short      var_count;     /* max # of unbound literals in this clause    */
                    /* 0 indicates the clause is satisfied         */
  char       cflag;         /* for marking a clause with                   */
};

typedef struct {
  Clause    *clause_p;
  ushort     prev_count;
}
SaveCount;

typedef struct TempImpl {
  struct TempImpl *next;
  Variable *var;
}
TempImpl;

typedef struct {
  VarPtr *save_vt;
  SaveCount *save_it;
  TempImpl  *save_im;
  VarPtr    *save_et;
}
StackStatus;

typedef struct SatRec {
  unsigned   n_binary_var;
  Variable  *pivot_var;
  Variable  *pos_var_array;
  Variable  *neg_var_array;
  Variable  *variable_array;
  unsigned   a_variable;
  unsigned   n_variable;
  VarPtr    *var_order;
  unsigned   n_free_var;
  Clause    *clause_array;
  unsigned   n_clause;
  unsigned   a_clause;
  VarPtr    *literal_array;
  unsigned   a_literal;
  ClausePtr *list_array;
  unsigned   a_list;
}
SatRec;


#define AllocSize(qty,type)     (qty * (unsigned) sizeof(type))
#define Malloc(qty,type)        (type *) malloc ((unsigned)(qty) * sizeof(type))
#define Calloc(qty,type)        (type *) calloc ((unsigned)(qty),sizeof(type))
#define Free(ptr)               free ((char *)(ptr))
#define Realloc(ptr,qty,type)   (type *) realloc ((char *)ptr,(unsigned)(qty) * sizeof(type))
#define SUM(a,b)                ((a) + (b))
#define SWAP(type,a,b)          {type x_swap=(a);(a)=(b);(b)=x_swap;}


#define push_vt(v) \
    { assert ((IM-var_trail) < n_vt);\
    assert (!(v)->opposite->value);\
    v->value = true; *(IM++) = v; }
#define next_impl()     (*(VT++))
#define more_impl()     (IM > VT)
#define push_et(v) \
    { assert ((ET-equal_trail) < n_et);\
    *(ET++) = v; }
#define push_it(i,C) \
    { assert((IT-incr_trail) < n_it);\
    IT->prev_count=i; (IT++)->clause_p = C; }
#define push_im(v) \
    { TempImpl *new_impl = free_impl; \
    if (new_impl == NULL) new_impl = Malloc(1,TempImpl); \
    else free_impl = new_impl->next; \
    assert (new_impl != NULL); \
    new_impl->var = (v); \
    new_impl->next = impl_trail; impl_trail = new_impl; }
#define Bound(v)  ((v)->next_equal!=NULL || (v)->value || (v)->opposite->value)
#define DEREF(v)  while ((v)->next_equal != NULL) (v) = (v)->next_equal


                                /* Static functions in satisfy.c */
#ifdef __STDC__
Boolean initialize (int *, unsigned, SatRec *, int);
void alloc_sat (SatRec *, unsigned,unsigned,unsigned);
void free_sat (SatRec *);
Boolean end_stuff (SatRec *);
int set_equal (Variable *, Variable *);
void drop_clause (Clause *);
int var_value (int, SatRec *);
void fix_impl (Variable *);
void subst_var (Variable *);
void add_clause (Variable *, Clause *);
Boolean branch_and_bound (SatRec *, sat_pass *);
Boolean branch (VarPtr *);
Boolean bound (Variable *);
Boolean find_nli (Variable *, Boolean);
Boolean do_1_clauses (void);
void save_stacks (StackStatus *);
void undo_assign (StackStatus *);
void estimate_sizes (int *, unsigned, unsigned *,unsigned *,unsigned *);
void check_assignment (int *, unsigned, SatRec *);
int find_all_loops (SatRec *);
int find_loops (Variable *, int (*loop_f)(), Boolean);
void set_useless_var (SatRec *);
Boolean find_all_nli (SatRec *);
int add_impl (Variable *, Variable *);
int compare_v (VarPtr *, VarPtr *);
void good_var_order (SatRec *, sat_pass *);
void init_variables (SatRec *);
void init_var_links (SatRec *);
Boolean perm_assign (Variable *);
int comp_var (Variable *, Variable *);
void summarize_sat (SatRec *);
Boolean init_clauses (int *, unsigned, SatRec *);
int levelize (Variable *);
char *clause_name (Clause *);
void copy_assignment (int *, int *, unsigned, SatRec *);
void print_summary (int *, unsigned);
int tellv (Variable *);
void dsort (char **, unsigned, int, int, int (*)());
int temp_equal (Variable *, Variable *);
void undo_equal (Variable *);
Boolean check_equivalence (Boolean);
Boolean drop_literal (Clause *);
Boolean is_descendant_of (Variable *, Variable *);
void clear_var_flags (SatRec *);
void clear_trace_flags (Variable *);
int tn_implies (Variable *, int);
void fill_lit_info (Variable *, sat_lit_info *);
void dump_var (Variable *);
void dump_impl (void);
#else
void dump_impl ();
void fill_lit_info ();
void clear_var_flags ();
void clear_trace_flags ();
int tn_implies ();
Boolean is_descendant_of ();
Boolean initialize ();
void add_clause ();
void alloc_sat ();
void free_sat ();
Boolean end_stuff ();
Boolean branch_and_bound ();
Boolean branch ();
Boolean bound ();
Boolean find_nli ();
void undo_equal ();
void undo_assign ();
void save_stacks ();
void estimate_sizes ();
void check_assignment ();
int find_all_loops ();
int find_loops ();
int set_equal ();
void set_useless_var ();
Boolean find_all_nli ();
int add_impl ();
int compare_v ();
void good_var_order ();
void init_variables ();
Boolean init_clauses ();
int levelize ();
char *clause_name ();
void dump_var ();
int var_value ();
void copy_assignment ();
void print_summary ();
void summarize_sat ();
Boolean perm_assign ();
void drop_clause ();
void init_var_links ();
void report_loop ();
void cat_impl ();
void dsort ();
int tellv ();
Boolean do_1_clauses ();
Boolean drop_literal ();
Boolean check_equivalence ();
int temp_equal ();
#endif
