#include "mis.h"
#include "atpg_int.h"

static void sim_eval_inverter();
static void sim_eval_nand2();
static void sim_eval_node();
static void sim_eval_pi();
static void sim_eval_po();
static void sim_extract_function();
static void sim_extract_node();
static void sim_free_node();
static int  sim_is_inverter();
static int  sim_is_nand2();
static int  sim_is_pi();

/* Basically a modification of network_dfs which also includes in the
 * returned array primary inputs and outputs which are not actually
 * used in the network.  Given a network returns a topological array
 * as per network_dfs. */
array_t *
network_dfs_complete(network)
network_t *network;
{
    array_t *nodevec, *temp;
    node_t  *pi;
    lsGen   gen;

    nodevec = network_dfs(network);
    foreach_primary_input(network, gen, pi) {
        if (node_num_fanout(pi) == 0) {
            temp = array_alloc(node_t *, 0);
            array_insert(node_t *, temp, 0, pi);
            array_append(temp,nodevec);
            array_free(nodevec);
            nodevec = temp;
        }
    }
    return nodevec;
}

/*
 * extract a network into an efficient data structure for simulation
 * also returns a hash table containing links between nodes and uid's
 * the hash table should be allocated already
 */
sim_node_t *
sim_extract_network(network, table)
network_t *network;
st_table *table;
{
    int i, uid;
    array_t *nodevec;
    sim_node_t *result;

    nodevec = network_dfs_complete(network);
    result = ALLOC(sim_node_t, array_n(nodevec));
    uid = 0;
    for (i = 0; i < array_n(nodevec); i++) {
        node_t *node = array_fetch(node_t *, nodevec, i);
        st_insert(table, (char *) node, (char *) uid);
        uid++;
    }
    assert(st_count(table) == array_n(nodevec));
    for (i = 0; i < array_n(nodevec); i++) {
        node_t *node = array_fetch(node_t *, nodevec, i);
        assert(st_lookup(table, (char *) node, (char **) &uid));
        sim_extract_node(node, &result[uid], table);
    }
    array_free(nodevec);
    return result;
}

void 
sim_free_nodes(sim_nodes, table)
sim_node_t *sim_nodes;
st_table *table;
{
    int i;
  
    for (i = 0; i < st_count(table); i++) {
        sim_free_node(&sim_nodes[i]);
    }
    FREE(sim_nodes);
}

static void 
sim_extract_node(node, sim_node, table)
node_t *node;
sim_node_t *sim_node;
st_table *table;
{
    int i, uid;

    sim_node->node = node;
    assert(st_lookup(table, (char *) node, (char **) &uid));
    sim_node->id = uid;
    sim_node->n_cubes = node_num_cube(node);
    sim_node->function = ALLOC(int *, sim_node->n_cubes);
    sim_node->n_inputs = node_num_fanin(node);
    for (i = 0; i < sim_node->n_cubes; i++) {
        sim_node->function[i] = ALLOC(int, sim_node->n_inputs);
    }
    sim_extract_function(sim_node);
    sim_node->input_uids = ALLOC(int, sim_node->n_inputs);
    for (i = 0; i < sim_node->n_inputs; i++) {
        int uid;
        node_t *fanin = node_get_fanin(node, i);
        if (st_lookup(table, (char *) fanin, (char **) &uid))
            sim_node->input_uids[i] = uid;
    }
    sim_node->inputs = ALLOC(int *, 3);
    for (i = 0; i < 3; i++) {
        sim_node->inputs[i] = ALLOC(int, sim_node->n_inputs);
    }
    for (i = 0; i < sim_node->n_inputs; i++) {
        sim_node->inputs[2][i] = ALL_ONES;
    }
    sim_node->or_output_mask = ALL_ZEROS;
    sim_node->or_input_masks = ALLOC(int, sim_node->n_inputs);
    for (i = 0; i < sim_node->n_inputs; i++) {
        sim_node->or_input_masks[i] = ALL_ZEROS;
    }
    sim_node->and_output_mask = ALL_ONES;
    sim_node->and_input_masks = ALLOC(int, sim_node->n_inputs);
    for (i = 0; i < sim_node->n_inputs; i++) {
        sim_node->and_input_masks[i] = ALL_ONES;
    }
    if (sim_node->node->type == PRIMARY_INPUT) {
        sim_node->eval = sim_eval_pi;
    }
    else
        if (sim_node->n_cubes == 0) {
            if (sim_node->node->type != PRIMARY_OUTPUT) {
                sim_node->eval = sim_eval_pi;
            }
            else {
                /*assert(sim_node->node->type == PRIMARY_OUTPUT);*/
                sim_node->eval = sim_eval_po;
            }
        }
        else
            if (sim_is_inverter(sim_node)) {
                sim_node->eval = sim_eval_inverter;
        } else
            if (sim_is_nand2(sim_node)) {
                sim_node->eval = sim_eval_nand2;
            }
            else {
                sim_node->eval = sim_eval_node;
            }
}

static void 
sim_extract_function(sim_node)
sim_node_t *sim_node;
{
    int c, l;
    node_t *node = sim_node->node;
    node_cube_t cube;

    for (c = 0; c < sim_node->n_cubes; c++) {
        cube = node_get_cube(node, c);
	for (l = 0; l < node_num_fanin(node); l ++) {
            int literal = node_get_literal(cube, l);
            assert(l >= 0 && l < sim_node->n_inputs);
            switch (literal) {
                case ZERO:
                    sim_node->function[c][l] = 0;
                    break;
                case ONE:
                    sim_node->function[c][l] = 1;
                    break;
                case TWO:
                    sim_node->function[c][l] = 2;
                    break;
                default:
                    fail("bad literal");
            }
        }
    }
}

static void 
sim_free_node(sim_node)
sim_node_t *sim_node;
{
    int i;

    FREE(sim_node->and_input_masks);
    FREE(sim_node->or_input_masks);
    for (i = 0; i < 3; i++) {
        FREE(sim_node->inputs[i]);
    }
    FREE(sim_node->inputs);
    FREE(sim_node->input_uids);
    for (i = 0; i < sim_node->n_cubes; i++) {
        FREE(sim_node->function[i]);
    }
    FREE(sim_node->function);
}

static int 
sim_is_inverter(sim_node)
sim_node_t *sim_node;
{
    if (sim_node->n_cubes != 1) return 0;
    if (sim_node->n_inputs != 1) return 0;
    return (sim_node->function[0][0] == 0);
}

 /* this is particularly ugly; should be simplified */
static int 
sim_is_nand2(sim_node)
sim_node_t *sim_node;
{
    int i, j;

    if (sim_node->n_cubes != 2)
        return 0;
    if (sim_node->n_inputs != 2) return 0;
    for (i = 0; i < 2; i++)
        for (j = 0; j < 2; j++)
            if (sim_node->function[i][j] == 1)
                return 0;
    if (sim_node->function[0][0] == 0) {
        if (sim_node->function[0][1] != 2) return 0;
        if (sim_node->function[1][0] != 2) return 0;
        if (sim_node->function[1][1] != 0) return 0;
    }
    else {
        if (sim_node->function[0][1] != 0) return 0;
        if (sim_node->function[1][0] != 0) return 0;
        if (sim_node->function[1][1] != 2) return 0;

    }
    return 1;
}

static void 
sim_eval_po(sim_nodes, uid)
sim_node_t *sim_nodes;
int uid;
{
    register int result;
    sim_node_t *sim_node = &sim_nodes[uid];
    int input_id = sim_node->input_uids[0];

    result = sim_nodes[input_id].value;
    result &= (sim_node->and_input_masks[0] & sim_node->and_output_mask);
    result |= (sim_node->or_input_masks[0] | sim_node->or_output_mask);
    sim_node->value = result;
}

static void 
sim_eval_pi(sim_nodes, uid)
sim_node_t *sim_nodes;
int uid;
{
    register int result;
    sim_node_t *sim_node = &sim_nodes[uid];

    result = sim_node->value;
    result &= sim_node->and_output_mask;
    result |= sim_node->or_output_mask;
    sim_node->value = result;
}

static void 
sim_eval_inverter(sim_nodes, uid)
sim_node_t *sim_nodes;
int uid;
{
    int input_id;
    register int result;
    sim_node_t *sim_node = &sim_nodes[uid];

    input_id = sim_node->input_uids[0];
    result = sim_nodes[input_id].value;
    result &= sim_node->and_input_masks[0];
    result |= sim_node->or_input_masks[0];
    result = ~ result;
    result &= sim_node->and_output_mask;
    result |= sim_node->or_output_mask;
    sim_node->value = result;
}

static void 
sim_eval_nand2(sim_nodes, uid)
sim_node_t *sim_nodes;
int uid;
{
    int id0, id1;
    register int result, result0, result1;
    sim_node_t *sim_node = &sim_nodes[uid];

    id0 = sim_node->input_uids[0];
    result0 = sim_nodes[id0].value;
    result0 &= sim_node->and_input_masks[0];
    result0 |= sim_node->or_input_masks[0];
    id1 = sim_node->input_uids[1];
    result1 = sim_nodes[id1].value;
    result1 &= sim_node->and_input_masks[1];
    result1 |= sim_node->or_input_masks[1];
    result = ~ result0 | ~ result1 | sim_node->or_output_mask;
    sim_node->value = (result & sim_node->and_output_mask);
}
  
static void 
sim_eval_node(sim_nodes, uid)
sim_node_t *sim_nodes;
int uid;
{
    register int i, j;
    int input_id;
    register int result, and_result;
    register sim_node_t *sim_node = &sim_nodes[uid];

    for (j = 0; j < sim_node->n_inputs; j++) {
        input_id = sim_node->input_uids[j];
        result = sim_nodes[input_id].value;
        result &= sim_node->and_input_masks[j];
        result |= sim_node->or_input_masks[j];
        sim_node->inputs[1][j] = result;
        sim_node->inputs[0][j] = ~result;
    }
    result = sim_node->or_output_mask;
    for (i = 0; i < sim_node->n_cubes; i++) {
        and_result = sim_node->and_output_mask;
        for (j = 0; j < sim_node->n_inputs; j++) {
            and_result &= sim_node->inputs[sim_node->function[i][j]][j];
        }
        result |= and_result;
    }
    sim_node->value = result;
}
