#include "mis.h"
#include "bdd_int.h"

static bdd_t *
bdd_get_ith_variable(bdd,i)
bdd_manager *bdd;
unsigned int i;
{
    bdd_node *n;
    bdd_t *f;
    
    n = bdd_find_or_add(bdd->table,i,BDD_ONE(bdd),BDD_ZERO(bdd));
    bdd_ref(n);
    if (i >= bdd->vars) {
        bdd->vars = i + 1;
    }
    f = ALLOC(bdd_t,1);
    f->node = n;
    f->bdd = bdd;
    return(f);
}

/*
 * Returns a bdd formula of the form (v,1,0), where 'v' is a unique new
 * variable.
 *
 * The reference count for the new variable is initially 1.
 */
bdd_t *
bdd_create_variable(bdd)
bdd_manager *bdd;
{
    return(bdd_get_ith_variable(bdd,bdd->vars));
}


/*
 * If the BDDs for the primary inputs in the order list of this bdd manager
 * are NIL or for other bdd_managers, change them so that they are for this
 * bdd_manager.
 */
static void
bdd_register_order(bdd,locality)
bdd_manager *bdd;
int locality;
{
    array_t *iton,*order;
    int size,i;
    unsigned int count;
    node_t *node;
    bdd_t *f;
    st_table *local;
    
    iton = bdd->iton;
    order = bdd->order;
    size = array_n(order);

    count = 0;
    if (locality == 0) {
        for (i = 0; i < size; i++) {
	    node = array_fetch(node_t *,order,i);
	    if (node_function(node) == NODE_PI) {
	        f = bdd_at_node(node);
		if (f == NIL(bdd_t) || f->bdd != bdd) {
		    bdd_set_at_node(node,bdd_get_ith_variable(bdd,count));
		    array_insert(node_t *,iton,count,node);
		}
		count++;
	    }
	}
    }
    else {
	if (bdd->local_order == NIL(st_table)) {
	    bdd->local_order = local = st_init_table(st_ptrcmp,st_ptrhash);
	    for (i = 0; i < size; i++) {
	        node = array_fetch(node_t *,order,i);
		if (node_function(node) != NODE_PO) {
		    f = bdd_get_ith_variable(bdd,count);
		    array_insert(node_t *,iton,count,node);
		    (void) st_insert(local,(char *) node,(char *) f);
		    count++;
		}
	    }
	}
    }
}

/*
 * Ands together all the elements of the given cube.
 *
 * When done with temporary result n, throws out temporary result n-1
 * Result comes back with non-zero ref count.
 */
static bdd_node *
solve_row(node,literals,row,bdd,get_bdd)
node_t *node;
int literals,row;
bdd_manager *bdd;
bdd_t *(*get_bdd)();
{
    int i,complement;
    bdd_node *t,*cur,*last;
    node_cube_t cube = node_get_cube(node,row);
    bdd_node *zero = BDD_ZERO(bdd);
    bdd_table *table = bdd->table;

    last = NIL(bdd_node);
    for (i = 0; i < literals; i++) {
	complement = 0;
	switch (node_get_literal(cube,i)) {
	    case ZERO:
	        complement = 1;
	    case ONE:
	        t = (*get_bdd)(node_get_fanin(node,i))->node;
		if (complement != 0) {
		    t = BDD_NOT(t);
		}
		if (last == NIL(bdd_node)) {
		    last = t;
		    bdd_ref(last);
		}
		else {
		    cur = ITE(bdd,last,t,zero);
		    bdd_ref(cur);
		    bdd_recursive_deref(last,table);
		    last = cur;
		}
	}
    }
    return(last);
}

/*
 * Recursive descent global bdd constructor.
 */
static bdd_t *
do_construct(node,bdd)
node_t *node;
bdd_manager *bdd;
{
    bdd_t *f;
    node_function_t func;
    bdd_node *last,*cur,*row,*one;
    bdd_table *table;
    int literals,i,j;
    node_t *fi;
    
    f = bdd_at_node(node);
    if (f != NIL(bdd_t)) {
        if (f->bdd == bdd) {
	    return(f);
	}
    }

    literals = node_num_fanin(node);
    if (literals == 0) {
        func = node_function(node);
	if (func == NODE_0) {
	    f = bdd_zero(bdd);
	}
	else if (func == NODE_1) {
	    f = bdd_one(bdd);
	}
	else {
	    f = bdd_create_variable(bdd);
	    array_insert(node_t *,bdd->iton,f->node->id,node);
	}
	bdd_set_at_node(node,f);
	return(f);
    }
    if (literals == 1 && node_function(node) == NODE_PO) {
        f = bdd_dup(do_construct(node_get_fanin(node,0),bdd));
	bdd_set_at_node(node,f);
	return(f);	
    }	

    foreach_fanin (node,i,fi) {
        (void) do_construct(fi,bdd);
    }

    one = BDD_ONE(bdd);
    table = bdd->table;
    j = node_num_cube(node);

/* OR together products */

    for (i = 0; i < j; i++) {
        row = solve_row(node,literals,i,bdd,bdd_at_node);
	if (i == 0) {
	    cur = last = row;
	}
	else {
	    cur = ITE(bdd,last,one,row);
	    bdd_ref(cur);
	    bdd_recursive_deref(last,table);
	    bdd_recursive_deref(row,table);
	    last = cur;
	}
    }

    f = ALLOC(bdd_t,1);
    f->bdd = bdd;
    f->node = cur;
    bdd_set_at_node(node,f);

    return(f);    
}    

static st_table *bdd_local;	/* what a gross way of doing it */

static bdd_t *
get_local(node)
node_t *node;
{
    bdd_t *f;

    (void) st_lookup(bdd_local,(char *) node,(char **) &f);
    return(f);
}

static bdd_t *
do_local_construct(node,bdd)
node_t *node;
bdd_manager *bdd;
{
    int literals,i,j;
    node_function_t func;
    bdd_t *f;
    st_table *fanins;

    bdd_node *one,*cur,*row,*last;
    bdd_table *table;

    fanins = bdd->local_order;

    literals = node_num_fanin(node);
    if (literals == 0) {
        func = node_function(node);
	if (func == NODE_0) {
	    return(bdd_zero(bdd));
	}
	if (func == NODE_1) {
	    return(bdd_one(bdd));
	}
	(void) st_lookup(fanins,(char *) node,(char **) &f);
	return(bdd_dup(f));		/* so user can't delete original */
    }
    if (literals == 1 && node_function(node) == NODE_PO) {
        return(do_local_construct(node_get_fanin(node,0),bdd));
    }

    bdd_local = fanins;

    one = BDD_ONE(bdd);
    table = bdd->table;
    j = node_num_cube(node);

    for (i = 0; i < j; i++) {
        row = solve_row(node,literals,i,bdd,get_local);
	if (i == 0) {
	    cur = last = row;
	}
	else {
	    cur = ITE(bdd,last,one,row);
	    bdd_ref(cur);
	    bdd_recursive_deref(last,table);
	    bdd_recursive_deref(row,table);
	    last = cur;
	}
    }

    f = ALLOC(bdd_t,1);
    f->bdd = bdd;
    f->node = cur;
    return(f);    
}
	

bdd_t *
bdd_do_construct(node,bdd,locality)
node_t *node;
bdd_manager *bdd;
int locality;
{
    bdd_register_order(bdd,locality);

    if (locality == 0) {
	return(do_construct(node,bdd));
    }
    return(do_local_construct(node,bdd));
}
