#include "mis.h"
#include "bdd_int.h"

#define BDD_MEM_CHUNK		1000

int bdd_garbage_collect();

/*
 * Faster storage allocation for bdd_nodes in the table.
 * The first 4 bytes of a chunk contain a pointer to the next block, and
 *  the rest contains BDD_MEM_CHUNK spaces for bdd_nodes.
 */
static bdd_node *
bdd_alloc_node(table)
bdd_table *table;
{
    int i;
    bdd_node **mem,*list,*node;

    if (table->next_free == NIL(bdd_node)) {	/* memory is all full */
        mem = (bdd_node **) malloc(BDD_MEM_CHUNK * sizeof(bdd_node) +
							sizeof(bdd_node *));
	if (mem == NIL(bdd_node *)) {
	    if (bdd_garbage_collect(table) == 0) {
	        fail("bdd_alloc_node: out-O-memory");
	    }
	}
	else {
#if BDD_STATS
	    memused += BDD_MEM_CHUNK * sizeof(bdd_node) + sizeof(bdd_node *);
#endif
	    mem[0] = (bdd_node *) table->memory_list;
	    table->memory_list = mem;

	    list = (bdd_node *) &mem[1];

	    i = 0;
	    do {
	        list[i].next = &list[i + 1];
	    } while (++i < BDD_MEM_CHUNK - 1);

	    list[BDD_MEM_CHUNK - 1].next = NIL(bdd_node);

	    table->next_free = list;
	}
    }
#if BDD_STATS
    nodesallocated++;
#endif
    node = table->next_free;
    table->next_free = node->next;
    return(node);
}

/*
 * Adds this to head of free list.
 * Does not deallocate memory chunks that become free.
 */
static void
bdd_dealloc_node(table,node)
bdd_table *table;
bdd_node *node;
{
    if (table->next_free == NIL(bdd_node)) {
        fail("bdd_dealloc_node: calling free before anything added");
    }
#if BDD_STATS
    nodesfreed++;
#endif
    node->next = table->next_free;
    table->next_free = node;
}

bdd_table *
bdd_init_table()
{
    bdd_table *unique = ALLOC(bdd_table,1);
    int i;

    unique->slots = BDD_UNIQUE_SLOTS;
    unique->keys = 0;
    unique->max_keys = unique->slots * MAX_TABLE_DENSITY;
    unique->dead = 0;
    unique->cache = ALLOC(bdd_cache,BDD_UNIQUE_SLOTS);
    unique->nodelist = ALLOC(bdd_node *,BDD_UNIQUE_SLOTS);

    for (i = BDD_UNIQUE_SLOTS - 1; i >= 0; i--) {
        unique->nodelist[i] = NIL(bdd_node);
	unique->cache[i].data = NIL(bdd_node);
    }

    unique->memory_list = NIL(bdd_node *);
    unique->next_free = NIL(bdd_node);

    return(unique);
}

void
bdd_free_table(table)
bdd_table *table;
{
    bdd_node **next;
    bdd_node **memlist = table->memory_list;

    while (memlist != NIL(bdd_node *)) {
        next = (bdd_node **) memlist[0];	/* link to next block */
	FREE(memlist);
	memlist = next;
    }
    table->next_free = NIL(bdd_node);
    table->memory_list = NIL(bdd_node *);
    
    FREE(table->nodelist);
    FREE(table->cache);
    FREE(table);
}

/*
 * Returns the next prime >= p
 */
static unsigned int
prime(p)
unsigned int p;
{
    int i,prime;

    p--;
    do {
        p++;
        if (p & 1) {
	    prime = 1;
	    i = 3;
	    while (i * i <= p) {
	        if (p % i == 0) {
		    prime = 0;
		    break;
		}
		i += 2;
	    }
	}
	else {
	    prime = 0;
	}
    } while (!prime);
    return(p);
}

static void
bdd_rehash(table)
bdd_table *table;
{
    unsigned int slots,oldslots;
    int i,pos;
    bdd_node **nodelist,**oldnodelist;
    bdd_node *node,*next;
    bdd_cache *cache,*oldcache,*entry,*old;

    oldslots = table->slots;
    oldnodelist = table->nodelist;
    oldcache = table->cache;

    table->slots = slots = prime(BDD_GROW_FACTOR * table->slots);
    table->max_keys = slots * MAX_TABLE_DENSITY;
#if BDD_STATS
    (void) fprintf(misout,"rehashing: new size %d\n",slots);
    
    memused += (slots - oldslots) * (sizeof(bdd_node *) + sizeof(bdd_cache));
#endif

    table->nodelist = nodelist = ALLOC(bdd_node *,slots);
    table->cache = cache = ALLOC(bdd_cache,slots);
    for (i = slots - 1; i >= 0; i--) {
        nodelist[i] = NIL(bdd_node);
	cache[i].data = NIL(bdd_node);
    }

    for (i = oldslots - 1; i >= 0; i--) {
        node = oldnodelist[i];

	while (node != NIL(bdd_node)) {
	    next = node->next;
	    pos = bdd_hash(node->id,(unsigned int) node->T,
	    		   (unsigned int) node->E) % slots;
	    node->next = nodelist[pos];
	    nodelist[pos] = node;
	    node = next;
	}

	old = &oldcache[i];
	pos = bdd_hash((unsigned int) old->f,(unsigned int) old->g,
			(unsigned int) old->h) % slots;
	entry = &cache[pos];
	entry->f = old->f;
	entry->g = old->g;
	entry->h = old->h;
	entry->data = old->data;	
    }

    FREE(oldnodelist);
    FREE(oldcache);
}

static int
bdd_garbage_collect(table)
bdd_table *table;
{
    bdd_cache *cache = table->cache;
    bdd_node **nodelist = table->nodelist;
    int i,deleted;
    bdd_cache *c;
    bdd_node *node,*last,*next;

    if (table->dead == 0) {
        return(0);
    }

#if BDD_STATS
    (void) fprintf(misout,"garbage collecting (%d dead out of %d)\n",
    		   table->dead,table->keys);
#endif
    deleted = 0;

    for (i = table->slots - 1; i >= 0; i--) {
        node = nodelist[i];
	last = NIL(bdd_node);
	while (node != NIL(bdd_node)) {
	    next = node->next;
	    if (node->ref == 0) {
	        deleted++;
	        bdd_dealloc_node(table,node);
		if (last != NIL(bdd_node)) {
		    last->next = next;
		}
		else {
		    table->nodelist[i] = next;
		}
	    }
	    else {
	        last = node;
	    }	    
	    node = next;
	}

        c = &cache[i];
	if (c->data == NIL(bdd_node)) {
	    continue;
	}
	if (BDD_REGULAR(c->f)->ref == 0 || BDD_REGULAR(c->g)->ref == 0 ||
	      BDD_REGULAR(c->h)->ref == 0 || BDD_REGULAR(c->data)->ref == 0) {
	    c->data = NIL(bdd_node);
	}
    }

    if (deleted != table->dead) {
        fail("bdd_garbage_collect: dead count != deleted");
    }

    table->keys -= deleted;
    table->dead = 0;
    return(deleted);
}    

/*
 * Does not modify the reference count of whatever is returned.  A newly
 *  created node comes back with a reference count 0.
 * For a newly created node, increments the reference counts of what E
 *  points to.
 *
 * When called, T should always be a regular pointer.
 * A regular pointer is always returned.
 */
bdd_node *
bdd_find_or_add(table,id,T,E)
bdd_table *table;
unsigned int id;
bdd_node *T,*E;
{
    int pos;
    bdd_node **nodelist;
    bdd_node *looking;

    if (table->keys > table->max_keys) {

        if (table->dead * GC_PERCENT > table->keys) { 	/* too many dead */
    	    /*
	     * E possibly has a ref count of 0.  Must ref it, so garbage
	     * collection will not think it's dead.
	     */
	    bdd_ref(E);
	    (void) bdd_garbage_collect(table);
	    bdd_deref(E); 
	}
	else {
	    bdd_rehash(table);
	}
    }	    

    pos = bdd_hash(id,(unsigned int) T,(unsigned int) E) % table->slots;
    nodelist = table->nodelist;
    looking = nodelist[pos];

    while (looking != NIL(bdd_node)) {
        if (looking->id == id && looking->T == T && looking->E == E) {
	    if (looking->ref == 0) {

		/*
		 * E may have a ref count of 0.
		 *
		 * If during bdd_reclaim it finds E, it will try to bring E
		 * back from the dead if its ref count is 0, even though E
		 * really isn't dead.
		 */

	        bdd_ref(E);
	        bdd_reclaim(looking,table);
		bdd_deref(E);
	    }
	    return(looking);
	}
	looking = looking->next;
    }

    table->keys++;

	/*
	 * Ref T & E before calling bdd_alloc_node, because it may garbage
	 * collect, and then T & E would be lost.
	 */
    bdd_ref(T);
    bdd_ref(E);
    looking = bdd_alloc_node(table);
    looking->ref = 0;
    looking->id = id;
    looking->T = T;
    looking->E = E;
    looking->next = nodelist[pos];
    nodelist[pos] = looking;

    return(looking);
}

