#include "mis.h"
#include "bdd_int.h"

/*
 * Frees the given bdd_t.
 * Don't call with BDDs assigned to a node -- use bdd_node_free.
 */
void
bdd_free(f)
bdd_t *f;
{
    if (f != NIL(bdd_t)) {
        bdd_recursive_deref(f->node,f->bdd->table);
	f->node = NIL(bdd_node);
	FREE(f);
    }
}

/*
 * Frees the BDD associated with this node.
 */
void
bdd_node_free(n)
node_t *n;
{
    bdd_set_at_node(n,NIL(bdd_t));		/* side effect: delete old one */
}

/*
 * Returns a copy of the given bdd.
 */
bdd_t *
bdd_dup(f)
bdd_t *f;
{
    bdd_t *g;

    if (f == NIL(bdd_t)) {
        return(NIL(bdd_t));
    }

    g = ALLOC(bdd_t,1);
    g->bdd = f->bdd;
    bdd_ref(g->node = f->node);
    return(g);
}

/*
 * Returns a new copy of the 'zero' constant.
 */
bdd_t *
bdd_zero(bdd)
bdd_manager *bdd;
{
    bdd_t *f;

    if (bdd == NIL(bdd_manager)) {
        fail("bdd_zero: bad bdd manager");
    }
    f = ALLOC(bdd_t,1);
    f->bdd = bdd;
    bdd_ref(f->node = BDD_ZERO(bdd));
    return(f);
}

/*
 * Returns a new copy of the 'one' constant.
 */
bdd_t *
bdd_one(bdd)
bdd_manager *bdd;
{
    bdd_t *f;

    if (bdd == NIL(bdd_manager)) {
        fail("bdd_one: bad bdd manager");
    }
    f = ALLOC(bdd_t,1);
    f->bdd = bdd;
    bdd_ref(f->node = BDD_ONE(bdd));
    return(f);
}

int			/* whether the two bdds are the same */
bdd_equal(f,g)
bdd_t *f,*g;
{
    if (f == NIL(bdd_t) || g == NIL(bdd_t)) {
        fail("bdd_equal: invalid BDD");
    }
    if (f->bdd != g->bdd) {
        fail("bdd_equal: different bdd managers");
    }
    return(f->node == g->node);
}

/*
 * Returns the not of the given bdd.
 */
bdd_t *
bdd_not(f)
bdd_t *f;
{
    bdd_t *g;

    if (f == NIL(bdd_t)) {
        fail("bdd_not: invalid BDD");
    }
    g = ALLOC(bdd_t,1);
    g->bdd = f->bdd;
    bdd_ref(g->node = BDD_NOT(f->node));
    return(g);
}

bdd_t *
bdd_and(f,g,p1,p2)		/* F * G  =  ITE(F,G,0) */
bdd_t *f,*g;
int p1,p2;
{
    bdd_t *h = ALLOC(bdd_t,1);
    bdd_manager *bdd;
    bdd_node *n1,*n2;

    if (f == NIL(bdd_t) || g == NIL(bdd_t)) {
        fail("bdd_and: invalid BDD");
    }
    bdd = f->bdd;
    if (bdd != g->bdd) {
        fail("bdd_and: different bdd managers");
    }
    h->bdd = bdd;

    n1 = (p1 == 0) ? BDD_NOT(f->node) : f->node;
    n2 = (p2 == 0) ? BDD_NOT(g->node) : g->node;

    bdd_ref(h->node = ITE(bdd,n1,n2,BDD_ZERO(bdd)));

    return(h);
}

bdd_t *
bdd_or(f,g,p1,p2)		/* F + G  =  ITE(F,1,G) */
bdd_t *f,*g;
int p1,p2;
{
    bdd_t *h = ALLOC(bdd_t,1);
    bdd_manager *bdd;
    bdd_node *n1,*n2;
    
    if (f == NIL(bdd_t) || g == NIL(bdd_t)) {
        fail("bdd_or: invalid BDD");
    }
    bdd = f->bdd;
    if (bdd != g->bdd) {
        fail("bdd_or: different bdd managers");
    }
    h->bdd = bdd;

    n1 = (p1 == 0) ? BDD_NOT(f->node) : f->node;
    n2 = (p2 == 0) ? BDD_NOT(g->node) : g->node;

    bdd_ref(h->node = ITE(bdd,n1,BDD_ONE(bdd),n2));

    return(h);
}

bdd_t *
bdd_xor(f,g)		/* F xor G  =  ITE(F,!G,G) */
bdd_t *f,*g;
{
    bdd_t *h = ALLOC(bdd_t,1);
    bdd_manager *bdd = f->bdd;

    if (bdd != g->bdd) {
        fail("bdd_xor: different bdd managers");
    }
    h->bdd = bdd;
    bdd_ref(h->node = ITE(bdd,f->node,BDD_NOT(g->node),g->node));

    return(h);
}

bdd_t *
bdd_xnor(f,g)		/* !(F xor G)  =  ITE(F,G,!G) */
bdd_t *f,*g;
{
    bdd_t *h = ALLOC(bdd_t,1);
    bdd_manager *bdd = f->bdd;

    if (bdd != g->bdd) {
        fail("bdd_xnor: different bdd managers");
    }
    h->bdd = bdd;
    bdd_ref(h->node = ITE(bdd,f->node,g->node,BDD_NOT(g->node)));

    return(h);
}

int
bdd_leq(f,g,p1,p2)		/* f implies g  <=>  ITE_constant(F,G,1) = 1 */
bdd_t *f,*g;
int p1,p2;			/* phases */
{
    bdd_node *n1,*n2;
    bdd_manager *bdd;

    if (f == NIL(bdd_t) || g == NIL(bdd_t)) {
        fail("bdd_leq: invalid BDD");
    }
    bdd = f->bdd;
    if (bdd != g->bdd) {
        fail("bdd_leq: different bdd managers");
    }
    n1 = (p1 == 0) ? BDD_NOT(f->node) : f->node;
    n2 = (p2 == 0) ? BDD_NOT(g->node) : g->node;
    return(ITE_constant(bdd,n1,n2,BDD_ONE(bdd)) == BDD_ONE(bdd));
}

int
bdd_is_tautology(f,phase)
bdd_t *f;
int phase;
{
    if (f == NIL(bdd_t)) {
        fail("bdd_is_tautology: invalid BDD");
    }
    if(phase==0)return(f->node == BDD_ZERO(f->bdd));
    else 	return(f->node == BDD_ONE(f->bdd));
}

bdd_t *
bdd_then(f)
bdd_t *f;
{
    bdd_t *g = ALLOC(bdd_t,1);
    bdd_node *n,*nr;
    
    if (f == NIL(bdd_t)) {
        fail("bdd_then: invalid BDD");
    }
    n = f->node;
    nr = BDD_REGULAR(n);
    if (nr == BDD_ONE(f->bdd)) {
        fail("bdd_then: constant BDD");
    }
    bdd_ref(g->node = (BDD_IS_COMPLEMENT(n) == 1) ? BDD_NOT(nr->T) : n->T);
    g->bdd = f->bdd;
    return(g);
}

bdd_t *
bdd_else(f)
bdd_t *f;
{
    bdd_t *g = ALLOC(bdd_t,1);
    bdd_node *n,*nr;
    
    if (f == NIL(bdd_t)) {
        fail("bdd_else: invalid BDD");
    }
    n = f->node;
    nr = BDD_REGULAR(n);
    if (nr == BDD_ONE(f->bdd)) {
        fail("bdd_else: constant BDD");
    }
    bdd_ref(g->node = (BDD_IS_COMPLEMENT(n) == 1) ? BDD_NOT(nr->E) : n->E);
    g->bdd = f->bdd;
    return(g);
}

bdd_t *
bdd_top_var(f)
bdd_t *f;
{
    bdd_manager *bdd = f->bdd;
    bdd_t *g;
    bdd_node *n,*one,*zero;

    if (f == NIL(bdd_t)) {
        fail("bdd_top_var: invlaid BDD");
    }
    n = f->node;
    if (n == (one = BDD_ONE(bdd))) {
        return(bdd_one(bdd));
    }
    if (n == (zero = BDD_ZERO(bdd))) {
        return(bdd_zero(bdd));
    }
    n = bdd_find_or_add(bdd->table,BDD_REGULAR(n)->id,one,zero);
    g = ALLOC(bdd_t,1);
    bdd_ref(g->node = n);
    g->bdd = bdd;
    return(g);
}
    

#define bang(f) ((BDD_IS_COMPLEMENT(f)) ? '!' : ' ')

static void
bp2(f,one,t)
bdd_node *f,*one;
st_table *t;
{
    bdd_node *g,*n;
    int T,E;
    
    if (f == NIL(bdd_node)) {
        return;
    }
    g = BDD_REGULAR(f);
    if (g == one) {
        (void) fprintf(misout,"ID =  %d\n",f == one);
	return;
    }
    if (st_is_member(t,(char *) g) == 1) {
        return;
    }
    (void) st_insert(t,(char *) g,NIL(char));
#if BDD_STATS
    (void) fprintf(misout,"ID = %c0x%x\tindex = %d\tr = %d\t",bang(f),
    		(int) g / sizeof(bdd_node),g->id,g->ref);
#else
    (void) fprintf(misout,"ID = %c0x%x\tindex = %d\t",bang(f),
    		(int) g / sizeof(bdd_node),g->id);
#endif
    n = g->T;
    if (n == one) {
        (void) fprintf(misout,"T =  1\t\t");
	T = 1;
    }
    else {
        (void) fprintf(misout,"T =  0x%x\t",(int) n / sizeof(bdd_node));
	T = 0;
    }

    n = g->E;
    if (BDD_REGULAR(n) == one) {
        (void) fprintf(misout,"E =  %d\n",(n == one));
	E = 1;
    }
    else {
        (void) fprintf(misout,"E = %c0x%x\n",bang(n),
		    (int) n / sizeof(bdd_node));
	E = 0;
    }
    if (E == 0) {
        bp2(BDD_REGULAR(g->E),one,t);
    }
    if (T == 0) {
        bp2(BDD_REGULAR(g->T),one,t);
    }
}

#if BDD_STATS
bddp(f,bdd)
bdd_node *f;
bdd_manager *bdd;
{
    st_table *table = st_init_table(st_ptrcmp,st_ptrhash);

    bp2(f,BDD_ONE(bdd),table);
    st_free_table(table);
    (void) fputc('\n',misout);
}
#endif

void
bdd_print(f)
bdd_t *f;
{
    array_t *iton;
    st_table *table;
    int size,j;
    
    if (f == NIL(bdd_t)) {
        return;
    }

    iton = f->bdd->iton;
    size = array_n(iton);

    for (j = 0; j < size; j++) {
        (void) fprintf(misout,"\tindex %d is %s\n",j,
		node_name(array_fetch(node_t *,iton,j)));
    }

    table = st_init_table(st_ptrcmp,st_ptrhash);
    bp2(f->node,BDD_ONE(f->bdd),table);
    st_free_table(table);

    (void) fputc('\n',misout);
}
    
static int
bdd_size_int(n,tab)
bdd_node *n;
st_table *tab;
{
    if (n == NIL(bdd_node)) {
        return(0);
    }
    n = BDD_REGULAR(n);

    if (st_is_member(tab,(char *) n) == 1) {
        return(0);
    }
    (void) st_insert(tab,(char *) n,NIL(char));
    return(1 + bdd_size_int(n->T,tab) + bdd_size_int(n->E,tab));
}

int
bdd_size(f)
bdd_t *f;
{
    bdd_node *n;
    bdd_manager *bdd;
    int i;
    st_table *table;

    if (f == NIL(bdd_t)) {
        fail("bdd_size: illegal BDD");
    }

    table = st_init_table(st_ptrcmp,st_ptrhash);

    n = f->node;
    bdd = f->bdd;
    if (n == BDD_ONE(bdd)) {
	st_free_table(table);
        return(1);
    }
    if (n == BDD_ZERO(bdd)) {
	st_free_table(table);
        return(0);
    }
    i = bdd_size_int(n,table);
    st_free_table(table);

    /* add one because 0 will not be distinguished from 1 in bdd_size_int */

    return(i + 1);
}

bdd_t *
bdd_at_node(node)
node_t *node;
{
    return((bdd_t *) node->bdd);
}

/*
 * Automatically frees old bdd if node gets assigned new one.
 */
void
bdd_set_at_node(node,bdd)
node_t *node;
bdd_t *bdd;
{
    bdd_t *f;
    
    f = (bdd_t *) node->bdd;
    
    if (f != bdd) {
	node->bdd = (char *) bdd;
        bdd_free(f);
    }
}    

