#include "mis.h"
#include "util.h"
#include "bdd_int.h"
#include "vov.h"

#if BDD_STATS
int memused = 0;
int nodesallocated = 0;
int nodesfreed = 0;
int nodesdropped = 0;
int nodesreclaimed = 0;
int cachehits = 0;
int cachecollisions = 0;
int cacheinserts = 0;
#endif

static int
com_bdd_test(network)
network_t **network;
{
    array_t *dfs = network_dfs(*network);

    (void) bdd_init(*network,dfs);

    bdd_print(bdd_constant(0,dfs));
    bdd_print(bdd_constant(1,dfs));
    
    return(0);
}

static int
com_bdd_create(network,argc,argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec,*order_list;
    int n,i,order_style,c,locality;
    node_t *node;
    bdd_t *f;
    long time;
    
    order_style = HEURISTIC;
    locality = 0;		/* global */
    getopt_reset();
    while ((c = getopt(argc,argv,"lfrh")) != EOF){
        switch (c) {
	    case 'f':
	    	order_style = FANIN;
		break;
	    case 'r':
	        order_style = RANDOM;
		break;
	    case 'h':
	        order_style = HEURISTIC;
		break;
	    case 'l':
	        locality = 1;		/* local */
		break;
	    default:
	        goto usage;
	}
    }
    node_vec = com_get_nodes(*network,argc - optind + 1,argv + optind - 1);

    n = array_n(node_vec);
    if (n == 0) {
        array_free(node_vec);
usage:
	(void) fprintf(miserr,"usage: bdd_create [-f] [-r] [-a] [-h] [-l] n1 n2 ...\n");
	return(1);
    }

    (void) fprintf(misout,"ordering nodes\n");

    if (locality == 1) {
        order_style = FANIN;	/* just pick an order */
    }
    switch (order_style) {
        case HEURISTIC:
	    order_list = order_nodes(node_vec,1);
	    break;
	case FANIN:
	    order_list = network_dfs(*network);
	    break;
	case RANDOM:
	    order_list = network_dfs(*network);
	    order_list = shuffle(order_list);
	    break;
    }

    (void) fprintf(misout,"constructing BDDs\n");
    time = util_cpu_time();
    for (i = 0; i < n; i++) {
        node = array_fetch(node_t *,node_vec,i);
	f = bdd_construct(node,order_list,locality);
	if (locality == 1) {
	    (void) fprintf(misout,"node: %s\n",node_name(node));
	    bdd_print(f);
	    bdd_free(f);
	}
    }
    (void) fprintf(misout,"CPU time used: %g sec\n",
    		(double) (util_cpu_time() - time) / 1000);
    array_free(order_list);
    array_free(node_vec);
    return(0);
}

static int
com_bdd_print(network,argc,argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec;
    int n,i;
    node_t *node;
    bdd_t *bdd;

    node_vec = com_get_nodes(*network,argc,argv);

    n = array_n(node_vec);
    if (n == 0) {
        array_free(node_vec);
	(void) fprintf(miserr,"usage: bdd_print n1 n2 ...\n");
	return(1);
    }

    for (i = 0; i < n; i++) {
        node = array_fetch(node_t *,node_vec,i);
	bdd = bdd_at_node(node);
	if (bdd != NIL(bdd_t)) {
	    (void) fprintf(misout,"node: %s\n",node_name(node));
	    bdd_print(bdd);
	}
    }
    array_free(node_vec);
    return(0);
}

static int
com_bdd_size(network,argc,argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec;
    int i,n;
    node_t *node;
    bdd_t *bdd;

    node_vec = com_get_nodes(*network,argc,argv);

    n = array_n(node_vec);
    if (n == 0) {
        array_free(node_vec);
	(void) fprintf(miserr,"usage: bdd_size n1 n1 ...\n");
	return(1);
    }
    for (i = 0; i < n; i++) {
        node = array_fetch(node_t *,node_vec,i);
	bdd = bdd_at_node(node);
	if (bdd != NIL(bdd_t)) {
	    (void) fprintf(misout,"node: %s, size %d\n",node_name(node),
			bdd_size(bdd));
	}
    }
    array_free(node_vec);
    return(0);
}

static int
com_bdd_implies(network,argc,argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec;
    node_t *n1,*n2;
    int p1,p2;

    if (argc != 5) {
        if (argc == 3) {
	    p1 = 1;
	    p2 = 1;
	}
	else {
usage:
	    (void) fprintf(miserr,
	    	"usage: bdd_implies n1 n2 [phase1 phase2]\n");
	    (void) fprintf(miserr,
	    	"\tIf phase1 & phase2 are left out they are assumed both 1\n");
	    return(1);
	}
    }
    else {
        p1 = atoi(argv[3]);
	p2 = atoi(argv[4]);
    }
    if ((p1 != 0 && p1 != 1) || (p2 != 0 && p2 != 1)) {
        goto usage;
    }
    node_vec = com_get_nodes(*network,3,argv);
    if (array_n(node_vec) != 2) {
        array_free(node_vec);
	goto usage;
    }
    n1 = array_fetch(node_t *,node_vec,0);
    n2 = array_fetch(node_t *,node_vec,1);

    array_free(node_vec);
    if (n1 == NIL(node_t) || n2 == NIL(node_t)) {
        goto usage;
    }
    (void) fprintf(misout,"%s set to %d %s ",node_name(n1),p1,
    		((bdd_leq(bdd_at_node(n1),bdd_at_node(n2),p1,p2) == 1) ?
			"forces" : "does not force"));
    (void) fprintf(misout,"%s to %d\n",node_name(n2),p2);
    return(0);
}

static int
com_bdd_cofactor(network,argc,argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec;
    node_t *n1,*n2;
    bdd_t *bdd;
    int stat;

    node_vec = com_get_nodes(*network,argc,argv);
    if (argc != 3 || array_n(node_vec) != 2) {
        stat = 1;
    }
    else {
        n1 = array_fetch(node_t *,node_vec,0);
	n2 = array_fetch(node_t *,node_vec,1);

	if (n1 == NIL(node_t) || n2 == NIL(node_t)) {
	    stat = 1;
	}
	else {
	    bdd = bdd_cofactor(bdd_at_node(n1),bdd_at_node(n2));
	    bdd_print(bdd);
	    bdd_free(bdd);
	    stat = 0;
	}
    }
    array_free(node_vec);
    return(stat);
}

static int
com_bdd_compose(network,argc,argv)
network_t **network;
int argc;
char **argv;
{
    array_t *node_vec;
    node_t *n1,*n2,*n3;
    bdd_t *bdd;
    int stat;

    node_vec = com_get_nodes(*network,argc,argv);
    if (argc != 4 || array_n(node_vec) != 3) {
        stat = 1;
    }
    else {
        n1 = array_fetch(node_t *,node_vec,0);
	n2 = array_fetch(node_t *,node_vec,1);
	n3 = array_fetch(node_t *,node_vec,2);

	if (n1 == NIL(node_t) || n2 == NIL(node_t) || n3 == NIL(node_t)) {
	    stat = 1;
	}
	else {
	    bdd = bdd_compose(bdd_at_node(n1),bdd_at_node(n2),bdd_at_node(n3));
	    bdd_print(bdd);
	    bdd_free(bdd);
	    stat = 0;
	}
    }
    array_free(node_vec);
    return(stat);
}

/*ARGSUSED*/
static int
com_bdd_verify(network,argc,argv)
network_t **network;
int argc;
char **argv;
{
    FILE *f1,*f2;
    network_t *net1,*net2;
    int stat;

    if (argc != 3) {
        return(1);
    }
    f1 = fopen(argv[1],"r");
    if (f1 == NIL(FILE)) {
        goto badfile;
    }
    VOVinput( VOV_UNIX_FILE, argv[1] );
    f2 = fopen(argv[2],"r");
    if (f2 == NIL(FILE)) {
        (void) fclose(f1);
badfile:
	(void) fprintf(miserr,"bad eqn input file\n");
	return(1);
    }
    VOVinput( VOV_UNIX_FILE, argv[2] );
    net1 = read_eqn(f1);
    net2 = read_eqn(f2);
    if (net1 == NIL(network_t) || net2 == NIL(network_t)) {
        (void) fprintf(miserr,"bad network\n");
        stat = 1;
    }
    else {
        (void) fprintf(misout,"%d\n",bdd_verify(net1,net2));
	stat = 0;
    }
    network_free(net1);
    network_free(net2);
    (void) fclose(f1);
    (void) fclose(f2);
    return(stat);
}

static void
bddAlloc(node)
node_t *node;
{
    node->bdd = NIL(char);
}

void
bddDealloc(node)
node_t *node;
{
    FREE(node->bdd);
}    

init_bdd()
{
    com_add_command("_bdd_test",com_bdd_test,0);
    com_add_command("_bdd_create",com_bdd_create,0);
    com_add_command("_bdd_print",com_bdd_print,0);
    com_add_command("_bdd_size",com_bdd_size,0);
    com_add_command("_bdd_implies",com_bdd_implies,0);
    com_add_command("_bdd_cofactor",com_bdd_cofactor,0);
    com_add_command("_bdd_compose",com_bdd_compose,0);
    com_add_command("_bdd_verify",com_bdd_verify,0);

    node_register_daemon(DAEMON_ALLOC,bddAlloc);
    node_register_daemon(DAEMON_FREE,bddDealloc);
}

end_bdd()
{
#if BDD_STATS
    (void) fprintf(misout,"memused: %d   nodes used: %d, freed: %d\n",memused,
    		   nodesallocated,nodesfreed);
    (void) fprintf(misout,"nodes dropped: %d, reclaimed: %d\n",
    		   nodesdropped,nodesreclaimed);
    (void) fprintf(misout,"cache hits: %d, inserts: %d, collisions: %d\n",
    		   cachehits,cacheinserts,cachecollisions);
    (void) fprintf(misout,"\nfinal nodes: %d, bytes/node: %g\n",
    		   nodesallocated - nodesfreed,
    		   (double) memused / (nodesallocated - nodesfreed));
#endif
}

