#include "mis.h"
#include "io_int.h"

/*
 *  bug list: 
 *    - fscanf() might read '\n' which we miss in our line number count
 *    - blif cannot describe a single gate which feeds multiple outputs
 */

static int read_node_list();
static int read_cover();
static node_t *read_input_node();
static node_t *read_output_node();
static int read_blif_delay();
static void read_blif_delay_cleanup();
static int read_blif_gate();

/* ### hack, reach in and grab this from the node package */
extern void node_replace_internal();


#define MAX_WORD  1024

#define GETC(c, fp)            \
    (c = getc(fp),             \
     c == '\n' ? read_lineno++ :     \
    (c == '\\' ?             \
        (c = getc(fp), c == '\n' ? (c = ' ', read_lineno++) : 0) : 0), \
    c)

#define UNGETC(c, fp)    if (c == '\n') read_lineno--; (void) ungetc(c, fp);

/*
 *  read_blif: read a blif description of a network.  
 *
 *  returns:
 *    -1  if end-of-file reached before a network has been read (not EOF !)
 *     0  if an error occured during reading
 *     1  if no errors
 *
 *  Always check error_string() for any error/warning messages
 *
 *  We return immediately on any fatal error with error_string() containing
 *  a somewhat useful error message.
 */

int 
read_blif(fp, network_p)
register FILE *fp;
network_t **network_p;
{
    register int c;
    int i, nlist, nin, nout, nterm, got_something;
    char word[MAX_WORD];
    node_t **list, *output;
    network_t *network;
    lsList po_list;

    /* Initialize the network */
    got_something = 0;
    network = network_alloc();
    read_filename_to_netname(network, read_filename);

    /* initialize the 'global' defaults */
    delay_set_default_parameter(network, DELAY_DEFAULT_DRIVE_RISE, DELAY_NOT_SET);
    delay_set_default_parameter(network, DELAY_DEFAULT_DRIVE_FALL, DELAY_NOT_SET);
    delay_set_default_parameter(network, DELAY_DEFAULT_ARRIVAL_RISE, DELAY_NOT_SET);
    delay_set_default_parameter(network, DELAY_DEFAULT_ARRIVAL_FALL, DELAY_NOT_SET);
    delay_set_default_parameter(network, DELAY_DEFAULT_REQUIRED_RISE, DELAY_NOT_SET);
    delay_set_default_parameter(network, DELAY_DEFAULT_REQUIRED_FALL, DELAY_NOT_SET);
    delay_set_default_parameter(network, DELAY_DEFAULT_OUTPUT_LOAD, DELAY_NOT_SET);

    /* we record the primary outputs on a special list because of the
     * difficulties of mapping blif into our network structure which includes
     * an extra 'PO' node 
     */
    po_list = lsCreate();

    while (GETC(c, fp) != EOF) {
    if (isspace(c)) {
        /* eat whitespace */
    } else if (c == '.') {
        if (read_word(fp, word, MAX_WORD) == EOF) {
        break;
        }

        if (strcmp(word, "names") == 0 || strcmp(word, "cover") == 0) {
        if (strcmp(word, "cover") == 0) {
            if (fscanf(fp, "%d %d %d", &nin, &nout, &nterm) != 3) {
            read_error(".cover requires nin, nout, nterm");
            return 0;
            }
            if (nout != 1) {
            read_error("only single-output .cover allowed");
            return 0;
            }
        }
        got_something = 1;
        nlist = read_node_list(fp, network, &list);
        if (nlist == 0) {
            read_error("Too few names following .names");
            return 0;
        }
        output = list[nlist-1];
        if (node_function(output) != NODE_UNDEFINED) {
            (void) sprintf(word, 
            "fatal: output '%s' is multiply defined", output->name);
            read_error(word);
            return 0;
        }

        i = read_cover(fp, output, list, nlist-1);
        if (i == EOF) {
            break;
        } else if (i == 0) {
            return 0;
        }

        } else if (strcmp(word, "gate") == 0) {
        got_something = 1;
        if (! read_blif_gate(fp, network)) {
            return 0;
        }

        } else if (strcmp(word, "model") == 0 || 
                        strcmp(word, "circuit") == 0) {
        got_something = 1;
        if (read_word(fp, word, MAX_WORD) == EOF) {
            break;
        }
        network_set_name(network, word);

        } else if (strcmp(word, "inputs") == 0) {
        got_something = 1;
        nlist = read_node_list(fp, network, &list);
        for(i = 0; i < nlist; i++) {
            if (node_function(list[i]) != NODE_UNDEFINED) {
            (void) sprintf(word, 
                "fatal: output '%s' is multiply defined", 
                list[i]->name);
            read_error(word);
            return 0;
            }
            network_change_node_type(network, list[i], PRIMARY_INPUT);
        }
        FREE(list);

        } else if (strcmp(word, "outputs") == 0) {
        got_something = 1;
        nlist = read_node_list(fp, network, &list);
        for(i = 0; i < nlist; i++) {
            lsNewEnd(po_list, (lsGeneric) list[i], LS_NH);
        }
        FREE(list);

        } else if (strcmp(word, "end") == 0) {
        /* skip to the end of this line ... */
        for( ; c != EOF && c != '\n'; GETC(c, fp)) {
        }
        break;

        } else {
        c = read_blif_delay(network, po_list, fp, word);
        if (c == -1) {
            /* echo line with unknown keyword */
            (void) fprintf(misout, ".%s", word);
            for(c = ' '; c != EOF && c != '\n'; GETC(c, fp)) {
            (void) putc(c, misout);
            }
            (void) putc('\n', misout);
        } else if (c == 0) {
            /* error parsing delay-related keyword */
            return 0;
        }
        }

    } else if (c == '#') {
        /* skip a comment line */ 
        for( ; c != EOF && c != '\n'; GETC(c, fp)) {
        }

    } else {
        /* ignore any line which is not a command line */
        for( ; c != EOF && c != '\n'; GETC(c, fp)) {
        (void) putc(c, misout);
        }
        (void) putc('\n', misout);
    }
    }

    if (! got_something) {
	lsDestroy(po_list, (void (*)()) 0);
	network_free(network);
	return -1;
    }

    read_check_io_list(network, po_list);
    read_hack_outputs(network, po_list);
    lsDestroy(po_list, (void (*)()) 0);

    if (! network_is_acyclic(network)) {
    read_error("read_blif: network contains a cycle");
    network_free(network);
    return 0;
    }

    read_blif_delay_cleanup(network);
    *network_p = network;
    return 1;
}

void
read_dc_blif(fp, network)
register FILE *fp;
network_t **network;
{
    int c,i;
    network_t *dc_network;
    network_t *net;
    char word[MAX_WORD];
    char *error;

    net= *network;
      while (GETC(c, fp) != EOF) {
    if (isspace(c)) {
    /* eat whitespace */
    } else if (c == '.') {
        if (read_word(fp, word, MAX_WORD) == EOF) {
            break;
        }
        if (strcmp(word, "exdc") == 0) {
            while (GETC(c, fp) != EOF && c != '\n' );
            i = read_blif(fp, &dc_network);
            if (i == EOF) {
		(void) fprintf(miserr, 
		"read_blif: no external don't care network found\n");
		network_free(net);
		net= 0;
            } else if (i == 1) {
		network_collapse(dc_network);
		net->dc_network = dc_network;
            }else{
		network_free(net);
		net= 0;
            }
            error = error_string();
            if (error[0] != '\0') {
            (void) fprintf(miserr, "External Don't Care Network \n");
            (void) fprintf(miserr, "%s", error);
            }
            break;
        }
    }
    }
}

static int
read_node_list(fp, network, list)
register FILE *fp;
network_t *network;
node_t ***list;
{
    register int c;
    int nlist, list_max;
    char word[MAX_WORD];
    node_t *node;

    list_max = 10;
    *list = ALLOC(node_t *, list_max);
    nlist = 0;
    do {
    /* Skip any leading white space */
    while (GETC(c, fp) != EOF && c != '\n' && isspace(c))
        ;

    /* Read another name (unless we're at end of the line) */
    if (c != EOF && c != '\n') {
        UNGETC(c, fp);
        c = read_word(fp, word, MAX_WORD);
        if (c != EOF) {
        node = read_find_or_create_node(network, word);
        if (nlist >= list_max) {
            list_max *= 2;
            *list = REALLOC(node_t *, *list, list_max);
        }
        (*list)[nlist++] = node;
        }
    }
    } while (c != EOF && c != '\n');
    return nlist;
}

/*
 *  read_word -- return a token from an input stream, skipping all whitespace
 *
 *  Return EOF when there are no more tokens on input stream.
 */

int 
read_word(fp, word, word_len)
register FILE *fp;
char *word;
int word_len;
{
    register int c, i;

    /* skip any leading white-space */
    while (GETC(c, fp) != EOF && isspace(c))
    ;

    if (c == EOF) {
    word[0] = '\0';
    return EOF;
    }
    i = 0;
    word[i++] = c;

    /* read until next white-space character */
    while (GETC(c, fp) != EOF && ! isspace(c)) {
    if (i < word_len-1) {
        word[i++] = c;
    }
    }
    if (c != EOF) {
    UNGETC(c, fp);
    }
    word[i] = '\0';
    return 0;
}

static int 
read_cover(fp, node, fanin, ninput)
register FILE *fp;
node_t *node;
node_t **fanin;
int ninput;
{
    register int i, c;
    register pset pdest;
    pset_family f;
    char word[MAX_WORD];
    int is_normal, is_compl;

    is_compl = 0;            /* records some table has "0" output */
    is_normal = 0;            /* records some table has "1" output */
    f = sf_new(8, ninput * 2);

    for (;;) {
    /* skip any leading white-space */
    while (GETC(c, fp) != EOF && isspace(c))
        ;
    if (c == EOF) break;            /* end table on end of file */

    UNGETC(c, fp);
    if (c == '.') break;            /* end table on a command */

    /* check for and ignore a comment */
    if (c == '#') {
        while (GETC(c, fp) != EOF && c != '\n')
        ;
        if (c == EOF) break;        /* end table on end of file */
        continue;
    }

    if (ninput > 0) {
        if (read_word(fp, word, MAX_WORD) == EOF) break;
        if (strlen(word) != ninput) goto bad_char;

        pdest = set_new(ninput*2);
        for(i = 0; i < ninput; i++) {
        switch (word[i]) {
        case '0':
            set_insert(pdest, 2*i);
            break;
        case '1':
            set_insert(pdest, 2*i+1);
            break;
        case '2':
        case '-':
            set_insert(pdest, 2*i);
            set_insert(pdest, 2*i+1);
            break;
        default:
            goto bad_char;
        }
        }

        if (read_word(fp, word, MAX_WORD) == EOF) goto bad_char;
        if (strcmp(word, "1") == 0) {
        if (is_compl) {
            read_error("must give F or R, but not both");
            return 0;
        }
        f = sf_addset(f, pdest);
        set_free(pdest);
        is_normal = 1;
        } else if (strcmp(word, "0") == 0) {
        if (is_normal) {
            read_error("must give F or R, but not both");
            return 0;
        }
        f = sf_addset(f, pdest);
        set_free(pdest);
        is_compl = 1;
        } else {
        goto bad_char;
        }

    } else {
        if (read_word(fp, word, MAX_WORD) == EOF) goto bad_char;
        if (strcmp(word, "1") == 0) {
        f->count = 1;        /* the '1' function */ 
        } else if (strcmp(word, "0") == 0) {
                        /* the '0' function */
        } else {
        goto bad_char;
        }
    }
    }

    node_replace_internal(node, fanin, ninput, f);
    if (is_compl) {
    node_replace(node, node_not(node));
    }
    node_scc(node);        /* make it scc-minimal, dup_free, etc. */
    return c == EOF ? EOF : 1;

bad_char:
    read_error("bad character in PLA table");
    return 0;
}

/*
 *  check a blif keyword for a delay-related keyword; returns -1
 *  if not delay related, 0 if delay-related and an error occured,
 *  or 1 if delay-related and no error.
 */

static int
read_blif_delay(network, po_list, fp, word)
network_t *network;
lsList po_list;
FILE *fp;
char word[];
{
    double br, bf, dr, df;
    double ar, af, rr, rf;
    double area, load, max_load, phase; 
    int c;
    node_t *node;

    if (strcmp(word, "area") == 0 || strcmp(word, "cost") == 0) {
    if (fscanf(fp, "%lf", &area) != 1) {
        read_error("error reading number after .area");
        return 0;
    }
    if (network->area_given) {
        read_error("area given twice for same model");
        return 0;
    }
    network->area_given = 1;
    network->area = area;
    return 1;

    } else if (strcmp(word, "delay") == 0) {
    node = read_input_node(fp, network, word);
    if (node == NIL(node_t)) {
        return 0;
    }
    if (read_word(fp, word, MAX_WORD) == EOF) {
        read_error("bad phase specification in .delay");
        return 0;
    }
    if (strcmp(word, "INV") == 0) {
        phase = DELAY_PHASE_INVERTING;
    } else if (strcmp(word, "NONINV") == 0) {
        phase = DELAY_PHASE_NONINVERTING;
    } else if (strcmp(word, "UNKNOWN") == 0) {
        phase = DELAY_PHASE_NEITHER;
    } else {
        read_error("bad phase specification in .delay");
        return 0;
    }
    if (fscanf(fp, "%lf %lf %lf %lf %lf %lf", 
        &load, &max_load, &br, &dr, &bf, &df) != 6) {
        read_error("need load, max_load, and 4 delay values in .delay");
        return 0;
    }
    delay_set_parameter(node, DELAY_BLOCK_RISE, br);
    delay_set_parameter(node, DELAY_BLOCK_FALL, bf);
    delay_set_parameter(node, DELAY_DRIVE_RISE, dr);
    delay_set_parameter(node, DELAY_DRIVE_FALL, df);
    delay_set_parameter(node, DELAY_INPUT_LOAD, load);
    delay_set_parameter(node, DELAY_MAX_INPUT_LOAD, max_load);
    delay_set_parameter(node, DELAY_PHASE, phase);
    return 1;

    } else if (strcmp(word, "input_arrival") == 0) {
    node = read_input_node(fp, network, word);
    if (node == NIL(node_t)) {
        return 0;
    }
    if (fscanf(fp, "%lf %lf", &ar, &af) != 2) {
        read_error("expect arrival time in .input_arrival");
        return 0;
    }
    delay_set_parameter(node, DELAY_ARRIVAL_RISE, ar);
    delay_set_parameter(node, DELAY_ARRIVAL_FALL, af);
    return 1;

    } else if (strcmp(word, "output_required") == 0) {
    node = read_output_node(fp, po_list, word);
    if (node == NIL(node_t)) {
        return 0;
    }
    if (fscanf(fp, "%lf %lf", &rr, &rf) != 2) {
        read_error("expect required time in .output_required");
        return 0;
    }
    delay_set_parameter(node, DELAY_REQUIRED_RISE, rr);
    delay_set_parameter(node, DELAY_REQUIRED_FALL, rf);
    return 1;

    } else if (strcmp(word, "input_drive") == 0) {
    node = read_input_node(fp, network, word);
    if (node == NIL(node_t)) {
        return 0;
    }
    if (fscanf(fp, "%lf %lf", &dr, &df) != 2) {
        read_error("expect drive values in .input_drive");
        return 0;
    }
    delay_set_parameter(node, DELAY_DRIVE_RISE, dr);
    delay_set_parameter(node, DELAY_DRIVE_FALL, df);
    return 1;

    } else if (strcmp(word, "output_load") == 0) {
    node = read_output_node(fp, po_list, word);
    if (node == NIL(node_t)) {
        return 0;
    }
    if (fscanf(fp, "%lf", &load) != 1) {
        read_error("expect load value in .output_load");
        return 0;
    }
    delay_set_parameter(node, DELAY_OUTPUT_LOAD, load);
    return 1;

    } else if (strcmp(word, "default_input_arrival") == 0) {
    if (fscanf(fp, "%lf %lf", &ar, &af) != 2) {
        read_error("expect arrival time in .default_input_arrival");
        return 0;
    }
    delay_set_default_parameter(network, DELAY_DEFAULT_ARRIVAL_RISE, ar);
    delay_set_default_parameter(network, DELAY_DEFAULT_ARRIVAL_FALL, af);
    return 1;

    } else if (strcmp(word, "default_output_required") == 0) {
    if (fscanf(fp, "%lf %lf", &rr, &rf) != 2) {
        read_error("expect required time in .default_output_required");
        return 0;
    }
    delay_set_default_parameter(network, DELAY_DEFAULT_REQUIRED_RISE, rr);
    delay_set_default_parameter(network, DELAY_DEFAULT_REQUIRED_FALL, rf);
    return 1;

    } else if (strcmp(word, "default_input_drive") == 0) {
    if (fscanf(fp, "%lf %lf", &dr, &df) != 2) {
        read_error("expect drive values in .default_input_drive");
        return 0;
    }
    delay_set_default_parameter(network, DELAY_DEFAULT_DRIVE_RISE, dr);
    delay_set_default_parameter(network, DELAY_DEFAULT_DRIVE_FALL, df);
    return 1;

    } else if (strcmp(word, "default_output_load") == 0) {
    if (fscanf(fp, "%lf", &load) != 1) {
        read_error("expect load value in .output_load");
        return 0;
    }
    delay_set_default_parameter(network, DELAY_DEFAULT_OUTPUT_LOAD, load);
    return 1;

    } else if (strcmp(word, "wire_load_slope") == 0) {
	if (fscanf(fp, "%lf", &load) != 1) {
	    read_error("expect load value in .wire_load_slope");
	    return(0);
	}
	delay_set_default_parameter(network, DELAY_WIRE_LOAD_SLOPE, load);
	return 1;

    } else if (strcmp(word, "wire") == 0 ) {
	/* discard any existing wire load table */
	delay_set_default_parameter(network, DELAY_ADD_WIRE_LOAD, -1.0);

	/* read the wire loads */
	do {
	    /* Skip any leading white space */
	    while (GETC(c, fp) != EOF && c != '\n' && isspace(c))
	    ;

	    /* Read another value (unless we're at end of the line) */
	    if (c != EOF && c != '\n') {
	    UNGETC(c, fp);
	    c = read_word(fp, word, MAX_WORD);
	    if (c != EOF) {
		load = atof(word);
		delay_set_default_parameter(network, DELAY_ADD_WIRE_LOAD, load);
	    }
	    }
	} while (c != EOF && c != '\n');
	return 1;
    }

    return -1;
}


static node_t *
read_input_node(fp, network, errmsg)
FILE *fp;
network_t *network;
char *errmsg;
{
    node_t *node;
    char word[MAX_WORD], error[MAX_WORD];

    if (read_word(fp, word, MAX_WORD) == EOF) {
    (void) sprintf(error, "missing name in .%s", errmsg);
    read_error(error);
    return NIL(node_t);
    }
    node = network_find_node(network, word);
    if (node == NIL(node_t) || node->type != PRIMARY_INPUT) {
    (void) sprintf(error, "node '%s' not defined as input node in .%s", 
        word, errmsg);
    read_error(error);
    return NIL(node_t);
    }
    return node;
}


static node_t *
read_output_node(fp, po_list, errmsg)
FILE *fp;
lsList po_list;
char *errmsg;
{
    node_t *node;
    lsGen gen;
    char word[MAX_WORD], error[MAX_WORD];

    if (read_word(fp, word, MAX_WORD) == EOF) {
    (void) sprintf(error, "missing name in .%s", errmsg);
    read_error(error);
    return NIL(node_t);
    }
    lsForeachItem(po_list, gen, node) {
    if (strcmp(node->name, word) == 0) {
        lsFinish(gen);
        return node;
    }
    }
    (void) sprintf(error, "node '%s' not defined as output node in .%s", 
    word, errmsg);
    read_error(error);
    return NIL(node_t);
}

static void
read_blif_delay_cleanup(network)
network_t *network;
{
    lsGen gen;
    node_t *node, *fanin;
    double value;

    /* copy any output information from its fanin to the output node */
    foreach_primary_output(network, gen, node) {
	fanin = node_get_fanin(node, 0);

	value = delay_get_parameter(fanin, DELAY_OUTPUT_LOAD);
	delay_set_parameter(node, DELAY_OUTPUT_LOAD, value);

	value = delay_get_parameter(fanin, DELAY_REQUIRED_RISE);
	delay_set_parameter(node, DELAY_REQUIRED_RISE, value);

	value = delay_get_parameter(fanin, DELAY_REQUIRED_FALL);
	delay_set_parameter(node, DELAY_REQUIRED_FALL, value);
    }
}

static int
read_blif_gate(fp, network)
FILE *fp;
network_t *network;
{
    register int c;
    char *actual, errmsg[MAX_WORD], word[MAX_WORD];
    array_t *formal_array, *actual_array;
    char **formal_data;
    node_t *node, **actual_data;
    library_t *library;
    lib_gate_t *gate;
    int i, tot;

    /* get the current library */
    library = lib_get_library();
    if (library == NIL(library_t)) {
    read_error("cannot process .gate without a current library");
    return 0;
    }

    /* read the gate name */
    c = read_word(fp, word, MAX_WORD);
    if (c == EOF) {
    read_error("missing gate name in .gate");
    return 0;
    }
    gate = lib_get_gate(library, word);
    if (gate == NIL(lib_gate_t)) {
    (void) sprintf(errmsg, "cannot find gate '%s' in the library", word);
    read_error(errmsg);
    return 0;
    }

    formal_array = array_alloc(char *, 10);
    actual_array = array_alloc(node_t *, 10);
    do {
    /* Skip any leading white space */
    while (GETC(c, fp) != EOF && c != '\n' && isspace(c))
        ;

    /* Read another token (unless we're at end of the line) */
    if (c != EOF && c != '\n') {
        UNGETC(c, fp);
        c = read_word(fp, word, MAX_WORD);
        if (c != EOF) {
        actual = strchr(word, '=');
        if (actual == 0) {
            read_error("bad format for formal=actual in .gate");
            return 0;
        }
        *actual++ = 0;
        node = read_find_or_create_node(network, actual);
        array_insert_last(char *, formal_array, util_strsav(word));
        array_insert_last(node_t *, actual_array, node);
        }
    }
    } while (c != EOF && c != '\n');

    formal_data = array_data(char *, formal_array);
    actual_data = array_data(node_t *, actual_array);
    tot = array_n(formal_array);
    array_free(formal_array);
    array_free(actual_array);

    node = actual_data[tot-1];    /* ### output pin name not checked ! */

    if (! lib_set_gate(node, gate, formal_data, actual_data, tot-1)) {
    (void) sprintf(errmsg, 
        "error loading gate '%s' (do you have the correct library ?)",
        word);
    read_error(errmsg);
    return 0;
    }

    /* free the formal strings */
    for(i = 0; i < tot; i++) {
    FREE(formal_data[i]);
    }
    FREE(formal_data);
    FREE(actual_data);
    return 1;
}
