#include "mis.h"
#include "io_int.h"
#include "vov.h"

static int write_slif();
char * replace_chars_in();
void io_write_slif_name();
static void write_slif_sop();
int own_strcmp();

int
com_write_slif(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c;
    bool file_found = FALSE;
    char *fname;
    FILE *fp;

    getopt_reset ();
    while ((c = getopt(argc, argv, "f:")) != EOF) {
	switch (c) {
        case 'f':
            fname = ALLOC (char, strlen (optarg) + 1);
	    (void) strcpy (fname, optarg);
	    file_found = TRUE;
	    break;
	default :
	    goto usage;
        }
    }
    if (file_found) {
	if ((fp = fopen (fname, "w")) == NULL) {
	    (void) fprintf(miserr, "unable to open slif output file %s\n", fname);
	    FREE (fname);
	    exit (-1);
	}
	VOVoutput( VOV_UNIX_FILE, fname );
	(void) write_slif (fp, *network);
	FREE (fname);
	fclose (fp);
    }
    else 
	(void) write_slif (misout, *network);
    return 0;

usage:
    (void) fprintf(miserr,
	    "write_slif [-f file]\n");
    return 1;
}

static int
write_slif(fp, network)
FILE *fp;
network_t *network;
{
    lsGen gen;
    node_t *node, *fanin;
    int t1, i, short_name;

    /* Slif can be written either as boolean equations or as a
     * mapped network. Dependend on the status of the network
     * a decision is made whether to write equations or gates.
     */

    if (! lib_network_is_mapped(network)) 
    {
       fprintf(miserr, "write_slif: network not mapped, writing equations\n");
    }
    else
    {
       fprintf(miserr, "write_slif: writing mapped network\n");
    }

    /* .model, .inputs and .outputs are similar in both cases */

    (void) fprintf(fp, ".model %s ;\n\n", network->net_name);

    (void) fprintf(fp, ".inputs");
    t1 = 0;
    foreach_primary_input(network, gen, node) 
    {
       (void) fprintf(fp, " %s", replace_chars_in(node->name));
       if (t1++ > 8) 
       {
           t1 = 0;
           fprintf(fp,"\n");
       }
    }
    (void) fprintf(fp, ";\n\n");

    (void) fprintf(fp, ".outputs");
    t1 = 0;
    foreach_primary_output(network, gen, node) 
    {
        (void) fprintf(fp, " %s", replace_chars_in(node->name));
        if (t1++ > 8) 
        {
           t1 = 0;
           fprintf(fp,"\n");
        }
    }
    (void) fprintf(fp, ";\n\n");

    /* if the network is mapped we are writing gates */
    if (lib_network_is_mapped(network)) 
    {
        foreach_node(network, gen, node) 
        {
           if (node->type == INTERNAL) 
           {
              /* slif does not support calls to  ZERO and ONE */
              if (own_strcmp(lib_gate_name(lib_gate_of(node)),"zero") != 0 &
                  own_strcmp(lib_gate_name(lib_gate_of(node)),"one") != 0 )
              {
                 fprintf(fp, ".call X%s%s", lib_gate_name(lib_gate_of(node)),
                               replace_chars_in(node->name));
                 fprintf(fp, "\t%s(", lib_gate_name(lib_gate_of(node)));
                 foreach_fanin(node, i, fanin) 
                 {
                    if (fanin->type == PRIMARY_INPUT)
                    {
                       fprintf(fp, " %s", replace_chars_in(fanin->name));
                    }
                    else
                    {
                       fprintf(fp, " %s%s", lib_gate_name(lib_gate_of(fanin)), 
                               replace_chars_in(fanin->name));
                    }
                    if (i < node->nin-1) fprintf(fp, ",");
                 }
                 fprintf(fp, ";; %s%s);\n\n", 
                         lib_gate_name(lib_gate_of(node)), 
                         replace_chars_in(node->name));
              }
           }
        }

        foreach_node(network, gen, node) 
        {
           if (node->type == PRIMARY_OUTPUT)
           {
              foreach_fanin(node, i, fanin) 
              {
                 if (own_strcmp(lib_gate_name(lib_gate_of(fanin)),"zero") == 0)
                 {
                    fprintf(fp, "%s = 0 ;\n", 
                            replace_chars_in(node->name));
                 }
                 else if (own_strcmp(lib_gate_name(lib_gate_of(fanin)),"one") == 0 )
                 {
                    fprintf(fp, "%s = 1 ;\n", 
                            replace_chars_in(node->name));
                 }
                 else 
                 {
                    fprintf(fp, "%s = ", replace_chars_in(node->name));
                    if (lib_gate_name(lib_gate_of(fanin)) != NULL)
                       fprintf(fp, "%s", lib_gate_name(lib_gate_of(fanin))); 
                    fprintf(fp, "%s;\n", replace_chars_in(fanin->name));
                 }
              }
           }
        }
     } /* endif write mapped network */

     else /* write equations */
     {
       short_name = 0; /* no short_names please */
       foreach_node(network, gen, node)
       {
          write_slif_sop(fp, node, short_name);
       }
     } /* end else writing equations */

    /* .endmodel is the same again */
    (void) fprintf(fp, "\n.endmodel %s ;\n", network->net_name);
    return 1;
}


static void
write_slif_sop(fp, node, short_name)
FILE *fp;
node_t *node;
int short_name;
{
    register int i, j;
    int first_cube, first_lit;
    node_cube_t cube;
    node_literal_t literal;
    node_t *fanin;

    /* never print inputs ... */
    if (node->type == PRIMARY_INPUT) return;

    /* print outputs only if feed by inputs, or if feed by > 1 node */
    if (node->type == PRIMARY_OUTPUT) {
    fanin = node_get_fanin(node, 0);
    if (fanin->type == INTERNAL && 
        io_po_fanout_count(fanin, NIL(node_t *)) == 1) {
        return;
    }
    }

    io_write_slif_name(fp, node, short_name);
    io_fputs_break(fp, " = ");

    switch(node_function(node)) {
    case NODE_0:
    io_fputs_break(fp, "0;\n");
    return;
    case NODE_1:
    io_fputs_break(fp, "1;\n");
    return;
    case NODE_PO:
    io_write_slif_name(fp, node_get_fanin(node, 0), short_name);
    io_fputs_break(fp, ";\n");
    return;
    }

    first_cube = 1;
    for(i = node_num_cube(node)-1; i >= 0; i--) {
    if (! first_cube) {
        io_fputs_break(fp, " + ");
    }
    cube = node_get_cube(node, i);

    first_lit = 1;
    foreach_fanin(node, j, fanin) {
        literal = node_get_literal(cube, j);
        switch(literal) {
        case ZERO:
        case ONE:
        if (! first_lit) {
            io_fputc_break(fp, ' ');
        }
        io_write_slif_name(fp, fanin, short_name);
                if (literal == ZERO) 
                {
                   io_fputc_break(fp, '\'');
                }
        first_lit = 0;
        break;
        }
    }
    first_cube = 0;
    }
    io_fputs_break(fp, ";\n");
}

char *
replace_chars_in(node_name)
char *node_name;
{
   /* MisII allows some charcaters in its variable names
    * that are not allowed in SLIF.
    * This routine scans variable names and replaces those characters
    * by other ones. It currently replaces 
    *      '-' by ':',
    *      '(' by '[' and 
    *      ')' by ']'.
    */
   static char *buffer = NULL;
   int t1 = 0;

   if (buffer == NULL)
   {
      buffer = ALLOC(char, strlen(node_name) + 2);
   }
   else if (strlen(buffer) < strlen(node_name))
   {
      if (buffer != NULL) FREE(buffer);
      buffer = ALLOC(char, strlen(node_name) + 2);
   }

   while (node_name[t1] != '\0')
   {
      if (node_name[t1] == '-')
      {
         buffer[t1] = ':';
      }
      else if (node_name[t1] == ')')
      {
         buffer[t1] = ']';
      }
      else if (node_name[t1] == '(')
      {
         buffer[t1] = '[';
      }
      else
      {
         buffer[t1] = node_name[t1];
      }
      t1++;
    }
    buffer[t1] = '\0';

    return(buffer);
}

void
io_write_slif_name(fp, node, short_flag)
FILE *fp;
node_t *node;
int short_flag;
{
    node_t *po;

    if (node->type == INTERNAL && io_po_fanout_count(node, &po) == 1) {
    io_fputs_break(fp, short_flag ? po->short_name : replace_chars_in(po->name));
    } else {
    io_fputs_break(fp, short_flag ? node->short_name : replace_chars_in(node->name));
    }
}

int
own_strcmp(p,q)
char *p, *q;
{
   if ((p == NULL) || (q == NULL)) return -1;

   return(strcmp(p,q));
}

