 /* file bin_delay.c release 1.4 */
 /* last modified: 9/9/90 at 17:31:29 */
#include <math.h>
#include "mis.h"
#include "map_macros.h"
#include "map_int.h"
#include "bin_int.h"
#include "map_delay.h"
#include "fanout_int.h"

static bin_global_t global;
#include "bin_delay_static.h"


 /* EXTERNAL INTERFACE */

void bin_delay_area_tree_match(network, library, globals)
network_t *network;
library_t *library;
bin_global_t globals;
{
  fail("ERROR: no longer supported at this moment\n");
}

void bin_delay_tree_match(network, library, globals)
network_t *network;
library_t *library;
bin_global_t globals;
{
  global = globals;
  global.library = library;
  init_delay_bucket_storage();
  bin_for_all_nodes_inputs_first(network, bin_alloc);
  bin_delay_initialize_pipo(network);
  bin_for_all_nodes_inputs_first(network, compute_best_match);
  bin_for_all_nodes_outputs_first(network, select_best_gate);
  bin_for_all_nodes_inputs_first(network, bin_free);
  free_delay_bucket_storage();
}

 /* some utilities */

void bin_for_all_nodes_inputs_first(network,fn)
network_t *network;
VoidFn fn;
{
  int i;
  array_t *nodevec;
  node_t *node;

  nodevec = network_dfs(network);
  for(i = 0; i < array_n(nodevec); i++) {
    node = array_fetch(node_t *, nodevec, i);
    (*fn)(node);
  }
  array_free(nodevec);
}

void bin_for_all_nodes_outputs_first(network,fn)
network_t *network;
VoidFn fn;
{
  int i;
  array_t *nodevec;
  node_t *node;

  nodevec = network_dfs_from_input(network);
  for(i = 0; i < array_n(nodevec); i++) {
    node = array_fetch(node_t *, nodevec, i);
    (*fn)(node);
  }
  array_free(nodevec);
}


 /* INTERNAL INTERFACE */

 /* if the node was obtained from a fanout tree, it should stay mapped the way it is */
 /* we just compute its pwl so that we get an uniform interface for the gate selection part */
 /* otherwise, standard tree mapping stuff apply */

static void compute_best_match(node)
node_t *node;
{
  lsGen gen;
  prim_t *prim;

  if (node->type != INTERNAL || node_num_fanin(node) == 0) return;
  if (MAP(node)->type == FANOUT_NODE) {
    compute_fanout_node_pwl(node);
  } else {
    lsForeachItem(global.library->patterns, gen, prim) {
      gen_all_matches(node, prim, find_best_delay, global.verbose > 5, global.allow_internal_fanout);
    }
  }
}

 /* goes through all the gates corresponding to the prim and record */
 /* their delay information in the BIN(node)->pwl routine */
 /* this means computing the arrival times at the input pins of the prim */
 /* then making a new pwl pair (rise/fall) per gate */
 /* then registrating it in the BIN(node)->pwl */
 /* NOTE: also computes the optimal area and sticks it into MAP entry */

static int find_best_delay(prim)
prim_t *prim;
{
  int i;
  lsGen gen;
  lib_gate_t *gate;
  node_t *input;
  node_t *node;
  int ninputs;
  node_t **binding;

  node = map_prim_get_root(prim);

 /* first make a copy of the input bindings */
  ninputs = prim->ninputs;
  binding = ALLOC(node_t *, ninputs);
  for (i = 0; i < ninputs; i++) {
    binding[i] = prim->inputs[i]->binding;
  }

 /* then try all gates associated to the prim */
  lsForeachItem(prim->gates, gen, gate) {
    if (global.verbose > 2) {
      (void) fprintf(misout, "trying gate %s at node %s...\n", gate->name, node->name);
    }
    if (strcmp(gate->name, "**wire**") == 0) {
      /* first arg passed by ref, second arg by value */
      compute_node_pwl_for_wire(&(BIN(node)->pwl), gate, ninputs, binding);
    } else {
      compute_node_pwl(&(BIN(node)->pwl), gate, ninputs, binding);
    }
    preserve_best_area(&(BIN(node)->area), gate, ninputs, binding);
  }
  FREE(binding);
  return 1;
}

 /* computes the optimal area */

static void preserve_best_area(bucket, gate, ninputs, binding)
area_bucket_t *bucket;
lib_gate_t *gate;
int ninputs;
node_t **binding;
{
  int i;
  node_t *input;
  double area = gate->area;
    
  for (i = 0; i < ninputs; i++) {
    input = binding[i];
    area += BIN(input)->area.area;
  }
  if (area >= bucket->area) return;
  bucket->area = area;
  bucket->gate = gate;
  if (bucket->ninputs > 0) FREE(bucket->save_binding);
  bucket->ninputs = ninputs;
  if (bucket->ninputs == 0) return;
  bucket->save_binding = ALLOC(node_t *, bucket->ninputs);
  for (i = 0; i < bucket->ninputs; i++) {
    bucket->save_binding[i] = binding[i];
  }
}

 /* copy the pwl of the input; update its buckets */

static void compute_node_pwl_for_wire(pwl, gate, ninputs, binding)
pwl_delay_t *pwl;
lib_gate_t *gate;
int ninputs;
node_t **binding;
{
  pwl_t *new_fn;
  delay_bucket_t *bucket;

  bucket = compute_wire_pwl(gate, ninputs, binding);
  compute_pwl_delay_min(pwl, bucket->pwl);
}

static delay_bucket_t *compute_wire_pwl(gate, ninputs, binding)
lib_gate_t *gate;
int ninputs;
node_t **binding;
{
  delay_bucket_t *bucket;

 /* alloc a bucket and set it up */
  bucket = delay_bucket_alloc();
  bucket->gate = gate;
  bucket->ninputs = 1;
  bucket->save_binding = ALLOC(node_t *, 1);
  bucket->save_binding[0] = binding[0];
  bucket->pwl.rise = pwl_dup(BIN(binding[0])->pwl.rise);
  bucket->pwl.fall = pwl_dup(BIN(binding[0])->pwl.fall);
  pwl_set_data(bucket->pwl.rise, (char *) bucket);
  pwl_set_data(bucket->pwl.fall, (char *) bucket);
  return bucket;
}

/* compute the pwl delay model at the node for that gate */

static void compute_node_pwl(pwl, gate, ninputs, binding)
pwl_delay_t *pwl;
lib_gate_t *gate;
int ninputs;
node_t **binding;
{
  int i;
  double pin_load;
  node_t *pin_node;
  delay_bucket_t *bucket;

  delay_time_t *input_arrivals = ALLOC(delay_time_t, ninputs);
  for (i = 0; i < ninputs; i++) {
    pin_node = binding[i];
    if (MAP(pin_node)->type == TREE_NODE) {
      pin_load = delay_get_load(gate->delay_info[i]);
    } else {
      pin_load = MAP(pin_node)->load;
    }
    if (pin_node->type == PRIMARY_INPUT) {
      input_arrivals[i] = compute_pwl_delay(BIN(pin_node)->pwl, pin_load);
    } else {
 /* why not just computing the pwl_delay here? */
 /* because of rise and fall delays. The best pwl_delay */
 /* may be obtained for different gates */
 /* compute_consistent_pwl_delay guarantees of realizable, if not always optimal, value */
 /* open question: is the cpu time worth the extra precision?*/
      input_arrivals[i] = compute_consistent_pwl_delay(BIN(pin_node)->pwl, pin_load);
    }
    if (global.verbose > 2) {
      pwl_print(misout, BIN(pin_node)->pwl.rise, delay_bucket_print);
      pwl_print(misout, BIN(pin_node)->pwl.fall, delay_bucket_print);
      (void) fprintf(misout, "a(%2.3f,%2.3f) l(%2.3f)\n", input_arrivals[i].rise,
		     input_arrivals[i].fall, pin_load);
		     
    }
  }
  bucket = compute_gate_pwl(gate, input_arrivals, ninputs, binding);
  FREE(input_arrivals);
  compute_pwl_delay_min(pwl, bucket->pwl);
}

 /* the easiest way to do it is to make a fake prim */
 /* the only thing really needed from the prim is the ninputs */
 /* and the list of inputs. What should be done is to change the routines */
 /* using the prim and pass them the ninputs and the list of inputs instead */
 /* but this would require more copying that the current implementation */

static void compute_fanout_node_pwl(node)
node_t *node;
{
  int ninputs;
  node_t **binding;

  assert(MAP(node)->ninputs == 1);
  assert(MAP(node)->gate != NIL(lib_gate_t));
  ninputs = 1;
  binding = ALLOC(node_t *, 1);
  binding[0] = MAP(node)->save_binding[0];
  compute_node_pwl(&(BIN(node)->pwl), MAP(node)->gate, ninputs, binding);
  FREE(binding);
}


 /* not easy: if rise and fall from same gate, OK */
 /* if from different gate, take the latest among rise and fall */
 /* take the gate that does the best (MAX(rise,fall)) */
 /* this is not necessarily optimal */

static delay_time_t compute_consistent_pwl_delay(pwl, load)
pwl_delay_t pwl;
double load;
{
  delay_time_t result;
  delay_bucket_t *bucket;

 /* first select the gate (bucket) that applies */
  bucket = pwl_select_best_gate(pwl, load);

 /* then compute the arrival time using that gate */
  return compute_pwl_delay(bucket->pwl, load);
}

static delay_bucket_t *pwl_select_best_gate(pwl, load)
pwl_delay_t pwl;
double load;
{
  pwl_point_t *p_rise, *p_fall;
  delay_bucket_t *b_rise, *b_fall;
  delay_time_t b_rise_delay, b_fall_delay;

  p_rise = pwl_lookup(pwl.rise, load);
  p_fall = pwl_lookup(pwl.fall, load);
  b_rise = (delay_bucket_t *) pwl_point_data(p_rise);
  b_fall = (delay_bucket_t *) pwl_point_data(p_fall);
  if (b_rise->gate == b_fall->gate) return b_rise;

 /* need to recompute the delay using the gate stored in bucket */
  if (global.verbose > 2) {
    (void) fprintf(misout, "!"); (void) fflush(misout);
  }
  b_rise_delay = compute_pwl_delay(b_rise->pwl, load);
  b_fall_delay = compute_pwl_delay(b_fall->pwl, load);
  return (GETMAX(b_rise_delay) < GETMAX(b_fall_delay)) ? b_rise : b_fall;
}

static delay_time_t compute_pwl_delay(pwl, load)
pwl_delay_t pwl;
double load;
{
  pwl_point_t *p_rise;
  pwl_point_t *p_fall;
  delay_time_t result;

  p_rise = pwl_lookup(pwl.rise, load);
  p_fall = pwl_lookup(pwl.fall, load);
  result.rise = pwl_point_eval(p_rise, load);
  result.fall = pwl_point_eval(p_fall, load);
  return result;
}

 /* given a gate and arrival times at its inputs, compute the pwl for the gate */
 /* that the only tricky part in this entire thing */
 /* should be put in map_delay.c */

static delay_bucket_t *compute_gate_pwl(gate, input_arrivals, ninputs, binding)
lib_gate_t *gate;
delay_time_t *input_arrivals;
int ninputs;
node_t **binding;
{
  int i, pt;
  int n_points;
  pwl_point_t *rise_points, *fall_points;
  delay_pin_t *pin_delay, **lib_delay_model;
  delay_bucket_t *bucket;

 /* alloc a bucket and set it up */
  bucket = delay_bucket_alloc();
  bucket->gate = gate;
  bucket->ninputs = ninputs;
  bucket->save_binding = ALLOC(node_t *, ninputs);
  for (i = 0; i < ninputs; i++) {
    bucket->save_binding[i] = binding[i];
  }

 /* count the number of points: two points per 'PHASE_NEITHER' pin */
  lib_delay_model = (delay_pin_t **) gate->delay_info;
  for (n_points = 0, i = 0; i < ninputs; i++) {
    pin_delay = lib_delay_model[i];
    n_points += (pin_delay->phase == PHASE_NEITHER) ? 2 : 1;
  }
  rise_points = ALLOC(pwl_point_t, n_points);
  fall_points = ALLOC(pwl_point_t, n_points);

 /* store info in points; there is no data field at this point */
  for (pt = 0, i = 0; i < ninputs; i++) {
    pin_delay = lib_delay_model[i];
    switch(pin_delay->phase) {
    case PHASE_INVERTING:
      rise_points[pt].x = 0.0;
      rise_points[pt].y = input_arrivals[i].fall + pin_delay->block.rise;
      rise_points[pt].slope = pin_delay->drive.rise;
      rise_points[pt].data = (char *) bucket;
      fall_points[pt].x = 0.0;
      fall_points[pt].y = input_arrivals[i].rise + pin_delay->block.fall;
      fall_points[pt].slope = pin_delay->drive.fall;
      fall_points[pt].data = (char *) bucket;
      pt++;
      break;
    case PHASE_NONINVERTING:
      rise_points[pt].x = 0.0;
      rise_points[pt].y = input_arrivals[i].rise + pin_delay->block.rise;
      rise_points[pt].slope = pin_delay->drive.rise;
      rise_points[pt].data = (char *) bucket;
      fall_points[pt].x = 0.0;
      fall_points[pt].y = input_arrivals[i].fall + pin_delay->block.fall;
      fall_points[pt].slope = pin_delay->drive.fall;
      fall_points[pt].data = (char *) bucket;
      pt++;
      break;
    case PHASE_NEITHER:
      rise_points[pt].x = 0.0;
      rise_points[pt].y = input_arrivals[i].rise + pin_delay->block.rise;
      rise_points[pt].slope = pin_delay->drive.rise;
      rise_points[pt].data = (char *) bucket;
      fall_points[pt].x = 0.0;
      fall_points[pt].y = input_arrivals[i].fall + pin_delay->block.fall;
      fall_points[pt].slope = pin_delay->drive.fall;
      fall_points[pt].data = (char *) bucket;
      pt++;
      rise_points[pt].x = 0.0;
      rise_points[pt].y = input_arrivals[i].fall + pin_delay->block.rise;
      rise_points[pt].slope = pin_delay->drive.rise;
      rise_points[pt].data = (char *) bucket;
      fall_points[pt].x = 0.0;
      fall_points[pt].y = input_arrivals[i].rise + pin_delay->block.fall;
      fall_points[pt].slope = pin_delay->drive.fall;
      fall_points[pt].data = (char *) bucket;
      pt++;
      break;
    }
  }
  assert(pt == n_points);
  bucket->pwl.rise = pwl_create_linear_max(n_points, rise_points);
  if (global.verbose > 2) {
    (void) fprintf(misout, "rise = ");
    pwl_print(misout, bucket->pwl.rise, delay_bucket_print);
  }
  bucket->pwl.fall = pwl_create_linear_max(n_points, fall_points);
  if (global.verbose > 2) {
    (void) fprintf(misout, "fall = ");
    pwl_print(misout, bucket->pwl.fall, delay_bucket_print);
  }
  FREE(rise_points);
  FREE(fall_points);
  return bucket;
}

static compute_pwl_delay_min(target, operand)
pwl_delay_t *target;
pwl_delay_t operand;
{
  pwl_t *new_fn;

  if (global.verbose > 2) {
    (void) fprintf(misout, "arg1 = ");
    pwl_print(misout, target->rise, delay_bucket_print);
    (void) fprintf(misout, "arg2 = ");
    pwl_print(misout, operand.rise, delay_bucket_print);
  }
  new_fn = pwl_min(target->rise, operand.rise);
  if (global.verbose > 2) {
    (void) fprintf(misout, "result = ");
    pwl_print(misout, new_fn, delay_bucket_print);
  }
  pwl_free(target->rise);
  target->rise = new_fn;

  if (global.verbose > 2) {
    (void) fprintf(misout, "arg1 = ");
    pwl_print(misout, target->fall, delay_bucket_print);
    (void) fprintf(misout, "arg2 = ");
    pwl_print(misout, operand.fall, delay_bucket_print);
  }
  new_fn = pwl_min(target->fall, operand.fall);  
  if (global.verbose > 2) {
    (void) fprintf(misout, "result = ");
    pwl_print(misout, new_fn, delay_bucket_print);
  }
  pwl_free(target->fall);
  target->fall = new_fn;
}

static void delay_bucket_print(fp, bucket)
FILE *fp;
delay_bucket_t *bucket;
{
  int i;
  node_t *fanin;

 /* case of PI */
  if (bucket == NIL(delay_bucket_t)) return;

  (void) fprintf(fp, "gate=%s [", bucket->gate->name);
  for (i = 0; i < bucket->ninputs; i++) {
    (void) fprintf(fp, "%s ", bucket->save_binding[i]->name);
  }
  (void) fprintf(fp, "] 0x%x 0x%x", (int) bucket->pwl.rise, (int) bucket->pwl.fall);
}


 /* GATE_SELECTION */

 /* where the final selection of gates takes place */
 /* go through the network and select the best gates, given the loads */
 /* at the outputs. Put the resulting selection into the MAP fields */
 /* a node participates if root or pin of a final gate */
 /* otherwise its load stays to INFINITY and we can ignore it */

static void select_best_gate(node)
node_t *node;
{
  char *type;
  delay_bucket_t *bucket;
  delay_time_t arrival;

  if (node->type == PRIMARY_OUTPUT || node->type == PRIMARY_INPUT || MAP(node)->load == INFINITY) {
    MAP(node)->gate = NIL(lib_gate_t);
    return;
  }
  bucket = pwl_select_best_gate(BIN(node)->pwl, MAP(node)->load);
  copy_bucket_to_map(node, bucket);
  assert(MAP(node)->gate != NIL(lib_gate_t));
  if (global.verbose > 0) {
    type = "unknown";
    if (MAP(node)->type == TREE_NODE) type = "TREE_NODE";
    else if (MAP(node)->type == FANOUT_NODE) type = "FANOUT_NODE";
    else if (MAP(node)->type == ROOT_NODE) type = "ROOT_NODE";
    (void) fprintf(miserr, "node %s of type %s\n", node->name, type);
  }
  if (global.verbose > 1) {
    (void) fprintf(miserr, "Best match at %s for load %2.2f is gate(%s) ", node->name, MAP(node)->load, bucket->gate->name); 
    arrival = MAP(node)->map_arrival;
    (void) fprintf(miserr, "area=%2.2f arrival=(%2.2f,%2.2f)\n", bucket->gate->area, arrival.rise, arrival.fall);
  }
  increment_input_loads(node, bucket);
}
 
/* install into MAP fields contents of selected bucket  */

static void copy_bucket_to_map(node, bucket)
node_t *node;
delay_bucket_t *bucket;
{
  int i;

  MAP(node)->gate = bucket->gate;
  MAP(node)->map_arrival = compute_consistent_pwl_delay(bucket->pwl, MAP(node)->load);
  MAP(node)->ninputs = bucket->ninputs;
  if (MAP(node)->save_binding != NIL(node_t *)) FREE(MAP(node)->save_binding);
  MAP(node)->save_binding = ALLOC(node_t *, bucket->ninputs);
  for (i = 0; i < bucket->ninputs; i++) {
    MAP(node)->save_binding[i] = bucket->save_binding[i];
  }
}

 /* MAP(node)->load = INFINITY means no load was assigned to node yet */
 /* we use INFINITY instead of the more natural 0.0 value because */
 /* we want the code to work even if the user specifies 0.0 as a load to PO's */

static void increment_input_loads(node, bucket)
node_t *node;
delay_bucket_t *bucket;
{
  int pin;
  double *loads;
  node_t *fanin;

  loads = ALLOC(double, bucket->ninputs);
  if (strcmp(bucket->gate->name, "**wire**") == 0) {
    loads[0] = MAP(node)->load;
  } else {
    for (pin = 0; pin < bucket->ninputs; pin++) {
      loads[pin] = delay_get_load(bucket->gate->delay_info[pin]);
    }
  }
  for (pin = 0; pin < bucket->ninputs; pin++) {
    fanin = bucket->save_binding[pin];
    if (MAP(fanin)->type != TREE_NODE) continue;
    if (MAP(fanin)->load == INFINITY) MAP(fanin)->load = 0.0;
    MAP(fanin)->load += loads[pin];
  }
  FREE(loads);
}


 /* should build pwl's for the PI's; just read the loads for the PO's */

static void bin_delay_initialize_pipo(network)
network_t *network;
{
  lsGen gen;
  node_t *pi, *po, *fanin;

  foreach_primary_input(network, gen, pi) {
    bin_init_pi(pi);
  }
  foreach_primary_output(network, gen, po) {
    fanin = node_get_fanin(po, 0);
    MAP(po)->load = MAP(fanin)->load = pipo_get_po_load(po);
  }
}

 /* would be nice to have this arbitrary delay model at PI's */
 /* right now, we simply have a single point */

static void bin_init_pi(node)
node_t *node;
{
  int i;
  pwl_point_t rise_point;
  pwl_point_t fall_point;
  delay_time_t drive, arrival;
  
  drive = pipo_get_pi_drive(node);
  arrival = pipo_get_pi_arrival(node);
  rise_point.x = 0.0;
  rise_point.y = arrival.rise;
  rise_point.slope = drive.rise;
  rise_point.data = NIL(char);
  fall_point.x = 0.0;
  fall_point.y = arrival.fall;
  fall_point.slope = drive.fall;
  fall_point.data = NIL(char);
  BIN(node)->pwl.rise = pwl_create_linear_max(1, &rise_point);
  BIN(node)->pwl.fall = pwl_create_linear_max(1, &fall_point);
}


 /* DEBUGGING ROUTINES */

static st_table *global_gate_stat_table;

 /* where we print statistics on which gate got matched how many times */

void bin_print_gate_stats(network)
network_t *network;
{
  global_gate_stat_table = st_init_table(st_numcmp, st_numhash);
  bin_for_all_nodes_outputs_first(network, gather_gate_statistics);
  print_gate_stat_table();
  st_free_table(global_gate_stat_table);
}

static void gather_gate_statistics(node)
node_t *node;
{
  int count;
  st_table *table = global_gate_stat_table;
  lib_gate_t *gate = MAP(node)->gate;

  if (gate == NIL(lib_gate_t)) return;
  if (node->type != INTERNAL) return;
  if (st_lookup(table, (char *) gate, (char **) &count) == 0) count = 0;
  (void) st_insert(table, (char *) gate, (char *) (count + 1));
}

static void print_gate_stat_table()
{
  st_table *table = global_gate_stat_table;
  st_generator *gen;
  char *key, *value;
  st_foreach_item(table, gen, &key, &value) {
    lib_gate_t *gate = (lib_gate_t *) key;
    int count = (int) value;
    (void) fprintf(miserr, "%s %2.2f %d\n", gate->name, gate->area, count);
  }
}

 /* should be visited inputs first */

static void bin_alloc(node)
node_t *node;
{
  int i;
  node_t *fanin;

  node->BIN_SLOT = (char *) ALLOC(bin_t, 1);
  BIN(node)->pwl.rise = pwl_create_empty();
  BIN(node)->pwl.fall = pwl_create_empty();
  BIN(node)->area.gate = NIL(lib_gate_t);
  BIN(node)->area.ninputs = 0;
  BIN(node)->area.save_binding = NIL(node_t *);
  BIN(node)->area.area = 0.0;

 /* this means that after fanout optimization we keep */
 /* the actual load of FANOUT_NODE and ROOT_NODE */
  if (MAP(node)->type == TREE_NODE) {
    MAP(node)->load = INFINITY;
  } else {
    MAP(node)->load = 0.0;
  }

 /* compute the load of fanout nodes */
  if (node->type == PRIMARY_OUTPUT) {
    fanin = node_get_fanin(node, 0);
    if (MAP(fanin)->type != TREE_NODE) {
      MAP(fanin)->load += pipo_get_po_load(node);
    }
  } else if (MAP(node)->gate != NIL(lib_gate_t)) {
    for (i = 0; i < MAP(node)->ninputs; i++) {
      fanin = MAP(node)->save_binding[i];
      if (MAP(fanin)->type != TREE_NODE) {
	MAP(fanin)->load += delay_get_load(MAP(node)->gate->delay_info[i]);
      }
    }
  }
}

static void bin_free(node)
node_t *node;
{
  pwl_free(BIN(node)->pwl.rise);
  pwl_free(BIN(node)->pwl.fall);
  if (BIN(node)->area.ninputs > 0) {
    FREE(BIN(node)->area.save_binding);
  }
  FREE(node->BIN_SLOT);
  node->BIN_SLOT = NIL(char);
}

 /* need some way to allocate and free the buckets with gate information */

static delay_bucket_t *global_head_bucket;
static void init_delay_bucket_storage()
{
  global_head_bucket = NIL(delay_bucket_t);
}

static void free_delay_bucket_storage()
{
  delay_bucket_t *head, *next;

  for (head = global_head_bucket; head != NIL(delay_bucket_t); head = next) {
    pwl_free(head->pwl.rise);
    pwl_free(head->pwl.fall);
    if (head->ninputs > 0) {
      FREE(head->save_binding);
    }
    next = head->next;
    FREE(head);
  }
}

static delay_bucket_t *delay_bucket_alloc()
{
  delay_bucket_t *bucket;

  bucket = ALLOC(delay_bucket_t, 1);
  bucket->pwl.rise = pwl_create_empty();
  bucket->pwl.fall = pwl_create_empty();
  bucket->gate = NIL(lib_gate_t);
  bucket->ninputs = 0;
  bucket->save_binding = NIL(node_t *);
  bucket->next = global_head_bucket;
  global_head_bucket = bucket;
  return bucket;
}
