 /* file cluster.c release 1.2 */
 /* last modified: 9/9/90 at 17:31:32 */
#include "mis.h"

 /* this file implements Lawler's clustering algorithm */
 /* for a given limit in size of clusters, find a minimum depth */
 /* clustering. All nodes are supposed to be of size 1. */
 /* When all nodes are of the same size, clustering guarantees */
 /* that each cluster has only one external output (root of cluster) */
 /* then we collapse each cluster into its root node */

#define array_fetch_p(type, a, i)			\
    (array_i = (i),				\
      (array_i < 0 || array_i >= (a)->num) ? array_abort(0) : 0,\
      ((type *) ((a)->space + array_i * (a)->obj_size)))

#define MAX_LEVEL 0x7fffffff


 /* a cluster node is a ROOT_NODE iff it fan outs to a node with a larger label */

typedef enum {
  NORMAL_NODE, ROOT_NODE
  } clust_node_type_t;

typedef struct {
  int number;
  st_table *nodes;		      /* tells the nodes that it contains; -> NORMAL_NODE, ROOT_NODE  */
  int label;
} cluster_t;

typedef struct {
  int label;
  int weight;
  int weight_same_label;
  int is_in_cluster;
  array_t *clusters;		      /* list of clusters to which this node belongs (cluster_t *) */
} clust_node_t;


#include "cluster_static.h"

 /* EXTERNAL INTERFACE */

 /* Lawler's algorithm: given a max cluster size, returns the best labelling */
 /* form and collapse the clusters */

network_t *cluster_under_size_constraint(network, cluster_size, verbose)
network_t *network;
int cluster_size;
int verbose;
{
  int max_label; 
  array_t *nodevec;
  array_t *clusters;		      /* array of cluster_t * */
  st_table *table;

  nodevec = network_dfs(network);
   
  table = clust_node_table_alloc();
  if (verbose >= 2) (void) printf("------ labelling nodes-------\n");
  max_label = label_nodes(table, nodevec, cluster_size, verbose);
  if (verbose >= 1) (void) printf("*** # levels = %d\n", max_label + 1);

  if (verbose >= 2) (void) printf("------ forming clusters-------\n");
  clusters = array_alloc(cluster_t *, 0);
  form_clusters(table, nodevec, clusters);
  clust_node_table_free(table);

  if (verbose >= 2) {
    (void) printf("------ printing clusters-------\n");
    print_clusters(clusters);
  }

  if (verbose >= 2) (void) printf("------ collapsing clusters-------\n");
  collapse_clusters(clusters, nodevec);

 /* remove intermediate nodes not roots of clusters */
  network_sweep(network);

  array_free(nodevec);
  clusters_free(clusters);

  return network;
}

 /* built above Lawler's algorithm: finds the minimum cluster size for which */
 /* the label can attain "depth"; select that cluster size and call Lawler's algorithm */

network_t *cluster_under_depth_constraint(network, depth, verbose)
network_t *network;
int depth;
int verbose;
{
  int cluster_size;
  int min_cluster_size, max_cluster_size;
  array_t *nodevec;
  int nsize;
  int max_label = depth - 1;
  
  nodevec = network_dfs(network);
  nsize = array_n(nodevec);
  
  min_cluster_size = 1;
  max_cluster_size = nsize;
  cluster_size = lazy_binary_search(min_cluster_size, max_cluster_size, max_label, nodevec, best_labelling);
  if (verbose >= 1) (void) printf("*** max cluster size = %d\n", cluster_size);
  array_free(nodevec);
  return cluster_under_size_constraint(network, cluster_size, verbose);
}

 /* built above the previous one. Find the minimum size giving the same */
 /* depth as the size passed as argument and use it for Lawler's algorithm */

network_t *cluster_under_size_as_depth_constraint(network, size, verbose)
network_t *network;
int size;
int verbose;
{
  int cluster_size;
  int min_cluster_size, max_cluster_size;
  array_t *nodevec;
  int max_label;
  
  nodevec = network_dfs(network);
  
  min_cluster_size = 1;
  max_cluster_size = size;
  max_label = best_labelling(nodevec, size);
  cluster_size = lazy_binary_search(min_cluster_size, max_cluster_size, max_label, nodevec, best_labelling);
  if (verbose >= 1) (void) printf("*** max cluster size = %d\n", cluster_size);
  array_free(nodevec);
  return cluster_under_size_constraint(network, cluster_size, verbose);
}


 /* INTERNAL INTERFACE */

 /* alloc and free the hash table containing info on nodes: used for labelling */

static st_table *clust_node_table_alloc()
{
  return st_init_table(st_numcmp, st_numhash);
}

static void clust_node_table_free(table)
st_table *table;
{
  node_t *node;
  st_generator *gen;
  clust_node_t *node_info;

  st_foreach_item(table, gen, (char **) &node, (char **) &node_info) {
    clust_node_free(node_info);
  }
  st_free_table(table);
}

static void clusters_free(clusters)
array_t *clusters;
{
  int i;
  cluster_t *cluster;

  for (i = 0; i < array_n(clusters); i++) {
    cluster = array_fetch(cluster_t *, clusters, i);
    cluster_free(cluster);
  }
  array_free(clusters);
}

 /* performs the labelling and returns the max label */

static int label_nodes(table, nodevec, cluster_size, verbose)
st_table *table;
array_t *nodevec;
int cluster_size;
int verbose;
{
  int i;
  int label;
  node_t *node;
  int max_label = -1;
  int nsize = array_n(nodevec);

  for (i = 0; i < nsize; i++) {
    node = array_fetch(node_t *, nodevec, i);
    if (node->type != INTERNAL) continue;
    clust_node_alloc(table, node);
    label = decide_label_node(table, node, cluster_size);
    if (label > max_label) max_label = label;
    if (verbose >= 3) {
      clust_node_t *node_info;
      assert(st_lookup(table, (char *) node, (char **) &node_info));
      printf("label of %s = %d, weight_same_label = %d\n", node_long_name(node), label, node_info->weight_same_label);
    }
  }
  return max_label;
}

static void form_clusters(table, nodevec, clusters)
st_table *table;
array_t *nodevec;
array_t *clusters;
{
  int i;
  node_t *node;
  cluster_t *new_cluster;
  clust_node_t *node_info;
  int nsize = array_n(nodevec);

  for (i = nsize - 1; i >= 0; i--) {
    node = array_fetch(node_t *, nodevec, i);
    if (node->type != INTERNAL) continue;
    assert(st_lookup(table, (char *) node, (char **) &node_info));
    if (node_info->is_in_cluster) continue;
    new_cluster = cluster_alloc();
    array_insert_last(cluster_t *, clusters, new_cluster);
    put_node_and_tfi_in_last_cluster(table, node, clusters);
  }
}

 /* just an interface to allow binary search without concern for memory leaks */
 /* returns the max label value obtained by Lawler's algorithm for the given cluster_size */

static int best_labelling(nodevec, cluster_size)
array_t *nodevec;
int cluster_size;
{
  int max_label;
  st_table *table = clust_node_table_alloc();

  max_label = label_nodes(table, nodevec, cluster_size, 0);
  clust_node_table_free(table);
  return max_label;
}

 /* given a function fn guaranteed to be non increasing */
 /* find the smallest integer x between min and max such that fn(x) >= value */
 /* the objective is to minimize the number of calls to fn */

static int lazy_binary_search(min, max, value, nodevec, fn)
int min;
int max;
int value;
array_t *nodevec;
int (*fn)();
{
  int middle;
  int min_value;
  int max_value;
  int middle_value;

  assert(min <= max);
  min_value = (*fn)(nodevec, min);
  if (min_value <= value) return min;
  max_value = (*fn)(nodevec, max);
  if (max_value > value) return max;
  for (;;) {
    middle = min + (max - min) / 2;
    if (middle == min) {
      assert(max_value <= value);
      return (min_value <= value) ? min : max;
    }
    middle_value = (*fn)(nodevec, middle);
    if (middle_value > value) {
      min = middle;
      min_value = middle_value;
    } else {
      max = middle;
      max_value = middle_value;
    }
  }
}

/*-------------------------------------------------------------------*
     The procedure looks at the node and its fanins and decides what 
     label should be given to the node. It puts the label and 
     weight_same_label in the node. Returns a label (int)
*---------------------------------------------------------------------*/

static int decide_label_node(table, node, cluster_size)
st_table *table;
node_t *node;
int cluster_size;
{
  int i;
  int max_label;
  node_t *fanin;
  int fanin_label;
  int total_weight, weight_nodes;
  clust_node_t *node_info;
  array_t *tfi;

  max_label = -1;		      /* not 0 */

  foreach_fanin(node, i, fanin) {
    if (fanin->type == PRIMARY_INPUT) continue;
    assert(st_lookup(table, (char *) fanin, (char **) &node_info));
    fanin_label = node_info->label;
    if (fanin_label <= max_label) continue;
    max_label = fanin_label;
  }     
  /* all fanins were PI's */
  /*----------------------*/
  assert(st_lookup(table, (char *) node, (char **) &node_info));
  if (max_label == -1) {
    node_info->label = 0;
    return 0;
  }
  /* this can  be put inside the loop */
  /*----------------------------------*/
  tfi = network_tfi(node, MAX_LEVEL);
  weight_nodes = sum_weight_nodes_with_max_label(table, tfi, max_label);
  total_weight = weight_nodes + node_info->weight;
  if (total_weight > cluster_size) {
    node_info->label = max_label + 1;
    node_info->weight_same_label = node_info->weight;
  } else {
    node_info->label = max_label;
    node_info->weight_same_label = total_weight;
  }
  array_free(tfi);
  return node_info->label;
}	

static void put_node_and_tfi_in_last_cluster(table, node, clusters)
st_table *table;
node_t *node;
array_t *clusters;
{
  int label;
  int i, tfi_size;
  node_t *tfi_node;
  array_t *tfi;
  cluster_t *last_cluster;
  int cluster_index;
  clust_node_t *node_info;
  clust_node_type_t type;

  if (node->type == PRIMARY_INPUT) return; /* making doubly sure */
  cluster_index = array_n(clusters) - 1;
  last_cluster = array_fetch(cluster_t *, clusters, cluster_index);
  assert(st_lookup(table, (char *) node, (char **) &node_info));
  label = node_info->label;
  array_insert_last(cluster_t *, node_info->clusters, last_cluster);
  node_info->is_in_cluster = 1;

  /* putting node in the cluster */
  /*-----------------------------*/
  last_cluster->number = cluster_index;
  type = ROOT_NODE;
  st_insert(last_cluster->nodes, (char *) node, (char *) type);

  /* put the appropriate predecessors in this cluster */
  /*--------------------------------------------------*/
  tfi = network_tfi(node, MAX_LEVEL);
  tfi_size = array_n(tfi);
  for (i = 0; i < tfi_size; i++) {
    tfi_node = array_fetch(node_t *, tfi, i);
    if (tfi_node->type == PRIMARY_INPUT) continue;
    assert(st_lookup(table, (char *) tfi_node, (char **) &node_info));
    if (node_info->label != label) continue;
    array_insert_last(cluster_t *, node_info->clusters, last_cluster);
    node_info->is_in_cluster = 1;
    type = get_clust_node_type(tfi_node, table);
    st_insert(last_cluster->nodes, (char *) tfi_node, (char *) type);
  }
  array_free(tfi);
}


 /* returns the sum of the weights of all nodes in tfi which have label 'label' */
static int sum_weight_nodes_with_max_label(table, tfi, label)
st_table *table;
array_t *tfi;
int label;
{
  int i, tfi_size;
  int weight;
  node_t *tfi_node;
  clust_node_t *node_info;

  weight = 0;
  tfi_size = array_n(tfi);
  for (i = 0; i < tfi_size; i++) {
    tfi_node = array_fetch(node_t *, tfi, i);
    if (tfi_node->type == PRIMARY_INPUT) continue;
    assert(st_lookup(table, (char *) tfi_node, (char **) &node_info));
    if (node_info->label == label) 
      weight += node_info->weight;
  }
  return weight;
}

 /* a node is of type ROOT_NODE iff it fanouts to a node with greater label */
 /* and of type NORMAL_NODE otherwise */

static clust_node_type_t get_clust_node_type(node, table)
node_t *node;
st_table *table;
{
  lsGen gen;
  node_t *fanout;
  int label;
  clust_node_t *node_info;
  clust_node_type_t type;

  assert(node_num_fanout(node) > 0);
  assert(st_lookup(table, (char *) node, (char **) &node_info));
  label = node_info->label;
  type = NORMAL_NODE;
  foreach_fanout(node, gen, fanout) {
    if (fanout->type == PRIMARY_OUTPUT) {
      type = ROOT_NODE;
      continue;
    }
    assert(st_lookup(table, (char *) fanout, (char **) &node_info));
    if (label < node_info->label) {
      type = ROOT_NODE;
    }
  }
  return type;
}


static cluster_t *cluster_alloc()
{
  cluster_t *result = ALLOC(cluster_t, 1);

  result->nodes = st_init_table(st_numcmp, st_numhash);
  result->number = 0;
  result->label = -1;
  return result;
}

static cluster_t *cluster_free(cluster)
cluster_t *cluster;
{
  st_free_table(cluster->nodes);
  FREE(cluster);
}

static void clust_node_alloc(table, node)
st_table *table;
node_t *node;
{
  clust_node_t *node_info;

  node_info = ALLOC(clust_node_t, 1);
  node_info->label = -1;
  node_info->weight = 1;
  node_info->weight_same_label  = 0;
  node_info->is_in_cluster = 0;
  node_info->clusters = array_alloc(cluster_t *, 0);
  st_insert(table, (char *) node, (char *) node_info);
}

static void clust_node_free(node_info)
clust_node_t *node_info;
{
  array_free(node_info->clusters);
  FREE(node_info);
}
  

 /* for debugging */

static void print_clusters(clusters)
array_t *clusters;
{
  int i;
  cluster_t *cluster;

  for (i = 0; i < array_n(clusters); i++) {
    cluster = array_fetch(cluster_t *, clusters, i);
    print_cluster(cluster);
  }
}

static void print_cluster(cluster)
cluster_t *cluster;
{
  clust_node_type_t type;
  node_t *node;
  int num_nodes;
  st_generator *gen;

  printf("cluster#%d ", cluster->number);
  num_nodes = st_count(cluster->nodes);
  printf("(number of nodes = %d) =>", num_nodes);
  st_foreach_item(cluster->nodes, gen, (char **) &node, (char **) &type) {
    switch (type) {
    case NORMAL_NODE:
      printf(" %s ", node->name);
      break;
    case ROOT_NODE:
      printf(" %s:root ", node->name);
      break;
    default:
      printf("\nunexpected clust_node type\n");
      break;
    }
  }
  printf("\n\n");
}

 /* collapsing the clusters */
 /* the lower the index in nodevec, the closer to the inputs */

static void collapse_clusters(clusters, nodevec)
array_t *clusters;
array_t *nodevec;
{
  int i;
  node_t *node;
  cluster_t *cluster;
  st_table *topological_order = st_init_table(st_numcmp, st_numhash);

  for (i = 0; i < array_n(nodevec); i++) {
    node = array_fetch(node_t *, nodevec, i);
    st_insert(topological_order, (char *) node, (char *) i);
  }
  for (i = 0; i < array_n(clusters); i++) {
    cluster = array_fetch(cluster_t *, clusters, i);
    collapse_cluster(cluster, topological_order);
  }
  st_free_table(topological_order);
}

typedef struct {
  int order;
  clust_node_type_t type;
  node_t *node;
} sorted_node_t;

 /* should put first those closer to PO */

static int sorted_node_cmp(obj1, obj2)
char *obj1;
char *obj2;
{
  sorted_node_t *c1 = (sorted_node_t *) obj1;
  sorted_node_t *c2 = (sorted_node_t *) obj2;

  return - (c1->order - c2->order);
}

 /* extract all the nodes in a cluster into an array */
 /* sort the array in such a way to put first the nodes closest to the outputs */
 /* for consistency, check that the first node in the array is marked "root" */
 /* collapse all the nodes in the array into the first node of the array */

/*
int
node_collapse(f, g)
node_t *f, *g;
	Assuming g is an input to f, re-express f without using g.
	Changes f in-place, g is unchanged.

*/

static void collapse_cluster(cluster, topological_order)
cluster_t *cluster;
st_table *topological_order;
{
  int i, j;
  st_generator *gen;
  node_t *net_node;
  int order;
  sorted_node_t new_node;
  sorted_node_t *root;
  sorted_node_t *node;
  clust_node_type_t type;
  int n_nodes = st_count(cluster->nodes);
  array_t *nodes = array_alloc(sorted_node_t, n_nodes);

  assert(n_nodes > 0);
  st_foreach_item(cluster->nodes, gen, (char **) &net_node, (char **) &type) {
    assert(st_lookup(topological_order, (char *) net_node, (char **) &order));
    new_node.node = net_node;
    new_node.order = order;
    new_node.type = type;
    array_insert_last(sorted_node_t, nodes, new_node);
  }
  array_sort(nodes, sorted_node_cmp);
  for (i = 0; i < n_nodes; i++) {
    root = array_fetch_p(sorted_node_t, nodes, i);
    if (root->type != ROOT_NODE) continue;
    for (j = i + 1; j < n_nodes; j++) {
      node = array_fetch_p(sorted_node_t, nodes, j);
      node_collapse(root->node, node->node);
    }
    for (j = 0; j < n_nodes; j++) {
      node = array_fetch_p(sorted_node_t, nodes, j);
      if (node_get_fanin_index(root->node, node->node) != -1) {
	(void) fprintf(miserr, "collapsing incorrect: node %s still appears in node %s\n", 
		       node->node->name, root->node->name);
	fail("");
      }
    }
  }
  array_free(nodes);
}

