 /* file com_map.c release 1.4 */
 /* last modified: 9/9/90 at 17:31:33 */
#include "mis.h"
#include "map_int.h"
#include "lib_int.h"
#include "bin_int.h"
#include "fanout_int.h"

static array_t *current_fanout_alg;
static int map_fill_options();
static void map_report_usage();

static int
com_replace(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  int c;
  int verbose = 0;

  getopt_reset();
  while ((c = getopt(argc, argv, "v:")) != EOF) {
    switch(c) {
    case 'v':
      verbose = atoi(optarg);
      break;
    default:
      goto usage;
    }
  }
  replace_2or(*network, verbose);
  return 0;
 usage:
  (void) fprintf(miserr, "_replace [-v verbosity level]\n");
  return 1;
}


 /* perform premapping, mapping, fanout optimization... */

static int
com_map(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  network_t *new_network;
  bin_global_t options;

  if (map_fill_options(&options, &argc, &argv)) {
    fprintf(miserr, "%s", error_string());
    map_report_usage("map");
    return 1;
  }
  new_network = complete_map_interface(*network, &options);
  if (new_network == 0) {
    fprintf(miserr, "%s", error_string());
    return 1;
  }
  network_free(*network);
  *network = new_network;
  return 0;
}

 /* builds a MAPPED network from an ANNOTATED network */

static int
com_build_map(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  network_t *new_network;
  bin_global_t globals;

  if (map_fill_options(&globals, &argc, &argv)) {
    fprintf(miserr, "%s", error_string());
    map_report_usage("map");
    return 1;
  }
  new_network = build_mapped_network_interface(*network, &globals);
  if (new_network == NIL(network_t)) {
    fprintf(miserr, "%s", error_string());
    return 1;
  }
  network_free(*network);
  *network = new_network;
  return 0;
}

/* takes a RAW or an ANNOTATED network and returns an ANNOTATED network */
 /* only do tree mapping */

static int
com_tree_map(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  network_t *new_network;
  bin_global_t options;

  if (map_fill_options(&options, &argc, &argv)) {
    fprintf(miserr, "%s", error_string());
    map_report_usage("map");
    return 1;
  }
  new_network = tree_map_interface(*network, &options);
  if (new_network == 0) {
    fprintf(miserr, "%s", error_string());
    return 1;
  }
  if (new_network != *network) {
    network_free(*network);
    *network = new_network;
  }
  return 0;
}

 /* takes an ANNOTATED network and returns an ANNOTATED network */
 /* performs fanout optimization and, if specified, area recovery */

static int
com_fanout_opt(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  network_t *new_network;
  bin_global_t options;

  if (map_fill_options(&options, &argc, &argv)) {
    fprintf(miserr, "%s", error_string());
    map_report_usage("map");
    return 1;
  }
  new_network = fanout_opt_interface(*network, &options);
  if (new_network == 0) {
    fprintf(miserr, "%s", error_string());
    return 1;
  }
  if (new_network != *network) {
    network_free(*network);
    *network = new_network;
  }
  return 0;
}

 /* can be used to add/remove fanout algorithms from the list of active alg's */ 

 /* ARGSUSED */
static int
com_fanout_alg(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  int no_options = 0;
  int verbose = 0;

  if (current_fanout_alg == NIL(array_t))
    current_fanout_alg = fanout_opt_get_fanout_alg(0, NIL(char *));
  argc--; argv++;
  if (argc == 0) no_options = 1;
  while (argc > 0 && argv[0][0] == '-') {
    switch (argv[0][1]) {
    case 'v':
      verbose = 1;
      break;
    default:
      goto usage;
    }
    argc--;
    argv++;
  }
  if (! no_options) current_fanout_alg = fanout_opt_get_fanout_alg(argc, argv);
  if (verbose || no_options) fanout_opt_print_fanout_alg(current_fanout_alg);
  return 0;
 usage:
  fprintf(miserr, "usage: fanout_alg [-v] fanout_alg1 fanout_alg2 ...\n");
  fprintf(miserr, "       -v\t verbose mode\n");
  return 1;
}

 /* can be used to give values to parameters of specific fanout algorithms */

 /* ARGSUSED */
static int
com_fanout_param(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  int fanout_alg_index;
  int verbose = 0;

  if (current_fanout_alg == NIL(array_t))
    current_fanout_alg = fanout_opt_get_fanout_alg(0, NIL(char *));
  argc--; argv++;
  while (argc > 0 && argv[0][0] == '-') {
    switch (argv[0][1]) {
    case 'v':
      verbose = 1;
      break;
    default:
      goto usage;
    }
    argc--;
    argv++;
  }
  fanout_opt_set_fanout_param(current_fanout_alg, argc, argv, &fanout_alg_index);
  if (fanout_alg_index == -1) goto usage;
  if (verbose) fanout_opt_print_fanout_param(current_fanout_alg, fanout_alg_index);
  return 0;
 usage:
  fprintf(miserr, "usage: fanout_param [-v] fanout_alg property value\n");
  fprintf(miserr, "       -v\t lists all (property, value) pairs associated with algorithm\n");
  fanout_opt_print_fanout_alg(current_fanout_alg);
  return 1;
}

static bin_global_t DEFAULT_OPTIONS = {
  0,			/* int new_mode; 	        // new_mode */
  0.0,			/* double new_mode_value;	// 0.0->area ... 1.0->delay */
  0,			/* int load_bins_count;		// historical */
  0,			/* int delay_bins_count;	// historical */
  0.0,			/* double old_mode;		// old map; area or delay optimize */
  1,			/* int inverter_optz;		// inverter optimization heuristics */
  -1,			/* int allow_internal_fanout;	// degrees of internal fanouts allowed */
  0,			/* int print_stat;		// print area/delay stats at the end */
  0,			/* int verbose;			// verbosity level */
  0,			/* int raw;			// raw mapping */
  0.0,			/* double thresh;		// threshold used in old tree mapper */
  0,			/* library_t *library;		// current library */
  0,			/* int fanout_optimize;		// if set, the fanout optimizer is called */
  0,			/* int area_recover;		// if set, area is recovered after fanout opt */
  0,			/* int peephole;		// if set, call fanout peephole optimizer */
  0,			/* array_t *fanout_alg;		// array of fanout algorithm descriptors */
  0,			/* int remove_inverters;	// if set, remove redundant inverters after mapping */
};

static int map_fill_options(options, argc_addr, argv_addr)
bin_global_t *options;
int *argc_addr;
char ***argv_addr;
{
  int c;
  double mode;
  int argc = *argc_addr;
  char **argv = *argv_addr;
  
  error_init();
  *options = DEFAULT_OPTIONS;
  if (current_fanout_alg == NIL(array_t)) {
    current_fanout_alg = fanout_opt_get_fanout_alg(0, NIL(char *));
  }
  options->fanout_alg = current_fanout_alg;
  if ((options->library = lib_get_library()) == NIL(library_t)) {
    error_append("map_interface: need to load a library with read_library\n");
    return 1;
  }
  getopt_reset();
  while ((c = getopt(argc, argv, "AFD:L:f:im:n:rst:v:")) != EOF) {
    switch(c) {
    case 'A':
      options->area_recover = 1;
      options->fanout_optimize = 1;
      break;
    case 'F':
      options->fanout_optimize = 1;
      break;
    case 'D':
      options->delay_bins_count = atoi(optarg);
      if (options->delay_bins_count < 2) {
	error_append("map_interface: number of delay bins should be >= 2\n");
	return 1;
      }
      break;
    case 'L':
      options->load_bins_count = atoi(optarg);
      if (options->load_bins_count < 2) {
	error_append("map_interface: number of load bins should be >= 2\n");
	return 1;
      }
      break;
    case 'f':
      options->allow_internal_fanout = atoi(optarg);
      break;
    case 'i':
      options->inverter_optz = 0;
      break;
    case 'm':
      options->old_mode = atof(optarg);
      if (options->old_mode > 1.0000 || options->old_mode < 0) {
	error_append("map_interface: -m n: n should be in interval (0,1)\n");
	return 1;
      }
      options->remove_inverters = (options->old_mode == 0.0);
      break;
    case 'n':
      options->new_mode = 1;
      options->new_mode_value = atof(optarg);
      mode = options->new_mode_value;
      if (mode < 0.0 || mode > 1.0) {
	error_append("map_interface: -n n: n should be in interval (0,1)\n");
	return 1;
      }
      options->remove_inverters = (mode == 0.0);
      break;
    case 'r':
      options->raw = 1;
      break;
    case 's':
      options->print_stat = 1;
      break;
    case 't':
      options->old_mode = 2.0;
      options->thresh = atof(optarg);
      break;
    case 'v':
      options->verbose = atoi(optarg);
      break;
    default:
      error_append("map_interface: unknown option\n");
      return 1;
    }
  }

 /* determine the default for internal fanout: 11 for min area, 0 otherwise */
  if (options->allow_internal_fanout == -1) {
    if (options->new_mode == 0) {
      options->allow_internal_fanout = (options->old_mode > 0.0) ? 0 : 11;
    } else {
      options->allow_internal_fanout = (options->new_mode_value > 0.0) ? 0 : 11;
    }
    if (options->fanout_optimize) options->allow_internal_fanout = 0;
  } else {
 /* if -f option was specified non zero with the fanout optimizer, disallow it */
    if (options->allow_internal_fanout != 0 && options->fanout_optimize) {
      options->allow_internal_fanout = 0;
      (void) fprintf(miserr, "WARNING: internal fanout disallowed with fanout optimization\n");
    }
  }
  *argc_addr = argc;
  *argv_addr = argv;
  return 0;
}

static void map_report_usage(command_name)
char *command_name;
{
  (void) fprintf(miserr, "usage: %s\n", command_name);
  (void) fprintf(miserr, "\t-v n\tselect verbosity level\n");
  (void) fprintf(miserr, "\t-A\trecover area after fanout optimization\n");
  (void) fprintf(miserr, "\t-F\tfanout optimize\n");
  (void) fprintf(miserr, "\t-D n\tset # delay_bins to n\n");
  (void) fprintf(miserr, "\t-L n\tset # load bins to n\n");
  (void) fprintf(miserr, "\t-f n\tallow internal fanout (n = 0: none)\n");
  (void) fprintf(miserr, "\t-i\tinverter optimization\n");
  (void) fprintf(miserr, "\t-m n\told map mode: n = 0.0 -> area, n = 1.0, delay\n");
  (void) fprintf(miserr, "\t-n n\tnew map mode: n = 0.0 -> area, n = 1.0, delay\n");
  (void) fprintf(miserr, "\t-r\traw map\n");
  (void) fprintf(miserr, "\t-s\tprint stats at the end\n");
  (void) fprintf(miserr, "\t-t n\told_mode with threshold n\n");
}


/* ARGSUSED */
static int
com_read_library(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  FILE *fp, *outfp;
  char *filename, *real_filename;
  int append, add_inverters, nand_flag, raw, c;

  append = 0; /* default is no-append */
  raw = 0; /* default is genlib format */
  add_inverters = 1; /* default adds inverters */
  nand_flag = 0; /* default is nor */
  getopt_reset();
  while ((c = getopt(argc, argv, "ainrbs")) != EOF) {
    switch(c) {
    case 'a':
      append = 1;
      break;
    case 'i':
      add_inverters = 0;
      break;
    case 'n':
      nand_flag = 1;
      break;
    case 'r':
      raw = 1;
      break;
    default:
      goto usage;
    }
  }
  if (argc - optind != 1) goto usage;

  filename = argv[optind];
  fp = com_open_file(filename, "r", &real_filename, /* silent */ 0);
  if (fp == NULL) goto error_return;

  /* free current library */
  if (! append) {
    if (lib_current_library != 0) {
      lib_free(lib_current_library);
      lib_current_library = 0;
    }
  }

  /* run the library though genlib if not reading it raw */
  if (! raw) {
    outfp = util_tmpfile();
    if (outfp == NULL) {
      fprintf(miserr, "Error opening temporary file\n");
      return 1;
    }
    if (! genlib_parse_library(fp, real_filename, outfp, ! nand_flag)) {
      fprintf(miserr, "%s", error_string());
      (void) fclose(fp);
      goto error_return; 
    }
    (void) fclose(fp);
    rewind(outfp);
    fp = outfp;
  }

  /* read a new library */
  if (! lib_read_blif(fp, real_filename, 
		      add_inverters, nand_flag, &lib_current_library)) {
    (void) fclose(fp);
    fprintf(misout, "%s", error_string());
    goto error_return;
  }

  (void) fclose(fp);
  FREE(real_filename);

  return 0;

 usage:
  fprintf(miserr, "usage: read_library [-adnr] filename\n");
  fprintf(miserr, "	-a\t\tappend to current library\n");
  fprintf(miserr, "	-i\t\tsuppress add inverters\n");
  fprintf(miserr, "	-n\t\tuse NAND gates (instead of NOR)\n");
  fprintf(miserr, "	-r\t\tread raw blif (rather than genlib)\n");
  return 1;

 error_return:
  FREE(real_filename);
  lib_current_library = 0;
  return 1;
}

typedef struct gate_count_struct gate_count_t;
struct gate_count_struct {
	char *name;
	int count;
	double area;
};


static void
gate_summary(node_vec)
array_t *node_vec;
{
  int i, total_gates;
  node_t *node;
  gate_count_t *count;
  lib_gate_t *gate;
  char *gatename, *key, *value;
  double total_area;
  avl_generator *avlgen;
  avl_tree *gate_count_table;

  gate_count_table = avl_init_table(strcmp);

  for(i = 0; i < array_n(node_vec); i++) {
    node = array_fetch(node_t *, node_vec, i);
    if (node->type == INTERNAL) {
      gate = lib_gate_of(node);
      if (gate == NIL(lib_gate_t)) {
	gatename = "<none>";
      } else {
	gatename = lib_gate_name(gate);
      }

      if (avl_lookup(gate_count_table, gatename, &value)) {
	count = (gate_count_t *) value;
	count->count++;
      } else {
	count = ALLOC(gate_count_t, 1);
	count->name = util_strsav(gatename);
	count->count = 1;
	count->area = lib_gate_area(gate);
	value = (char *) count;
	(void) avl_insert(gate_count_table, count->name, value);
      }
    }
  }

  total_area = 0;
  total_gates = 0;
  avl_foreach_item(gate_count_table, avlgen, AVL_FORWARD, &key, &value) {
    count = (gate_count_t *) value;
    fprintf(misout, "%-10s : %4d  (area=%4.2f)\n", 
	    count->name, count->count, count->area);
    total_area += count->count * count->area;
    total_gates += count->count;
  }
  fprintf(misout, 
	  "Total: %d gates, %4.2f area\n", total_gates, total_area);
  avl_free_table(gate_count_table, free, free);
}

static void
print_lib_gate_info(node, print_pins)
node_t *node;
int print_pins;
{
  int j;
  node_t *fanin;
  lib_gate_t *gate;
  char *pinname, *gatename;
  double area;

  if (node->type == INTERNAL) {
    gate = lib_gate_of(node);
    if (gate == NIL(lib_gate_t)) {
      gatename = "<none>";
      area = 0.0;
    } else {
      gatename = lib_gate_name(gate);
      area = lib_gate_area(gate);
    }

    fprintf(misout, "%-10s %-10s %4.2f",
	    node_name(node), gatename, area);
    if (print_pins && gate != NIL(lib_gate_t)) {
      fprintf(misout, "  (");
      foreach_fanin(node, j, fanin) {
	pinname = lib_gate_pin_name(gate, j, /* inflag */ 1);
	if (j != 0) fprintf(misout, " ");
	fprintf(misout, "%s=%s", pinname, node_name(fanin));
      }
      fprintf(misout, ")");
    }
    fprintf(misout, "\n");
  }
}

/* ARGSUSED */
static int
com_print_gate(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  node_t *node;
  array_t *node_vec;
  int i, print_pins, summary, c;

  print_pins = 0;
  summary = 0;
  getopt_reset();
  while ((c = getopt(argc, argv, "ps")) != EOF) {
    switch(c) {
    case 'p':
      print_pins = 1;
      break;
    case 's':
      summary = 1;
      break;
    default:
      goto usage;
    }
  }

  node_vec = com_get_nodes(*network, argc-optind+1, argv+optind-1);
  if (array_n(node_vec) < 1) goto usage;

  if (summary) {
    gate_summary(node_vec);
  } else {
    for(i = 0; i < array_n(node_vec); i++) {
      node = array_fetch(node_t *, node_vec, i);
      print_lib_gate_info(node, print_pins);
    }
  }
  array_free(node_vec);
  return 0;

 usage:
  fprintf(miserr, "usage: print_gate [-ps] n1 n2 ...\n");
  return 1;
}

/* ARGSUSED */
static int
com_treesize(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  if (argc - optind != 0) goto usage;

  map_print_tree_size(misout, *network);
  return 0;

 usage:
  fprintf(miserr, "usage: _treesize\n");
  return 1;
}

/* ARGSUSED */
static int
com_eat_buffer(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  lsGen gen;
  node_t *node;

  if (argc != 1) goto usage;
  foreach_primary_output(*network, gen, node) {
    (void) network_sweep_node(node);
  }
  (void) network_cleanup(*network);
  return 0;

 usage:
  fprintf(miserr, "usage: _eat_buffer\n");
  return 1;
}

/* ARGSUSED */
static int
com_premap(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  library_t *library;
  library = lib_get_library();

  if (library == 0) {
    fprintf(miserr, "_premap: no current library\n");
    return 1;
  }
  if (! map_check_form(*network, library->nand_flag)) {
    return 0;
  }

  error_init();
  if (! do_tree_premap(*network, library)) {
    fprintf(miserr, "%s", error_string());
  }
  return 0;
}

static int
com_map_addinv(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  int at_pipo, c;

  at_pipo = 0;
  getopt_reset();
  while ((c = getopt(argc, argv, "p")) != EOF) {
    switch(c) {
    case 'p':
      at_pipo = 1;
      break;
    default:
      goto usage;
    }
  }
  if (argc - optind != 0) goto usage;

  map_add_inverter(*network, at_pipo);
  return 0;

 usage:
  fprintf(miserr, "usage: _addinv [-p]\n");
  return 1;
}

/* ARGSUSED */
static int
com_map_reminv(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
	if (argc != 1) {
	fprintf(miserr, "usage: _reminv\n");
	return 1;
	}
	map_remove_inverter(*network, 0);
	return 0;
}

/* ARGSUSED */
static int
com_buffer(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  if (argc != 1) {
    fprintf(miserr, "usage: _buffer\n");
    return 1;
  }
  if (lib_get_library() == 0) {
    fprintf(miserr, "print_library: no current library\n");
    return 1;
  }
  buffer_inputs(*network, lib_get_library());
  return 0;
}

/* ARGSUSED */
static int
com_dump_patterns(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  if (lib_get_library() == 0) {
    fprintf(miserr, "print_library: no current library\n");
    return 1;
  }

  lib_dump(misout, lib_get_library(), /* detail */ 1);
  return 0;
}

/* ARGSUSED */
static int
com_print_library(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  lib_class_t *class;
  network_t *net1;

  if (lib_get_library() == 0) {
    fprintf(miserr, "print_library: no current library\n");
    return 1;
  }

  if (argc == 1) {
    lib_dump(misout, lib_get_library(), /* detail */ 0);

  } else if (argc == 2) {
    net1 = read_eqn_string(argv[1]);
    if (net1 == 0) {
      fprintf(miserr, "%s", error_string());
      return 1;
    }
    class = lib_get_class(net1, lib_get_library());
    if (class == 0) {
      fprintf(miserr, 
	      "print_library: cannot find class for given function\n");
      return 0;
    }
    lib_dump_class(misout, class);
    network_free(net1);

  } else {
    fprintf(miserr, "usage: print_library [function-string]\n");
    return 1;
  }

  return 0;
}

 /* extract data from a network. Contribution of Tzvi Ben-Tzur */

 /* ARGSUSED */
com_print_map_stats(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  lsGen gen;
  int num_out=0, count=0, num_inv=0, num_buf=0;
  double area=0.0;
  delay_time_t t;
  double temp, min_slack, total_slack;
  node_t *node;

  if (! lib_network_is_mapped(*network)) {
    (void)fprintf(misout,"Network not mapped\n");
    return 1;
  }
  assert(delay_trace(*network, DELAY_MODEL_LIBRARY));

  min_slack= 100000.0;
  total_slack = 0.0;

  foreach_node(*network, gen, node){
    if (node->type == INTERNAL){
      count++;
      area += lib_gate_area(lib_gate_of(node));
      if (node_function(node) == NODE_INV){
	num_inv++;
      } else if (node_function(node) == NODE_BUF){
	num_buf++;
      }
    } else if (node->type == PRIMARY_OUTPUT){
      t = delay_slack_time(node);
      temp = MIN(t.rise, t.fall);
      if (temp < 0){
	num_out++;
	min_slack = MIN(temp, min_slack);
	total_slack += temp;
      }
    }
  }
  (void) fprintf(misout,"Total Area\t\t= %6.2f\n", area);
  (void) fprintf(misout,"Gate Count\t\t= %d\n", count);
  (void) fprintf(misout,"Buffer Count\t\t= %d\n", num_buf);
  (void) fprintf(misout,"Inverter Count \t\t= %d\n", num_inv);
  min_slack = (min_slack < 0 ? min_slack : 0);
  total_slack = (total_slack < 0 ? total_slack : 0);
  (void) fprintf(misout,"Most Negative Slack\t= %6.2f\n", min_slack);
  (void) fprintf(misout,"Sum of Negative Slacks\t= %6.2f\n", total_slack);
  (void) fprintf(misout,"Number of Critical PO\t= %d\n", num_out);
  return 0;
}


 /* reduces depth by clustering using Lawler's algorithm */

extern network_t *cluster_under_size_constraint();
extern network_t *cluster_under_depth_constraint();
extern network_t *cluster_under_size_as_depth_constraint();

typedef enum {SIZE_CONSTRAINT, DEPTH_CONSTRAINT, SIZE_AS_DEPTH_CONSTRAINT} clustering_type_t;

com_reduce_depth(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  int c;
  int depth;
  int verbose;
  int cluster_size;
  int max_cluster_size;
  network_t *new_network;
  clustering_type_t type = SIZE_CONSTRAINT;
 
  /* default */
  verbose = 0;
  cluster_size = 8;

  getopt_reset();
  while ((c = getopt(argc, argv, "d:s:S:v:")) != EOF) {
    switch(c) {
    case 'd':
      type = DEPTH_CONSTRAINT;
      depth = atoi(optarg);
      if (depth <= 0) goto usage;
      break;
    case 's':
      type = SIZE_CONSTRAINT;
      cluster_size = atoi(optarg);
      if (cluster_size <= 0) goto usage;
      break;
    case 'S':
      type = SIZE_AS_DEPTH_CONSTRAINT;
      max_cluster_size = atoi(optarg);
      if (max_cluster_size <= 0) goto usage;
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    default:
      goto usage;
    }
  }
  if (argc - optind != 0) goto usage;
  switch (type) {
  case SIZE_CONSTRAINT:
    new_network = cluster_under_size_constraint(*network, cluster_size, verbose);
    break;
  case DEPTH_CONSTRAINT:
    new_network = cluster_under_depth_constraint(*network, depth, verbose);
    break;
  case SIZE_AS_DEPTH_CONSTRAINT:
    new_network = cluster_under_size_as_depth_constraint(*network, max_cluster_size, verbose);
    break;
  default:
    goto usage;
  }
  if (new_network == NIL(network_t)) return 1;
  *network = new_network;
  return 0;
 usage:
  (void) fprintf(miserr, "usage: _reduce_depth [-S max_cluster_size | -s cluster_size | -d max_level] -v verbosity_level\n");
  (void) fprintf(miserr, "     -d max_level\tmaximum level of nodes in final network\n");
  (void) fprintf(miserr, "     -s cluster_size\tmaximum number of nodes in a cluster (Default = 8)\n");
  (void) fprintf(miserr, "     -S max_cluster_size\tmaximum value of cluster_size; \
					pick up the smallest one yielding the same depth\n");
  return 1;
}

 /* timeout command: set up an alarm and send an interrupt to the process */
 /* when alarm is raised. With "-k" option, forces a fault and kill the process */

#include <signal.h>
static void timeout_interrupt()
{
  int pid = getpid();

  fprintf(miserr, "time out has occurred: process %d interrupted\n", pid);
  kill(pid, SIGINT);
}
static void timeout_kill()
{
  int pid = getpid();

  fprintf(miserr, "time out has occurred: process %d killed\n", pid);
  kill(pid, SIGKILL);
}
static void timeout_noop()
{}

 /* ARGSUSED */
static int
com_timeout(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
  int c;
  int timeout;
  int kill_process = 0;

  if (argc == 1) {
    (void) signal(SIGALRM, timeout_noop);
    return 0;
  }
  getopt_reset();
  while ((c = getopt(argc, argv, "t:k")) != EOF) {
    switch(c) {
    case 't':
      timeout = atoi(optarg);
      if (timeout <= 0 || timeout > 365 * 24 * 3600) goto usage;
      break;
    case 'k':
      kill_process = 1;
      break;
    default:
      goto usage;
    }
  }
  if (kill_process) {
    (void) signal(SIGALRM, timeout_kill);
  } else {
    (void) signal(SIGALRM, timeout_interrupt);
  }
  (void) alarm(timeout);
  return 0;
 usage:
  (void) fprintf(miserr, "usage: timeout [-t time_limit (seconds)] [-k]\n");
  (void) fprintf(miserr, "     -k\t kill the process instead of coming back to the prompt\n");
  return 1;
  
}

init_map()
{
  com_add_command("_replace", com_replace, /* changes_network */ 1);
  com_add_command("_reduce_depth", com_reduce_depth, /* changes_network */ 1);
  com_add_command("map", com_map, /* changes_network */ 1);
  com_add_command("_build_map", com_build_map, /* changes_network */ 1);
  com_add_command("_tree_map", com_tree_map, /* changes_network */ 1);
  com_add_command("_fanout_opt", com_fanout_opt, /* changes_network */ 1);
  com_add_command("fanout_alg", com_fanout_alg, /* changes_network */ 0);
  com_add_command("fanout_param", com_fanout_param, /* changes_network */ 0);
  com_add_command("read_library", com_read_library, /* changes_network */ 0);
  com_add_command("print_gate", com_print_gate, /* changes_network */ 0);
  com_add_command("print_library", com_print_library, /*changes_network*/ 0);
  com_add_command("print_map_stats", com_print_map_stats, 0);
  com_add_command("_dump_patterns", com_dump_patterns, /*changes_network*/ 0);
  com_add_command("_addinv", com_map_addinv, /* changes_network */ 1);
  com_add_command("_eat_buffer", com_eat_buffer, /* changes_network */ 1);
  com_add_command("_reminv", com_map_reminv, /* changes_network */ 1);
  com_add_command("_treesize", com_treesize, /* changes_network */ 0);
  com_add_command("_premap", com_premap, /* changes_network */ 1);
  com_add_command("_buffer", com_buffer, /* changes_network */ 1);
  com_add_command("timeout", com_timeout, /* changes_network */ 0);

  node_register_daemon(DAEMON_FREE, map_free);
  node_register_daemon(DAEMON_INVALID, map_invalid);
  node_register_daemon(DAEMON_DUP, map_dup);
}

end_map()
{
	if (lib_current_library != 0) {
		lib_free(lib_current_library);
	}
}
