 /* file fanout_log.c release 1.2 */
 /* last modified: 9/9/90 at 17:31:36 */
#include "mis.h"
#include "map_macros.h"
#include "map_int.h"
#include "fanout_int.h"
#include "gate_link.h"

 /* this file keeps a log of all modifications to MAP(node)->save_binding */
 /* during fanout optimization, and use it at the end to restore a network */
 /* where all nodes  */

typedef struct fanout_log_struct fanout_log_t;
struct fanout_log_struct {
  node_t *node;
  int pin;
  node_t *input;
};

static array_t *source_log;
static array_t *sink_log;
static void fanout_process_log();
static void fanout_process_sink_log();
static void fanout_process_source_log();
static void fanout_remove_unused_nodes();
static void network_delete_node_rec();
static node_t *get_matching_fanout();


 /* EXTERNAL INTERFACE */

void fanout_init_log()
{
  source_log = array_alloc(fanout_log_t, 0);
  sink_log = array_alloc(fanout_log_t, 0);
}

void fanout_add_to_log(node, pin, input, type)
node_t *node;
int pin;
node_t *input;
log_type_t type;
{
  fanout_log_t log_entry;

  log_entry.node = node;
  log_entry.pin = pin;
  log_entry.input = input;
  switch (type) {
  case SOURCE_UPDATE:
    array_insert_last(fanout_log_t, source_log, log_entry);
    break;
  case SINK_UPDATE:
    array_insert_last(fanout_log_t, sink_log, log_entry);
    break;
  default:
    fail("illegal type of log entry");
    break;
  }
}

void fanout_cleanup_log(network)
network_t *network;
{
  network_type_t type;
  
  fanout_process_source_log(source_log);
  fanout_process_sink_log(network, sink_log);
  fanout_remove_unused_nodes(network);
  array_free(source_log);
  array_free(sink_log);
  type = map_get_network_type(network);
  assert(type == ANNOTATED_NETWORK);
}


 /* INTERNAL INTERFACE */


 /* just a consistency check. Should be removed ultimately */

static void fanout_process_source_log(log)
array_t *log;
{
  int i;
  fanout_log_t *entry;

  for (i = 0; i < array_n(log); i++) {
    entry = array_fetch_p(fanout_log_t, log, i);
    assert(entry->input == NIL(node_t));
    assert(MAP(entry->node)->gate == NIL(lib_gate_t));
    assert(gate_link_is_empty(entry->node));
  }
}


 /* if node in the log is no longer mapped */
 /* we can safely ignore it: nothing to modify */

 /* ARGSUSED */
static void fanout_process_sink_log(network, log)
network_t *network;
array_t *log;
{
  int i;
  fanout_log_t *entry;
  node_t **addr;
  char *dummy;
  st_table *visited;
  node_t *old_source, *new_source, *gate_pin_node;

  if (array_n(log) == 0) return;
  visited = st_init_table(st_ptrcmp, st_ptrhash);
  for (i = 0; i < array_n(log); i++) {
    entry = array_fetch_p(fanout_log_t, log, i);
    old_source = entry->input;
    assert(old_source != NIL(node_t));
    if (entry->node->type == PRIMARY_OUTPUT) continue;
    if (MAP(entry->node)->gate == NIL(lib_gate_t)) continue;
    addr = &(MAP(entry->node)->save_binding[entry->pin]);
    if (! st_lookup(visited, (char *) addr, (char **) &dummy)) {
      st_insert(visited, (char *) addr, NIL(char));
      gate_pin_node = get_matching_fanout(old_source, entry->node);
      new_source = *addr;
      assert(new_source != NIL(node_t));
      node_patch_fanin(gate_pin_node, old_source, new_source);
    }
  }
  st_free_table(visited);
}

static void fanout_remove_unused_nodes(network)
network_t *network;
{
  lsGen gen;
  node_t *node;
  
  foreach_node(network, gen, node) {
    if (node->type == PRIMARY_OUTPUT) continue;
    if (node_num_fanout(node) > 0) continue;
    network_delete_node_rec(network, node);
  }
}

static node_t *get_matching_fanout(source, root)
node_t *source;
node_t *root;
{
  int i;
  node_t **gate_pins;
  node_t **trans_fanouts;
  int n_fanouts = node_num_fanout(source);

  assert(n_fanouts > 0);
  if (n_fanouts == 1) return node_get_fanout(source, 0);
  gate_pins = ALLOC(node_t *, n_fanouts);
  trans_fanouts = ALLOC(node_t *, n_fanouts);
  for (i = 0; i < n_fanouts; i++) {
    trans_fanouts[i] = gate_pins[i] = node_get_fanout(source, i);
    if (trans_fanouts[i] == root) return gate_pins[i];
  }
  for (;;) {
    int nothing_changed = 1;
    for (i = 0; i < n_fanouts; i++) {
      if (node_num_fanout(trans_fanouts[i]) != 1) continue;
      trans_fanouts[i] = node_get_fanout(trans_fanouts[i], 0);
      nothing_changed = 0;
      if (trans_fanouts[i] == root) return gate_pins[i];
    }
    assert(! nothing_changed);
  }
}


static void network_delete_node_rec(network, node)
network_t *network;
node_t *node;
{
  node_t *fanin;

  assert(node_function(node) == NODE_INV);
  assert(MAP(node)->gate == NIL(lib_gate_t));
  assert(node->type != PRIMARY_INPUT && node->type != PRIMARY_OUTPUT);
  fanin = node_get_fanin(node, 0);
  network_delete_node(network, node);
  if (node_num_fanout(fanin) > 0) return;
  network_delete_node_rec(network, fanin);
}
