 /* file fanout_opt.c release 1.3 */
 /* last modified: 9/9/90 at 17:31:37 */
#include <math.h>
#include "mis.h"
#include "map_macros.h"
#include "map_int.h"
#include "fanout_int.h"
#include "fanout_delay.h"

static int compare_gate_links();
static int extract_fanout_problem();
static int fanout_map_optimal();
static node_t *keep_external_source_only();
static st_table *extract_links();
static void do_fanout_opt();
static void fanout_opt_end();
static void fanout_opt_init();



static bin_global_t global;


 /* EXTERNAL INTERFACE */

void fanout_opt_update_peephole_flag(alg, property)
fanout_alg_t *alg;
alg_property_t *property;
{
  alg->peephole_flag = property->value;
}

void fanout_opt_update_size_threshold(alg, property)
fanout_alg_t *alg;
alg_property_t *property;
{
  alg->min_size = property->value;
}

 /* called by the command line option */

array_t *fanout_opt_get_fanout_alg(n_entries, entries)
int n_entries;
char **entries;
{
  int i, j;
  fanout_alg_t fanout_alg;
  fanout_alg_t *alg;
  alg_property_t property;
  static array_t *list_of_algs = NIL(array_t);

  if (list_of_algs == NIL(array_t)) {
    list_of_algs = array_alloc(fanout_alg_t, 0);
#define ALG_DEF(NAME,STRING)\
    fanout_alg.on_flag = 0;\
    fanout_alg.name = STRING;\
    fanout_alg.properties = array_alloc(alg_property_t, 0);\
    fanout_alg.init = name2(NAME,_init);\
    array_insert_last(fanout_alg_t, list_of_algs, fanout_alg);\
    alg = array_fetch_p(fanout_alg_t, list_of_algs, array_n(list_of_algs) - 1);
#define ALG_PROPERTY(FN,STRING,DEFAULT_VALUE)\
    property.name = STRING;\
    property.value = DEFAULT_VALUE;\
    property.update_fn = FN;\
    array_insert_last(alg_property_t, alg->properties, property);
#include "fanout_alg.macro.h"
#undef ALG_DEF
#undef ALG_PROPERTY
  }
  for (j = 0; j < array_n(list_of_algs); j++) {
    alg = array_fetch_p(fanout_alg_t, list_of_algs, j);
    alg->on_flag = (strcmp(alg->name, "noalg") == 0);
  }
  for (i = 0; i < n_entries; i++) {
    for (j = 0; j < array_n(list_of_algs); j++) {
      alg = array_fetch_p(fanout_alg_t, list_of_algs, j);
      if (strcmp(alg->name, entries[i]) == 0)
	alg->on_flag = 1;
    }
  }    
  return list_of_algs;
}

void fanout_opt_print_fanout_alg(array)
array_t *array;
{
  int i;
  fanout_alg_t *alg;

  fprintf(misout, "fanout algs in use: ");
  for (i = 0; i < array_n(array); i++) {
    alg = array_fetch_p(fanout_alg_t, array, i);
    if (alg->on_flag) fprintf(misout, "\"%s\" ", alg->name);
  }
  fprintf(misout, "\n");

  fprintf(misout, "fanout algs not in use: ");
  for (i = 0; i < array_n(array); i++) {
    alg = array_fetch_p(fanout_alg_t, array, i);
    if (! alg->on_flag) fprintf(misout, "\"%s\" ", alg->name);
  }
  fprintf(misout, "\n");
}

void fanout_opt_set_fanout_param(array, n_entries, entries, alg_index)
array_t *array;
int n_entries;
char **entries;
int *alg_index;
{
  int i;
  int entry_index;
  fanout_alg_t *alg;
  alg_property_t *property;

  *alg_index = -1;
  if (n_entries == 0) return;
  for (i = 0; i < array_n(array); i++) {
    alg = array_fetch_p(fanout_alg_t, array, i);
    if (strcmp(alg->name, entries[0]) == 0) {
      *alg_index = i;
      n_entries--;
      entries++;
      break;
    }
  }
  if (n_entries == 0 || *alg_index == -1) return;
  entry_index = -1;
  for (i = 0; i < array_n(alg->properties); i++) {
    property = array_fetch_p(alg_property_t, alg->properties, i);
    if (strcmp(property->name, entries[0]) == 0) {
      entry_index = i;
      n_entries--;
      entries++;
      break;
    }
  }
  if (n_entries != 1 || entry_index == -1) {
    *alg_index = -1;
    return;
  }
  property->value = atoi(entries[0]);
}

void fanout_opt_print_fanout_param(array, index)
array_t *array;
int index;
{
  int i;
  alg_property_t *property;
  fanout_alg_t *alg = array_fetch_p(fanout_alg_t, array, index);
  
  fprintf(misout, "property list of alg \"%s\"\n", alg->name);
  for (i = 0; i < array_n(alg->properties); i++) {
    property = array_fetch_p(alg_property_t, alg->properties, i);
    fprintf(misout, "%s --> %d\n", property->name, property->value);
  }
}


void fanout_optimization(network, globals)
     network_t *network;
     bin_global_t globals;
{
  global = globals;
  if (! global.fanout_optimize && ! global.area_recover) return;
  fanout_opt_init(network);
  virtual_network_setup_gate_links(network);
  virtual_network_remove_wires(network);
  virtual_network_flatten(network);
  virtual_delay_compute_arrival_times(network);
  virtual_delay_set_po_negative_required(network);
  bin_for_all_nodes_outputs_first(network, do_fanout_opt);
  if (global.area_recover) {
    virtual_delay_compute_arrival_times(network);
    virtual_network_flatten(network);
    virtual_delay_set_po_required(network);
    bin_for_all_nodes_outputs_first(network, do_fanout_opt);
  }
  fanout_opt_end(network);
}


  /* INTERNAL INTERFACE */

static void fanout_opt_init(network)
network_t *network;
{
  int i;
  int j;
  fanout_alg_t *alg;
  alg_property_t *property;

  for (i = 0; i < array_n(global.fanout_alg); i++) {
    alg = array_fetch_p(fanout_alg_t, global.fanout_alg, i);
    if (alg->on_flag) {
      for (j = 0; j < array_n(alg->properties); j++) {
	property = array_fetch_p(alg_property_t, alg->properties, j);
	(*property->update_fn)(alg, property);
      }
      (*alg->init)(network, alg);
    }
  }
  fanout_delay_init();
  fanout_init_log();
}


static void fanout_opt_end(network)
network_t *network;
{
  fanout_cleanup_log(network);
  fanout_delay_end();
}


static void do_fanout_opt(node_x)
    node_t *node_x;
{
  switch (node_function(node_x)) {
  case NODE_UNDEFINED:
    /* former po removed */
    break;
  case NODE_0:
  case NODE_1:
    (void) fprintf(miserr, "WARNING: constants should have been removed earlier\n");
    break;
  case NODE_PO:
 /* already processed */
    break;
  case NODE_BUF:
    fail("should not be any buffers in underlying network");
    break;
  case NODE_INV:
    if (MAP(node_x)->gate != NIL(lib_gate_t)) 
      virtual_delay_compute_node_required_time(node_x);
    break;
  case NODE_PI:
  case NODE_AND:
  case NODE_OR:
  case NODE_COMPLEX:
    if (fanout_map_optimal(node_x) == 0) {
      if (node_x->type == PRIMARY_INPUT || MAP(node_x)->gate != NIL(lib_gate_t))
	virtual_delay_compute_node_required_time(node_x);
    }
    break;
  }
}

 /* needed for K.J.'s code */
static node_t *current_root_node;

 /* returns 1 iff fanout optimization was performed */


static int fanout_map_optimal(node_x)
     node_t *node_x;
{
  int i, p;
  node_t *sources[2];
  opt_array_t fanout_info[2];
  array_t **trees;
  fanout_cost_t *costs;
  fanout_alg_t *alg;
  int best_sol;
  fanout_cost_t best_cost;
  int n_algs = array_n(global.fanout_alg);

 /* extract sink information */
  if (! extract_fanout_problem(node_x, sources, fanout_info)) return 0;
  current_root_node = node_x;
  if (sources[POLAR_X] != NIL(node_t))
    fanout_delay_add_source(sources[POLAR_X], POLAR_X);
  if (sources[POLAR_Y] != NIL(node_t))
    fanout_delay_add_source(sources[POLAR_Y], POLAR_Y);

 /* iterate through algorithms */
  trees = ALLOC(array_t *, n_algs);
  costs = ALLOC(fanout_cost_t, n_algs);
  for (i = 0; i < n_algs; i++) {
    alg = array_fetch_p(fanout_alg_t, global.fanout_alg, i);
    trees[i] = fanout_tree_alloc();
    costs[i].slack = MINUS_INFINITY;
    costs[i].area = INFINITY;
    if (alg->min_size > fanout_info[POLAR_X].n_elts + fanout_info[POLAR_Y].n_elts) continue;
    if (alg->on_flag && (*alg->optimize)(fanout_info, trees[i], &costs[i])) {
      fanout_tree_add_edges(trees[i]);
      fanout_tree_check(trees[i], &costs[i], fanout_info);
      if (alg->peephole_flag) 
	fanout_tree_peephole_optimize(trees[i], &costs[i], fanout_info, global.verbose);
    }
  }

 /* remove unused nodes from virtual network */
  if (MAP(node_x)->node_y != NIL(node_t) && sources[POLAR_Y] == NIL(node_t))
    virtual_network_remove_node(MAP(node_x)->node_y, 0);
  if (sources[POLAR_X] == NIL(node_t))
    virtual_network_remove_node(node_x, 0);

 /* select best solution and build it */
  best_sol = -1;
  best_cost.slack = MINUS_INFINITY;
  best_cost.area = INFINITY;
  for (i = 0; i < n_algs; i++) {
    if (fanout_opt_is_better_cost(&costs[i], &best_cost)) {
      best_cost = costs[i];
      best_sol = i;
    }
  }
  assert(best_sol != -1);
  fanout_tree_build(trees[best_sol]);

 /* clean up */
  for (i = 0; i < n_algs; i++)
    fanout_tree_free(trees[i]);
  FREE(trees);
  FREE(costs);
  fanout_delay_free_sources();
  foreach_polarity(p) { fanout_info_free(&fanout_info[p], 1); }
  return 1;
}


  /* just a routine to call initialization routines */
  /* returns 0 and cleanup if there is not enough to optimize */

static int extract_fanout_problem(root, sources, fanout_info)
     node_t *root;
     node_t **sources;
     opt_array_t *fanout_info;
{
  int p, q;

  sources[POLAR_X] = root;
  sources[POLAR_Y] = MAP(root)->node_y;
  foreach_polarity(p) {
    q = POLAR_INV(p);
    fanout_info[p].links = extract_links(sources[p], sources[q]);
    fanout_info_preprocess(&fanout_info[p]);
  }
  foreach_polarity(p) {
    sources[p] = keep_external_source_only(sources[p]);
  }
  if (fanout_info[POLAR_X].n_elts + fanout_info[POLAR_Y].n_elts >= 1) return 1;
  foreach_polarity(p) { fanout_info_free(&fanout_info[p], 1); }
  return 0;
}

static st_table *extract_links(source, exception)
node_t *source;
node_t *exception;
{
  gate_link_t link, *link_ptr;
  st_table *result = st_init_table(st_ptrcmp, st_ptrhash);

  if (source == NIL(node_t)) return result;
  if (gate_link_is_empty(source)) return result;
  gate_link_first(source, &link);
  do {
    if (link.node == exception) continue;
    link_ptr = ALLOC(gate_link_t, 1);
    *link_ptr = link;
    st_insert(result, (char *) link_ptr, NIL(char));
  } while (gate_link_next(source, &link));
  return result;
}


 /* extract the sinks of same polarity as source */
 /* except if sink is the only possible local_node */

void fanout_info_preprocess(fanout_info)
     opt_array_t *fanout_info;
{
  int n_elts;
  int i = 0;
  gate_link_t *key;
  double acc_load;
  delay_time_t required;
  char *dummy_value = NIL(char);
  st_generator *gen;

  n_elts = fanout_info->n_elts = st_count(fanout_info->links);
  if (n_elts == 0) {
    fanout_info->total_load = 0;
    return;
  }
  fanout_info->required = array_alloc(gate_link_t *, n_elts);
  st_foreach_item(fanout_info->links, gen, (char **) &key, &dummy_value) {
    array_insert(gate_link_t *, fanout_info->required, i, key);
    i++;
  }
  array_sort(fanout_info->required, compare_gate_links);

  fanout_info->cumul_load = ALLOC(double, n_elts + 1);
  acc_load = 0.0;
  for (i = 0; i < n_elts; i++) {
    gate_link_t *link = array_fetch(gate_link_t *, fanout_info->required, i);
    fanout_info->cumul_load[i] = acc_load;
    acc_load += link->load;
  }
  fanout_info->total_load = fanout_info->cumul_load[n_elts] = acc_load;

  fanout_info->min_required = ALLOC(delay_time_t, n_elts);
  required = PLUS_INFINITY;
  for (i = n_elts - 1; i >= 0; i--) {
    gate_link_t *link = array_fetch(gate_link_t *, fanout_info->required, i);
    SETMIN(required, required, link->required);
    fanout_info->min_required[i] = required;
  }
}


 /* keeps node if it is a source external to fanout tree */
static node_t *keep_external_source_only(node)
     node_t *node;
{
  if (node == NIL(node_t)) return node;
  if (node->type == PRIMARY_INPUT) return node;
  if (MAP(node)->gate == NIL(lib_gate_t)) return NIL(node_t);
  if (MAP(node)->ninputs > 1) return node;
  return NIL(node_t);
}


  /* compares two gate links, by required times */

static int compare_gate_links(obj1, obj2)
     char *obj1, *obj2;
{
  gate_link_t *g1 = *(gate_link_t **) obj1;
  gate_link_t *g2 = *(gate_link_t **) obj2;
  double cmp = GETMIN(g1->required) - GETMIN(g2->required);
  if (cmp < 0) return -1;
  if (cmp > 0) return 1;
  return 0;
}


 /* the "free_link_flag" is for bottom_up */
 /* it should be 1 if used here in fanout_opt.c */
 /* and 0 if used in bottom_up.c to avoid freeing twice */

void fanout_info_free(fanout_info, free_link_flag)
     opt_array_t *fanout_info;
     int free_link_flag;
{
  gate_link_t *key;
  char *dummy_value = NIL(char);
  st_generator *gen;

  if (free_link_flag) {
    st_foreach_item(fanout_info->links, gen, (char **) &key, &dummy_value) {
      FREE(key);
    }
  }
  st_free_table(fanout_info->links);
  if (fanout_info->n_elts == 0) return;
  array_free(fanout_info->required);
  FREE(fanout_info->cumul_load);
  FREE(fanout_info->min_required);
}

int fanout_opt_is_better_cost(cost1, cost2)
fanout_cost_t *cost1;
fanout_cost_t *cost2;
{
  if (GETMIN(cost2->slack) < 0) {
    return (GETMIN(cost1->slack) > GETMIN(cost2->slack)) ? 1 : 0;
  } else if (GETMIN(cost1->slack) < 0) {
    return 0;
  } else {
    return (cost1->area < cost2->area) ? 1 : 0;
  }
}


 /* exported to K.J.'s buffering algorithm */

node_t *fanout_opt_get_root()
{
  int p;
  node_t *source;
  node_t *sources[2];

  sources[POLAR_X] = current_root_node;
  sources[POLAR_Y] = MAP(current_root_node)->node_y;
  foreach_polarity(p) {
    source = keep_external_source_only(sources[p]);
    if (source != NIL(node_t)) return source;
  }
  return NIL(node_t);
}

node_t *fanout_opt_get_root_inv()
{
  int p;
  node_t *source;
  node_t *sources[2];

  sources[POLAR_X] = current_root_node;
  sources[POLAR_Y] = MAP(current_root_node)->node_y;
  source = fanout_opt_get_root();
  foreach_polarity(p) {
    if (source == sources[p]) 
      return sources[POLAR_INV(p)];
  }
  return NIL(node_t);
}
