 /* file library.c release 1.2 */
 /* last modified: 9/9/90 at 17:31:42 */
#include "mis.h"
#include "map_int.h"

static network_t *lib_wire();
static int process_pattern();
static int process_constant();
static prim_t *find_or_create_pattern();
static lib_gate_t *find_gate();
static lib_gate_t *find_or_create_gate();
static lib_class_t *find_class();
static lib_class_t *find_or_create_class();
static void derive_dual_classes();
static network_t *dual_network();

int
lib_read_blif(fp, filename, add_inverter, nand_flag, library_p)
FILE *fp;
char *filename;
int add_inverter;
int nand_flag;
library_t **library_p;
{
    int new_gates, c;
    library_t *library;
    network_t *network, *wire;

    lib_reset_default_inverter();
    library = *library_p;
    if (library == 0) {
	library = ALLOC(library_t, 1);
	library->gates = lsCreate();
	library->classes = lsCreate();
	library->patterns = lsCreate();
	library->nand_flag = nand_flag;
	library->add_inverter = add_inverter;
    }

    new_gates = 0;
    read_register_filename(filename);
    while ((c = read_blif(fp, &network)) != -1) {
	if (! c) return 0;

	if (network_num_pi(network) == 0) {
	    /* process constant 0 or 1 cell */
	    if (! process_constant(library, network)) {
		return 0;
	    }

	} else {
	    /* make sure its a 2-input NAND (NOR) gate network */
	    if (! map_check_form(network, library->nand_flag)) {
		return 0;
	    }

	    /* add double-inverter pair to all internal nodes */
	    if (library->add_inverter) {
		map_add_inverter(network, /* add_at_pipo */ 0);
	    }

	    /* add the pattern and gate to the library */
	    if (! process_pattern(library, network)) {
		return 0;
	    }
	}
	new_gates = 1;
    }

    /* Make sure the library has a wire primitive */
    wire = lib_wire();
    if (! find_gate(library, wire)) {
	if (! process_pattern(library, wire)) {
	    return 0;
	}
	new_gates = 1;
    } else {
	network_free(wire);
    }

    if (new_gates) {
	derive_dual_classes(library);
    }

    *library_p = library;
    return 1;
}

static int
process_pattern(library, network)
library_t *library;
network_t *network;
{
    lib_class_t *class;
    lib_gate_t *gate;
    prim_t *prim;

    map_setup_network(network);

    prim = find_or_create_pattern(library, network);
    if (! prim) return 0;

    gate = find_or_create_gate(library, network);
    if (! gate) return 0;

    class = find_or_create_class(library, gate);
    if (! class) return 0;

    lsNewEnd(prim->gates, (char *) gate, LS_NH);

    return 1;
}



static int
process_constant(library, network)
library_t *library;
network_t *network;
{
    node_t *node;
    node_function_t func;
    lib_gate_t *gate;
    lib_class_t *class;

    if (network_num_po(network) != 1) {
	error_append("no-input function is not single-output ?\n");
	return 0;
    }
    node = node_get_fanin(network_get_po(network, 0), 0);
    func = node_function(node);
    if (func == NODE_0 || func == NODE_1) {
	gate = find_or_create_gate(library, network);
	if (! gate) return 0;

	class = find_or_create_class(library, gate);
	if (! class) return 0;
    } else {
	error_append("no-input function is not a constant ?\n");
	return 0;
    }
    return 1;
}

static int match_flag;


static int 
found_match(prim)
prim_t *prim;
{
    int i;

    for(i = 0; i < prim->ninputs; i++) {
	if (prim->inputs[i]->binding->type != PRIMARY_INPUT) {
	    return 1;	/* not a match, continue search ... */
	}
    }
    match_flag = 1;
    return 0;		/* stop search now ... */
}


int
lib_pattern_matches(network, prim)
network_t *network;
prim_t *prim;
{
    node_t *po;

    if (network_num_pi(network) != prim->ninputs) {
	return 0;
    }
    if (network_num_po(network) != prim->noutputs) {
	return 0;
    }

    po = node_get_fanin(network_get_po(network, 0), 0);
    match_flag = 0;
    gen_all_matches(po, prim, found_match, /* debug */  0, /* allow_internal_fanout */ 0);
    return match_flag;
}


static prim_t *
find_or_create_pattern(library, network)
library_t *library;
network_t *network;
{
    lsGen gen;
    prim_t *prim;

    /* see if this is a new pattern, or duplicates some existing pattern */ 
    lsForeachItem(library->patterns, gen, prim) {
	if (lib_pattern_matches(network, prim)) {
	    (void) lsFinish(gen);
	    return prim;
	}
    }

    if (! map_network_to_prim(network, &prim)) {
	return 0;
    }
    lsNewEnd(library->patterns, (char *) prim, LS_NH);
    return prim;
}

static lib_gate_t *
find_gate(library, network)
library_t *library;
network_t *network;
{
    lsGen gen;
    lib_gate_t *gate;
    char *netname;

    netname = network_name(network);
    lsForeachItem(library->gates, gen, gate) {
	if (strcmp(gate->name, netname) == 0) {
	    /* ### should check that area, delay, function are the same ... */
	    (void) lsFinish(gen);
	    return gate;
	}
    }
    return 0;
}


static lib_gate_t *
find_or_create_gate(library, network)
library_t *library;
network_t *network;
{
    int i;
    lib_gate_t *gate;
    char errmsg[1024];

    gate = find_gate(library, network);
    if (gate == 0) {
	/* collapse network, re-order fanin of each node */
	node_lib_process(network);

	gate = ALLOC(lib_gate_t, 1);
	gate->name = util_strsav(network_name(network));
	gate->network = network;
	gate->area = network->area;
	gate->delay_info = delay_giveaway_pin_delay(network);
	gate->class = 0;
	lsNewEnd(library->gates, (char *) gate, LS_NH);

	if (! network->area_given) {
	    (void) sprintf(errmsg, "'%s': area not given", gate->name);
	    read_error(errmsg);
	    return 0;
	}
	if (strcmp(network_name(network), "**wire**") != 0) {
	    for(i = network_num_pi(network)-1; i >= 0; i--) {
		if (gate->delay_info[i] == 0) {
		    (void) sprintf(errmsg, 
			"'%s': missing delay information for pin '%s'", 
			gate->name, node_name(network_get_pi(network, i)));
		    read_error(errmsg);
		    return 0;
		}
	    }
	}
    } else {
	network_free(network);
    }
    return gate;
}

/*
 *  compare two networks for equality (by name) of all outputs
 *  the networks should already be collapsed.
 */

static int
compare_network(net1, net2)
network_t *net1, *net2;
{
    lsGen gen;
    node_t *po1, *po2;

    if (network_num_pi(net1) != network_num_pi(net2)) {
	return 1;
    }
    if (network_num_po(net1) != network_num_po(net2)) {
	return 1;
    }

    foreach_primary_output(net1, gen, po1) {
	po2 = network_find_node(net2, po1->name);
	if (po2 == 0) {
	    (void) lsFinish(gen);
	    return 1;
	}
	po1 = node_get_fanin(po1, 0);
	po2 = node_get_fanin(po2, 0);
	if (! node_equal_by_name(po1, po2)) {
	    (void) lsFinish(gen);
	    return 1;
	}
    }
    return 0;
}

static lib_class_t *
find_class(library, network)
library_t *library;
network_t *network;
{
    lsGen gen;
    lib_class_t *class;

    lsForeachItem(library->classes, gen, class) {
	if (compare_network(network, class->network) == 0) {
	    (void) lsFinish(gen);
	    return class;
	}
    }
    return 0;
}


static lib_class_t *
find_or_create_class(library, gate)
library_t *library;
lib_gate_t *gate;
{
    lib_class_t *class;

    /* see if gate is already assigned to a class */
    if (gate->class != 0) {
	return gate->class;
    }

    class = find_class(library, gate->network);
    if (class == NIL(lib_class_t)) {
	class = ALLOC(lib_class_t, 1);
	class->network = gate->network;
	class->gates = lsCreate();
	class->dual_class = 0;
	class->name = 0;
	lsNewEnd(library->classes, (char *) class, LS_NH);
	class->library = library;
    } else {
	network_free(gate->network);
	gate->network = class->network;
    }

    lsNewEnd(class->gates, (char *) gate, LS_NH);
    gate->class = class;
    return class;
}

static void
derive_dual_classes(library)
library_t *library;
{
    lsGen gen;
    lib_class_t *class;
    network_t *dual;

    lsForeachItem(library->classes, gen, class) {
	dual = dual_network(class->network);
	(void) network_collapse(dual);
	class->dual_class = find_class(library, dual);
	network_free(dual);
    }
}


static network_t *
dual_network(network)
network_t *network;
{
    node_t *pi, *po, *node, *inv;
    network_t *new_network;
    lsGen gen, gen1;

    new_network = network_dup(network);

    foreach_primary_input(new_network, gen, pi) {
	inv = node_literal(pi, 0);
	foreach_fanout(pi, gen1, node) {
	    assert(node_patch_fanin(node, pi, inv));
	}
	network_add_node(new_network, inv);
    }

    foreach_primary_output(new_network, gen, po) {
	node = node_get_fanin(po, 0);
	inv = node_literal(node, 0);
	assert(node_patch_fanin(po, node, inv));
	network_add_node(new_network, inv);
    }

    return new_network;
}

static network_t *
lib_wire()
{
    network_t *wire;

    wire = read_eqn_string("NAME = \"**wire**\"; Z = ! o1; o1 = ! a;");
    wire->area_given = 1;
    wire->area = 0.0;
    assert(wire != 0);
    return wire;
}


static double get_gate_area(gate)
lib_gate_t *gate;
{
  if (gate == NIL(lib_gate_t)) return INFINITY;
  return lib_gate_area(gate);
}

 /* default inverter; cached for efficiency but need to be reset */

static lib_gate_t *default_inverter = NIL(lib_gate_t);

void lib_reset_default_inverter()
{
  default_inverter = NIL(lib_gate_t);
}

lib_gate_t *lib_get_default_inverter()
{
  library_t *library;
  lsGen gen;
  lib_gate_t *gate;
  lib_gate_t *smallest_gate;
  lib_gate_t *second_smallest_gate;
  network_t *network;
  lib_class_t *class;

  if (default_inverter != NIL(lib_gate_t)) return default_inverter;
  assert(library = lib_get_library());
  network = read_eqn_string("f=!a;");
  class = lib_get_class(network, library);
  gen = lib_gen_gates(class); 
  smallest_gate = second_smallest_gate = NIL(lib_gate_t);
  while (lsNext(gen, (char **) &gate, LS_NH) == LS_OK) {
    if (get_gate_area(gate) < get_gate_area(second_smallest_gate)) {
      if (get_gate_area(gate) < get_gate_area(smallest_gate)) {
	second_smallest_gate = smallest_gate;
	smallest_gate = gate;
      } else {
	second_smallest_gate = gate;
      }
    }
  }
  (void) lsFinish(gen);
  network_free(network);
  if (second_smallest_gate != NIL(lib_gate_t)) {
    default_inverter = second_smallest_gate;
  } else {
    assert(smallest_gate != NIL(lib_gate_t));
    default_inverter = smallest_gate;
  }
  return default_inverter;
}
