 /* file map_interface.c release 1.5 */
 /* last modified: 9/9/90 at 17:31:46 */
#include <math.h>
#include "mis.h"
#include "map_macros.h"
#include "map_int.h"
#include "lib_int.h"
#include "bin_int.h"
#include "map_delay.h"
#include "fanout_int.h"

#include "map_interface_static.h"

 /* EXTERNAL INTERFACE */

 /* takes a RAW_NETWORK as input and produces a MAPPED_NETWORK */

network_t *complete_map_interface(network, options)
network_t *network;
bin_global_t *options;
{
  network_type_t type;
  library_t *library = options->library;

  error_init();
  type = map_get_network_type(network);
  switch (type) {
  case RAW_NETWORK:
    break;
  case ANNOTATED_NETWORK:
    error_append("map_interface: network is already annotated\n");
    return NIL(network_t);
    break;
  case MAPPED_NETWORK:
    error_append("map_interface: network is already mapped\n");
    return NIL(network_t);
  case EMPTY_NETWORK:
    return NIL(network_t);
  default:
    error_append("map_interface: bad network type\n");
    return NIL(network_t);
  }

 /* remove unnecessary nodes */
  network_sweep(network);

  /* initialize things properly for delay */
  map_reset_default_delay_info(network);
  (void) pipo_get_pi_arrival(NIL(node_t));
  (void) pipo_get_pi_drive(NIL(node_t));
  (void) pipo_get_po_load(NIL(node_t));
  (void) pipo_get_po_required(NIL(node_t));

 /* initialize the network */
  if (! map_premap_network(network, options)) {
    error_append("map_interface: can't premap the network\n");
    return NIL(network_t);
  }

 /* perform tree mapping */
  if (options->new_mode == 0 || (options->new_mode == 1 && FP_EQUAL(options->new_mode_value, 0.0))) {
    if (! tree_map(network, library, *options)) {/* old mapper; use it for minimum area mapping */
      error_append("map_interface: can't treemap the network\n");
      return NIL(network_t);
    }
  } else if (FP_EQUAL(options->new_mode_value, 1.0)) {
    bin_delay_tree_match(network, library, *options);
  } else {
    bin_delay_area_tree_match(network, library, *options);
  }

  if (options->fanout_optimize) fanout_optimization(network, *options);

  return build_network_from_annotations(network, options);
}

 /* this routine takes an annotated network and maps it */
 /* it also performs a basic consistency check for robustness */

network_t *build_mapped_network_interface(network, options)
network_t *network;
bin_global_t *options;
{
  network_type_t type = map_get_network_type(network);

  error_init();
  if (type != ANNOTATED_NETWORK) {
    error_append("build_annotated_network:  expects an ANNOTATED network\n");
    return NIL(network_t);
  }
  return build_network_from_annotations(network, options);
}

/* takes a RAW or an ANNOTATED network and returns an ANNOTATED network */
 /* only do tree mapping */

network_t *tree_map_interface(network, options)
network_t *network;
bin_global_t *options;
{
  network_type_t type;
  library_t *library = options->library;

  error_init();
  type = map_get_network_type(network);
  switch (type) {
  case RAW_NETWORK:
  case ANNOTATED_NETWORK:
    break;
  case MAPPED_NETWORK:
    error_append("map_interface: network is already mapped\n");
    return NIL(network_t);
  case EMPTY_NETWORK:
    return NIL(network_t);
  default:
    error_append("map_interface: bad network type\n");
    return NIL(network_t);
  }
  
    /* initialize things properly for delay */
  map_reset_default_delay_info(network);
  (void) pipo_get_pi_arrival(NIL(node_t));
  (void) pipo_get_pi_drive(NIL(node_t));
  (void) pipo_get_po_load(NIL(node_t));
  (void) pipo_get_po_required(NIL(node_t));

 /* initialize the network */
  if (type == RAW_NETWORK) {
    if (! map_premap_network(network, options)) {
      error_append("map_interface: can't premap the network\n");
      return NIL(network_t);
    }
  }

 /* perform tree mapping */
  if (options->new_mode == 0 || (options->new_mode == 1 && FP_EQUAL(options->new_mode_value, 0.0))) {
    if (! tree_map(network, library, *options)) {/* old mapper; use it for minimum area mapping */
      error_append("map_interface: can't treemap the network\n");
      return NIL(network_t);
    }
  } else if (FP_EQUAL(options->new_mode_value, 1.0)) {
    bin_delay_tree_match(network, library, *options);
  } else {
    bin_delay_area_tree_match(network, library, *options);
  }

 /* cleanup annotations */
  map_remove_unnecessary_annotations(network);
  if (options->print_stat) map_report_data_annotated(network);
  return network;
}

 /* takes an ANNOTATED network and returns an ANNOTATED network */
 /* performs fanout optimization and, if specified, area recovery */

network_t *fanout_opt_interface(network, options)
network_t *network;
bin_global_t *options;
{
  network_type_t type;

  error_init();
  type = map_get_network_type(network);
  if (type == EMPTY_NETWORK) return NIL(network_t);
  if (type != ANNOTATED_NETWORK) {
    error_append("fanout_opt_interface: expects an annotated network\n");
    return NIL(network_t);
  }
  
    /* initialize things properly for delay */
  map_reset_default_delay_info(network);
  (void) pipo_get_pi_arrival(NIL(node_t));
  (void) pipo_get_pi_drive(NIL(node_t));
  (void) pipo_get_po_load(NIL(node_t));
  (void) pipo_get_po_required(NIL(node_t));

  fanout_optimization(network, *options);
  map_remove_unnecessary_annotations(network);
  if (options->print_stat) map_report_data_annotated(network);
  return network;
}

 /* returns the type of a network */

network_type_t map_get_network_type(network)
network_t *network;
{
  int is_mapped = 1;
  int is_annotated = 1;
  lsGen gen;
  node_t *node;

  if (network == NIL(network_t)) return EMPTY_NETWORK;
  foreach_node(network, gen, node) {
    if (node->type == INTERNAL) {
      if (node_num_fanin(node) == 0) continue; /* handle the case of constants */
      if (MAP(node) == 0) {
	is_mapped = is_annotated = 0;
	(void) lsFinish(gen);
	break;
      } else if (MAP(node)->gate == NIL(lib_gate_t)) {
	is_mapped = 0;
      } else if (MAP(node)->gate != NIL(lib_gate_t) && MAP(node)->save_binding == NIL(node_t *)) {
	is_annotated = 0;
      } else if (MAP(node)->gate != NIL(lib_gate_t) && MAP(node)->save_binding != NIL(node_t *)) {
	is_mapped = 0;
      }
    }
  }
  if (is_mapped) return MAPPED_NETWORK;
  if (is_annotated) return ANNOTATED_NETWORK;
  return RAW_NETWORK;
}

 /* INTERNAL INTERFACE */

 /* decompose into 2-input NAND gates if needed */
 /* add inverters if required */
 /* and premap with 2-input NAND gates */

static network_t *map_premap_network(network, options)
network_t *network;
bin_global_t *options;
{
  library_t *library = options->library;

  /* If not 'raw' mapping, get it into 2-input nand gate form */
  if (! options->raw) {
    map_prep(network, library);
  }
  /* make sure it is the correct form */
  if (! map_check_form(network, library->nand_flag)) {
    return NIL(network_t);
  }
  /* add redundant inverters: should be consistent with the library */
  if (library->add_inverter) {
    map_add_inverter(network, /* add_at_pipo */ 1);
  }
  /* make sure there's at most 1 0-cell and 1 1-cell */
  patch_constant_cells(network);

  /* pre-map the circuit */
  if (! do_tree_premap(network, library)) {
    return NIL(network_t);
  }
  return network;
}

 /* takes an ANNOTATED network, builds it, removes inverters if necessary */
 /* and report stats if required */

static network_t *build_network_from_annotations(network, options)
network_t *network;
bin_global_t *options;
{
  network_t *new_network = map_build_network(network);

  if (options->remove_inverters) {
    if (options->print_stat) {
      map_remove_inverter(new_network, map_report_data_mapped);
    } else {
      map_remove_inverter(new_network, (VoidFn) 0);
    }
  }
  if (options->print_stat) map_report_data_mapped(new_network);
  return new_network;
}

 /* traverse the network from PO to PI through the virtual network */
 /* and record all the nodes encountered there */
 /* unmap the others */

static void map_remove_unnecessary_annotations(network)
network_t *network;
{
  lsGen gen;
  node_t *node, *fanin, *output;
  st_table *visited = st_init_table(st_ptrcmp, st_ptrhash);
  
  foreach_primary_output(network, gen, output) {
    st_insert(visited, (char *) output, NIL(char));
    fanin = node_get_fanin(output, 0);
    visit_mapped_nodes(fanin, visited);
  }
  foreach_node(network, gen, node) {
 /* isolated FANOUT_NODE are reset to TREE_NODE */
 /* because the tree mapper can handle them */
    if (MAP(node)->type == FANOUT_NODE) {
      if (node_num_fanout(node) == 1) {
	MAP(node)->type = TREE_NODE;
      }
    }
    if (st_lookup(visited, (char *) node, NIL(char *))) continue;

 /* all nodes that would not appear in the mapping are reset */
    MAP(node)->type = TREE_NODE;
    MAP(node)->gate = NIL(lib_gate_t);
    MAP(node)->ninputs = 0;
    FREE(MAP(node)->save_binding);
    MAP(node)->save_binding = NIL(node_t *);
    gate_link_delete_all(node);
  }
  st_free_table(visited);
}

static void visit_mapped_nodes(node, visited)
node_t *node;
st_table *visited;
{
  int i;

  st_insert(visited, (char *) node, NIL(char));
  if (node->type == PRIMARY_INPUT) return;
  assert(MAP(node)->save_binding != NIL(node_t *));
  for (i = 0; i < MAP(node)->ninputs; i++)
    visit_mapped_nodes(MAP(node)->save_binding[i], visited);
}

static void map_report_data_mapped(network)
network_t *network;
{
  node_t *node;
  lsGen gen;

  assert(delay_trace(network, DELAY_MODEL_LIBRARY));

  foreach_primary_output(network, gen, node) {
    MAP(node)->map_arrival = delay_arrival_time(node);
  }
  map_report_data(network);
}


 /* the network should have been cleaned up before calling this routine */
 /* every gate is considered part of the final circuit */

static void map_report_data_annotated(network)
network_t *network;
{
  map_compute_annotated_arrival_times(network);
  map_report_data(network);
}

static void map_report_data(network)
network_t *network;
{
  node_t *node;
  lsGen gen;
  double area = 0.0;
  int n_failing_outputs;
  delay_time_t arrival, required, slack;
  delay_time_t max_slack, min_slack, sum_slack, max_arrival;

  n_failing_outputs = 0;
  max_slack = MINUS_INFINITY;
  min_slack = PLUS_INFINITY;
  sum_slack = ZERO_DELAY;
  max_arrival = MINUS_INFINITY;

  foreach_primary_output(network, gen, node) {
    arrival = MAP(node)->map_arrival;
    required = pipo_get_po_required(node);
    SETSUB(slack, required, arrival);
    if (GETMIN(slack) < 0) {
      n_failing_outputs++;
      sum_slack.rise += (slack.rise < 0) ? slack.rise : 0.0;
      sum_slack.fall += (slack.fall < 0) ? slack.fall : 0.0;
    }
    SETMAX(max_slack, max_slack, slack);
    SETMIN(min_slack, min_slack, slack);
    SETMAX(max_arrival, max_arrival, arrival);
  }

  foreach_node(network, gen, node) {
    if (MAP(node)->gate != NIL(lib_gate_t)) {
      area += lib_gate_area(MAP(node)->gate);
    }
  }	
  (void) fprintf(misout, "# of outputs:          %d\n", network_num_po(network));
  (void) fprintf(misout, "total gate area:       %2.2f\n", area);
  (void) fprintf(misout, "maximum arrival time: (%2.2f,%2.2f)\n", max_arrival.rise, max_arrival.fall);
  (void) fprintf(misout, "maximum po slack:     (%2.2f,%2.2f)\n", max_slack.rise, max_slack.fall);
  (void) fprintf(misout, "minimum po slack:     (%2.2f,%2.2f)\n", min_slack.rise, min_slack.fall);
  (void) fprintf(misout, "total neg slack:      (%2.2f,%2.2f)\n", sum_slack.rise, sum_slack.fall);
  (void) fprintf(misout, "# of failing outputs:  %d\n", n_failing_outputs);
}

static void map_compute_annotated_arrival_times(network)
network_t *network;
{
  lsGen gen;
  node_t *output, *fanin;
  st_table *visited = st_init_table(st_ptrcmp, st_ptrhash);

  map_compute_loads(network);
  visited = st_init_table(st_ptrcmp, st_ptrhash);
  foreach_primary_output(network, gen, output) {
    fanin = node_get_fanin(output, 0);
    virtual_delay_arrival_times_rec(visited, fanin);
    MAP(output)->map_arrival = MAP(fanin)->map_arrival;
  }
  st_free_table(visited);
}

static void map_compute_loads(network)
network_t *network;
{
  int i;
  lsGen gen;
  node_t *fanin, *node;
  char *dummy;
  st_generator *wire_gen;
  st_table *wire_nodes = st_init_table(st_ptrcmp, st_ptrhash);

  foreach_node(network, gen, node) {
    MAP(node)->load = 0.0;
  }
  foreach_node(network, gen, node) {
    if (node->type == PRIMARY_OUTPUT) {
      MAP(node)->load = pipo_get_po_load(node);
      fanin = node_get_fanin(node, 0);
      MAP(fanin)->load += MAP(node)->load;
    } else if (MAP(node)->gate == NIL(lib_gate_t)) {
      /* do nothing */
    } else if (strcmp(MAP(node)->gate->name, "**wire**") == 0) {
      st_insert(wire_nodes, (char *) node, NIL(char));
    } else {
      for (i = 0; i < MAP(node)->ninputs; i++) {
	fanin = MAP(node)->save_binding[i];
	MAP(fanin)->load += delay_get_load(MAP(node)->gate->delay_info[i]);
      }
    }
  }
  st_foreach_item(wire_nodes, wire_gen, (char **) &node, &dummy) {
    fanin = node;
    while (st_lookup(wire_nodes, (char *) fanin, NIL(char *))) {
      fanin = MAP(fanin)->save_binding[0];
    }
    MAP(fanin)->load += MAP(node)->load;
  }  
  st_free_table(wire_nodes);
}
