 /* file treemap.c release 1.3 */
 /* last modified: 9/9/90 at 17:51:02 */
#include <math.h>
#include "mis.h"
#include "map_macros.h"
#include "map_int.h"
#include "lib_int.h"
#include "bin_int.h"
#include "map_delay.h"
#include "fanout_int.h"

static int do_tree_match();
static network_t *do_new_tree_match();
static int find_best_match();
static void find_match_cost();
static void match_cost();
static node_t *build_network_recur();
static int inverter_optimize();

static int global_debug, global_inverter_optz;
static double global_area_mode, global_thresh;
static int g_allow_internal_fanout;
static bin_global_t global;

/* debugger support ... till we get something better than DBX */
/* typedef char * STRING;
 * static char myname[512];
 * static char *mystr;
 * static lib_gate_t *mygate;
 * static lsGen mygen;
 * static node_t *mynode;
 */

static double max_arrival_estimate;
static double area_estimate;

/*
 *  exported interface ...
 */

network_t *
map_network(network, library, mode, inv_optz, allow_internal_fanout)
network_t *network;
library_t *library;
int inv_optz, allow_internal_fanout;
double mode;
{
  bin_global_t globals;

  globals.raw = 0;
  globals.new_mode = 0;
  globals.load_bins_count = globals.delay_bins_count = 0;
  globals.old_mode = mode;
  globals.remove_inverters = 1;
  globals.inverter_optz = inv_optz;
  globals.allow_internal_fanout = allow_internal_fanout;
  globals.print_stat = 0;
  globals.verbose = 0;
  globals.thresh = 0.0;
  globals.library = library;
  globals.fanout_optimize = 0;
  globals.area_recover = 0;
  globals.peephole = 0;
  globals.fanout_alg = NIL(array_t);

  return complete_map_interface(network, &globals);
}

/*
 *  tree_map -- implement 'network' using the tree-matching algorithm
 */


int tree_map(network, library, globals)
network_t *network;
library_t *library;
bin_global_t globals;
{
  global = globals;
  global.library = library;
  global_debug = globals.verbose;	     /* needed for find_best_match() ! */
  global_area_mode = globals.old_mode;	     /* needed for find_best_match() ! */
  global_thresh = globals.thresh;
  global_inverter_optz = globals.inverter_optz;	/* needed for find_best_match() ! */
  g_allow_internal_fanout = globals.allow_internal_fanout;

  area_estimate = 0.0;			     /* for final heuristics checking */
  max_arrival_estimate = 0.0;		     /* for final heuristics checking */

  return do_tree_match(network, library, globals.verbose);
}

/*
 *  premap: set each node to be a nand gate or an inverter
 *  we do this regardless of whether we are using 'nand' gates or 'nor'
 *  gates in order to time the circuit based on nand gates; note that
 *  this may change the logic function (if we stopped now); however,
 *  all pattern matching, etc. is strictly topological based.
 */

int
do_tree_premap(network, library)
network_t *network;
library_t *library;
{
    lsGen gen;
    network_t *libnet;
    node_t *pi, *node;
    lib_gate_t *gate, *zero, *one, *inv, *nand2;
    char **inv_formal, **nand2_formal;
    delay_time_t slack;
    double real_slack;
    int i;

    /* Give each node a 'map' record */
    map_setup_network(network);

    /* find standard gates */
    nand2 = choose_smallest_gate(library, "f=!(a*b);");
    if (nand2 == NIL(lib_gate_t)) {
	nand2 = choose_smallest_gate(library, "f=!(a+b);");
	if (nand2 == NIL(lib_gate_t)) {
	    nand2 = choose_smallest_gate(library, "f=a*b;");	/* hack */
	    if (nand2 == NIL(lib_gate_t)) {
		nand2 = choose_smallest_gate(library, "f=a+b;");/* hack */
		if (nand2 == NIL(lib_gate_t)) {
		    error_append(
			"library error: cannot find 2-input nand-gate\n");
		    return 0;
		}
	    }
	}
    }
    inv = choose_smallest_gate(library, "f=!a;");
    if (inv == NIL(lib_gate_t)) {
	error_append("library error: cannot find inverter\n");
	return 0;
    }
    zero = choose_smallest_gate(library, "f=0;");
    if (zero == NIL(lib_gate_t)) {
	error_append("library error: cannot find constant 0 cell\n");
	return 0;
    }
    one = choose_smallest_gate(library, "f=1;");
    if (one == NIL(lib_gate_t)) {
	error_append("library error: cannot find constant 1 cell\n");
	return 0;
    }

    /* create the formal arglist for inv and nand2 */
    inv_formal = ALLOC(char *, 1);
    i = 0;
    libnet = lib_class_network(lib_gate_class(inv));
    foreach_primary_input(libnet, gen, pi) {
	inv_formal[i++] = pi->name;
    }

    /* create the formal arglist for inv and nand2 */
    nand2_formal = ALLOC(char *, 2);
    i = 0;
    libnet = lib_class_network(lib_gate_class(nand2));
    foreach_primary_input(libnet, gen, pi) {
	nand2_formal[i++] = pi->name;
    }

    /* set the default mapping for each node */
    foreach_node(network, gen, node) {
	if (node->type == INTERNAL) {
	    switch(node_num_fanin(node)) {
	    case 0:
		gate = node_function(node) == NODE_0 ? zero : one;
		assert(lib_set_gate(node, gate, NIL(char *), NIL(node_t *), 0));
		break;
	    case 1:
		assert(lib_set_gate(node, inv, inv_formal, node->fanin, 1));
		break;
	    case 2:
		assert(lib_set_gate(node, nand2, nand2_formal, node->fanin, 2));
		break;
	    default:
		fail("bad node degree; should have been caught before here");
		break;
	    }
	}
    }
    FREE(inv_formal);
    FREE(nand2_formal);

    /* perform delay trace using unit-delay model */
    assert(delay_trace(network, DELAY_MODEL_LIBRARY));

    /* assign the initial area and arrival times for each node */
    foreach_node(network, gen, node) {
	if (node->type == PRIMARY_INPUT) {
	    MAP(node)->map_area = 0;
	    MAP(node)->map_arrival = delay_arrival_time(node);
	} else {
	    MAP(node)->map_area = -1;
	    MAP(node)->map_arrival.rise = INFINITY;
	    MAP(node)->map_arrival.fall = INFINITY;
	}
	MAP(node)->load = delay_load(node);
	slack = delay_slack_time(node);
	real_slack = MIN(slack.rise, slack.fall);
	MAP(node)->slack_is_critical = real_slack < global_thresh;
    }

    return 1;
}

static int
do_tree_match(network, library, debug)
network_t *network;
library_t *library;
int debug;
{
  int i, some_output;
  node_t *node, *fanout;
  prim_t *prim;
  array_t *nodevec;
  lsGen gen, gen1;
  char errmsg[1024];

  /* clear all bindings for all patterns */
  lsForeachItem(library->patterns, gen, prim) {
    prim_clear_binding(prim);
  }

  /* Match each node in a depth-first order */
  nodevec = network_dfs(network);
  for (i = 0; i < array_n(nodevec); i++) {
    node = array_fetch(node_t *, nodevec, i);
    if (node->type != INTERNAL) continue;

    /*   Try each primitive rooted at this node */ 
    if (node_num_fanin(node) == 0) continue; /* avoid constants */

    MAP(node)->gate = NIL(lib_gate_t);
    lsForeachItem(library->patterns, gen, prim) {
      gen_all_matches(node, prim, find_best_match, debug > 5, g_allow_internal_fanout);
    }
    if (MAP(node)->map_area < 0.0 || MAP(node)->gate == 0) {
      (void) sprintf(errmsg, 
		     "library error: no gate matches at %s\n", node_name(node));
      error_append(errmsg);
      return 0;
    }

    /* apply inverter optimization at the root of a tree */
    if (global_inverter_optz && is_tree_leaf(node)) {
      inverter_optimize(node);
    }

    /* compute the area and timing estimates */
    some_output = 0;
    foreach_fanout(node, gen1, fanout) {
      if (fanout->type == PRIMARY_OUTPUT) {
	some_output = 1;
      }
    }
    if (some_output) {
      area_estimate += MAP(node)->map_area / node_num_fanout (node);
      max_arrival_estimate = MAX (max_arrival_estimate, MAP(node)->map_arrival.rise);
      max_arrival_estimate = MAX (max_arrival_estimate, MAP(node)->map_arrival.fall);
    }

    if (debug > 0) {
      (void) fprintf(misout,
		     "Best match at %s is %s area=%5.2f arrival=(%5.2f %5.2f)\n",
		     node_name(node), MAP(node)->gate->name, MAP(node)->map_area,
		     MAP(node)->map_arrival.rise, MAP(node)->map_arrival.fall);
    }
  }
  array_free(nodevec);
  return 1;
}

/*
 *  find_best_match: simulate all gates which match a given pattern;
 *  maintain at the node the best match seen to this point
 */

static int
find_best_match(prim)
prim_t *prim;
{
    lsGen gen;
    lib_gate_t *gate;

    /* simulate each gate which matches this pattern */
    lsForeachItem(prim->gates, gen, gate) {
      find_match_cost(prim, gate);
    }
    return 1;
}

/* compare_cost contains the cost function: it returns 1 if THIS is better than
 * MATCH.
 */

int
compare_cost (this_area, this_arrival, match_area, match_arrival, slack_is_critical)

double match_area, this_area, match_arrival, this_arrival;
int slack_is_critical;

{
	int optimize_area, optimize_delay;
	double this_cost, match_cost;

	/* old cost function for 0.0, 1.0 and 2.0 */
	if (FP_EQUAL (global_area_mode, 2)) {
		optimize_area = ! slack_is_critical;
		optimize_delay = slack_is_critical;
	}
	else {
		optimize_area = FP_EQUAL (global_area_mode, 0.0);
		optimize_delay = FP_EQUAL (global_area_mode, 1.0);
	}

	if (optimize_area) {
	    if (match_area < this_area || 
	    (FP_EQUAL(match_area, this_area) && match_arrival < this_arrival)) {
			return 1;
	    }
	} else if (optimize_delay) {
	    if (match_arrival < this_arrival || 
	    (FP_EQUAL(match_arrival, this_arrival) && match_area < this_area)) {
			return 1;
	    }
	}
	else {
	/* cost = (1 - mode) * area + mode * delay
	 * use area then delay to break ties
	 */
		this_cost = (1.0 - global_area_mode) * this_area + global_area_mode * this_arrival;
		match_cost = (1.0 - global_area_mode) * match_area + global_area_mode * match_arrival;
		if (match_cost < this_cost ||
	    (FP_EQUAL(match_cost, this_cost) && match_area < this_area) ||
	    (FP_EQUAL(match_cost, this_cost) && FP_EQUAL(match_area, this_area) && match_arrival < this_arrival)) {
			return 1;
		}
	}

	return 0;
}


/*
 *  find_match_cost: simulate a single gate; keep it if its the best
 */

static void
find_match_cost(prim, gate)
prim_t *prim;
lib_gate_t *gate;
{
    int i, new_best;
    double match_area, this_area, match_arrival, this_arrival;
    delay_time_t arrival;
    node_t *node;

    /* get the root in the subject network */
    node = map_prim_get_root(prim);

    /* Determine the value of this match */
    match_cost(node, prim, gate, &match_area, &arrival);
    match_arrival = MAX(arrival.rise, arrival.fall);

    if (global_debug > 1) {
	(void) fprintf(misout, 
	    "    match of %s at %s (area=%5.2f arrival=%5.2f)\n", 
	    gate->name, node_name(node), match_area, match_arrival);
	if (global_debug > 2) {
	    for(i = 0; i < prim->ninputs; i++) {
		(void) fprintf(misout, 
		    "        Graph node=%10s <---> (in)  prim_node=%10s\n",
		    prim->inputs[i]->binding->name, prim->inputs[i]->name);
	    }
	    for(i = 0; i < prim->noutputs; i++) {
		(void) fprintf(misout, 
		    "        Graph node=%10s <---> (out) prim_node=%10s\n",
		    prim->outputs[i]->binding->name, prim->outputs[i]->name);
	    }
	}
    }

    if (MAP(node)->gate == 0) {
	/* first match is always the best */
	new_best = 1;

    } else {
	this_area = MAP(node)->map_area;
	this_arrival = MAX(MAP(node)->map_arrival.rise,
					    MAP(node)->map_arrival.fall);
	new_best = compare_cost (this_area, this_arrival, match_area, match_arrival,
		MAP(node)->slack_is_critical);
	}

    if (new_best) {
	/* save this match as minimum cost for this node */
		MAP(node)->gate = gate;
		MAP(node)->map_area = match_area;
		MAP(node)->map_arrival.rise = arrival.rise;
		MAP(node)->map_arrival.fall = arrival.fall;

		FREE(MAP(node)->save_binding);
		MAP(node)->ninputs = prim->ninputs;
		MAP(node)->save_binding = ALLOC(node_t *, prim->ninputs);
		for(i = 0; i < prim->ninputs; i++) {
	    	MAP(node)->save_binding[i] = prim->inputs[i]->binding;
		}
	}
}

static void
match_cost(node, prim, gate, area, arrival)
node_t *node;
prim_t *prim;
lib_gate_t *gate;
double *area;
delay_time_t *arrival;
{
    register int i;
    register node_t *fanin;
    register delay_time_t **arrival_times;
    register prim_node_t **leafs;

    leafs = prim->inputs;

    /* check -- is the match an inverter which is already paid for ? */
    if (global_inverter_optz) {
	if (match_is_inverter(node, prim)) {
	    fanin = leafs[0]->binding;
	    if (MAP(fanin)->inverter_paid_for) {
		*area = 0.0;
		*arrival = MAP(fanin)->inverter_arrival;
		return;
	    }
	}
	}


    /* 
     *  Compute area cost of this match 
     */
    *area = gate->area;
    for(i = prim->ninputs - 1; i >= 0; i--) {
      fanin = leafs[i]->binding;
      if (g_allow_internal_fanout & 1) {
	if (fanin->type != PRIMARY_INPUT) {
	  *area += MAP(fanin)->map_area / node_num_fanout(fanin);
	}
      } else {
	if (! is_tree_leaf(fanin)) {
	  *area += MAP(fanin)->map_area;
	}
      }
    }

    /* 
     *  Compute arrival times (if this match chosen)
     */
    if (strcmp(gate->name, "**wire**") == 0) {
	*arrival = MAP(prim->inputs[0]->binding)->map_arrival;

    } else {
	arrival_times = ALLOC(delay_time_t *, prim->ninputs);
	for(i = prim->ninputs - 1; i >= 0; i--) {
	    fanin = leafs[i]->binding;
	    arrival_times[i] = &(MAP(fanin)->map_arrival);
	}
	*arrival = delay_map_simulate(prim->ninputs, arrival_times, 
					    gate->delay_info, MAP(node)->load);
	FREE(arrival_times);
    }
}

network_t *
map_build_network(network)
network_t *network;
{
  node_t *node, *new_node, **fanin;
  network_t *new_network;
  lsGen gen;

  /* clear the 'matching' node for each node in the old network */
  foreach_node(network, gen, node) {
    MAP(node)->match_node = NIL(node_t);
  }

  /* allocate a new network */
  new_network = network_alloc();
  network_set_name(new_network, network_name(network));
  delay_network_dup(new_network, network);

  /* copy the PI's first */
  foreach_primary_input(network, gen, node) {
    delay_time_t drive;
    delay_time_t arrival;

    new_node = node_dup(node);
    MAP(new_node)->gate = NIL(lib_gate_t);
    if (MAP(new_node)->ninputs > 0) {
      MAP(new_node)->ninputs = 0;
      FREE(MAP(new_node)->save_binding);
    }
    network_add_primary_input(new_network, new_node); 
    MAP(node)->match_node = new_node;
    if (delay_get_pi_drive(node, &drive)) {
      delay_set_parameter(new_node, DELAY_DRIVE_RISE, drive.rise);
      delay_set_parameter(new_node, DELAY_DRIVE_FALL, drive.fall);
    }
    if (delay_get_pi_arrival_time(node, &arrival)) {
      delay_set_parameter(new_node, DELAY_ARRIVAL_RISE, arrival.rise);
      delay_set_parameter(new_node, DELAY_ARRIVAL_FALL, arrival.fall);
    }
  }

  /* copy the PO's and recursively from the PO's */
  foreach_primary_output(network, gen, node) {
    delay_time_t required;
    double load;

    new_node = node_dup(node);
    MAP(new_node)->gate = NIL(lib_gate_t);
    if (MAP(new_node)->ninputs > 0) {
      MAP(new_node)->ninputs = 0;
      FREE(MAP(new_node)->save_binding);
    }
    fanin = ALLOC(node_t *, 1);
    fanin[0] = build_network_recur(new_network, node_get_fanin(node, 0));
    node_replace_internal(new_node, fanin, 1, NIL(set_family_t));
    network_add_node(new_network, new_node);
    MAP(node)->match_node = new_node;
    if (delay_get_po_required_time(node, &required)) {
      delay_set_parameter(new_node, DELAY_REQUIRED_RISE, required.rise);
      delay_set_parameter(new_node, DELAY_REQUIRED_FALL, required.fall);
    }
    if (delay_get_po_load(node, &load)) {
      delay_set_parameter(new_node, DELAY_OUTPUT_LOAD, load);
    }
  }
  return new_network;
}

static node_t *
build_network_recur(new_network, node)
network_t *new_network;
node_t *node;
{
    int i, nin;
    node_t *pi, *new_node, **fanin;
    char **formals;
    lsGen gen;

    if (MAP(node)->match_node == NIL(node_t)) {
	assert(node->type == INTERNAL);

	/* minor hack -- remove 'wires' when building the output graph */

	assert (MAP(node)->gate != NIL(lib_gate_t));
	if (strcmp(MAP(node)->gate->name, "**wire**") == 0) {
	    new_node = build_network_recur(new_network, 
					    MAP(node)->save_binding[0]);
	    MAP(node)->match_node = new_node;

	} else {
	    new_node = node_alloc();
	    new_node->name = util_strsav(node->name);
	    nin = network_num_pi(MAP(node)->gate->network);

	    fanin = ALLOC(node_t *, nin);
	    for(i = 0; i < nin; i++) {
		fanin[i] = build_network_recur(new_network, 
					    MAP(node)->save_binding[i]);
	    }

	    formals = ALLOC(char *, nin);
	    i = 0;
	    foreach_primary_input(MAP(node)->gate->network, gen, pi) {
		formals[i++] = pi->name;
	    }
	    if (!lib_set_gate(new_node, MAP(node)->gate, formals, fanin, nin)) {
		fail("lib_set_gate cannot fail");
	    }
	    network_add_node(new_network, new_node); 
	    MAP(node)->match_node = new_node;
	    FREE(formals);
	    FREE(fanin);
	}
    }
    return MAP(node)->match_node;
}

static void
set_leaf_inverter_flag(node)
node_t *node;
{
    int i;
    node_t *fanin;

    if (is_tree_leaf(node)) {
	/* found leaf -- exit recursion */
	return;
    }

    /* see if an inverter matched at the leaf */
    if (best_match_is_inverter(node)) {
	fanin = MAP(node)->save_binding[0];
	if (is_tree_leaf(fanin) && ! MAP(fanin)->inverter_paid_for) {
	    MAP(fanin)->inverter_paid_for = 1;
	    MAP(fanin)->inverter_arrival = MAP(node)->map_arrival;
	}
	}

    for(i = MAP(node)->ninputs-1; i >= 0; i--) {
	set_leaf_inverter_flag(MAP(node)->save_binding[i]);
    }
}


static int
inverter_optimize(node)
register node_t *node;
{
    int i;

    /* see if an inverter matched at the root of the tree */
    if (best_match_is_inverter(node)) {
	MAP(node)->inverter_paid_for = 1;
	MAP(node)->inverter_arrival = MAP(node)->map_arrival;
    }

    /* walk back to the leaves, seeing if inverters matched at the inputs */
    for(i = MAP(node)->ninputs-1; i >= 0; i--) {
	set_leaf_inverter_flag(MAP(node)->save_binding[i]);
    }
}

/*
 *  search the library for the 'smallest' gate with a given function
 */

lib_gate_t *
choose_smallest_gate(library, string)
library_t *library;
char *string;
{
    network_t *network;
    lib_class_t *class;
    lib_gate_t *gate, *best_gate;
    lsGen gen;
    double best_area;

    network = read_eqn_string(string);
    if (network == 0) return 0;
    class = lib_get_class(network, library);
    if (class == 0) return 0;

    best_area = -1;
    best_gate = 0;
    gen = lib_gen_gates(class);
    while (lsNext(gen, (char **) &gate, LS_NH) == LS_OK) {
	if (best_area < 0 || lib_gate_area(gate) < best_area) {
	    best_area = lib_gate_area(gate);
	    best_gate = gate;
	}
    }
    (void) lsFinish(gen);
    network_free(network);
    return best_gate;
}

/*
 *  patch the netlist so that all 0,1 cells are driven by a single node
 */

void
patch_constant_cells(network)
network_t *network;
{
    lsGen gen, gen1;
    node_t *node, *fanout, *const0_node, *const1_node;

    const0_node = const1_node = NIL(node_t);

    foreach_node(network, gen, node) {
	switch(node_function(node)) {
	case NODE_0:
	    if (const0_node == 0) {
		const0_node = node;
	    } else {
		/* patch all other 0 nodes to be driven by the 0-cell */
		foreach_fanout(node, gen1, fanout) {
		    assert(node_patch_fanin(fanout, node, const0_node));
		}
		/* delete this redundant 0-cell */
		network_delete_node_gen(network, gen);
	    }
	    break;

	case NODE_1:
	    if (const1_node == 0) {
		const1_node = node;
	    } else {
		/* patch all other 1 nodes to be driven by the 1-cell */
		foreach_fanout(node, gen1, fanout) {
		    assert(node_patch_fanin(fanout, node, const1_node));
		}
		/* delete this redundant 1-cell */
		network_delete_node_gen(network, gen);
	    }
	    break;
	}
    }
}
