#ifdef OCT

/*
 * routine for reading MISII logic networks (mapped and unmapped) from
 * Oct facets
 */
#include "copyright.h"
#include "port.h"
#undef assert
#include "mis.h"
#include "st.h"
#undef REALLOC
#undef FREE
#include "errtrap.h"
#include "oct.h"
#include "oh.h"
#include "vov.h"
#include "octio.h"

extern node_t *read_find_or_create_node();

#define node_oct_name(node)   (node)->name

#define STREQ(a, b)  (strcmp(a, b) == 0)
  
#define OCTMISCHECK(fnc) \
    { \
        octStatus octmischeck_status; \
	octmischeck_status = fnc; \
        if (octmischeck_status < OCT_OK) { \
	    errRaise(MIS_PKG_NAME, octmischeck_status, octErrorString()); \
	} \
    }			     


static void
uniqNames(container, mask)
octObject *container;
octObjectMask mask;
{
    octObject obj;
    octGenerator gen;
    octStatus status;
    char *name, dummyname[1024];
    st_table *name_table;
    int cnt, count;

    name_table = st_init_table(strcmp, st_strhash);
    cnt = 0;

    OCTMISCHECK(octInitGenContents(container, mask, &gen));
    while ((status = octGenerate(&gen, &obj)) == OCT_OK) {
	name = ohGetName(&obj);
	if (name == NIL(char)) {
	    errRaise(MIS_PKG_NAME, OCT_ERROR, "ohGetName returned NIL");
	}
	if (name[0] == '\0') {
	    /* make sure the name is not empty */
	    (void) sprintf(dummyname, "%s-%d", ohTypeName(&obj), cnt++);
	    OCTMISCHECK(ohPutName(&obj, dummyname));
	    OCTMISCHECK(octModify(&obj));
	    OCTMISCHECK(ohGetById(&obj, obj.objectId));
	    name = ohGetName(&obj);
	}
	if (st_lookup(name_table, name, NIL(char *))) {
	    /* loop until we can generate a unique name */
	    count = 2;
	    do {
		(void) sprintf(dummyname, "%s-{%d}", name, count++);
	    } while (st_lookup(name_table, dummyname, NIL(char *)));

	    /* change the objects name */
	    OCTMISCHECK(ohPutName(&obj, dummyname));
	    OCTMISCHECK(octModify(&obj));
	    OCTMISCHECK(ohGetById(&obj, obj.objectId));
	    name = ohGetName(&obj);
	}

	(void) st_insert(name_table, name, NIL(char));
    }
    OCTMISCHECK(status);
    (void) st_free_table(name_table);
    return;
}


static node_t *
build_new_node(node, network, instance)
node_t *node;
network_t *network;
octObject *instance;
{
    node_t *result, *temp, *oldfanin, *newfanin, *nodeliteral, *and, *or;
    node_cube_t cube;
    node_literal_t literal;
    int i, j;
    octObject term, net;
    octGenerator gen;
    st_table *table;
    char *netname;
    octStatus status;

    /* build up terminal/net table */
    table = st_init_table(strcmp, st_strhash);
    OCTMISCHECK(octInitGenContents(instance, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &term) == OCT_OK) {
	status = octGenFirstContainer(&term, OCT_NET_MASK, &net);
	if (status == OCT_OK) {
	    (void) st_insert(table, term.contents.term.name, net.contents.net.name);
	}
	if (status < OCT_OK) {
	    errRaise(MIS_PKG_NAME, status,
		     "oct error (%s)",
		     octErrorString());
	}
    }

    result = node_constant(0);

    for (i = node_num_cube(node) - 1; i >= 0; i--) {
        cube = node_get_cube(node, i);
        temp = node_constant(1);
        foreach_fanin(node, j, oldfanin) {
            literal = node_get_literal(cube, j);

            if (literal == TWO) {
                continue;
            }

	    /* get the net attached to the terminal with the name of the oldfanin */
	    if (!st_lookup(table, node_oct_name(oldfanin), &netname)) {
		errRaise(MIS_PKG_NAME, OCT_ERROR,
			 "FATAL: missing connection to terminal %s on %s\n",
			 node_oct_name(oldfanin), ohFormatName(instance));
	    }
	    newfanin = read_find_or_create_node(network, netname);

	    nodeliteral = node_literal(newfanin, ((literal == ZERO) ? 0 : 1));
	    and = node_and(temp, nodeliteral);
	    node_free(temp);		/* garbage collection */
	    node_free(nodeliteral);
            temp = and;
        }
        or = node_or(result, temp);
        node_free(result);
	node_free(temp);
        result = or;
    }

    st_free_table(table);
    return(result);
}


struct termInfo {
    char *name;
    network_t *logicFunction;
};

static st_table *MasterTable = NIL(st_table);

static void
initialize_master_table()
{
    MasterTable = st_init_table(strcmp, st_strhash);
    return;
}

static void
termFree(terminfo)
struct termInfo *terminfo;
{
    (void) free(terminfo->name);
    network_free(terminfo->logicFunction);
    return;
}


static void
clean_master_table()
{
    st_generator *gen;
    char *key, *value;
    lsList the_list;

    st_foreach_item(MasterTable, gen, &key, &value) {
	the_list = (lsList) value;
	if (the_list != (lsList) 0) {
	    (void) lsDestroy(the_list, termFree);
	}
    }
    st_free_table(MasterTable);
    return;
}


static lsList
is_a_logic_cell(instance)
octObject *instance;
/*
 * determine if an instance has a logic representation
 *
 *    returns: lsList of the logic function(s), 0 if not a logic cell,
 */
{
    char *ptr;
    octGenerator gen;
    octObject term, prop, master;
    lsList outputs;
    struct termInfo *termInfo;
    int termTypeExists;

    if (st_lookup(MasterTable, instance->contents.instance.master, &ptr)) {
	/* it has been seen before */
	return((lsList) ptr);
    }

    if ((ohOpenMaster(&master, instance, "interface", "r") < OCT_OK) ||
	(ohGetByPropName(&master, &prop, "CELLTYPE") < OCT_OK) ||
	!STREQ(prop.contents.prop.value.string, "COMBINATIONAL")) {
	
	(void) st_insert(MasterTable, instance->contents.instance.master, NIL(char));
	return((lsList) 0);
    }
	
    VOVinputFacet( &master );
    if (octInitGenContents(&master, OCT_TERM_MASK, &gen) < OCT_OK) {
	errRaise(MIS_PKG_NAME, OCT_ERROR, octErrorString());
    }

    outputs = lsCreate();
    
    while (octGenerate(&gen, &term) == OCT_OK) {
	termTypeExists = (ohGetByPropName(&term, &prop, "TERMTYPE") == OCT_OK);
	/* if no TERMTYPE, defaults to SIGNAL */
	if (termTypeExists && !STREQ(prop.contents.prop.value.string, "SIGNAL")) {
	    continue;
	}
	if ((ohGetByPropName(&term, &prop, "DIRECTION") < OCT_OK) ||
	    !STREQ(prop.contents.prop.value.string, "OUTPUT")) {
	    continue;
	}
	termInfo = ALLOC(struct termInfo, 1);
	termInfo->name = util_strsav(term.contents.term.name);
	if (ohGetByPropName(&term, &prop, "LOGICFUNCTION") < OCT_OK) {
	    termInfo->logicFunction = NIL(network_t);
	} else {
	    if ((termInfo->logicFunction = read_eqn_string(prop.contents.prop.value.string)) == NIL(network_t)) {
		errRaise(MIS_PKG_NAME, OCT_ERROR,
			 "bad equation %s (%s)",
			 prop.contents.prop.value.string,
			 error_string());
	    }
	}
	(void) lsNewEnd(outputs, (lsGeneric) termInfo, NIL(lsHandle));
    }

    (void) st_insert(MasterTable, instance->contents.instance.master, (char *) outputs);

    if (octCloseFacet(&master) < OCT_OK) {
	errRaise(MIS_PKG_NAME, OCT_ERROR, octErrorString());
    }
    return(outputs);
}


	
network_t *
read_oct(facet, mappedp)
octObject *facet;
int mappedp;
{

    network_t *network, *newnetwork;
    node_t *output, *fanout, *newnode;
    octObject instance, term, net, prop, bag;
    octGenerator gen, tgen;
    octStatus status;

    lsList outputList;
    lsGen outputGen;
    lsGeneric ptr;
    struct termInfo *termInfo;
    int instanceLogicFunction;

    library_t *library;

    node_t *node;

    /* open the facet */
    VOVinputFacet( facet );
    if (octOpenFacet(facet) < OCT_OK) {
	errRaise(MIS_PKG_NAME, OCT_ERROR,
		 "can not open the facet %s (%s)",
		 ohFormatName(facet),
		 octErrorString());
    }

    /*
     * XXX read the required properties to make sure that it 
     * is a good facet 
     */

    /* force the net names to be unique */
    uniqNames(facet, OCT_NET_MASK);

    /* 
     * make sure formal term names are not the same as net names
     * (really only a problem for primary outputs)
     */

    OCTMISCHECK(octInitGenContents(facet, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &term) == OCT_OK) {
	OCTMISCHECK(octGenFirstContainer(&term, OCT_NET_MASK, &net));
	if (strcmp(term.contents.term.name, net.contents.net.name) == 0) {
	    /* build a new name */
	    char buffer[1024];
	    static int count = 0;
	    do {
	        (void) sprintf(buffer, "%s-%d",
				 term.contents.term.name, count++);
	        net.contents.net.name = buffer;
	    } while (octGetByName(facet, &net) == OCT_OK);

	    OCTMISCHECK(octModify(&net));
	}
    }

    network = network_alloc();

    /* is there a library set? */
    if (mappedp) {
        library = lib_get_library();
    } else {
	library = NIL(library_t);
    }
    
    if (ohGetByBagName(facet, &bag, "INSTANCES") != OCT_OK) {
	errRaise(MIS_PKG_NAME, OCT_ERROR,
		 "Missing INSTANCES bag, facet does not follow policy");
    }
    
    OCTMISCHECK(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen));

    initialize_master_table();
    
    while (octGenerate(&gen, &instance) == OCT_OK) {

	if (library) {
	    /* see if this is a mapped gate */
	    char *lastslash = strrchr(instance.contents.instance.master, '/');
	    if (lastslash) {
		char gateName[1024];
 		lib_gate_t *gate;

                (void) sprintf(gateName, "%s:%s",
				lastslash + 1,
			 	instance.contents.instance.view);
		if ((gate = lib_get_gate(library, gateName))) {
		    array_t *formal_array, *actual_array;
		    char **formal_data;
		    node_t **actual_data;
		    node_t *outnode;
		    int i, size = lib_gate_num_in(gate);

		    formal_array = array_alloc(char *, size);
		    actual_array = array_alloc(node_t *, size);

		    for (i = 0; i < size; i++) {
			OCTMISCHECK(ohGetByTermName(&instance, &term, lib_gate_pin_name(gate, i, 1)));
			OCTMISCHECK(octGenFirstContainer(&term, OCT_NET_MASK, &net));
			node = read_find_or_create_node(network, net.contents.net.name);
			array_insert_last(char *, formal_array, term.contents.term.name);
			array_insert_last(node_t *, actual_array, node);
		    }

		    for (i = 0; i < lib_gate_num_out(gate); i++) {
			OCTMISCHECK(ohGetByTermName(&instance, &term, lib_gate_pin_name(gate, i, 0)));
			OCTMISCHECK(octGenFirstContainer(&term, OCT_NET_MASK, &net));
			outnode = read_find_or_create_node(network, net.contents.net.name);
		    }
		    formal_data = array_data(char *, formal_array);
		    actual_data = array_data(node_t *, actual_array);
		    array_free(formal_array);
		    array_free(actual_array);
		    if (!lib_set_gate(outnode, gate, formal_data, actual_data, size)) {
			
			errRaise(MIS_PKG_NAME, -1, "error loading gate `%s'", gateName);
		    }

		    free(formal_data);
		    free(actual_data);
		    continue;
		}
	    }
        }

	outputList = is_a_logic_cell(&instance);

	/* no logic functions - how about a latch */
	if (outputList == (lsList) 0) {
	    continue;
	}

	outputGen = lsStart(outputList);

	while (lsNext(outputGen, &ptr, NIL(lsHandle)) == LS_OK) {

	    termInfo = (struct termInfo *) ptr;

	    /* see if there is a LOGICFUNCTION on the instance */
	    if (ohGetByTermName(&instance, &term, termInfo->name) < OCT_OK) {
		errRaise(MIS_PKG_NAME, OCT_ERROR,
			 "can not get the actual terminal %s of %s (%s)",
			 termInfo->name,
			 ohFormatName(&instance),
			 octErrorString());
	    }

	    if (ohGetByPropName(&term, &prop, "LOGICFUNCTION") == OCT_OK) {
		if ((newnetwork = read_eqn_string(prop.contents.prop.value.string)) == NIL(network_t)) {
		    errRaise(MIS_PKG_NAME, OCT_ERROR,
			     "bad equation %s (%s)",
			     prop.contents.prop.value.string,
			     error_string());
		}
		instanceLogicFunction = 1;
	    } else {
		newnetwork = termInfo->logicFunction;
		instanceLogicFunction = 0;
		if (newnetwork == NIL(network_t)) {
		    continue;
		}
	    }

	    /* get the primary output of the network */
	    output = network_get_po(newnetwork, 0);

	    status = octGenFirstContainer(&term, OCT_NET_MASK, &net);
	    if (status == OCT_GEN_DONE) {
		if (instanceLogicFunction) {
		    network_free(newnetwork);
		}
		continue;
	    }
	    if (status < OCT_OK) {
		errRaise(MIS_PKG_NAME, status, octErrorString());
	    }
	    
	    fanout = read_find_or_create_node(network, net.contents.net.name);

	    /* build the new node */
	    /* get the internal node - should be the only fanin the the primary output  */
	    newnode = build_new_node(node_get_fanin(output, 0), network, &instance);

	    /* put in the new function */
	    if (node_function(fanout) == NODE_UNDEFINED) {
		node_replace(fanout, newnode);
	    } else {
		errRaise(MIS_PKG_NAME, OCT_ERROR,
			 "two outputs driving the same node: %s",
			 node_oct_name(fanout));
	    }

	    if (instanceLogicFunction) {
		/* garbage collection */
		network_free(newnetwork);
	    }
	}

	(void) lsFinish(outputGen);
    }

    /* determine PIs and POs */
	
    OCTMISCHECK(octInitGenContents(facet, OCT_TERM_MASK, &gen));
    
    while (octGenerate(&gen, &term) == OCT_OK) {
	node_function_t funct;
	
	status = octGenFirstContainer(&term, OCT_NET_MASK, &net);
	if (status == OCT_GEN_DONE) {
	    continue;
	}
	if (status < OCT_OK) {
	    errRaise(MIS_PKG_NAME, status, "oct error (%s)", octErrorString());
	}

	if ((node = network_find_node(network, net.contents.net.name)) == NIL(node_t)) {
	    continue;
	}

	funct = node_function(node);

	if (funct == NODE_UNDEFINED) {
	    /* primary input */
	    network_change_node_type(network, node, NODE_PI);
	    network_change_node_name(network, node, util_strsav(term.contents.term.name));
	} else {
	    node_t *new;
	    char *save;
	    /*
	     * network_add_primary_output takes the name of 'node' and
	     * gives it to 'new'; it then gensyms a name for 'node'
	     *
	     * this is okay if 'node' is an internal node, but it 'node'
	     * is a primary input, this is trouble
	     *
	     * thus the name of node must be saved and replaced
	     */
	    save = util_strsav(node_oct_name(node));
	    new = network_add_primary_output(network, node);

	    /*
	     * change the name of the output to the name stored in Oct
	     */
	    network_change_node_name(network, new, util_strsav(term.contents.term.name));
	    network_change_node_name(network, node, save);
	}
    }

    /* garbage collect */
    clean_master_table();

    if (!network_check(network)) {
	errRaise(MIS_PKG_NAME, OCT_ERROR, "network check failed (%s)", error_string());
    }

    OCTMISCHECK(octCloseFacet(facet));

    network_set_name(network, ohFormatName(facet));

    return(network);
}



static jmp_buf mis_oct_error_buf;

static void
mis_oct_error_handler(pkgName, code, message)
char *pkgName;
int code;
char *message;
{
    (void) fprintf(stderr, "%s: error (%d)\n\t%s\n", pkgName, code, message);
    (void) longjmp(mis_oct_error_buf, -1);
    return;
}

  
int
external_read_oct(network, argc, argv)
network_t **network;
int argc;
char **argv;
/*
 * read_oct: read a network from an Oct facet
 *
 */
{
    octObject facet;
    int mappedp = 0;
    
    if ((argc != 2) && (argc != 3)) {
	(void) fprintf(miserr, "usage: read_oct [-m] cell[:view]\n");
	return(1);
    }

    octBegin();

    if (setjmp(mis_oct_error_buf) == 0) {
	errPushHandler(mis_oct_error_handler);
	
	(void) ohUnpackDefaults(&facet, "r", ":logic:contents");
	if (ohUnpackFacetName(&facet, argv[argc - 1]) < OCT_OK) {
	    errRaise(MIS_PKG_NAME, OCT_ERROR, "usage: read_oct cell[:view]");
	}

	error_init();
	if (argc == 3) {
	    mappedp = 1;
	}
	*network = read_oct(&facet, mappedp);
	errPopHandler();
    } else {
	return(1);
    }

    octEnd();
    
    return(0);
}

#endif /* OCT */

