#ifdef OCT

/*
 * routine for writing MISII logic networks (mapped and unmapped) into
 * Oct facets
 */
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "mis.h"
#include "farray.h"
#include "st.h"
#include "errtrap.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "vov.h"
#include "octio.h"

#define STREQ(a, b)  (strcmp(a, b) == 0)

#define OCTMISCHECK(fnc) \
    { \
        octStatus octmischeck_status; \
	octmischeck_status = fnc; \
        if (octmischeck_status < OCT_OK) { \
	    errRaise(MIS_PKG_NAME, octmischeck_status, octErrorString()); \
	} \
    }			     

#define node_oct_name(node)   (node)->name
static char *octCellPath, *octCellView;

/* name of the program - for errtrap */
/* char *optProgName = "misII"; */



/*
 * special routines to make octGetByName faster (use hash tables)
 *
 *   why: octGetByName does a linear search to find an object,
 *   as the number of objects increases, the search times becomes
 *   dominant.  The special functions surround octGetByName and
 *   hash information as it is found/created.
 */

static st_table *NetTable = NIL(st_table);

static void
initialize_net_table()
{
    NetTable = st_init_table(strcmp, st_strhash);
    return;
}

static octStatus
specialGetByNetName(facet, net, name)
octObject *facet, *net;
char *name;
{
    char *ptr;
    
    if (st_lookup(NetTable, name, &ptr)) {
	net->objectId = *(octId *) ptr;
	OCTMISCHECK(octGetById(net));
	return(OCT_OK);
    } else {
	if (ohGetByNetName(facet, net, name) == OCT_OK) {
	    octId *id = ALLOC(octId, 1);
	    *id = net->objectId;
	    (void) st_insert(NetTable, name, (char *) id);
	    return(OCT_OK);
	}
	return(OCT_ERROR);
    }
}

static void
specialCreateNet(facet, net, name)
octObject *facet, *net;
char *name;
{
    octId *id;

    if (st_lookup(NetTable, name, NIL(char *))) {
	errRaise(MIS_PKG_NAME, OCT_ERROR,
		 "trying to create a net that already exists: %s",
		 name);
    }
    if (ohCreateNet(facet, net, name) != OCT_OK) {
	errRaise(MIS_PKG_NAME, OCT_ERROR,
		 "can not create net %s (%s)",
		 name,
		 octErrorString());
    }
    id = ALLOC(octId, 1);
    *id = net->objectId;
    (void) st_insert(NetTable, name, (char *) id);
    return;
}

static void
specialGetOrCreateNet(facet, net, name)
octObject *facet, *net;
char *name;
{
    if (specialGetByNetName(facet, net, name) == OCT_OK) {
	return;
    }
    specialCreateNet(facet, net, name);
    return;
}



/*
 * routines to attach required SYMBOLIC policy properties
 */

static void
attachPolicyTermProperties(term, direction, termtype)
octObject *term;
char *direction, *termtype;
{
    octObject prop;
    
    OCTMISCHECK(ohCreatePropStr(term, &prop, "DIRECTION", direction));
    OCTMISCHECK(ohCreatePropStr(term, &prop, "TERMTYPE", termtype));
    return;
}

static void
attachPolicyProperties(facet)
octObject *facet;
{
    char *technology, *editstyle, *viewtype;
    octObject prop;
    
    /* attach POLICY properties */
    if ((technology = com_get_flag("OCT-TECHNOLOGY")) == NIL(char)) {
	OCTMISCHECK(ohCreateOrModifyPropStr(facet, &prop, "TECHNOLOGY", "scmos"));
    } else {
	OCTMISCHECK(ohCreateOrModifyPropStr(facet, &prop, "TECHNOLOGY", technology));
    }
    if ((editstyle = com_get_flag("OCT-EDITSTYLE")) == NIL(char)) {
	OCTMISCHECK(ohCreateOrModifyPropStr(facet, &prop, "EDITSTYLE", "SYMBOLIC"));
    } else {
	OCTMISCHECK(ohCreateOrModifyPropStr(facet, &prop, "EDITSTYLE", editstyle));
    }
    if ((viewtype = com_get_flag("OCT-VIEWTYPE")) == NIL(char)) {
	OCTMISCHECK(ohCreateOrModifyPropStr(facet, &prop, "VIEWTYPE", "SYMBOLIC"));
    } else {
	OCTMISCHECK(ohCreateOrModifyPropStr(facet, &prop, "VIEWTYPE", viewtype));
    }
    return;
}

static void
createInterface(facet)
octObject *facet;
{
    octObject interface, prop, term, iterm, dummy;
    octGenerator tgen;

    interface.type = OCT_FACET;
    interface.contents.facet.cell = facet->contents.facet.cell;
    interface.contents.facet.view = facet->contents.facet.view;
    interface.contents.facet.facet = "interface";
    interface.contents.facet.version = OCT_CURRENT_VERSION;
    interface.contents.facet.mode = "w";
    
    VOVoutputFacet( &interface );
    if (octOpenFacet(&interface) < OCT_OK) {
	errRaise(MIS_PKG_NAME, OCT_ERROR,
		 "can not open the facet %s (%s)",
		 ohFormatName(&interface),
		 octErrorString());
    }

    attachPolicyProperties(&interface);

    OCTMISCHECK(octInitGenContents(facet, OCT_TERM_MASK, &tgen));
    while (octGenerate(&tgen, &term) == OCT_OK) {
	if (ohGetByTermName(&interface, &iterm, term.contents.term.name) < OCT_OK) {
	    errRaise(MIS_PKG_NAME, OCT_ERROR,
		     "can not get the formal terminal %s on %s (%s)",
		     term.contents.term.name,
		     ohFormatName(&interface),
		     octErrorString());
	}
	if (ohGetByPropName(&term, &prop, "DIRECTION") < OCT_OK) {
	    errRaise(MIS_PKG_NAME, OCT_ERROR,
		     "can not get the DIRECTION property on terminal %s of %s (%s)",
		     term.contents.term.name,
		     ohFormatName(&interface),
		     octErrorString());
	}
	OCTMISCHECK(ohCreatePropStr(&iterm, &dummy, "DIRECTION", prop.contents.prop.value.string));
	if (ohGetByPropName(&term, &prop, "TERMTYPE") < OCT_OK) {
	    errRaise(MIS_PKG_NAME, OCT_ERROR,
		     "can not get the TERMTYPE property on terminal %s of %s (%s)",
		     term.contents.term.name,
		     ohFormatName(&interface),
		     octErrorString());
	}
	OCTMISCHECK(ohCreatePropStr(&iterm, &dummy, "TERMTYPE", prop.contents.prop.value.string));
    }
    OCTMISCHECK(ohCreatePropStr(&interface, &prop, "CELLCLASS", "MODULE"));

    OCTMISCHECK(octCloseFacet(&interface));

    return;
}
    


/*
 * replacements for the 'library' package functions to handle
 * generic instances
 */

static int IsGeneric = 0;

static void
createGenericMaster(node, name, view)
node_t *node;
char *name, *view;
/*
 * create a generic master
 */
{
    octObject facet, interface, prop, term;
    octStatus status;
    int i;
    char buffer[32];

    status = ohOpenFacet(&facet, name, view, "contents", "a");
    VOVinputFacet( &facet );
    VOVoutputFacet( &facet );
    
    if ((status == OCT_OLD_FACET) || (status == OCT_ALREADY_OPEN)) {
	return;
    } else if (status == OCT_NEW_FACET) {
	/* new cell */
	attachPolicyProperties(&facet);
	OCTMISCHECK(ohCreatePropStr(&facet, &prop, "CELLCLASS", "LEAF"));
	OCTMISCHECK(ohCreatePropStr(&facet, &prop, "CELLTYPE", "COMBINATIONAL"));
	/* put some terminals down */
	for (i = 0; i < node_num_fanin(node); i++) {
	    (void) sprintf(buffer, "in%d", i);
	    OCTMISCHECK(ohCreateTerm(&facet, &term, buffer));
	    attachPolicyTermProperties(&term, "INPUT", "SIGNAL");
	}
	OCTMISCHECK(ohCreateTerm(&facet, &term, "out0"));
	attachPolicyTermProperties(&term, "OUTPUT", "SIGNAL");
	OCTMISCHECK(octFlushFacet(&facet));

	/* create the interface facet */
	
	OCTMISCHECK(ohOpenInterface(&interface, &facet, "w"));
	VOVoutputFacet( &interface );

	attachPolicyProperties(&interface);
	OCTMISCHECK(ohCreatePropStr(&interface, &prop, "CELLCLASS", "LEAF"));
	OCTMISCHECK(ohCreatePropStr(&interface, &prop, "CELLTYPE", "COMBINATIONAL"));
	/* put some terminals down */
	for (i = 0; i < node_num_fanin(node); i++) {
	    (void) sprintf(buffer, "in%d", i);
	    if (ohGetByTermName(&interface, &term, buffer) < OCT_OK) {
		errRaise(MIS_PKG_NAME, OCT_ERROR,
			 "can not get formal terminal %s on %s (%s)",
			 buffer,
			 ohFormatName(&interface),
			 octErrorString());
	    }
	    attachPolicyTermProperties(&term, "INPUT", "SIGNAL");
	}
	if (ohGetByTermName(&interface, &term, "out0") < OCT_OK) {
	    errRaise(MIS_PKG_NAME, OCT_ERROR,
		     "can not get formal terminal %s on %s (%s)",
		     "out0",
		     ohFormatName(&interface),
		     octErrorString());
	}
	attachPolicyTermProperties(&term, "OUTPUT", "SIGNAL");
	OCTMISCHECK(octFlushFacet(&interface));
	
	return;
    }
    errRaise(MIS_PKG_NAME, status,
	     "can not open the generic master %s (%s)",
	     ohFormatName(&facet),
	     octErrorString());
    return;
}

/* XXX TOO MUCH STRING SAVING GOING ON HERE */

static void
unpackname(name, cell, view, defaultView)
char *name;
char **cell;
char **view;
char *defaultView;
{
    char *ptr;
    char buffer[4096];

    (void) strcpy(buffer, name);
    ptr = strchr(buffer, ':');
    if (ptr == NIL(char)) {
	*cell = util_strsav(buffer);
	*view = util_strsav(defaultView);
    } else {
	*ptr = '\0';
	*cell = util_strsav(buffer);
	*view = util_strsav(ptr + 1);
    }
    return;
}

static void
oct_lib_get_gate_cell_and_view(node, facet, cell, view)
node_t *node;
octObject *facet;
char **cell, **view;
{
    static char master[1024];
    char *gateName;
    char path[1024];
    int pinCount;

    if ((gateName = lib_get_gate_name(node)) == NIL(char)) {
	/* generic node */
	pinCount = node_num_fanin(node);
	(void) sprintf(path, "%s/generic-%d", facet->contents.facet.cell,
		       pinCount);
	createGenericMaster(node, path, octCellView);
	IsGeneric = 1;
	(void) sprintf(master, "generic-%d", pinCount);
	*cell = master;
	*view = octCellView;
	return;
	
    } else {
	IsGeneric = 0;
	unpackname(gateName, cell, view, octCellView);
	if ((**cell != '/') && (**cell != '~') && (octCellPath != NIL(char))) {
	    /* prepend octCellPath to cell */
	    (void) sprintf(master, "%s/%s", octCellPath, *cell);
	    *cell = master;
	}
	return;
    }	
}



static void
factored_form(network, orig_node, node, fform)
network_t *network;
node_t *orig_node, *node;
array_t *fform;
{
    node_cube_t cube;
    node_literal_t literal;
    int i, j, cube_count, fanin_count;
    char buffer[32];

    if (node_network(node) == network) {
	int size;

	/* the fanin_index stuff should be cached ... */
	(void) sprintf(buffer, "in%d", node_get_fanin_index(orig_node, node));
	size = strlen(buffer);
	for (i = 0; i < size; i++) {
	    f_array_insert_last(char, fform, buffer[i]);
	}
	return;
    }

    cube_count = node_num_cube(node);
    if (cube_count == 0) {
	return;
    }

    if (cube_count > 1) {
	f_array_insert_last(char, fform, '(');
	f_array_insert_last(char, fform, '+');
    }

    for (i = 0; i < cube_count; i++) {
	cube = node_get_cube(node, i);
	fanin_count = node_num_fanin(node);
	if (fanin_count == 0) {
	    continue;
	}
	if (fanin_count > 1) {
	    if (i > 0) {
		f_array_insert_last(char, fform, ' ');
	    }
	    f_array_insert_last(char, fform, '(');
	    f_array_insert_last(char, fform, '*');
	}

	for (j = 0; j < fanin_count; j++) {
	    literal = node_get_literal(cube, j);
	    switch (literal) {
	    case ZERO:
		f_array_insert_last(char, fform, ' ');
		f_array_insert_last(char, fform, '(');
		f_array_insert_last(char, fform, '!');
		f_array_insert_last(char, fform, ' ');
		factored_form(network, orig_node, node_get_fanin(node, j), fform);
		f_array_insert_last(char, fform, ')');
		break;
	    case ONE:
		f_array_insert_last(char, fform, ' ');
		factored_form(network, orig_node, node_get_fanin(node, j), fform);
		break;
	    }
	}
	if (fanin_count > 1) {
	    f_array_insert_last(char, fform, ')');
	}
    }

    if (cube_count > 1) {
	f_array_insert_last(char, fform, ')');
    }

    return;
}


static array_t *
oct_lib_get_factored_function(node)
node_t *node;
{
    network_t *network;
    array_t *nodes;
    int i;
    array_t *fform;

    network = node_network(node);
    nodes = factor_to_nodes(node);
    fform = array_alloc(char, 1024);
    factored_form(network, node, array_fetch(node_t *, nodes, 0), fform);
    f_array_insert_last(char, fform, '\0');
    for (i = 0; i < array_n(nodes); i++) {
	node_free(array_fetch(node_t *, nodes, i));
    }
    return fform;
}


static char *
oct_lib_get_pin_name(node, index)
node_t *node;
int index;
{
    static char buffer[32];
    
    if (IsGeneric) {
	(void) sprintf(buffer, "in%d", index);
	return(buffer);
    }
    return(lib_get_pin_name(node, index));
}


static char *
oct_lib_get_out_pin_name(node, index)
node_t *node;
int index;
{
    static char buffer[32];
    
    if (IsGeneric) {
	(void) sprintf(buffer, "out%d", index);
	return(buffer);
    }
    return(lib_get_out_pin_name(node, index));
}


static void
attachLogicFunction(term, node)
octObject *term;
node_t *node;
{
    node_function_t funct;
    
    octObject prop;
    if (IsGeneric) {
	funct = node_function(node);
	switch (funct) {
	    case NODE_UNDEFINED:
	    errRaise(MIS_PKG_NAME, OCT_ERROR,
		     "function for a node %s is undefined",
		     node_oct_name(node));
	    return;

	    case NODE_0:
	    OCTMISCHECK(ohCreatePropStr(term, &prop, "LOGICFUNCTION", "(0)"));
	    break;
	    
	    case NODE_1:
	    OCTMISCHECK(ohCreatePropStr(term, &prop, "LOGICFUNCTION", "(1)"));
	    break;

	    default:
	    {
		array_t *fform = oct_lib_get_factored_function(node);
		char *data = array_data(char, fform);

		OCTMISCHECK(ohCreatePropStr(term, &prop, "LOGICFUNCTION", data));
		array_free(fform);
	    }
	    break;
	}
    }
    return;
}



void
write_oct(network, facet)
network_t **network;
octObject *facet;
/*
 * write_oct: write a network out into an Oct facet (procedural interface)
 *
 */
{
    octObject instance, net, term, prop, bag;
    char *termName, *nodeName, *inTermName, *inNetName;
    char *masterCell, *masterView;
    int input, i;
    array_t *nodes;
    node_t *node, *fanin;
    node_function_t funct;

    VOVoutputFacet( facet );
    if (octOpenFacet(facet) < OCT_OK) {
	errRaise(MIS_PKG_NAME, OCT_ERROR,
		 "can not open the facet %s (%s)",
		 ohFormatName(facet),
		 octErrorString());
    }

    attachPolicyProperties(facet);

    /* flush the facet (side effect to create directory for GENERIC masters) */
    OCTMISCHECK(octFlushFacet(facet));
    
    OCTMISCHECK(ohCreatePropStr(facet, &prop, "CELLCLASS", "MODULE"));

    /* get a few needed items */
    octCellPath = com_get_flag("OCT-CELL-PATH");
    if ((octCellView = com_get_flag("OCT-CELL-VIEW")) == NIL(char)) {
	octCellView = "physical";
    }

    OCTMISCHECK(ohGetOrCreateBag(facet, &bag, "INSTANCES"));

    initialize_net_table();
    
    /* loop over all nodes in the network from the inputs */
    
    nodes = network_dfs(*network);

    for (i = 0; i < array_n(nodes); i++) {
	node = array_fetch(node_t *, nodes, i);

	nodeName = node_oct_name(node);
	
	funct = node_function(node);

	switch (funct) {
	    
	    case NODE_PO:
	    OCTMISCHECK(ohCreateTerm(facet, &term, nodeName));
	    attachPolicyTermProperties(&term, "OUTPUT", "SIGNAL");
	    specialGetOrCreateNet(facet, &net, node_oct_name(node_get_fanin(node, 0)));
	    OCTMISCHECK(octAttach(&net, &term));
	    break;
		
	    case NODE_PI:
	    OCTMISCHECK(ohCreateTerm(facet, &term, nodeName));
	    attachPolicyTermProperties(&term, "INPUT", "SIGNAL");
	    specialGetOrCreateNet(facet, &net, nodeName);
	    OCTMISCHECK(octAttach(&net, &term));
	    break;

	    default:
	    oct_lib_get_gate_cell_and_view(node, facet, &masterCell, &masterView);

	    /* instantiate the master */
	    OCTMISCHECK(ohCreateInst(facet, &instance, nodeName, masterCell, masterView));
	    OCTMISCHECK(octAttach(&bag, &instance));

	    /* connect up the inputs */

	    foreach_fanin(node, input, fanin) {
		inTermName = oct_lib_get_pin_name(node, input);
		inNetName = node_oct_name(fanin);
		if (ohGetByTermName(&instance, &term, inTermName) < OCT_OK) {
		    errRaise(MIS_PKG_NAME, OCT_ERROR,
			     "can not get the actual terminal %s of %s (%s)",
			     inTermName,
			     ohFormatName(&instance),
			     octErrorString());
		}
		/*
		 * the network has been ordered such that all inputs to a
		 * cell will have been created before this point
		 */
		if (specialGetByNetName(facet, &net, inNetName) < OCT_OK) {
		    errRaise(MIS_PKG_NAME, OCT_ERROR,
			     "can not get the net %s (%s)",
			     inNetName,
			     octErrorString());
		}
		OCTMISCHECK(octAttach(&net, &term));
	    }

	    specialGetOrCreateNet(facet, &net, nodeName);
	
	    termName = oct_lib_get_out_pin_name(node, 0);
	    if (ohGetByTermName(&instance, &term, termName) < OCT_OK) {
		errRaise(MIS_PKG_NAME, OCT_ERROR,
			 "can not get the terminal %s of %s (%s)",
			 termName,
			 ohFormatName(&instance),
			 octErrorString());
	    }
	    OCTMISCHECK(octAttach(&net, &term));

	    attachLogicFunction(&term, node);
	}
    }

    createInterface(facet);
    
    OCTMISCHECK(octCloseFacet(facet));

    /* free up some space */
    st_free_table(NetTable);

    return;
}



static jmp_buf mis_oct_error_buf;

static void
mis_oct_error_handler(pkgName, code, message)
char *pkgName;
int code;
char *message;
{
    (void) fprintf(stderr, "%s: error (%d)\n\t%s\n", pkgName, code, message);
    (void) longjmp(mis_oct_error_buf, -1);
    return;
}


int
external_write_oct(network, argc, argv)
network_t **network;
int argc;
char **argv;
/*
 * write_oct: write a network out into an Oct facet
 *
 */
{
    octObject facet;
    
    
    if (argc != 2) {
	(void) fprintf(miserr, "usage: write_oct cell[:view]\n");
	return(1);
    }

    octBegin();
    
    if (setjmp(mis_oct_error_buf) == 0) {
	errPushHandler(mis_oct_error_handler);
	
	(void) ohUnpackDefaults(&facet, "w", ":logic:contents");
	if (ohUnpackFacetName(&facet, argv[1]) != OCT_OK) {
	    errRaise(MIS_PKG_NAME, OCT_ERROR, "usage: write_oct cell[:view]");
	}

	error_init();
	write_oct(network, &facet);
	errPopHandler();
    } else {
	return(1);
    }
    
    octEnd();
    
    return(0);
}

#endif /* OCT */
