/************************************************************************************************
  Mapping for Actel architecture. Can be used for other mux based architectures too: patterns 
  in map_act() need be changed only. Also, throughout the file ACT and BDD are used 
  interchangeably.
*************************************************************************************************/

#include "mis.h"
#include "pld_int.h"
extern int MAXOPTIMAL;
static COST_STRUCT *act_evaluate_map_cost();

static network_t *decomp_big_nodes();
static network_t *improve_network();
static network_t *act_quick_phase();
static network_t *iterative_improvement();

static node_t *mux_node();
static node_t *basic_block_node();
static node_t *act_get_function();
static node_t *get_node_of_vertex();
static void free_node_if_possible();
static void my_traverse_act();
static void put_node_names_in_act();
static void print_vertex();
static int OR_pattern();
static array_t *OR_literal_order();
static array_t *single_cube_order();
static int minimum_cost_index();
static char *formulate_IP();
static void read_LINDO_file();
static void partial_scan_lindoFile();
static int is_anyfo_PO();
static int compute_cost_array();
static int cost_of_node();
static void cost_update();
static void print_output_node_with_string();
static void set_mark_act();
static void free_nodes_in_act();
static void init_bdnet();
static enum st_retval free_table_entry();
static void put_nodes();
static ACT_VERTEX_PTR my_create_act();
static ACT_VERTEX_PTR my_create_act_general();
static node_t *construct_node_from_cube();
static int all_fanins_positive();

static char *act_get_node_fn_type();
static void  print_node_with_string();

array_t *multiple_fo_array; /* holds the nodes that have multiple_fo in the
			       act */
array_t *num_mux_struct_array; /* holds the cost of each multiple_fo act */

array_t *my_node_array;

int WHICH_ACT;
ACT_VERTEX * PRESENT_ACT; /* for temporary storage */

static int MARK_VALUE, MARK_COMPLEMENT_VALUE;
int iteration_num;
int num_or_patterns;
static int instance_num;

network_t *
act_map_network(network, HEURISTIC_NUM, NUM_ITER, FANIN_COLLAPSE, GAIN_FACTOR, DECOMP_FANIN, DISJOINT_DECOMP, 
			 QUICK_PHASE, SIMPLIFY, cost_table)
  network_t *network;
  int  HEURISTIC_NUM, NUM_ITER, FANIN_COLLAPSE, DECOMP_FANIN, DISJOINT_DECOMP, QUICK_PHASE, SIMPLIFY;
  float GAIN_FACTOR;
  st_table *cost_table;

{
  node_t *node, *node_simpl;
  COST_STRUCT *cost_node;
  char *name;
  char *name1;
  int i, size;
  array_t *nodevec;
  
  /* Do a disjoint decomposition - can also be done in the script - added on Nov. 14 */
  /*---------------------------------------------------------------------------------*/
  if (DISJOINT_DECOMP == 1) decomp_disj_network(network);

  /* simplification is amust if not done an optimization phase. This is because
     otherwise bdnet file generated may not be correct. (minimum-base problem)		 */
  /*-------------------------------------------------------------------------------------*/
  if (SIMPLIFY) {
      nodevec = network_dfs(network);
      size = array_n(nodevec);
      for (i = 0; i < size; i++) {
	    node = array_fetch(node_t *, nodevec, i);
	    if (node->type != INTERNAL) continue;
	    node_simpl = node_simplify(node, (node_t *) NULL, NODE_SIM_ESPRESSO);
	    node_replace(node, node_simpl);	    
      }
      array_free(nodevec);
  }

  nodevec = network_dfs(network);
  size = array_n(nodevec);
  for (i = 0; i < size; i++) {
     node = array_fetch(node_t *, nodevec, i);
     if (node->type == INTERNAL) {
	cost_node =   act_evaluate_map_cost(node, HEURISTIC_NUM);
	name1 = node_long_name(node);
	name = ALLOC(char, strlen(name1) + 1);
        (void) strcpy(name, name1);
        assert (!st_insert(cost_table, name, (char *)cost_node));
    } /* if INTERNAL */
  }
  array_free(nodevec);
  iteration_num = 0;

  network =  iterative_improvement(network, cost_table, NUM_ITER, HEURISTIC_NUM,  
			 DECOMP_FANIN, FANIN_COLLAPSE, GAIN_FACTOR, QUICK_PHASE );	
  return network; 
}

/*******************************************************************************************************************
 improve the network using repeated partial collapse and decomp till NUM_ITER are done, or no further improvement.
 then do a quick-phase to determine which nodes should be realized as complemented. 
********************************************************************************************************************/

static
network_t *
iterative_improvement(network, cost_table, NUM_ITER, HEURISTIC_NUM, DECOMP_FANIN, FANIN_COLLAPSE, GAIN_FACTOR, QUICK_PHASE)
   network_t *network;
   st_table  *cost_table;
   int 	     NUM_ITER;
   int 	     HEURISTIC_NUM;
   int 	     DECOMP_FANIN;
   int	     FANIN_COLLAPSE;
   float     GAIN_FACTOR;
   int 	     QUICK_PHASE;
{
 if (NUM_ITER > 0)
    network = improve_network(network, cost_table, NUM_ITER, HEURISTIC_NUM, 
		              FANIN_COLLAPSE, DECOMP_FANIN, GAIN_FACTOR);
 if (QUICK_PHASE == 1) network =  act_quick_phase(network, cost_table, HEURISTIC_NUM);
 return network;
}

/****************************************************************************
act_evaluate_map_cost() - summary 
    It maps the given node onto the actel architecture. 
    For each node, the (BDD or OBDD - act) is generated;
    at multiple fanout points in the act,
    the act is broken and a set of acts is 
    constructed. For each small act(a tree), the pattern_match algorithm is
    used to compute the cost of the act in terms of the mux structure.
*****************************************************************************/

static
COST_STRUCT * act_evaluate_map_cost(node, HEURISTIC_NUM)
 node_t *node;
 int HEURISTIC_NUM;
{

  ACT_VERTEX_PTR act_of_node, act1_of_node, act2_of_node; /* act is for a multiple fo vertex */
  COST_STRUCT *cost_node, *cost_node1, *cost_node2;
  node_function_t node_fun;

  if (node->type != INTERNAL) return NIL(COST_STRUCT); 
  node_fun  = node_function(node);
  switch(node_fun) {
	case NODE_0:
    	   cost_node = ALLOC(COST_STRUCT, 1);
	   cost_node->act = NULL;
	   cost_node->cost = 0;
	   cost_node->fn_type = "GND";
	   return cost_node;
	case NODE_1:
    	   cost_node = ALLOC(COST_STRUCT, 1);
	   cost_node->act = NULL;
	   cost_node->cost = 0;
	   cost_node->fn_type = "Vdd";
	   return cost_node;
        default:;
     }
       /* if bdd was constructed using h 2, vertex->my_type = IMPROPER */
		
       if (ACT_DEBUG) node_print(stdout, node); 

       /* first construct OBDD for the case when it is known how to construct
	  optimum one */
       /*-------------------------------------------------------------------*/
       if((node_num_literal(node) == node_num_cube(node)) || 
	  (node_num_cube(node) == 1)) {
	   act_of_node = my_create_act(node);
	   act_of_node->my_type = PROPER; 
	   WHICH_ACT = PROPER;
           act_of_node->node = node;
	   cost_node = make_tree_and_map(node, act_of_node);
	   act_act_free(node);
	   return cost_node;
       } 

       switch(HEURISTIC_NUM){
	     case 1:
	       /* make OBDD */
	       /*-----------*/
	       act_of_node = my_create_act(node);
	       act_of_node->my_type = PROPER; 
	       WHICH_ACT = PROPER;
               act_of_node->node = node;
   	       cost_node = make_tree_and_map(node, act_of_node);
	       act_act_free(node);
	       return cost_node;

	     case 2:
		/* make unordered BDD (act) */
		/*--------------------------*/
	        make_act(node);
       		ACT_SET(node)->LOCAL_ACT->act->node_list = NIL(array_t );
       		act_of_node = ACT_SET(node)->LOCAL_ACT->act->root;
		act_of_node->my_type = IMPROPER; 
	        WHICH_ACT = IMPROPER;
                act_of_node->node = node;
    	        cost_node = make_tree_and_map(node, act_of_node);
		act_act_free(node);
	        return cost_node;

	     case 3:
	       /* decide which subject-graph to use based on fanin info */
	       /*-------------------------------------------------------*/
	       if ((node_num_fanin(node) <= MAXOPTIMAL) || 
		   (node_num_literal(node) == node_num_fanin(node))
		   ) {
		   act_of_node = my_create_act(node);
		   act_of_node->my_type = PROPER; 
		   WHICH_ACT = PROPER;
                   act_of_node->node = node;

	       } else {
		   make_act(node);
		   ACT_SET(node)->LOCAL_ACT->act->node_list = NIL(array_t ); 
		   act_of_node = ACT_SET(node)->LOCAL_ACT->act->root;
		   act_of_node->my_type = IMPROPER; 
		   WHICH_ACT = IMPROPER;
                   act_of_node->node = node;
	       }
   	       cost_node = make_tree_and_map(node, act_of_node);
   	       act_act_free(node);
	       return cost_node;


	     case 4:
		   /* make both the subject graphs and pick the one with minimum cost */
		   /* ----------------------------------------------------------------*/
	           make_act(node);
		   ACT_SET(node)->LOCAL_ACT->act->node_list = NIL(array_t );
		   act2_of_node = ACT_SET(node)->LOCAL_ACT->act->root;
		   act2_of_node->my_type = IMPROPER; 
		   WHICH_ACT = IMPROPER;
                   act2_of_node->node = node;
	           cost_node2 = make_tree_and_map(node, act2_of_node);		   
		   act_act_free(node);

		   if (cost_node2->cost <= 1) {
			return cost_node2;
    		   }
		   act1_of_node = my_create_act(node);
		   act1_of_node->my_type = PROPER; 
		   WHICH_ACT = PROPER;
                   act1_of_node->node = node;
	           cost_node1 = make_tree_and_map(node, act1_of_node);
		   act_act_free(node);
	 		   	
		   if (cost_node1->cost <= cost_node2->cost) {
			(void) free_cost_struct(cost_node2);
			return	cost_node1;
		   }		
		   (void) free_cost_struct(cost_node1);
		   return  cost_node2;

	     default:
	       (void) printf("Error: act_map_evaluate(): %d is not a valid heuristic number\n", HEURISTIC_NUM);
	       exit(1);
	   }
 return (COST_STRUCT *) NULL;
}


/************************************************************************************************
  Given a node and its subject_graph (ACT, or OACT), snap at multiple fanout points and maps each 
  tree using the pattern-set.
*************************************************************************************************/


COST_STRUCT *
make_tree_and_map(node, act_of_node)
   node_t *node;
   ACT_VERTEX_PTR act_of_node;

{
 int 		num_acts;
 int 		i;
 int 		num_mux_struct;
 int 		TOTAL_MUX_STRUCT; /* cost of the network */
 COST_STRUCT 	*cost_node;
 ACT_VERTEX_PTR 	act;
 static void initialize_act(); 

       TOTAL_MUX_STRUCT = 0;
       cost_node = ALLOC(COST_STRUCT, 1);
       init_multiple_fo_array(act_of_node); 
       if (ACT_DEBUG) my_traverse_act(act_of_node);

       initialize_act(act_of_node); /* initialize the required structures */
       num_acts = array_n(multiple_fo_array);
       for (i = 0; i < num_acts; i++) {
      	  act = array_fetch(ACT_VERTEX_PTR, multiple_fo_array, i);
          PRESENT_ACT = act;
    	  num_mux_struct = map_act( act); 
 	  TOTAL_MUX_STRUCT += num_mux_struct;
       }
       if (ACT_DEBUG) 
	  (void) printf("number of total mux_structures used for the node %s = %d\n",
			 node_long_name(node), TOTAL_MUX_STRUCT);


       array_free(multiple_fo_array);
       cost_node->cost = TOTAL_MUX_STRUCT;
       cost_node->act  = act_of_node; 
       return cost_node;
}

/****************************************************************************************************************
 Creates OBDD for a given node.  If the node has <= MAXOPTIMAL fanins, optimum OBDD (with min. number of basic
 blocks) would be constructed. 
*****************************************************************************************************************/

static
ACT_VERTEX_PTR my_create_act(node)
  node_t *node;

{
  array_t *OR_literal_order();
  array_t *single_cube_order();
  array_t *nodevec;
  array_t *order_list; 
  array_t *decomp_array;
  node_t  *n, *fanin;
  node_t  *node_of_cube;
  node_cube_t cube;
  int 	  size;
  int     num_cube_n, num_literal_n;
  int	  i;
  ACT_VERTEX_PTR   act;
  int     act_constructed; /* in case number of cubes = 2 */
  int 	  num_cubes;
  int	  num_literals;
  int     num_fanins;
  array_t *cube_order_list, *order_list_new;
  extern  ACT_VERTEX * p_act_construct();

   act_constructed = 0;

  /* if the number of cubes is 1, then make sure that the bottom ones are
  not inverted. If the number is 2 and the possibility that there
  is a factor of type (a+b), then order a,b towards the end(bottom) */
  /*----------------------------------------------------------------*/
  num_cubes = node_num_cube(node);
  if (num_cubes == 0) {

        nodevec = array_alloc(node_t *, 0);
        array_insert_last(node_t *, nodevec, node);
	p_actCreate4Set(nodevec, NIL(array_t), 1, OPTIMAL);
	act = ACT_SET(node)->LOCAL_ACT->act->root;
	array_free(nodevec);
	return act;
  }
  num_literals = node_num_literal(node);
  num_fanins = node_num_fanin(node);
  if (num_cubes == num_literals) {
    if (num_fanins == num_literals) {
	order_list = OR_literal_order(node);
	act = (ACT_VERTEX *) p_act_construct(node, order_list, 1);
	put_node_names_in_act(ACT_SET(node)->LOCAL_ACT->act->root, order_list);
	array_free(order_list);
	ACT_SET(node)->LOCAL_ACT->act->node_list = NIL (array_t);
	return act;
    }
    else {
	act = my_create_act_general(node);
	return act;
    }
  }
  /* take into account the fact that OR gate is there for single cube*/
  /*-----------------------------------------------------------------*/

  if (num_cubes == 1) {
    if  (num_fanins == num_literals) {
       order_list = single_cube_order(node);
       act = p_act_construct(node, order_list, 1);
       put_node_names_in_act(ACT_SET(node)->LOCAL_ACT->act->root, order_list);
       array_free(order_list);
       ACT_SET(node)->LOCAL_ACT->act->node_list = NIL (array_t);
       return act;
   }
   else {
       act = my_create_act_general(node);
       return act; 
   }
  }

  /* if num_literals = num_fanin_node, then it means it is a
   disjoint support tree, then the ordering is obtained by
   concatenating the optimal ordering of the individual cubes.
   ordering of the individual cubes is carried on in the same way as 
   ordering of the nodes with a single cube. case like f = ab + c'    */
   /*-----------------------------------------------------------------*/

  if (num_literals == num_fanins) {
       order_list = array_alloc(node_t *, 0);  
       for (i = num_cubes - 1; i >= 0; i--) {
           cube = node_get_cube(node, i);
	   node_of_cube = construct_node_from_cube(cube, node);
	   cube_order_list = single_cube_order(node_of_cube);
	   node_free(node_of_cube);
	   order_list_new = array_join(order_list, cube_order_list);
	   array_free(order_list);
	   array_free(cube_order_list);
  	   order_list = array_dup(order_list_new);
	   array_free(order_list_new);
       }
       act = p_act_construct(node, order_list, 1);
       put_node_names_in_act(ACT_SET(node)->LOCAL_ACT->act->root, order_list);
       array_free(order_list);
       ACT_SET(node)->LOCAL_ACT->act->node_list = NIL (array_t);
       return act;              
  }

  /* if num_cubes == 2, then check for a special condition -  basically a filter */
  /*-----------------------------------------------------------------------------*/
  if (num_cubes == 2) {
       decomp_array = decomp_good(node);
       size = array_n(decomp_array);
       if (size == 2) {
          n = array_fetch(node_t *, decomp_array, 1);
  	  num_cube_n = node_num_cube(n);
	  num_literal_n = node_num_literal(n);
	  if ((num_cube_n == num_literal_n) && (all_fanins_positive(n))){
	     order_list = array_alloc(node_t *, 0);
	     put_nodes(array_fetch(node_t *, decomp_array, 0), n, order_list);
	     for (i = num_literal_n - 1; i >= 0; i--) {
	       fanin = node_get_fanin(n, i);
	       array_insert_last(node_t *, order_list, fanin);
	     }
	     act_constructed = 1;
             act = p_act_construct(node, order_list, 1);
   	     put_node_names_in_act(ACT_SET(node)->LOCAL_ACT->act->root, order_list);
	     array_free(order_list);
  	     ACT_SET(node)->LOCAL_ACT->act->node_list = NIL (array_t);

          }
        }
        for (i = 0; i < size; i++) {
	  node_free(array_fetch(node_t *, decomp_array, i));
        }
        array_free(decomp_array);
        if (act_constructed) return act;

  }  /* if num_cubes == 2 */

  /* create general act */
  /*--------------------*/
  act = my_create_act_general(node);
  return act;
}

static 
ACT_VERTEX_PTR
my_create_act_general(node)
  node_t *node;
{
  array_t *nodevec;
  ACT_VERTEX_PTR   act;
  array_t *order_list;
  array_t *pld_order_nodes();
  extern  ACT_VERTEX * p_act_construct();
  
  nodevec = array_alloc(node_t *, 0);
  array_insert_last(node_t *, nodevec, node);
  
  /* if number of fanins is leq than MAXOPTIMAL, construct an optimal ACT */
  /*----------------------------------------------------------------------*/ 
  if (node_num_fanin(node) <=  MAXOPTIMAL) {
	p_actCreate4Set(nodevec, NIL(array_t), 1, OPTIMAL);
	act = ACT_SET(node)->LOCAL_ACT->act->root;
	if (ACT_SET(node)->LOCAL_ACT->act->node_list == NULL) {
		(void) printf(" Error: my_create_act(): optimal order list not\
				returned\n");
		exit(1);
	}
        put_node_names_in_act(ACT_SET(node)->LOCAL_ACT->act->root, 
			      ACT_SET(node)->LOCAL_ACT->act->node_list);

  }
  else {		
     order_list = pld_order_nodes(nodevec, 0); /* include internal nodes also */
     act =  p_act_construct(node, order_list, 1);  	/* local ACT */
     put_node_names_in_act(ACT_SET(node)->LOCAL_ACT->act->root, order_list);
     array_free( order_list);
     ACT_SET(node)->LOCAL_ACT->act->node_list = NIL (array_t);

  }
  array_free (nodevec); 
  return  act;

 
}


/************************************************************************************************
 Given a cube, constructs a node that has the same function as the cube 
************************************************************************************************/

static
node_t * construct_node_from_cube(cube, node)
   node_cube_t cube;
   node_t *node;

{

   node_t  *n, *nnew, *literal_node, *fanin;
   int     j;
   node_literal_t   literal;

   n = node_constant(1);
   foreach_fanin(node, j, fanin) {
     literal = node_get_literal(cube, j);
     switch(literal) {
       case ZERO:
	    literal_node = node_literal(fanin, 0);
	    nnew = node_and(n, literal_node);
	    node_free(literal_node);
	    node_free(n);
	    n = node_dup(nnew);
	    node_free(nnew);
	    break;

       case ONE:
	    literal_node = node_literal(fanin, 1);
	    nnew = node_and(n, literal_node);
	    node_free(literal_node);
	    node_free(n);
	    n = node_dup(nnew);
	    node_free(nnew);
	    break;
       default:;
     } /* switch */
  } /* for loop */

  return n;
}
   
           
   

/******************************************************************************
  given a node, checks if all the fanins have a positive phase.
  Returns 1 if so, else returns 0.
******************************************************************************/

static
int all_fanins_positive(node)
    node_t *node;
{
 
  int 	 num_fanin;
  node_t *fanin;
  input_phase_t phase;
  int 	 i;

  num_fanin = node_num_fanin(node);
  for (i = num_fanin - 1; i >= 0; i--) {
    fanin = node_get_fanin(node, i);
    phase = node_input_phase(node, fanin);
    if (phase != POS_UNATE) return 0;
  }
  return 1;
}

/****************************************************************************
  puts all the fanins of the node "node" in the order_list, 
  except for the node "n".
*****************************************************************************/

static void
put_nodes(node, n, order_list)
   node_t 	*node;
   node_t 	*n;
   array_t 	*order_list;
{
   int 		num_fanin;
   node_t 	*fanin;
   int		i;
   int 		check;  /* for debugging */
 
   check = 0;
   num_fanin = node_num_fanin(node); 
   for (i = num_fanin -1; i>= 0; i--) {
     fanin = node_get_fanin(node, i);
     if (fanin == n) {
       check = 1;
       continue;
     }
     array_insert_last(node_t *, order_list, fanin);
   }

   /* for debugging */
   /*---------------*/
   if (check == 0) {
    (void) printf("put_order: node %s does not have the required fanin\n", 
	    node_long_name(node));
    exit(1);
   }
}

/*************************************************************
  Initializes the array for the multiple_fo_vertices in the
  act for node under consideration. The first entry is the 
  root itself, although it is not multiple fan_out vertex.
************************************************************/

static
void init_multiple_fo_array(act_of_node)
   ACT_VERTEX_PTR act_of_node;
{
  multiple_fo_array = array_alloc(ACT_VERTEX_PTR, 0);
  array_insert_last(ACT_VERTEX_PTR, multiple_fo_array, act_of_node);
}

/**********************************************************************
    recursively initialize the required structure of each act vertex.
   also finds out multiple fan out vertices and stores them in the 
   multiple_fo_array.  
************************************************************************/

static
void
initialize_act(vertex)
   ACT_VERTEX_PTR vertex;
{
 /* to save repetition, mark the vertex as visited */
 /*------------------------------------------------*/
  if (vertex->mark == 0) vertex->mark = 1;
  else vertex->mark = 0;

  vertex->pattern_num = -1;
  vertex->cost = 0;
  vertex->mapped = 0;
  vertex->multiple_fo = 0;

  /* do not care for a terminal vertex- nothing is to be done */
  /*----------------------------------------------------------*/
  if (vertex->value != 4) return;

  /* if the left(right) son not visited earlier, initialize it, else
     mark it as a multiple fo vertex */
  /*---------------------------------------------------------------*/

  if (vertex->mark != vertex->low->mark)   initialize_act(vertex->low);
  else {
      if (vertex->low->multiple_fo == 0) {
         vertex->low->multiple_fo = 1;
         array_insert_last(ACT_VERTEX_PTR, multiple_fo_array, vertex->low);
     }
  }      

  if (vertex->mark != vertex->high->mark)  initialize_act(vertex->high);
  else {
      if (vertex->high->multiple_fo == 0) {
          vertex->high->multiple_fo = 1;
          array_insert_last(ACT_VERTEX_PTR, multiple_fo_array, vertex->high);
      }
  }

}

/******************************************************************************************
  Just prints out the BDD (OBDD) starting at the vertex. It first visits the low child and 
  then the right child. Marks all the nodes that are visited so that in future, it does not
  traverse an already traversed (marked) node. 
******************************************************************************************/
static
void my_traverse_act(vertex)
  ACT_VERTEX_PTR vertex;
{

   if (vertex->mark == 0) vertex->mark = 1;
   else vertex->mark = 0;
   print_vertex(vertex);
 
   if (vertex->value != 4) return;

   if (vertex->mark == vertex->low->mark){
	 print_vertex(vertex->low);
         if (vertex->mark == vertex->high->mark){
	    print_vertex(vertex->high);
   	    return;
         }
         my_traverse_act(vertex->high);
         return;
   }
   my_traverse_act(vertex->low);
   if (vertex->mark == vertex->high->mark){
	print_vertex(vertex->high);
   	return;
   }
   my_traverse_act(vertex->high);
   return;

}


/***************************************************************************
   Annotates the act vertex with the name of the corresponding index-node.
****************************************************************************/

static
void put_node_names_in_act(vertex, list)
  ACT_VERTEX_PTR vertex;
  array_t *list;
{
   node_t *node;

   if (vertex->mark == 0) vertex->mark = 1;
   else vertex->mark = 0;
 
   if (vertex->value != 4) return;

   node = array_fetch(node_t *, list, vertex->index);
   vertex->name = node_long_name(node);
   if (vertex->mark != vertex->low->mark)  /* low child already visited */
	  put_node_names_in_act(vertex->low, list);
   if (vertex->mark != vertex->high->mark)
	  put_node_names_in_act(vertex->high, list);	
}


/*****************************************************************************************
 Prints information on a vertex of the ACT (OACT) 
*****************************************************************************************/
static void
print_vertex(vertex)
ACT_VERTEX_PTR vertex;
{

   /* terminal vertex */
   /*-----------------*/
   if (vertex->value != 4) {
      (void) printf ("value = %d, id = %d, index = %d\n", vertex->value, 
	       vertex->id, vertex->index); 
      return;
   }

   /* nonterminal vertex */
   /*--------------------*/
      if (WHICH_ACT == PROPER)
         (void) printf ("id = %d, index = %d\n", vertex->id, vertex->index);
      else 
         (void) printf ("name = %s, index = %d\n", vertex->name, vertex->index);
}

/***********************************************************************
   Returns the cost of mapping a act vertex onto the patterns 
   The cost of 1 mux structure is 1, hence in all configurations the mux 
   is equally costly. This results in a very simple method of mapping.
   The function looks at the root of the tree_act, looks at two levels below
   and calculates the cost of the children there, if any, and adds one
   to the cost
***********************************************************************/
        
static
int map_act(vertex)
   ACT_VERTEX_PTR vertex;

{


   int TOTAL_PATTERNS = 12;
   int pat_num;
   int cost[12]; /* for each of the 12 patterns derived from the mux struct */
   int vlow_cost;
   int vhigh_cost;
   int vlowlow_cost;
   int vlowhigh_cost;
   int vhighlow_cost;
   int vhighhigh_cost;
   int best_pat_num;   

  /* if cost already computed, don't compute again */
  /*-----------------------------------------------*/
  if (vertex->mapped) return vertex->cost;


   /* if multiple fanout (but not the root of the present multiple_fo vertex) 
   return 0, but do not set any value, as this vertex would be/ or was
   visited earlier as a root. Basically zero is not its true cost. So do not set 
   vertex->mapped = 1 also */

  if ((vertex->multiple_fo == 1) && (vertex != PRESENT_ACT)) {
     vertex->cost = 0;
     return 0;
  }

  vertex->mapped = 1; 

  if (vertex->value != 4) {
     vertex->cost = 0;
     return 0;
  }  

  /* if a simple input, cost = 0 */
  /*-----------------------------*/
  if ((vertex->low->value == 0) && (vertex->high->value == 1)) {
      vertex->cost = 0;
      return 0;
  }
 

  /* initialize cost vector to a very high value for all the patterns*/
  /*-----------------------------------------------------------------*/
  for (pat_num = 0; pat_num < TOTAL_PATTERNS; pat_num++) {
     cost[pat_num] = MAXINT;
  }

  /***********recursively calculate relevant costs *************/
  
  if (vertex->low->multiple_fo == 1) vlow_cost = 0;
  else vlow_cost = map_act(vertex->low);

  if (vertex->high->multiple_fo == 1) vhigh_cost = 0;
  else   vhigh_cost = map_act(vertex->high);

  if (vertex->low->value == 4) {
    if (vertex->low->multiple_fo == 1) 
	vlowlow_cost = vlowhigh_cost = MAXINT;
    else {
      if (vertex->low->low->multiple_fo == 1) vlowlow_cost = 0;
      else  vlowlow_cost = map_act(vertex->low->low);
      if (vertex->low->high->multiple_fo == 1) vlowhigh_cost = 0;
      else  vlowhigh_cost = map_act(vertex->low->high);
    }
  }

  if (vertex->high->value == 4) {
    if (vertex->high->multiple_fo == 1)
	vhighlow_cost = vhighhigh_cost = MAXINT;
    else {
      if (vertex->high->low->multiple_fo == 1) vhighlow_cost = 0;      
      else  vhighlow_cost = map_act(vertex->high->low);
      if (vertex->high->high->multiple_fo == 1) vhighhigh_cost = 0;
      else  vhighhigh_cost = map_act(vertex->high->high);
    }
  }

  /**********enumerate possible cases************************/

 /* l.c. (left child) + r.c.(right child) + 1 */
 /*-------------------------------------------*/
 cost[0] = vlow_cost + vhigh_cost + 1;
  
  if (vertex->low->value == 4) 
     cost[1] = vlowlow_cost + vlowhigh_cost + vhigh_cost + 1;

  if (vertex->high->value == 4)
     cost[2] = vlow_cost + vhighlow_cost + vhighhigh_cost + 1;
  
  if ((vertex->low->value == 4) && (vertex->high->value == 4))
     cost[3] = vlowlow_cost + vlowhigh_cost +
		     vhighlow_cost + vhighhigh_cost + 1;

  /* OR patterns - pretty involved -> for more explanation, see doc. */


  if ((OR_pattern(vertex))
      && (vertex->low->low->multiple_fo == 0)
      && (vertex->low->low->value == 4)
     ) 
         cost[4] = map_act(vertex->low->low->low) + 
		   map_act(vertex->low->low->high)
         	   + 1;

  if ((OR_pattern(vertex->high)) && (vertex->high->multiple_fo == 0)) {
     if (vertex->low->value == 0) 
       cost[5] = map_act(vertex->high->high) + map_act(vertex->high->low->low)
	         + 1;
     else {
        if ((vertex->low->value == 1) && (vertex->high->high->value == 0) 
   	    && (vertex->high->low->low->value == 1))
      	  cost[6] = 1;
     }
  }

  if ((OR_pattern(vertex->low)) && (vertex->low->multiple_fo == 0)) {
     if (vertex->high->value == 0) 
       cost[7] = map_act(vertex->low->high) + map_act(vertex->low->low->low)
	         + 1;
  }

  /************find minimum cost************************/
  best_pat_num = minimum_cost_index(cost, TOTAL_PATTERNS);
  vertex->cost = cost[best_pat_num];
  vertex->pattern_num = best_pat_num;
  return vertex->cost;
}


/**************************************************************************
   if an OR pattern (or NOR pattern) is rooted at the root, return 1, else
   return 0 
***************************************************************************/
static
int OR_pattern(root)
    ACT_VERTEX_PTR root;
{
  if ((root->value == 4) && 
      (root->low->multiple_fo == 0) && (root->low->value == 4) &&
      (root->high == root->low->high)
     )
     return 1;

  return 0;
}

/************************************************************************
 returns the index of the pattern which results in the minimum cost
 at the present node.
*************************************************************************/
static
int minimum_cost_index(cost, num_entries)
 int cost[];
 int num_entries;
{

  int i;
  int min_index = 0;

  for (i = 1; i < num_entries; i++) {
    if (cost[i] < cost[min_index]) min_index = i;
  }
  return min_index;
}

static
network_t *
improve_network(network, cost_table, NUM_ITER, HEURISTIC_NUM, FANIN_COLLAPSE, DECOMP_FANIN, GAIN_FACTOR)
  network_t *network;
  st_table *cost_table;
  int NUM_ITER, HEURISTIC_NUM, FANIN_COLLAPSE, DECOMP_FANIN;
   float GAIN_FACTOR;
{

  int nogain;
  network_t *dup_net;
  node_t *dup_node;
  lsGen genfo;
  node_t *node, *fanout; 
  /* node_t *fanout_simpl; */
  COST_STRUCT  ***focost;
  int **focost_ind;
  char *dummy;
  char buf[BUFSIZE];  /* to hold the system command */
  int i, j;
  array_t *nodevec;
  st_table *node2num_table;
  int size_net;
  int clust_num;
  int numnodes;  /*number of internal nodes */

  int **M; /* coeff. for Integer Programming formulation */
  int *num_ones; /* number of ones in a row of M[][] */
  int *g; /*gain for a cluster*/
  int *collapsed_node; /* for a cluster, stores the id of the 
  			  collapsed node */

  int *variable; /* to hold the value of the variables after IP */

  int iteration_gain; /* tells the gain at the end of present iteration */  
  int gain; /* gain for the node being collapsed */
  int decomp_gain; /* gain for network_decomposition at an iteration */
  int collapse_gain; /*gain from the collapse operation */
  int old_cost; /* old_cost of a fan_out node */
  int cost_network; /* cost of the network in terms of mux_structs*/

  int num_fo;   /* number of fan_outs of a ndoe */
  int clust_num_add1;

  char *lindopathname;
  char *name, *name1;
  char *infile, *outfile;
  
nogain = FALSE;

while (nogain == FALSE) { 

  cost_network =   print_network(network, cost_table);

  /* decomposing big nodes in the network if it is profitable*/
  /*---------------------------------------------------------*/

  network =  decomp_big_nodes(network, cost_table, &decomp_gain, DECOMP_FANIN, HEURISTIC_NUM);

  /* check */
  /*-------*/
 if (!network_check(network)) {
    (void) printf("improve_network:%s\n", error_string());
    return network;
 }
  if (ACT_DEBUG && (decomp_gain > 0))  
       (void) print_network(network, cost_table);
  iteration_gain = decomp_gain;
  iteration_num++;
  if (iteration_num > NUM_ITER) return network;
  nogain = TRUE; 
  clust_num = -1;

 /* check if collapsing is desired. If not, continue if gain sufficient */
 /*---------------------------------------------------------------------*/

 if (FANIN_COLLAPSE == 0) {
        if (iteration_gain >= (GAIN_FACTOR * cost_network)) {
		nogain = FALSE;
	}
	continue;
  }  

  /* search for lindo in the path name. If not found, do not 
     do collapsing */
  /*--------------------------------------------------------*/
  lindopathname = util_path_search("lindo");
  if (lindopathname == NULL) {
	(void) printf("YOU DO NOT HAVE LINDO INTEGER PROGRAMMING PACKAGE\n");
	(void) printf("\tIN YOUR PATH.\n");
	(void) printf("CONTINUING WITHOUT LINDO.....\n");
	(void) printf("PLEASE OBTAIN LINDO FROM\n");

        (void) printf("The Scientific Press, 540 University Ave.\n");
        (void) printf("Palo Alto, CA 94301, USA\n");
        (void) printf("TEL : (415) 322-5221\n");
        if (iteration_gain >= (GAIN_FACTOR * cost_network)) {
		nogain = FALSE;
	}
	continue;
  }  
  FREE(lindopathname);

  /*Initialization */
  /*--------------*/
  numnodes = network_num_internal(network) + network_num_po(network) 
		     + network_num_pi(network);

  focost = ALLOC(COST_STRUCT **, numnodes);
  focost_ind = ALLOC(int *, numnodes);
  M = ALLOC(int *, numnodes);
  num_ones = ALLOC(int, numnodes);
  g = ALLOC(int, numnodes);
  collapsed_node = ALLOC(int, numnodes);
  variable = ALLOC(int, numnodes);

  for (i = 0; i < numnodes; i++) {
   focost[i] = ALLOC(COST_STRUCT *, numnodes);
   M[i] = ALLOC(int, numnodes);
   focost_ind[i] = ALLOC(int, numnodes);
   for (j = 0; j < numnodes; j++) {
     M[i][j] = 0;
     focost_ind[i][j] = 0;
   }
   num_ones[i] = 0;
  }
  
  node2num_table =  st_init_table(st_ptrcmp, st_ptrhash);    

  nodevec = network_dfs(network);
  size_net = array_n(nodevec);
  /* establish a link from node to its number */
  /*------------------------------------------*/
  for (i = 0; i <size_net; i++) {
     node = array_fetch(node_t *, nodevec, i);
     (void) st_insert(node2num_table, node_long_name(node), (char *)i);
  }

  for (i = 0; i < size_net; i++) {
        node = array_fetch(node_t *, nodevec, i);  
        if (node->type != INTERNAL) continue;
	if (node_num_fanin(node) > FANIN_COLLAPSE) continue;
        if (is_anyfo_PO(node)) {
	   continue; 
        }
        name = node_long_name(node);
        dup_net = network_dup(network);
        dup_node = network_find_node(dup_net, name);
 	if (dup_node == NULL) {
	    (void) printf("Error:dup_node not there %s\n", name);
	    exit(1);
	}
	name1 = node_long_name(node);
        if (!st_lookup(cost_table, name1, &dummy)) {
	   (void) printf("%s cost_struct not found in cost_table\n", 
		   name1);
	   exit(1);
        }

	/* compute the gain if the node "dup_node" is collapsed into each of its 
	   fanouts and is then removed from the network. The cost is stored in the
	   focost for each of the fanouts of the node "dup_node" */
	/*------------------------------------------------------------------------*/
        gain =  ((COST_STRUCT *)dummy)->cost;
	num_fo = 0;
	clust_num_add1 = clust_num + 1;

        /* for each fanout, recompute the cost */
 	/*-------------------------------------*/
        foreach_fanout(dup_node, genfo, fanout) {
           (void) node_collapse(fanout, dup_node);
	   /* fanout_simpl = node_simplify(fanout, (node_t *) NULL, NODE_SIM_ESPRESSO);
	   node_replace(fanout, fanout_simpl); */
	   focost[clust_num_add1][num_fo] = act_evaluate_map_cost(fanout, HEURISTIC_NUM);
	   focost_ind[clust_num_add1][num_fo] = 1;
	   /* free the act structure of the fanout node, it has been copied into the focost */
 	   /*-------------------------------------------------------------------------------*/
	   if (!st_lookup(cost_table, node_long_name(fanout), &dummy)) {
		(void) printf(" node %s not found in cost_table\n", 
			 node_long_name(fanout));
		exit(1);
	   }
	   old_cost = ((COST_STRUCT *)dummy)->cost;
	   gain = gain + old_cost - (focost[clust_num_add1][num_fo]->cost);
	   num_fo++;
        }

	if (gain > 0) {
	   clust_num++;

           /* assign proper value to g and M for the cluster "clust_num" */
	   /*------------------------------------------------------------*/
	   collapsed_node[clust_num] = i;
           g[clust_num] = gain;
	   M[i][clust_num] = 1; /* node (ie. with id i) in clust_num */
           num_ones[i]++;
           foreach_fanout(node, genfo, fanout) {
              if (!st_lookup(node2num_table, node_long_name(fanout), &dummy)) {
		(void) printf(" node %s not in node2num_table\n", node_long_name(fanout));
	 	exit(1);
	      }
  	      M[(int) dummy][clust_num] = 1;
	      num_ones[(int)dummy]++;
           }  	     
        } /* if (gain > 0) */

        /* if gain <= 0, free the storage in focost associated with node "dup_node" */

	else {
 	   for (j = 0; j < num_fo; j++) {
	      (void) free_cost_struct( focost[clust_num_add1][j]);
	      focost_ind[clust_num_add1][j] = 0;
	   }
	   if (!st_lookup(cost_table, name1, &dummy)) {
		(void) printf( "gain <= 0 =>%s not in cost_table\n", name1);
		exit(1);
	   }

        }
       /* free the duplicated node and the network */
       /*-------------------------------------------*/
        network_delete_node(dup_net, dup_node);
	network_free(dup_net);

  } /* for i = .. */
  
   /* if no cluster (ie with gain > 0) can be formed check here */
   /*-----------------------------------------------------------*/
   if (clust_num < 0) {
	    if (ACT_DEBUG)
	       (void) printf("***no improvement in collapse***\n");
	    for(i=0; i< numnodes; i++){
		for(j=0; j< numnodes; j++){
		    if (focost_ind[i][j] == 1) free_cost_struct( focost[i][j]);
		}
		FREE( focost[i]);
                FREE( focost_ind[i]);
	       	FREE( M[i]);
	    }
	    FREE( focost);
	    FREE( focost_ind);
	    FREE( variable);
	    FREE( collapsed_node);
	    FREE( g);
	    FREE( num_ones);
	    FREE(M);
	    array_free(nodevec);
	    st_free_table(node2num_table);
        if (iteration_gain < (GAIN_FACTOR * cost_network)) {
	    if (ACT_DEBUG) {
		(void) printf("****no significant improvement in iteration_gain***\n");
	        (void) printf("cost_network before iter. = %d, iteration_gain = %d\n",
		               cost_network, iteration_gain);
            }
	    return network;
        }
	nogain = FALSE;
  	continue;
   } /* if clust_num < 0 */

  /* else formulate the IP - using LINDO  */
  /*---------------------------------*/

  infile = formulate_IP(nodevec, M, g, num_ones, clust_num);
  outfile = ALLOC(char, 100);
  (void) sprintf (outfile, "/usr/tmp/actel.lindo.out%d.XXXXXX", iteration_num);
  (void) mktemp(outfile);
  (void) sprintf(buf, "lindo < %s > %s", infile, outfile);

  (void) system(buf);

  /* read the solution and change the  network */
  /*-------------------------------------------*/
   read_LINDO_file(outfile, variable, &collapse_gain);
   
   FREE(infile);
   FREE(outfile);
   iteration_gain = collapse_gain + decomp_gain;

  /* update the cost of affected nodes in the new network */
  /*------------------------------------------------------*/
  for (j = 0; j <= clust_num; j++) {
    if (variable[j] == 0) continue;
    node = array_fetch(node_t *, nodevec, collapsed_node[j]);
    num_fo = 0;
    foreach_fanout(node, genfo, fanout) {
       (void) node_collapse(fanout, node);
       /* fanout_simpl = node_simplify(fanout, (node_t *) NULL, NODE_SIM_ESPRESSO);
       node_replace(fanout, fanout_simpl); */
       name1 = node_long_name(fanout);
       name = ALLOC(char, strlen(name1) + 1);  /* name needed */
       (void) strcpy(name, name1);
       if (!st_delete(cost_table, &name1, &dummy)) {
	  (void) printf(" node %s can't be deleted - it is not in cost_table\n",
		   name);
	  exit(1);
       }
       (void) free_cost_struct((COST_STRUCT *) dummy);
       FREE(name1);

       /* checking */
       /*----------*/
       if (node_long_name(node) == NULL) {
	   (void) printf("Error:improve_ names---\n");
	    exit(1);
       }

       /* this focost is in the cost_table, unavailable for free.
	  Therefore set the indicator to 0. */
       /*-------------------------------------------------------*/
       focost_ind[j][num_fo] = 0;
       if (st_insert(cost_table, name,
		 (char *) focost[j][num_fo])) {
	  (void) printf("Error: while updating %s was already there\n", name);
	  exit(1);
       }
       num_fo++;
    } /* foreach_fanout */

    /* delete the collapsed node */  
    /*---------------------------*/
    name1 = node_long_name(node);
    /* delete the act storage from the struct */
    /*----------------------------------------*/
    if (!st_lookup(cost_table, name1, &dummy)) {
	(void) printf(" Error:%s could not be found (before deleting)\n", name1);
	exit(1);
    }
    if (!st_delete(cost_table, &name1,  &dummy)) {
	(void) printf("Error: %s could not be deleted\n", name1);
	exit(1);
    }
    (void) free_cost_struct((COST_STRUCT *) dummy);
    FREE(name1);

    /* checking */
    /*----------*/
    if (node_long_name(node) == NULL) {
	(void) printf("Error:improve_ names---\n");
	 exit(1);
    }

    network_delete_node(network, node); 
    
  } /* for j = loop */
  

 /* free storage */
 /*--------------*/
 for (i = 0; i < numnodes; i++) {
   for (j = 0; j < numnodes; j++) 
      if (focost_ind[i][j] == 1) free_cost_struct( focost[i][j]);
   FREE( focost[i]);
   FREE( focost_ind[i]);
   FREE( M[i]);
 }
 FREE( focost);
 FREE( focost_ind);
 FREE( M);
 FREE( variable);
 FREE( collapsed_node);
 FREE( g);
 FREE( num_ones);
 array_free(nodevec);
 st_free_table(node2num_table);
 if (iteration_gain >= (GAIN_FACTOR * cost_network)) nogain = FALSE;
 if (ACT_DEBUG)  
    (void) print_network(network, cost_table);

 } /* while (nogain == FALSE) */

 return network;
}

/********************************************************************************
  Generates a file for the program lindo and solves the integer program,
  The solution of the integer program generates nodes that should be collapsed.
*********************************************************************************/
static char *
formulate_IP(nodevec, M, g, num_ones, clust_num)
  array_t *nodevec;
  int **M;
  int g[];
  int num_ones[];
  int clust_num;
{

  int i, j;
  int num_constr;
  FILE *lindoFile;
  int FLAG;
  int termsPerLine = 5;
  int numTerm;
  char *filename;
  int net_size;

  if (ACT_DEBUG) (void) printf("formulating IP - creating lindo file *****\n");
  filename = ALLOC(char, 100);
  (void) sprintf(filename, "/usr/tmp/actel.lindo.in%d.XXXXXX", iteration_num);
  (void) mktemp(filename);
  lindoFile = fopen(filename, "w");

  /* generate objective function : MAX g[0]x0 + .... g[clust_num]xclust_num */
  /*------------------------------------------------------------------------*/
   (void) fprintf(lindoFile, "MAX ");
   for (numTerm = 0, i = 0; i <= clust_num; ++numTerm, ++i) {
     if (i == 0) (void) fprintf (lindoFile, "%d x0 ", g[0]);
     else {
        if (numTerm < termsPerLine) 
	   (void) fprintf(lindoFile, "+ %d x%d ", g[i], i);
        else {
	   numTerm = 0;
           (void) fprintf(lindoFile, "\n + %d x%d ", g[i], i);
        }
     }
   }
   (void) fprintf(lindoFile, "\n");

   /* Generate constraints for every node in the network, only if at least
      two entries for the corresponding row is 1(as told by num_ones.
      constraint for a node i is of the form mi,1 x1 + mi,2 x2 + .... <= 1 */
   /*----------------------------------------------------------------------*/
   net_size = array_n(nodevec);
   num_constr = 0;
   for (i = 0; i < net_size; i++) {
   /* if num_ones = 1, obviously constraint would be satisfied - so do
      not put the constraint at all */
   /*------------------------------------------------------------------*/
     if (num_ones[i] < 2) continue;
     num_constr++;
     if (num_constr == 1) (void) fprintf(lindoFile, "ST\n");
     
     numTerm = 0;
     FLAG = FALSE;
     for (j = 0; j <= clust_num; j++) {
       if (M[i][j] == 0) continue;
       if (FLAG == FALSE) {
	 FLAG = TRUE;
	 ++numTerm;
	 (void) fprintf(lindoFile, "x%d ", j);
         continue;
       }
       if (numTerm < termsPerLine) {
          ++numTerm;
          (void) fprintf(lindoFile, "+ x%d ", j);
          continue;
       }
       numTerm = 1;
       (void) fprintf(lindoFile, "\n + x%d ", j);
     }
     (void) fprintf(lindoFile, "<= 1\n");
  } /* for i = .. */ 
  (void) fprintf(lindoFile, "END\n");

  /* declaration of integrality - integrality would confirm 0-1 IP for lindo */
  /*-------------------------------------------------------------------------*/

  for ( j = 0; j <= clust_num; j++)
    (void) fprintf(lindoFile, "INTEGER x%d\n", j);

  (void) fprintf(lindoFile, "GO\n");
  (void) fprintf(lindoFile, "QUIT\n");
  (void) fclose(lindoFile);
  return filename;
}

/***************************************************************************
  reads the solution generated by the lindo from the file. Sets the 
  corresponding "variable" values 1. Also reads the gain that resulted 
  from the collapse operation for the network, stores it in iteration_gain.
    -"IP OPTIMUM" means that optimum solution was found.
    -"NO FEASIBLE" indicates that no feasible solution exists.
         (This normally should not be the case).
    -If none of the above two found, then the only other possibility is 
      "NEW INTEGER", which means optimal answer could not be found, 
      but some answer was found. So live with that, buddy. For this,
      however, the file is scanned again.
****************************************************************************/
static void
read_LINDO_file(filename, variable, iteration_gain)
 char *filename;
 int *variable;
 int *iteration_gain;

{
  FILE *lindoFile;
  float f_gain;
  int variable_id;
  float result;
  char word[BUFSIZE];
  int FLAG;

  FLAG = 0;



  lindoFile = fopen(filename, "r");
  while(TRUE) {  
    if (fscanf(lindoFile, "%s", word) == EOF) {
      if (FLAG == 1) {
	 FLAG = 2;
	 break;
      }
      else {
         (void) printf(" sudden end of file lindo.out\n");
         exit(1);
      }
    }

    if (strcmp(word, "NO") == 0) {
       (void) fscanf(lindoFile, "%s", word);
       if (strcmp(word, "FEASIBLE") == 0) {
         (void) printf(" NO FEASIBLE SOLUTION OBTAINED\n");
  	 exit(1);
       }
       (void) printf("Error:NO %s in lindo.out\n", word);
       exit(1);
    }
    if (strcmp(word, "NEW") == 0) {
       (void) fscanf(lindoFile, "%s", word);
       if (strcmp(word, "INTEGER") == 0) {
	  FLAG = 1;
       }
       else {
          (void) printf("Error:NEW %s in lindo.out\n", word);
          exit(1);
       }	
       continue;
    }
   /* searching for "IP OPTIMUM" - to get the value of the objective function*/
   /*------------------------------------------------------------------------*/
    if (strcmp(word, "IP") == 0) {
	(void) fscanf(lindoFile, "%s", word);
	if (strcmp(word, "OPTIMUM") == 0) 
	  break;
        else {
	   (void) printf("Error:unaccounted apperance of IP in lindo file\n");
	   exit(1);
	}
    }
  } /* while */

  /* Optimum solution was not found, but do with the best possible */
  /*---------------------------------------------------------------*/
  if (FLAG == 2) {
     (void) fclose(lindoFile);
     lindoFile =  fopen(filename, "r");
     partial_scan_lindoFile(lindoFile);
  }
  
  /* iteration_gain = value of the objective function */
  /*--------------------------------------------------*/
  while (TRUE) {
       (void) fscanf(lindoFile, "%s", word);
       if (strcmp(word, "1)") == 0) {
	  (void) fscanf(lindoFile, "%f", &f_gain);
	  *iteration_gain = (int) f_gain;
	  if (ACT_DEBUG) (void) printf("total gain in IP = %d\n", *iteration_gain);
          break;
       }
  }
  /* "REDUCED COST" is a precursor of the solution */    
  /*-----------------------------------------------*/
  while (TRUE) {
      (void) fscanf(lindoFile, "%s", word);
      if (strcmp(word, "REDUCED") == 0) {
         (void) fscanf(lindoFile, "%s", word);
         break;
      }
  }  
    
  /* now get the result */
  /*--------------------*/  
  for(; fscanf(lindoFile, "%s", word) != EOF; ) {

      /* if "ROW" appears solution set section ends */       
      /*--------------------------------------------*/
     if (strcmp(word, "ROW") == 0) break;
     if (word[0] == 'X') {   
        (void) sscanf(word, "X%d", &variable_id);
        (void) fscanf(lindoFile, "%f", &result);
        variable[variable_id] = (int) result;
     }
 }
 (void) fclose(lindoFile);
}

/***********************************************************************************
  now it is known that the solution is not Optimum; so search for the "NEW INTEGER"
  in the file. As soon as it is found(that is, the first instance) return.
*************************************************************************************/
static void
partial_scan_lindoFile(lindoFile)
  FILE 	*lindoFile;
{
  char word[BUFSIZE];

  while(TRUE) {  
    (void) fscanf(lindoFile, "%s", word);
    if (strcmp(word, "NEW") == 0) {
       (void) fscanf(lindoFile, "%s", word);
       if (strcmp(word, "INTEGER") == 0)  break;
    }
  } /* while */

}




/*************************************************************************
  return 1 if any fanout of the node is a Primary output. Else return 0.
*************************************************************************/
static
int is_anyfo_PO(node)
  node_t *node;
{
 lsGen genfo;
 node_t *fanout;
 node_function_t node_fun;

 foreach_fanout(node, genfo, fanout) {
   node_fun = node_function(fanout);
   if (node_fun == NODE_PO){
       (void) lsFinish(genfo);
       return  1;
   }
 }
 return 0;
}

int print_network(network, cost_table)
  network_t *network;
  st_table *cost_table;
{
  lsGen gen;
  node_t *node;
  char *dummy;
  int TOTAL_COST;
  char *name;

  TOTAL_COST = 0;
  if (ACT_DEBUG) (void) printf("printing network %s\n", network_name(network));
  foreach_node(network, gen, node) {
       if (node->type != INTERNAL) continue;
       if (ACT_DEBUG) node_print(stdout, node); 
       name =  node_long_name(node);

       if (!st_is_member(cost_table, name)) {
	 (void) printf("%s not there in the cost_table\n", name);
	 exit(1);
       }

       if (!st_lookup(cost_table, name, &dummy)) {
		(void) printf( "%s not in cost_table\n", name);
		exit(1);
       }
       if (ACT_DEBUG) (void) printf("cost = %d\n\n", ((COST_STRUCT *)dummy)->cost); 
       TOTAL_COST += ((COST_STRUCT *) dummy)->cost;
   }

   if (ACT_DEBUG)  (void) printf("**** total cost of network after iteration %d is %d\n***", 
	    iteration_num, TOTAL_COST);
    return TOTAL_COST;
}
  

/*************************************************************************
  Decompose nodes of the network. If decomposition results in a 
  lesser cost node, accept the decomposition, and replace the node by
  the decomposition.
  think about how to decompose the nodes in the network. Should I use the 
  mux structure and do a resub. this is because it possibly means that the
  decomposition would be in terms of a mux structure => Boolean div not 
  implemented in misII.
**************************************************************************/

static
network_t *
decomp_big_nodes(network, cost_table, gain, DECOMP_FANIN, HEURISTIC_NUM)
  network_t *network;
  st_table *cost_table;
  int      *gain;
  int 	   DECOMP_FANIN;
  int 	    HEURISTIC_NUM;
{
  network_t	*dup_net, *save_network;
  array_t 	*nodevec;
  int 		size_net;
  int 		i, j;
  node_t 	*node, *decomp_node, *dup_node, *first_node;
  COST_STRUCT   *decomp_node_cost_struct;
  array_t	*decomp_node_array, *cost_array;
  char		*nodename;     
  int		num_fanin;
  int 		new_cost, old_cost;
  char 		*dummy, *decomp_node_name1, *decomp_node_name;
  int		size_decomp_array;
  int 		FLAG; 		/* = ON if anytime gain is there */


  save_network = network;  /*save the network */
  FLAG = OFF;
  *gain = 0;

  nodevec = network_dfs(network);
  size_net = array_n(nodevec);
  for (i = 0; i < size_net; i++) {

    node = array_fetch(node_t *, nodevec, i);
    if (node->type != INTERNAL) continue;

    /* if the num_fanin is less than the limit, continue */
    /*---------------------------------------------------*/
    num_fanin = node_num_fanin(node);
    if (num_fanin < DECOMP_FANIN) continue;

     /*  if the node requires upto 2 muxes, do not split it.
         Reason being that if split, it would form 2 nodes
         (at least) and hence the cost would be >= 2 always */
     /*-----------------------------------------------------*/
    if (cost_of_node(node, cost_table) <= 2) continue;

    /* if constructed optimally, do not decompose */
    /*--------------------------------------------*/
    if ((node_num_cube(node) == 1) || 
        (node_num_literal(node) == node_num_fanin(node))
       ) continue;

    nodename = node_long_name(node);
    dup_net = network_dup(network);
    dup_node = network_find_node(dup_net, nodename);
    if (dup_node == NULL) {
       (void) printf("decomp_big_nodes(): NULL node found\n");
       exit(1);
    }

    /*  do a good decomposition on the node */
    /*--------------------------------------*/
    decomp_node_array = decomp_good(dup_node);
    size_decomp_array = array_n(decomp_node_array);

    /*  if no decomposition was found, continue */
    /*------------------------------------------*/
    if (size_decomp_array == 1) {
       node_free(array_fetch(node_t *, decomp_node_array, 0));
       array_free(decomp_node_array);
       network_free(dup_net);
       continue;
    }
    
   /*  insert the nodes formed after decomposition in the duplicate network */
   /*-----------------------------------------------------------------------*/
    for (j = 1; j < size_decomp_array; j++) {
         decomp_node = array_fetch(node_t *, decomp_node_array, j);
         network_add_node(dup_net, decomp_node);
    } /* for j = loop */
    first_node = array_fetch(node_t *, decomp_node_array, 0);
    node_replace(dup_node, first_node);

    /* because the earlier one was freed by replace */
    /*----------------------------------------------*/
    array_insert(node_t *, decomp_node_array, 0, dup_node);

    /*  compute the cost of the added nodes and compare with the old cost*/
    /*-------------------------------------------------------------------*/
    cost_array = array_alloc(COST_STRUCT *, 0);
    new_cost = compute_cost_array(decomp_node_array, dup_net, cost_array, HEURISTIC_NUM);
    if (!st_lookup(cost_table, nodename, &dummy)) {
	(void) printf("decomp_big_nodes(): node %s(%s) not in cost_table\n",
		nodename, node_name(node));
	exit(1);
    }
    old_cost = ((COST_STRUCT *)dummy)->cost;

    /*  if the new cost is less, then store  the costs in the cost_table */
    /*-------------------------------------------------------------------*/
    if (new_cost < old_cost) {
       *gain = *gain + (old_cost - new_cost);
       for (j = 1; j < size_decomp_array; j++) {
         decomp_node = array_fetch(node_t *, decomp_node_array, j);
         decomp_node_cost_struct = array_fetch(COST_STRUCT *, cost_array, j);
   	 decomp_node_name1 = node_long_name(decomp_node);
	 decomp_node_name = ALLOC(char, strlen(decomp_node_name1) + 1);
	 (void) strcpy(decomp_node_name, decomp_node_name1);
         if (st_insert(cost_table, decomp_node_name,
		 (char *) decomp_node_cost_struct)) {
	    (void) printf(" decomp_big_node():while inserting %s was already there\n", 
		   decomp_node_name1);
	    exit(1);
         }
       } /* for j = loop */
       cost_update(dup_node, cost_table, 
		   array_fetch(COST_STRUCT *, cost_array, 0));
       
        /* FLAG keeps track of if it is the first time network is being replaced.
 	   if yes(OFF), then do not free the network(as it is the original network) */
	/*--------------------------------------------------------------------------*/
	if (FLAG == OFF) {
		FLAG = ON;
	        network = dup_net;
        }
        else {
		network_free(network);
	        network = dup_net;
	}

        /* debugging */
 	/*-----------*/

	if (!network_check(network)) {
	   (void) printf("before_free:decompose_big_nodes():%s\n", error_string());
	   exit(1);
        }
    } /* if new_cost < ... */        

    /* if new_cost is greater, then just retain the original network, free the 
       duplicate one */
    /*------------------------------------------------------------------------*/
    else {
       for (j = 0; j < size_decomp_array; j++) {
  	  free_cost_struct( array_fetch(COST_STRUCT *, cost_array, j) );
       }
       network_free(dup_net);
       if (!network_check(network)) {
	   (void) printf("decompose_big_nodes(1):%s\n", error_string());
	   exit(1);
        }

    }
    array_free(decomp_node_array); 
    array_free(cost_array);
  
  } /* for i = loop*/    

  /* if FLAG = ON, then free the storage associated with the first network */
  /*-----------------------------------------------------------------------*/
  if (FLAG == ON) 
    network_free(save_network);

  array_free(nodevec);
  return network;

}

/*************************************************************************
  Compute the costs of the nodes in the node_array and store them in the 
  cost_array. node_array actually has nodes from dup_net. These are the same 
  nodes that were obtained after the decomposition of a node.
***************************************************************************/

 static
 int
 compute_cost_array(node_array, dup_net, cost_array, HEURISTIC_NUM)
   array_t 	*node_array;
   network_t  	*dup_net;
   array_t	*cost_array;
   int		HEURISTIC_NUM;
{
   node_t 	*node, *n;
   int 		TOTAL_COST;
   int 		this_cost;
   int 		check_cost;
   COST_STRUCT  *cost_node, *check_cost_struct;
   int 		size_array;
   int 		i;

   TOTAL_COST = 0;
   size_array = array_n(node_array);
   for (i = 0; i < size_array; i++) {
     node = array_fetch(node_t *, node_array, i);
     n    = network_find_node(dup_net, node_long_name(node));
     cost_node = act_evaluate_map_cost(n, HEURISTIC_NUM);
     array_insert_last(COST_STRUCT *, cost_array, cost_node);
     this_cost = cost_node->cost;
     check_cost_struct = array_fetch(COST_STRUCT *, cost_array, i);
     check_cost = check_cost_struct->cost;
     if (this_cost != check_cost) {
        (void) printf(" compute_cost_array():this_cost (%d) != check_cost(%d)\n",
		 this_cost, check_cost);
 	exit(1);
     }
     TOTAL_COST += this_cost;
   }

   /* for debugging */
   /*---------------*/

   i = array_n(cost_array);
   if (i != size_array) {
      (void) printf("compute_cost_array():strange!!!!!\n");
      exit(1);
   }

   return TOTAL_COST;
}

/***************************************************************************
  given the node, finds the cost of the node from the cost table.
***************************************************************************/

  static
  int 
  cost_of_node(node, cost_table)
     node_t 	*node;
     st_table   *cost_table;
{
  char	*dummy;

  if (!st_lookup(cost_table, node_long_name(node), &dummy)) {
     (void) printf("node %s not found in the cost table\n", node_long_name(node));
     exit(1);
  }
  return ((COST_STRUCT *) dummy)->cost;

}
  
/****************************************************************************
   given the node "node", and its new cost in the cost_struct, delete the 
   original cost in the cost_table and insert the new one 
****************************************************************************/

 static void
 cost_update(node, cost_table, cost_struct)
    node_t	 *node;
    st_table	 *cost_table;
    COST_STRUCT  *cost_struct;
{
    char        *nodename, *name_node;
    char   	*dummy;

    nodename = node_long_name(node);
    name_node = ALLOC(char, strlen(nodename) + 1);
    (void) strcpy(name_node, nodename);
    if (!st_delete(cost_table, &nodename,  &dummy)) {
	(void) printf(" %s could not be deleted\n", nodename);
	exit(1);
    }
    free_cost_struct((COST_STRUCT *) dummy);
    FREE(nodename);
    (void) st_insert(cost_table, name_node, (char *) cost_struct);
}

/***********************************************************************
   This procedure generates act ordering st. a function which has all the 
   cubes of single literal, is implemented with minimum no. of mux 
   structures. First separates the NEG and POS phase fanins. Rest is 
   complicated!!!!
  
*************************************************************************/

static
array_t *
OR_literal_order(node)
  node_t *node;

{

  array_t 	*order_list, *order_rev;
  array_t	*pos_list, *neg_list;
  int 		size_pos, size_neg;
  int		FLAG_POS, FLAG_NEG;
  int 		pointer_pos, pointer_neg;
  int 		num_fanin;
  node_t 	*fanin, *n;
  input_phase_t phase;
  int  		i, j, stage, position;

  num_fanin = node_num_fanin(node);
  order_list = array_alloc(node_t *, 0);
  order_rev = array_alloc(node_t *, 0);
  pos_list = array_alloc(node_t *, 0);
  neg_list = array_alloc(node_t *, 0);

  /* separate the fanins with positive phase and negative phase */
  /*------------------------------------------------------------*/
  for (i = num_fanin - 1; i >= 0; i--) {
    fanin = node_get_fanin(node, i);
    phase = node_input_phase(node, fanin);
    if (phase == NEG_UNATE) 
       array_insert_last(node_t *, neg_list, fanin);
    else if (phase == POS_UNATE)
       array_insert_last(node_t *, pos_list, fanin);  	
  }

  size_pos = array_n(pos_list);
  size_neg = array_n(neg_list);

  /* Get the optimal ordering now for STRUCT module - look at the paper for details*/
  /*-------------------------------------------------------------------------------*/

  pointer_neg = 0;
  pointer_pos = 0;
  FLAG_NEG = OFF;
  FLAG_POS = OFF;
  stage = 0;

  for ( i = 0; i < num_fanin; i++) {
    switch (stage) {
      case 1 : 
	  if (pointer_neg >= size_neg) {
	     FLAG_NEG = ON;
	     break;
          }
	  n = array_fetch(node_t *, neg_list, pointer_neg);
          array_insert_last(node_t *, order_rev, n);
	  pointer_neg++;
          break;
      case 0:;
      case 2:;
      case 3: 
	  if (pointer_pos >= size_pos) {
	     FLAG_POS = ON;
	     break;
          }
	  n = array_fetch(node_t *, pos_list, pointer_pos);
          array_insert_last(node_t *, order_rev, n);
  	  pointer_pos++;
          break;      
    }  /* switch */

   /* if all neg phase fanins finished, put positive now */
   /*----------------------------------------------------*/
    if (FLAG_NEG == ON) {
	for (j = i; j < num_fanin; j++) {
          if (pointer_pos >= size_pos) {
		(void) printf(" Error: pointer_pos = %d??\n", pointer_pos);
		exit(1);
 	  }
	  n = array_fetch(node_t *, pos_list, pointer_pos);
          array_insert_last(node_t *, order_rev, n);
          pointer_pos++;
        }
        break; /* from i loop */
    } 

   /* if all  positive phased fanins finished, put negative now */
   /*-----------------------------------------------------------*/
    if (FLAG_POS == ON) {
	for (j = i; j < num_fanin; j++) {
	  if (pointer_neg >= size_neg) {
		(void) printf("Error: pointer_neg = %d??\n", pointer_neg);
		exit(1);
 	  }
	  n = array_fetch(node_t *, neg_list, pointer_neg);
          array_insert_last(node_t *, order_rev, n);
	  pointer_neg++;
        }
        break; /* from i loop */
    }	   

   /* for first mux structure, there are 4 inputs that can be assigned,
      for the rest, just 3, because the rest of them have as their inputs 
      the outputs from the previous muxes */
    /*------------------------------------------------------------------*/
    if ((stage % 3) == 0) stage = 1;
    else stage = stage + 1;

  } /* for */    

 /*  order_rev has everything in reverse order,as it was constructed 
    bottom_up. Reverse it to get order_list */
 /*------------------------------------------------------------------*/
  for ( i = 0; i < num_fanin; i++) {
    position = num_fanin - ( i + 1);
    n = array_fetch(node_t *, order_rev, position);
    array_insert_last(node_t *, order_list, n);

  }

  /* check */
  /*-------*/
  if (array_n(order_list) != num_fanin) {
    (void) printf("single_cube_order():something is binate for node %s\n", 
	    node_long_name(node));
    exit(1);
  }

  /* free the lists */
  /*----------------*/
  array_free(order_rev);
  array_free(pos_list);
  array_free(neg_list);

  return order_list;

}

/***********************************************************************
   This procedure generates act ordering st. a function which has a
   single cube  is implemented with minimum no. of mux 
   structures. First separates the NEG and POS phase fanins. Rest is 
   complicated!!!!
  
*************************************************************************/

static
array_t *
single_cube_order(node)
  node_t *node;

{

  array_t 	*order_list, *order_rev;
  array_t	*pos_list, *neg_list;
  int 		size_pos, size_neg;
  int		FLAG_POS, FLAG_NEG;
  int 		pointer_pos, pointer_neg;
  int 		num_fanin;
  node_t 	*fanin, *n;
  input_phase_t phase;
  int  		i, j, stage, position;

  num_fanin = node_num_fanin(node);
  order_list = array_alloc(node_t *, 0);
  order_rev = array_alloc(node_t *, 0);
  pos_list = array_alloc(node_t *, 0);
  neg_list = array_alloc(node_t *, 0);

  for (i = num_fanin - 1; i >= 0; i--) {
    fanin = node_get_fanin(node, i);
    phase = node_input_phase(node, fanin);
    if (phase == NEG_UNATE) 
       array_insert_last(node_t *, neg_list, fanin);
    else if (phase == POS_UNATE)
       array_insert_last(node_t *, pos_list, fanin);  	
  }

  size_pos = array_n(pos_list);
  size_neg = array_n(neg_list);
  pointer_neg = 0;
  pointer_pos = 0;
  FLAG_NEG = OFF;
  FLAG_POS = OFF;
  stage = 0;

  for ( i = 0; i < num_fanin; i++) {
    switch (stage) {
      case 2 :
      case 3 : 
	  if (pointer_neg >= size_neg) {
	     FLAG_NEG = ON;
	     break;
          }
	  n = array_fetch(node_t *, neg_list, pointer_neg);
          array_insert_last(node_t *, order_rev, n);
	  pointer_neg++;
          break;
      case 0:;
      case 1:
	  if (pointer_pos >= size_pos) {
	     FLAG_POS = ON;
	     break;
          }
	  n = array_fetch(node_t *, pos_list, pointer_pos);
          array_insert_last(node_t *, order_rev, n);
  	  pointer_pos++;
          break;      
    }  /* switch */

    /* if all neg phase fanins finished, put positive now */
    /*----------------------------------------------------*/
    if (FLAG_NEG == ON) {
	for (j = i; j < num_fanin; j++) {
          if (pointer_pos >= size_pos) {
		(void) printf(" pointer_pos = %d??\n", pointer_pos);
		exit(1);
 	  }
	  n = array_fetch(node_t *, pos_list, pointer_pos);
          array_insert_last(node_t *, order_rev, n);
          pointer_pos++;
        }
        break; /* from i loop */
    } 

    /* if all  positive phased fanins finished, put negative now */
    /*-----------------------------------------------------------*/
    if (FLAG_POS == ON) {
	for (j = i; j < num_fanin; j++) {
	  if (pointer_neg >= size_neg) {
		(void) printf(" pointer_neg = %d??\n", pointer_neg);
		exit(1);
 	  }
	  n = array_fetch(node_t *, neg_list, pointer_neg);
          array_insert_last(node_t *, order_rev, n);
	  pointer_neg++;
        }
        break; /* from i loop */
    }	   

   /* for first mux structure, there are 4 inputs that can be assigned,
      for the rest, just 3  */
    /*----------------------------------------------------------------*/
    if ((stage % 3) == 0) stage = 1;
    else stage = stage + 1;

  } /* for */    

 /*  order_rev has everything in reverse order,as it was constructed 
    bottom_up. Reverse it to get order_list */
  /*-----------------------------------------------------------------*/
  for ( i = 0; i < num_fanin; i++) {
    position = num_fanin - ( i + 1);
    n = array_fetch(node_t *, order_rev, position);
    array_insert_last(node_t *, order_list, n);

  }

  /* check */
  /*-------*/
  if (array_n(order_list) != num_fanin) {
    (void) printf("single_cube_order():something is binate for node %s\n", 
	    node_long_name(node));
    exit(1);
  }

  /* free the lists */
  /*----------------*/
  array_free(order_rev);
  array_free(pos_list);
  array_free(neg_list);

  return order_list;

}

/****************************************************************************************
  Checks if it is beneficial to implement the node in negative phase. "Beneficial" is
  gotten by recomputing the cost of the node and the fanouts of the node. Greedily selects 
  the node, if it improves the cost. 
******************************************************************************************/
static
network_t *
act_quick_phase(network, cost_table, HEURISTIC_NUM)
   network_t 	*network;
   st_table	*cost_table;
   int 		HEURISTIC_NUM;
{

   array_t	*cost_array, *nodevec;
   COST_STRUCT  *cost_struc;
   int 		i, j;
   int 		size, TOTAL_GAIN, gain, new_cost, old_cost;
   lsGen	gen;
   node_t	*node, *fanout;  
   char 	*name, *dummy;

   if (ACT_DEBUG) (void) printf("-------entering act_quick_phase--------\n");
   nodevec = network_dfs(network);
   size = array_n(nodevec);

   TOTAL_GAIN = 0;
   for (i = 0; i < size; i++) {
    
     node = array_fetch(node_t *, nodevec, i);
     if (ACT_DEBUG) node_print(stdout, node);

     /* if node cannot be inverted, get the next node */
     /*-----------------------------------------------*/
     if (node->type != INTERNAL) continue;
     if (is_anyfo_PO(node)) continue;

     /* invert the node, if possible */
     /*------------------------------*/
     if (!node_invert(node)) {
	continue;
     }

     /* checking */
     /*----------*/
     if (!st_lookup(cost_table, node_long_name(node), &dummy)) {
	(void) printf("act_quick:phase: node %s not found in cost_table\n",
		node_long_name(node));
        exit(1);
     }

     /* evaluate the cost of the inverted node, store in cost_array */
     /*-------------------------------------------------------------*/
     gain = 0;
     cost_array = array_alloc(COST_STRUCT *, 0);

     old_cost = ((COST_STRUCT *)dummy)->cost;
     cost_struc = act_evaluate_map_cost(node, HEURISTIC_NUM);
     array_insert_last(COST_STRUCT *, cost_array, cost_struc);
     new_cost = cost_struc->cost;
     gain = old_cost - new_cost;

     /* evaluate the cost of all the fanouts of the node */
     /*--------------------------------------------------*/
     if (ACT_DEBUG) (void) printf(" num_fo = %d\n", node_num_fanout(node));
     foreach_fanout(node, gen, fanout) {
	if (!st_lookup(cost_table, node_long_name(fanout), &dummy)) {
   	   (void) printf("act_quick:phase: node %s not found in cost_table\n",
		   node_long_name(fanout));
           exit(1);
        }
	old_cost =  ((COST_STRUCT *)dummy)->cost;
	cost_struc = act_evaluate_map_cost(fanout, HEURISTIC_NUM);
        array_insert_last(COST_STRUCT *, cost_array, cost_struc);
	new_cost = cost_struc->cost;
	gain = gain + old_cost - new_cost; 
     }

     if (ACT_DEBUG) 
	(void) printf(" gain from node %s = %d\n", node_long_name(node), gain);

     /* if it is beneficial to invert the node, change the corresponding entries 
	of the cost_table */
     /*-------------------------------------------------------------------------*/
     if (gain > 0) {
	TOTAL_GAIN += gain;
	name = node_long_name(node);
        if (!st_delete(cost_table, &name, &dummy)) {
	   (void) printf("act_quick_phase:%s could not be deleted from cost_table\n",
		   name);
	   exit(1);
        }
	free_cost_struct((COST_STRUCT *) dummy);
	cost_struc = array_fetch(COST_STRUCT *, cost_array, 0);
	(void) st_insert(cost_table, name, (char *)cost_struc);
	j = 1;
	foreach_fanout(node, gen, fanout) {
  	   name = node_long_name(fanout);
           if (!st_delete(cost_table, &name, &dummy)) {
  	      (void) printf("act_quick_phase:%s could not be deleted from cost_table\n",
		      name);
	      exit(1);
           }
	   free_cost_struct((COST_STRUCT *) dummy);
	   cost_struc = array_fetch(COST_STRUCT *, cost_array, j);
	   j++;
	   (void) st_insert(cost_table, name, (char *)cost_struc);
        }
     }  /* if gain > 0 */

     /* if it is not beneficial, invert the node back, since earlier
	inversion was done in place */
     /*-------------------------------------------------------------*/
     else {
	/* if gain <= 0, then do not change the network. Restore it first */
	/*----------------------------------------------------------------*/
	(void) node_invert(node);

        for (j = array_n(cost_array) - 1; j >= 0; j--) {
	   cost_struc = array_fetch(COST_STRUCT *, cost_array, j);
	   free_cost_struct((COST_STRUCT *) cost_struc);
        }
     } /* else */
     
     array_free(cost_array);
  } /* for i = */
  if (ACT_DEBUG) (void) printf("*******TOTAL_GAIN from act_quick_phase = %d******\n", TOTAL_GAIN);
  array_free(nodevec);
  return network;
}     


void
free_cost_struct(cost_node)
  COST_STRUCT *cost_node;
{
  my_free_act(cost_node->act); 
  FREE(cost_node);
}

/*ARGSUSED*/
void 
free_cost_table(table)
  st_table *table;
{
   char *arg;
   st_foreach(table, free_table_entry, arg);
}

/*ARGSUSED */
static
enum st_retval
free_table_entry(key, value, arg)
   char *key, *value, *arg;      
{
   (void) free_cost_struct((COST_STRUCT *) value);
   FREE(key);
   return ST_DELETE;
}

   

/************************************************************************************
  This function returns a node that implements mux function, i.e.
		(control' left + control right).
*************************************************************************************/

static
node_t *
mux_node(control, left, right)
 node_t *control, *left, *right;
{
  node_t *not_control, *and1, *and2, *answer, *answer_simpl;

  not_control = node_not(control);
  and1 = node_and(not_control, left);
  and2 = node_and(control, right);
  answer = node_or (and1, and2);
  node_free(not_control);
  node_free(and1);
  node_free(and2);
  answer_simpl = node_simplify(answer, (node_t *) NULL, NODE_SIM_ESPRESSO);
  node_free(answer);
  return answer_simpl;
}

/********************************************************************************
  This function returns the function implemented by the Actel basic block.
  The naming convention of the inputs is the same as reported in the 
  Actel paper.
*********************************************************************************/
static
node_t *
basic_block_node(A0, A1, SA, B0, B1, SB, S0, S1, cost_table)
   node_t *A0, *A1, *SA, *B0, *B1, *SB, *S0, *S1;
   st_table *cost_table;
{
  node_t *out1, *out2, *or_control, *out, *out_simpl;
 
  (void) fprintf(BDNET_FILE, "\n INSTANCE \"BASIC_BLOCK\":physical NAME = INST%d;\n", instance_num);  
  instance_num++;

  print_node_with_string(A0, "A0", cost_table);
  print_node_with_string(A1, "A1", cost_table);
  print_node_with_string(SA, "SA", cost_table);
  print_node_with_string(B0, "B0", cost_table);
  print_node_with_string(B1, "B1", cost_table);
  print_node_with_string(SB, "SB", cost_table);
  print_node_with_string(S0, "S0", cost_table);
  print_node_with_string(S1, "S1", cost_table);

  out1 = mux_node(SA, A0, A1);
  out2 = mux_node(SB, B0, B1);
  or_control = node_or(S0, S1);
  out = mux_node(or_control, out1, out2);
  node_free(out1);
  node_free(out2);
  node_free(or_control);
  out_simpl = node_simplify(out, (node_t *) NULL, NODE_SIM_ESPRESSO);
  node_free(out);
  return out_simpl;
}
  
static void
print_output_node_with_string(node, name)
   node_t *node;
   char *name;

{
   node_function_t node_fun;

   (void) fprintf(BDNET_FILE, "        \"%s\"           :         \"", name);   
   node_fun = node_function(node);
   switch (node_fun) {
       case NODE_0 : (void) fprintf(BDNET_FILE, "GND");
		     break;
       case NODE_1 : (void) fprintf(BDNET_FILE, "Vdd");
		     break;
       default     : node_print_rhs(BDNET_FILE, node);
   }
   (void) fprintf(BDNET_FILE, "\";\n");


}


static void
print_node_with_string(node, name, cost_table)
   node_t *node;
   char *name;
   st_table *cost_table;

{
   node_t *fanin;
   char *fanin_fn_type;
   node_function_t node_fun;

   (void) fprintf(BDNET_FILE, "        \"%s\"           :         \"", name);   

   /* hack */
   /*------*/
   node_fun = node_function(node);
   switch(node_fun) {
      case NODE_0: fanin_fn_type = "GND";
		   (void) fprintf(BDNET_FILE, "%s", fanin_fn_type);
		   break;
      case NODE_1: fanin_fn_type = "Vdd";
		   (void) fprintf(BDNET_FILE, "%s", fanin_fn_type);
		   break;
      default:
		   fanin = node_get_fanin(node, 0);
		   fanin_fn_type = act_get_node_fn_type(fanin, cost_table);
		   if (fanin_fn_type == NULL) (void) fprintf(BDNET_FILE, "%s", node_long_name(fanin)); 
		   else (void) fprintf(BDNET_FILE, "%s", fanin_fn_type);

   }      
   (void) fprintf(BDNET_FILE, "\";\n");


}
     
/********************************************************************************
  Given a network and the cost_table, breaks up each node into nodes that can be 
  realized by one block of STRUCT module. Also writes the bdnet file.
*********************************************************************************/
network_t *
act_break_network(network, cost_table)
  network_t *network;
  st_table  *cost_table;
{
  array_t *nodevec;
  int size;
  int i, j;
  node_t *node, *node1, *node_new, *node_returned;
  char *dummy;
  COST_STRUCT *cost_struct;
  int i1;

  num_or_patterns = 0; /* keeps track of number of patterns which use OR gate */
  instance_num = 0;    /* keeps track of the instance numbers in the bdnet file */

  nodevec = network_dfs(network);
  size = array_n(nodevec);

  /* setting up the node_function_info */
  /*-----------------------------------*/  
  for (i = 0; i < size; i++) {
      node = array_fetch(node_t *, nodevec, i);
      if (node->type != INTERNAL) continue;
      if (!st_lookup(cost_table, node_long_name(node), &dummy)) {
	(void) printf(" act_break_network(): entry not found in the cost_table\n");	
	exit(1);
      }
      cost_struct = (COST_STRUCT *) dummy;      
      cost_struct->fn_type = act_get_node_fn_type(node, cost_table);

      /* check - just to be sure */
      /*-------------------------*/
      i1 = (cost_struct->fn_type == NULL);
      if (cost_struct->cost != 0) assert (i1);
  }

  init_bdnet(network, cost_table); /* print the beginning remarks */

  for (i = 0; i < size; i++) {
      node = array_fetch(node_t *, nodevec, i);
      if (node->type != INTERNAL) continue;
      if (!st_lookup(cost_table, node_long_name(node), &dummy)) {
	(void) printf(" act_break_network(): entry not found in the cost_table\n");	
	exit(1);
      }
      cost_struct = (COST_STRUCT *) dummy;
      if (cost_struct->cost == 0) continue;
      PRESENT_ACT = cost_struct->act;
      set_mark_act(PRESENT_ACT);
      my_node_array = array_alloc(node_t *, 0);
      node_returned =  act_get_function(network, node, PRESENT_ACT, cost_table);
      node_free(node_returned);
      (void) fprintf(BDNET_FILE, "        \"out\"          :          \"%s\"; \n\n", 
		     node_long_name(node));   

      if (array_n(my_node_array) == 0); 
      else {
         for (j = 0; j < array_n(my_node_array) - 1; j++) {
   	    node_new = array_fetch(node_t *, my_node_array, j);
	    network_add_node(network, node_new);
         }
         node1 = array_fetch(node_t *, my_node_array, (array_n(my_node_array) - 1));
         node_replace(node, node1);
      }

      set_mark_act(PRESENT_ACT);
      free_nodes_in_act(PRESENT_ACT);
      array_free(my_node_array);
  }
  (void) fprintf(BDNET_FILE, "ENDMODEL;\n");

  if (!network_check(network)) {
	   (void) printf("in network_break():%s\n", error_string());
	   exit(1);
  }
  array_free(nodevec);
  return network;
}

static void
set_mark_act(PRESENT_ACT)
  ACT_VERTEX_PTR PRESENT_ACT;
{
      MARK_VALUE = PRESENT_ACT->mark;
      if (MARK_VALUE == 1) MARK_COMPLEMENT_VALUE = 0;
      else MARK_COMPLEMENT_VALUE = 1;
}

/**************************************************************************************
  Given a node and its corresponding ACT(OACT), breaks the node into nodes, each 
  new node realizing a function that can be implemented by one basic block. Puts all 
  these nodes in the network. Original node is changed. 
***************************************************************************************/
static
node_t *
act_get_function(network, node, vertex, cost_table)
  network_t *network;
  node_t *node;
  ACT_VERTEX_PTR vertex;
  st_table *cost_table;
{

  node_t  *node1, *node2;
  node_t *A0, *A1, *SA, *B0, *B1, *SB, *S0, *S1;  

  /* vertex->node corresponds to the node at the vertex - only for the  
     multiple_fanout vertex */
  /*-----------------------------------------------------------------*/


  if (vertex->value == 0) {
      vertex->mark = MARK_COMPLEMENT_VALUE;
      return node_constant(0);
  }
  if (vertex->value == 1) {
      vertex->mark = MARK_COMPLEMENT_VALUE;
      return node_constant(1);
  }


  if (vertex->mark == MARK_COMPLEMENT_VALUE) {
	if (vertex->multiple_fo == 0) {
		(void) printf("error: act_get_function()\n");
		exit(1);
	}
	return vertex->node;
   }

  vertex->mark = MARK_COMPLEMENT_VALUE;

  if ((vertex->low->value == 0) && (vertex->high->value == 1)) {
      node1 = get_node_of_vertex(vertex, network);
      if (vertex->multiple_fo == 1) vertex->node = node1;  
      return node1;
  }

  if (vertex->pattern_num > 3) num_or_patterns++;

  switch(vertex->pattern_num) {

     case 0: 

        A0 = node_constant(0);
	A1 = act_get_function(network, node, vertex->low, cost_table);
        SA = node_constant(1);

	B0 = A0;
	B1 = act_get_function(network, node, vertex->high, cost_table);
	SB = SA;

        S0 = get_node_of_vertex(vertex, network);
        S1 = A0;

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);

	free_node_if_possible(A1, vertex->low);
 	free_node_if_possible(B1, vertex->high);

	node_free(A0);
	node_free(SA);
	node_free(S0); 

        node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
        if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;

     case 1:

	vertex->low->mark = MARK_COMPLEMENT_VALUE;
	A0 = act_get_function(network, node, vertex->low->low, cost_table);
	A1 = act_get_function(network, node, vertex->low->high, cost_table);
	SA = get_node_of_vertex(vertex->low, network);

	B0 = node_constant(0);
	B1 = act_get_function(network, node, vertex->high, cost_table);
 	SB = node_constant(1);

	S0 = get_node_of_vertex(vertex, network);
	S1 = B0;

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);
	free_node_if_possible(A0, vertex->low->low);
	free_node_if_possible(A1, vertex->low->high);
 	free_node_if_possible(B1, vertex->high);

	node_free(B0);
	node_free(SB);
  	node_free(S0);
   	node_free(SA); 

	node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
	if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;

     case 2:
	vertex->high->mark = MARK_COMPLEMENT_VALUE;
	A0 = node_constant(0);
	A1 = act_get_function(network, node, vertex->low, cost_table);
 	SA = node_constant(1);

	B0 = act_get_function(network, node, vertex->high->low, cost_table);
	B1 = act_get_function(network, node, vertex->high->high, cost_table);
	SB = get_node_of_vertex(vertex->high, network);

	S0 = get_node_of_vertex(vertex, network);
	S1 = A0;

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);

	free_node_if_possible(A1, vertex->low);
	free_node_if_possible(B0, vertex->high->low);
 	free_node_if_possible(B1, vertex->high->high);
	
	node_free(A0);
	node_free(SA);
	node_free(S0);
	node_free(SB); 

        node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
        if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;

     case 3:
	vertex->low->mark = MARK_COMPLEMENT_VALUE;
	vertex->high->mark = MARK_COMPLEMENT_VALUE;

	A0 = act_get_function(network, node, vertex->low->low, cost_table);
	A1 = act_get_function(network, node, vertex->low->high, cost_table);
	SA = get_node_of_vertex(vertex->low, network);

	B0 = act_get_function(network, node, vertex->high->low, cost_table);
	B1 = act_get_function(network, node, vertex->high->high, cost_table);
	SB = get_node_of_vertex(vertex->high, network);

	S0 = get_node_of_vertex(vertex, network);
	S1 = node_constant(0);

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);

	free_node_if_possible(A0, vertex->low->low);
	free_node_if_possible(A1, vertex->low->high);
 	free_node_if_possible(B0, vertex->high->low);	
 	free_node_if_possible(B1, vertex->high->high);	

	node_free(S1);
	node_free(SA);
	node_free(SB);
	node_free(S0); 
	
        node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
        if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;

     case 4:
	vertex->low->low->mark = MARK_COMPLEMENT_VALUE;
	vertex->low->mark = MARK_COMPLEMENT_VALUE;

	A0 = act_get_function(network, node, vertex->low->low->low, cost_table);
	A1 = act_get_function(network, node, vertex->low->low->high, cost_table);
	SA = get_node_of_vertex(vertex->low->low, network);

	B0 = node_constant(0);
	B1 = act_get_function(network, node, vertex->high, cost_table);
 	SB = node_constant(1);

	S0 = get_node_of_vertex(vertex, network);
	S1 = get_node_of_vertex(vertex->low, network);

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);

	free_node_if_possible(A0, vertex->low->low->low);
	free_node_if_possible(A1, vertex->low->low->high);
 	free_node_if_possible(B1, vertex->high);	

	node_free(SA);
	node_free(B0);
	node_free(SB);
	node_free(S0); 
	node_free(S1);
	
        node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
        if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;

     case 5:
	vertex->high->mark = MARK_COMPLEMENT_VALUE;
	vertex->high->low->mark = MARK_COMPLEMENT_VALUE;

	A0 = node_constant(0);
	A1 = act_get_function(network, node, vertex->high->low->low, cost_table);
	SA = get_node_of_vertex(vertex, network);

	B0 = A0;
	B1 = act_get_function(network, node, vertex->high->high, cost_table);
 	SB = SA;

	S0 = get_node_of_vertex(vertex->high, network);
	S1 = get_node_of_vertex(vertex->high->low, network);

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);

	free_node_if_possible(A1, vertex->high->low->low);
	free_node_if_possible(B1, vertex->high->high);

	node_free(A0);
	node_free(SA);
	node_free(S0); 
	node_free(S1);
	
        node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
        if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;

     case 6:

	vertex->high->mark = MARK_COMPLEMENT_VALUE;
	vertex->high->low->mark = MARK_COMPLEMENT_VALUE;

	A0 = node_constant(0);
	A1 = node_constant(1);
	SA = A1;

	B0 = A1;
	B1 = A0;
 	SB = get_node_of_vertex(vertex, network);

	S0 = get_node_of_vertex(vertex->high, network);
	S1 = get_node_of_vertex(vertex->high->low, network);

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);

	node_free(A0);
	node_free(A1);
	node_free(SB);
	node_free(S0); 
	node_free(S1);
	
        node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
        if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;

     case 7:
	vertex->low->mark = MARK_COMPLEMENT_VALUE;
	vertex->low->low->mark = MARK_COMPLEMENT_VALUE;

	A0 = act_get_function(network, node, vertex->low->low->low, cost_table);
	A1 = node_constant(0);
	SA = get_node_of_vertex(vertex, network);

	B0 = act_get_function(network, node, vertex->low->high, cost_table);
	B1 = A1;
 	SB = SA;

	S0 = get_node_of_vertex(vertex->low, network);
	S1 = get_node_of_vertex(vertex->low->low, network);

        node1 = basic_block_node (A0, A1, SA, B0, B1, SB, S0, S1, cost_table);

	free_node_if_possible(A0, vertex->low->low->low);
	free_node_if_possible(B0, vertex->low->high);

	node_free(A1);
	node_free(SA);
	node_free(S0); 
	node_free(S1);
	
        node2 = node_literal(node1, 1);
        if (vertex != PRESENT_ACT) 
	      print_output_node_with_string(node2, "out");
        if (vertex->multiple_fo == 1) vertex->node = node2;
	array_insert_last(node_t *, my_node_array, node1);
        return node2;
     default:
       (void) printf(" act_get_function(): unknown pattern_num %d found\n", vertex->pattern_num);
       exit(1);
  } /* switch */        
  
  return NULL;
}

static
node_t *
get_node_of_vertex(vertex, network)
  ACT_VERTEX_PTR vertex;
  network_t *network;
{
  node_t *node;

  node = network_find_node(network, vertex->name);
  if (node == NULL) {
	(void) printf(" get_node_of_network(): node is NULL\n");
	exit(1);
  }
  /* changed here July 20 '90 */
  /*--------------------------*/
  return (node_literal(node, 1));
}

/***********************************************************************************
   If the vertex is just an input, then the node cannot be freed. If the vertex is
   a multiple_fo point, then also it cannot be freed. This is because these nodes 
   are needed in the network. 
************************************************************************************/
static
void
free_node_if_possible(node, vertex)
  node_t *node; 
  ACT_VERTEX_PTR vertex;
{
  node_function_t node_fn;

  node_fn = node_function(node);
  if ((node_fn == NODE_0) || (node_fn == NODE_1)){
       node_free(node);
       return;
  }
  /* should be after the node_fn check, because 0, 1 vertices 
     do not store anything */
  /*---------------------------------------------------------*/
  if (vertex->multiple_fo == 1) return; 
  node_free(node); 
}  

/*--------------------------------------------------------------
 Frees all the associated storage for the act, except the act
 itself
----------------------------------------------------------------*/
void
act_act_free(node)
  node_t *node;
{
    
    ACT_SET(node)->LOCAL_ACT->act->root = NULL; 
    local_actFree(node);

}

static void
free_nodes_in_act(vertex)
  ACT_VERTEX_PTR vertex;
{
  if (vertex->value != 4) return;
  if (vertex->mark == MARK_COMPLEMENT_VALUE) return;
  vertex->mark = MARK_COMPLEMENT_VALUE;
  if (vertex->multiple_fo == 1) node_free(vertex->node);
  free_nodes_in_act(vertex->low);
  free_nodes_in_act(vertex->high);

}

/*********************************************************************
   writes input and output of the network in the bdnet file.
**********************************************************************/

static void
init_bdnet(network, table)
network_t *network;
st_table *table;

{

 lsGen genpi, genpo;
 node_t *pi, *po, *po_fanin;
 char *pi_name, *po_name, *po_fanin_fn_type;

 (void) fprintf(BDNET_FILE, "MODEL \"%s\";\n\n", network_name(network));

 (void) fprintf(BDNET_FILE, "TECHNOLOGY scmos;\n");
 (void) fprintf(BDNET_FILE, "VIEWTYPE SYMBOLIC;\n");
 (void) fprintf(BDNET_FILE, "EDITSTYLE SYMBOLIC;\n\n");
 /* writing inputs */
 /*----------------*/
 foreach_primary_input(network, genpi, pi) {
    pi_name = node_long_name(pi);
    (void) fprintf(BDNET_FILE, "INPUT \t\"%s\"	:	\"%s\";\n",
	           pi_name, pi_name);
 }
 (void) fprintf(BDNET_FILE, "\n"); 

 /* writing outputs */
 /*-----------------*/
 foreach_primary_output(network, genpo, po) {
    po_name = node_long_name(po);
    po_fanin = node_get_fanin(po, 0);

    /* hack for bdnet */
    /*----------------*/   
    /* if po_fanin has 0 cost, then print the transitive input (or Vdd, GND) that influences it */
    /*------------------------------------------------------------------------------------------*/
    po_fanin_fn_type = act_get_node_fn_type(po_fanin, table);
    if (po_fanin_fn_type == NULL)
	    (void) fprintf(BDNET_FILE, "OUTPUT \t\"%s\"	:	\"%s\";\n",
	  	           po_name,  node_long_name(po_fanin) );
    else
	    (void) fprintf(BDNET_FILE, "OUTPUT \t\"%s\"	:	\"%s\";\n",
	  	           po_name,  po_fanin_fn_type);
 }
 (void)  fprintf(BDNET_FILE, "\n"); 
 
}


/*-----------------------------------------------------------------
  This is a "hack" for generating bdnet file. If a node has NODE_0
  (NODE_1) function, returns "GND" (Vdd"). If it is a NODE_BUF
  (single input), then returns the name of the farthest
   transitive fanin node   which is either NODE_0, NODE_1, 
   or NODE_BUF, appropriately.
------------------------------------------------------------------*/  
static char *
act_get_node_fn_type(node, cost_table)
node_t *node;
st_table *cost_table;
{
  node_t *fanin;
  char *dummy;
  COST_STRUCT *cost_node;
  node_function_t node_fun;

  node_fun = node_function(node);
  switch(node_fun) {
     case NODE_0 : return "GND";
     case NODE_1 : return "Vdd";
     case NODE_BUF:
	  fanin = node_get_fanin(node, 0);
	  if (fanin->type == PRIMARY_INPUT) return node_long_name(fanin);
	  assert(st_lookup(cost_table, node_long_name(fanin), &dummy));
	  cost_node = (COST_STRUCT *) dummy;
	  if (cost_node->fn_type == NULL) return node_long_name(fanin);
	  else return cost_node->fn_type;
     default: return NIL (char);
  }
}
  
