#include "mis.h"
#include "pld_int.h"
#include <math.h>

struct my_cube {
   int *variable; /* maximum cardinality of LAMBDA-set */
 };

static node_t **ALPHA;  /* for storing ALPHA functions */
static node_t *G_node; /* used for finding G_node */

static void karp_decomp();
static void form_incompatibility_graph();
static void form_u_intersection();
static void get_lambda_cube();
static int val_literal();
static int** init_incompatible();
static void alloc_my_cube();
static void end_incompatible();
static void end_intersection();
static void free_my_cube();
static void make_incompatible();
static void generate_all_minterms();
static void generate_minterm_with_index();
static void tree_init();
static void form_compatibility_classes();
static void mark_roots_of_trees();
static void get_combination();
static void mark_all_inputs();
static int is_fanin_in_lambda();
static int marked();
static int find_alphas_and_G(); /* returns number of alpha nodes */
static node_t *product_node_for_minterm();
static node_t *find_G();
static void free_tree();
static enum st_retval k_decomp_free_table_entry();
static void replace_node_decomposed();

void
karp_decomp_network(network, SUPPORT, one_node, user_node)
  network_t *network;
  int SUPPORT;
  int one_node;
  node_t *user_node;
{

  node_t *node, *node_simpl;
  array_t *nodevec;
  int i;
  int num_nodes;
  int **intersection;
  int *lambda_indices;
  int NUM_NODES;

  NUM_NODES = (int)(pow((double)2, (double)SUPPORT));

  /* initializing intersection function */
  /*------------------------------------*/
  intersection = ALLOC(int *, 3);
  for (i = 0; i < 3; i++) {
 	intersection[i] = ALLOC(int, 3);
  }
  intersection[0][1] = -1;
  intersection[1][0] = -1;
  intersection[0][0] = intersection[0][2] = intersection[2][0] = 0;
  intersection[1][1] = intersection[1][2] = intersection[2][1] = 1;
  intersection[2][2] = 2;

  lambda_indices = ALLOC(int, SUPPORT);
  if (one_node) {   /* = 1 if there is just one node to be decomposed */
     /* simplify so that while taking complement, the support remains same for
        complemented node */
     /*---------------------------------------------------------------------*/
     node_simpl = node_simplify(user_node, (node_t *) NULL, NODE_SIM_ESPRESSO);
     node_replace(user_node, node_simpl);
     karp_decomp(network, user_node, intersection, lambda_indices, NUM_NODES, SUPPORT); 
     return;
  }

  nodevec = network_dfs_from_input(network);
  num_nodes = array_n(nodevec);
  /* simplify so that while taking complement, the support remains same for
     complemented node */
  /*---------------------------------------------------------------------*/
  for (i = 0; i < num_nodes; i++) {
     node = array_fetch(node_t *, nodevec, i);
     if (node->type != INTERNAL) continue;
     node_simpl = node_simplify(node, (node_t *) NULL, NODE_SIM_ESPRESSO);
     node_replace(node, node_simpl);
  }
  for (i = 0; i < num_nodes; i++) {
     node = array_fetch(node_t *, nodevec, i);
     (void) karp_decomp(network, node, intersection, lambda_indices, NUM_NODES, SUPPORT);
  }

  (void) network_sweep(network);

  /* freeing */
  /*---------*/ 
  array_free(nodevec);
  for (i = 0; i < 3; i++){
      FREE(intersection[i]);
  }
  FREE(intersection);

  FREE(lambda_indices);  
}

/**************************************************************
  decomposes a node in the network such that the bound set
  LAMBDA has a cardinality of SUPPORT. Based on karp's method of
  decomposition. However, to speed up things, presently we do 
  not go for the best decomposition. We just go for "a" 
  decomposition. 
**************************************************************/
static void
karp_decomp(network, node, intersection, lambda_indices, NUM_NODES, SUPPORT)
network_t *network;
node_t *node;
int **intersection;
int *lambda_indices;
int NUM_NODES, SUPPORT;

{

  int num_fanin;
  node_t *node_inv;
  int i;
  int max_limit;
  st_table *root_table;
  st_table *lambda_table;
  tree_node **treenode;
  int **incompatible;  /* stores whether minterms i & j are incompatible */
  int CLASS_NUM,  NUM_ALPHAS;

  num_fanin = node_num_fanin(node);
  if (num_fanin <= SUPPORT) return;
  incompatible = init_incompatible(NUM_NODES);

 /* to store the info. about the root of the trees. CLASS_NUM tells the 
   class_num of each tree (but is valid only at the end of mark_root */

  root_table = st_init_table(st_ptrcmp, st_ptrhash);
  lambda_table = st_init_table(st_ptrcmp, st_ptrhash); 
  CLASS_NUM = 0;
  treenode = ALLOC (tree_node *, NUM_NODES);
  tree_init(treenode, NUM_NODES);

  node_inv = node_not(node);

  /* for debugging purpose only - limit on the number of combinations
   to be examined for the LAMBDA set */
  /*-----------------------------------------------------------------*/
  max_limit = 1;


  for ( i = 1; i <= max_limit; i++) {
    get_combination(node, i-1, lambda_indices, SUPPORT);
    form_incompatibility_graph(node, node_inv, intersection, incompatible, lambda_table, lambda_indices, NUM_NODES, SUPPORT);
    form_compatibility_classes(treenode, incompatible, NUM_NODES);           
    mark_roots_of_trees(treenode, root_table, NUM_NODES, &CLASS_NUM);
    NUM_ALPHAS = find_alphas_and_G(node, lambda_table, treenode, lambda_indices, NUM_NODES, SUPPORT, CLASS_NUM);
  }

  /* node is changed */
  /*-----------------*/ 
  replace_node_decomposed(network, node,  NUM_ALPHAS);
  node_free(node_inv);
  free_tree(treenode, root_table, lambda_table, NUM_NODES);
  end_incompatible(incompatible, NUM_NODES);  
  /* if there is just one node needed to be decomposed, then recursively call the fn. */
  /*----------------------------------------------------------------------------------*/
  karp_decomp(network, node, intersection, lambda_indices, NUM_NODES, SUPPORT);
}
  
static void
form_incompatibility_graph(node, node_inv, intersection, incompatible, lambda_table, lambda_indices, NUM_NODES, SUPPORT)
node_t *node;
node_t *node_inv;
int **intersection,  **incompatible;
st_table *lambda_table;
int *lambda_indices;
int NUM_NODES, SUPPORT;

{
   int node_numcubes;
   int i;
   node_cube_t cube_n;

   /* if the input in is in lambda part, mark it 1, else mark it 0 */
   /* assume static (global) storage for the marking arrays */
   /*---------------------------------------------------------------*/
   mark_all_inputs(node, lambda_table, lambda_indices, SUPPORT);
   node_numcubes = node_num_cube(node);
   for (i = node_numcubes - 1; i >= 0; i--) {
       cube_n = node_get_cube(node, i);
       form_u_intersection(node, cube_n, node_inv, intersection, incompatible, lambda_table, NUM_NODES, SUPPORT);
   }
 }

/********************************************************************
   for a cube of the node, form its u-intersection with every
   cube of node_inv.
*********************************************************************/

static void
form_u_intersection(node, cube, node_inv, intersection, incompatible, lambda_table, NUM_NODES, SUPPORT)
node_t *node;
node_cube_t cube;
node_t *node_inv;
int **intersection;
int **incompatible;
st_table *lambda_table;
int NUM_NODES, SUPPORT;
{
   int NULL_INT;
   int node_inv_numcubes;
   int i;
   int j;
   node_t *fanin;
   node_t *fanin_inv;
   node_literal_t literal;
   int lit_val;
   node_literal_t literal_inv;
   int lit_inv_val;
   struct my_cube lambda_cube;
   struct my_cube lambda_cube_inv;
   node_cube_t cube_inv;

   node_inv_numcubes = node_num_cube(node_inv);
   
   for (i = node_inv_numcubes -1; i >= 0; i--) {
        cube_inv = node_get_cube(node_inv, i);
        NULL_INT = FALSE;
        foreach_fanin(node, j, fanin) {
          if (marked(fanin, lambda_table)) continue;
          literal = node_get_literal(cube, j);
	  lit_val = val_literal(literal);

          fanin_inv = node_get_fanin(node_inv, j);
          if (marked(fanin_inv, lambda_table)) {
            (void) fprintf(misout, "Fatal error: cube ordering erros\n");
	    exit(1);
	  }
          literal_inv = node_get_literal(cube_inv, j);
	  lit_inv_val = val_literal(literal_inv);
          if (intersection[lit_val][lit_inv_val] == -1) {
              NULL_INT = TRUE;
	      break;
	  }
	}
        if (NULL_INT) continue;

	/* form the incompatibity matrix from the lambda parts */
	/*-----------------------------------------------------*/        
	alloc_my_cube(&lambda_cube, SUPPORT);
	alloc_my_cube(&lambda_cube_inv, SUPPORT);
	get_lambda_cube(node, cube, &lambda_cube, lambda_table);
	get_lambda_cube(node_inv, cube_inv, &lambda_cube_inv, lambda_table);
	make_incompatible(lambda_cube, lambda_cube_inv, incompatible, NUM_NODES, SUPPORT);
	free_my_cube(&lambda_cube);
	free_my_cube(&lambda_cube_inv);
  }
}
	 
static void
get_lambda_cube(node, cube, lambda_cube, lambda_table)
node_t *node;
node_cube_t cube;
struct my_cube *lambda_cube;
st_table *lambda_table;
 {
   int var_num;
   node_t *fanin_node;
   int j;
   node_literal_t literal;

   var_num = 0;
   foreach_fanin(node, j, fanin_node) {
      literal = node_get_literal(cube, j);
      if (marked(fanin_node, lambda_table)) 
        lambda_cube->variable[var_num++] = val_literal(literal);
    }
}

static
int val_literal(literal)
    node_literal_t literal;
{    
    int value;

    switch(literal) {
         case ZERO : /* complemented */
	    value = 0;
	    break;
         case ONE : /* uncomplemented */
	    value = 1;
	    break;
	 case TWO : /* does not appear */
	    value = 2; 
	    break;
    }
    return value;
}


static int**
init_incompatible(NUM_NODES)
int NUM_NODES;
{
 int i;
 int j;
 int **incompatible;

 incompatible = ALLOC(int *, NUM_NODES);
 for (i = 0; i < NUM_NODES; i++) {
     incompatible[i] = ALLOC(int, NUM_NODES);
 }
 
 for (i =0 ; i < NUM_NODES; i++) {
   for (j = 0; j < NUM_NODES; j++) {
     incompatible[i][j] = FALSE;
   }
 } 
 return incompatible;
}

static void
end_incompatible(incompatible, NUM_NODES)
int **incompatible;
int NUM_NODES;
{
 int i;

 for (i = 0; i < NUM_NODES; i++) {
   FREE(incompatible[i]);
 } 
 FREE(incompatible);
}

static void
make_incompatible(a, b, incompatible, NUM_NODES, SUPPORT)
struct my_cube a, b;
int **incompatible;
int NUM_NODES, SUPPORT;
{

  int i;
  int j;
  int *min_a;
  int *min_b;
  int num_min_a = 0;
  int num_min_b = 0;

  min_a = ALLOC(int, NUM_NODES);
  min_b = ALLOC(int, NUM_NODES);

  generate_all_minterms(min_a, &num_min_a, a, SUPPORT);
  generate_all_minterms(min_b, &num_min_b, b, SUPPORT);
  for (i = 0; i < num_min_a; i++) {
    for (j = 0; j < num_min_b; j++) {
      incompatible[min_a[i]][min_b[j]] = incompatible[min_b[j]][min_a[i]]
      = TRUE;
      
    }
  }
  FREE(min_a);
  FREE(min_b);
}

/**************************************************************************
  returns in min_arr all the minterms in the cube cube. Returns the number
  of these minterms in num_min.
****************************************************************************/
static 
void
generate_all_minterms(min_arr, num_min, cube, support)
  int min_arr[];
  int *num_min;
  struct my_cube cube;
  int support;
{

  generate_minterm_with_index(min_arr, num_min, cube, support, 0, 0);
}

static
void 
generate_minterm_with_index(min_arr, num_min, cube, support, value, i)
  int min_arr[];
  int *num_min;
  struct my_cube cube;
  int support;
  int value;
  int i;
{
   if (i == support - 1) {
     switch (cube.variable[i]) {
        case 0: 
                min_arr[*num_min] = 2*value; *num_min = *num_min + 1; 
	        return;
        case 1: 
	        
	        min_arr[*num_min] = 2*value + 1; *num_min = *num_min + 1; 
	        return;
	case 2: 
	        
	        min_arr[*num_min] = 2*value; *num_min = *num_min + 1; 
	        min_arr[*num_min] = 2*value + 1;*num_min = *num_min + 1; 
	        return;
     };
   }
   switch (cube.variable[i]) {
      case 0 : generate_minterm_with_index(min_arr, num_min, cube, support, 2*value, i+1);
	       return;
      case 1 : generate_minterm_with_index(min_arr, num_min, cube, support, 2*value + 1, i+1);
 	       return;
      case 2 : generate_minterm_with_index(min_arr, num_min, cube, support, 2*value, i+1);
               generate_minterm_with_index(min_arr, num_min, cube, support, 2*value + 1, i+1);
	       return;
   }

}

static  void
tree_init(treenode, NUM_NODES)
tree_node **treenode;
int NUM_NODES;
{

 int i;
 
 for (i = 0; i < NUM_NODES; i++) {
   treenode[i] = ALLOC(tree_node, 1);
   treenode[i]->parent = treenode[i];
   treenode[i]->index = i;
   treenode[i]->num_child = 0;
   treenode[i]->class_num = -1;

 }


}

static void
form_compatibility_classes(treenode, incompatible, NUM_NODES)
tree_node **treenode;
int **incompatible;
int NUM_NODES;
{

  int i;
  int j;
  struct tree_node *root1;
  struct tree_node *root2;

  for (i = 0; i< NUM_NODES; i++) {
    for (j = i+1; j < NUM_NODES; j++) {
      if (incompatible[i][j] == FALSE ) /* i and j are compatible */
	{
	  root1 = find_tree(treenode[i]);
	  root2 = find_tree(treenode[j]);
	  if (root1->index == root2->index) continue;
	  unionop(root1, root2);

	}
    }
  }
}

static void
mark_roots_of_trees(treenode, root_table, NUM_NODES, CLASS_NUM) 
tree_node **treenode;
st_table *root_table;
int NUM_NODES, *CLASS_NUM;

{

  int i;
  tree_node *root;
  char *dummy;
  int value;

 for (i = 0; i <  NUM_NODES; i++) {
  root = find_tree(treenode[i]);
  treenode[i]->parent = root;  
  if (st_is_member(root_table, (char *)root->index)) {
     (void) st_lookup(root_table, (char *)root->index, &dummy);
     value = (int) dummy;
     treenode[i]->class_num = value;
   }
  else {
     (void) st_insert(root_table, (char*)root->index, (char *)(*CLASS_NUM));  
     treenode[i]->class_num = (*CLASS_NUM);
     *CLASS_NUM = (*CLASS_NUM) + 1;

   }
  
 }
}

static void
get_combination(node, comb_num, lambda_indices, SUPPORT)
node_t *node;
int comb_num;
int *lambda_indices;
int SUPPORT;
 
{
  int num_fanin;
  /* adhoc routine - can only return a few combinations*/  
  int i;
  num_fanin = node_num_fanin(node);
  for (i = 0; i < SUPPORT; i++) 
    lambda_indices[i] = (comb_num + i) % num_fanin;

}

/* marks all the inputs of the node as 1 if it corresponds to lambda part,
   else 0 */

static
void
mark_all_inputs(node, lambda_table, lambda_indices, SUPPORT)
node_t *node;
st_table *lambda_table;  
int *lambda_indices, SUPPORT;
{

  int islambda_fanin;
  int j;
  node_t *fanin;

  foreach_fanin(node, j, fanin) {
      islambda_fanin =  is_fanin_in_lambda(node, fanin, lambda_indices, SUPPORT);
      (void) 	st_insert(lambda_table, (char *)fanin, (char *)islambda_fanin);
    }
}

static
int is_fanin_in_lambda(node, fanin, lambda_indices, SUPPORT)
node_t *fanin ;
node_t *node;
int *lambda_indices, SUPPORT;
{

    int i;
    node_t *fanin_my;

    for (i = 0; i < SUPPORT; i++) {
      fanin_my = node_get_fanin(node, lambda_indices[i]);
      if (fanin_my == fanin) return 1;
    }
    return 0;
}

static
int marked(fanin, lambda_table)
node_t *fanin;
st_table *lambda_table;
{
   char *dummy;  
   int found;


   found = st_lookup(lambda_table, (char *)fanin, &dummy);
   if (!found) {
     (void) fprintf(misout,  " Fatal error: marked(fanin) : error in finding node\n", 
	     node_name(fanin));
     exit(1);
   }
   return (int)dummy;
}

/**********************************************************************
  f = g(alpha1, alpha2, ...., alphat, ....);
  Number of ALPHA functions is log2 of number of compatibility classes.
  fn_array stores for a minterm binary value of classnum that minterm
  belongs to. 
  From fn_array, the ALPHA functions are constructed by ORing the 
  minterms that result in 1 for that function.
**********************************************************************/

static int
find_alphas_and_G(node_to_decomp, lambda_table, treenode, lambda_indices, NUM_NODES, SUPPORT, CLASS_NUM)
node_t *node_to_decomp;
st_table *lambda_table;
tree_node **treenode;
int *lambda_indices;
int NUM_NODES, SUPPORT, CLASS_NUM;
{

 char **fn_array;
 node_t *fn, *fn_simplified;
 node_t *fn_node1;
 node_t *node_for_minterm;
 int minterm;
 int fn_num;
 int FLAG;
 int NUM_ALPHAS;

 fn_array = ALLOC(char *, NUM_NODES);

 if (CLASS_NUM == 0) {
    (void) fprintf(misout, "Fatal error: no class?\n");
    exit(1);
 }
 if (CLASS_NUM == 1) {
    (void) fprintf(misout, "Fatal error: strange!!\n");
    exit(1);
 }

 NUM_ALPHAS = intlog2(CLASS_NUM);
 ALPHA = ALLOC(node_t *, NUM_ALPHAS);
 for (minterm = 0; minterm < NUM_NODES; minterm++) {
   fn_array[minterm] = ALLOC(char, NUM_ALPHAS + 1);
   xl_binary1(treenode[minterm]->class_num, NUM_ALPHAS, fn_array[minterm]);
 }
 /* (void) printf( "NUM_ALPHAS = %d\n", NUM_ALPHAS); */

 for (fn_num = 0; fn_num < NUM_ALPHAS; fn_num++) {
   fn = node_constant(0);
   FLAG = 0;
   for (minterm =0; minterm < NUM_NODES; minterm++) {
      if (fn_array[minterm][fn_num] == '0') continue;
      FLAG = 1;
      node_for_minterm = product_node_for_minterm(minterm, node_to_decomp, lambda_indices, SUPPORT);
      fn_node1 = node_or(node_for_minterm, fn);
      node_free(fn);
      node_free(node_for_minterm);
      fn = fn_node1;
   };
   if (FLAG) {
	  fn_simplified = node_simplify(fn, (node_t *) NULL, NODE_SIM_ESPRESSO);
          ALPHA[fn_num] = fn_simplified;
          node_free(fn); 
   }
   else ALPHA[fn_num] = fn;
 }
 G_node =  find_G(node_to_decomp, lambda_table, treenode, NUM_NODES, SUPPORT, NUM_ALPHAS);

 for (minterm = 0; minterm < NUM_NODES; minterm++) {
     FREE(fn_array[minterm]);
 }
 FREE(fn_array);
 return  NUM_ALPHAS;
}

static 
node_t *
product_node_for_minterm(minterm, node_to_decomp, lambda_indices, SUPPORT)
int minterm;
node_t *node_to_decomp;
int *lambda_indices, SUPPORT;
{

  node_t *node;
  node_t *fn;
  node_t *fn1, *literal_fn, *fanin;
  char char_minterm[BUFSIZE];
  int length, i;
  int phase;

  fn = node_constant(1);
  xl_binary1(minterm, SUPPORT, char_minterm);
  length = strlen (char_minterm);
  if (length != SUPPORT) {
    (void) fprintf(misout,  "Fatal error: product_node_for_minterm(): length not SUPPORT\n");
    exit(1);
  }
  for (i = 0; i < length; i++) {
     fanin = node_get_fanin(node_to_decomp, lambda_indices[i]); 
     if (char_minterm[i] == '0') phase = 0;
     else phase = 1;
     literal_fn = node_literal(fanin, phase);
     fn1 = node_and (literal_fn, fn);
     node_free(fn);
     node_free(literal_fn);
     fn = fn1;
  }
  node = fn;
  return node;
}

static
node_t *
find_G( node_to_decomp, lambda_table, treenode, NUM_NODES, SUPPORT,  NUM_ALPHAS)
node_t *node_to_decomp;
st_table *lambda_table; 
tree_node **treenode;
int NUM_NODES, SUPPORT,  NUM_ALPHAS;

{
  char encoded_char[BUFSIZE];
  node_t *Gnode;  /* to be returned */
  node_t *fanin;
  node_t *product_fn, *product_fn1, *literal_fn;
  node_t *nodeG, *nodeG1, *nodeG2;
  node_t *u_fn, *u_fn1, *cube_fn, *cube_fn1;
  int i, j, k, l;
  int num_min;
  int *min_arr, minterm;
  int lit_val, phase;
  node_cube_t cube;
  struct my_cube lambda_cube;
  node_literal_t literal;

  min_arr = ALLOC(int, NUM_NODES);
  nodeG = node_constant(0);
  i = node_num_cube(node_to_decomp);
  if (i == 0) {
    (void) fprintf(misout,  "Fatal error:  node %s has no cubes??\n", 
            node_name(node_to_decomp));
    exit(1);
  }
  for (i = node_num_cube(node_to_decomp) - 1 ; i>= 0; i--) {
     cube = node_get_cube(node_to_decomp, i);
     alloc_my_cube(&lambda_cube, SUPPORT);
     get_lambda_cube(node_to_decomp, cube, &lambda_cube, lambda_table);
     num_min = 0;
     generate_all_minterms(min_arr, &num_min, lambda_cube, SUPPORT);

     /* compute fn. corresponding to the u_part only once for the
     cube */
     /*---------------------------------------------------------*/
     u_fn = node_constant(1);
     foreach_fanin(node_to_decomp, j, fanin) {
        if (marked(fanin, lambda_table)) continue;
        literal = node_get_literal(cube, j);
	lit_val = val_literal(literal);
	if (lit_val != 2) {
           literal_fn = node_literal(fanin, lit_val);
	   u_fn1 = node_and(literal_fn, u_fn);
	   node_free(literal_fn);
	   node_free(u_fn);
           u_fn = u_fn1;
        }
     }                

    /* for each minterm in the lambda_space, compute the 
     alpha_functions; OR them together, then AND with u_fn */
    /*-----------------------------------------------------*/

     cube_fn = node_constant(0);

     for (k = 0; k < num_min; k++) {
        product_fn = node_constant(1);
        minterm = min_arr[k];
	xl_binary1(treenode[minterm]->class_num, NUM_ALPHAS, 
                encoded_char);
 	for (l = 0; l < NUM_ALPHAS; l++) {
           if (encoded_char[l] == '0') phase = 0;
           else phase = 1;
           literal_fn = node_literal(ALPHA[l], phase);
	   product_fn1 = node_and (literal_fn, product_fn);
	   node_free(literal_fn);
	   node_free(product_fn);
	   product_fn = product_fn1;
        }
        cube_fn1 = node_or(cube_fn, product_fn);
        node_free(cube_fn);
	node_free(product_fn);
        cube_fn = cube_fn1;
    }
    
    nodeG1 = node_and(cube_fn, u_fn);
    node_free(cube_fn);
    node_free(u_fn);

    nodeG2 = node_or(nodeG1, nodeG);
    node_free(nodeG1);
    node_free(nodeG);
    nodeG = nodeG2;   
    free_my_cube(&lambda_cube);
 } /* for i */
 
  Gnode = node_simplify(nodeG, (node_t *) NULL, NODE_SIM_ESPRESSO);
  node_free(nodeG);
  FREE(min_arr);
  return Gnode;
} 
	    
/* frees the storage allocated for the nodes treenode[] after a node of the 
   network has been decomposed  */

/*ARGSUSED*/
static void
free_tree(treenode, root_table, lambda_table, NUM_NODES)
tree_node **treenode;
st_table *root_table, *lambda_table;
int NUM_NODES;
{

  int i;
  char *arg;

  for (i = 0; i < NUM_NODES; i++) {
    FREE(treenode[i]);
  }
  FREE(treenode);
  st_foreach(root_table, k_decomp_free_table_entry, arg); 
  st_free_table(root_table);
  st_foreach(lambda_table, k_decomp_free_table_entry, arg); 
  st_free_table(lambda_table);

}

/*ARGSUSED*/
static
enum st_retval 
k_decomp_free_table_entry(key, value, arg)
  char *key, *value, *arg;
{
  return ST_DELETE;
}

/******************************************************************
  The decomposed node is replaced by the set of ALPHA nodes and
  G node. The network changes as a result.
********************************************************************/
static void
replace_node_decomposed(network, node_to_replace,  NUM_ALPHAS)
network_t *network;
node_t *node_to_replace;
int  NUM_ALPHAS;
{

  int i; 
  
  /* add nodes corresponding to all ALPHA nodes and G node */
  /*-------------------------------------------------------*/
  for (i = 0; i < NUM_ALPHAS; i++) {
      network_add_node(network,  ALPHA[i]);
  };
  FREE(ALPHA);
  node_replace(node_to_replace, G_node);

}

static void
alloc_my_cube(cube, support)
  struct my_cube *cube;
  int  support;
{
    cube->variable = ALLOC(int, support);
}

static void
free_my_cube(cube)
 struct my_cube *cube;

{
   FREE(cube->variable);
}
