/*
 *  merge.c
 *  *******
 *
 *  Copyright (C) 1989  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for merging cells into one unit block
 *	by max cardinality matching method.
 *
 *  functions contained in this file:
 *	merge_node()
 *
 *
 *  Import descriptions:
 *	mis.h		: macros for misII
 *	ULM_int.h	: macros for universal logic module package
 *
 */

#include "mis.h"
#include "pld_int.h"




/*
 *  merge_node()
 *
 *  DESCRIPTION
 *  -----------
 *	Merge two (or more) intermediate nodes in the graph into one ULM node.
 *	Formulate the problem as an integer programming problem 
 *	and solve it by Lindo linear programming package.
 *
 *	This problem should be formulated as a max cardinality matching of a 
 *	general graph, if the # nodes to be merged is 2.
 *	But for the generality of the routine, we use integer programming.
 *	This does not depend on the number of nodes to be merged.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	network		: pointer to a network
 *
 *  Local variables:
 *  ---------------
 *	gen		: generator
 *	cand_node_array	: array of candidate node for merging
 *	fanin_array	: array of pointers to an array of fanin nodes a node
 *	fanin_nodes	: pointer to an array of fanin nodes
 *	fanin1		: pointer to an array of fanin nodes
 *	fanin2		: pointer to an array of fanin nodes
 *	node		: pointer to a node in the network
 *	fanin		: pointer to a fanin node of a node in the network
 *	node1		: pointer to a node in the network
 *	node2		: pointer to a node in the network
 *	intsec		: size of intersection of two sets
 *	union		: size of union of two sets
 *	match1_array	: array of pointers to left side nodes in the mathing
 *	match2_array	: array of pointers to right side nodes in the mathing
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	None
 *
 */


extern void merge_node();
void
merge_node(network, MAX_FANIN, MAX_COMMON_FANIN, MAX_UNION_FANIN, filename, verbose)

network_t	*network;
int  		MAX_FANIN, MAX_COMMON_FANIN, MAX_UNION_FANIN;
char 		*filename;
int 		verbose;
{

    lsGen	gen;
    array_t	*cand_node_array;
    array_t	*fanin_array;
    array_t	*fanin_nodes;
    array_t	*fanin1, *fanin2;
    node_t	*node;
    node_t	*fanin;
    node_t	*node1, *node2;
    int		i, j;
    int		num_intsec;
    int		num_union;
    array_t	*match1_array;
    array_t	*match2_array;
    sm_matrix	*coeff;
    int		col;
    int		num_match;
    char	*merge_file;
    char 	*buf;
    FILE	*mergeFile, *outputFile;
    char 	*lindopathname, *lindo_in, *lindo_out;


  /* search for lindo in the path name. If not found, return */
  /*---------------------------------------------------------*/
  lindopathname = util_path_search("lindo");
  if (lindopathname == NULL) {
	(void) printf("YOU DO NOT HAVE LINDO INTEGER PROGRAMMING PACKAGE\n");
	(void) printf("\tIN YOUR PATH.\n\n");
	(void) printf("EXITING FROM merge.....\n");
	(void) printf("PLEASE OBTAIN LINDO FROM\n\n");

        (void) printf("The Scientific Press, 540 University Ave.\n");
        (void) printf("Palo Alto, CA 94301, USA\n");
        (void) printf("TEL : (415) 322-5221\n");
	return;
  }
  FREE(lindopathname);

    /* Initialization */
    /*----------------*/
    cand_node_array = array_alloc(node_t *, 0);
    fanin_array = array_alloc(array_t *, 0);
    match1_array = array_alloc(node_t *, 0);
    match2_array = array_alloc(node_t *, 0);
    coeff = sm_alloc();

    /* Extract the intermediate nodes with fanins <= MAX_FANIN */
    /*---------------------------------------------------------*/
    foreach_node(network, gen, node) {
	if (node->type != INTERNAL) {
	    continue;
	}
	if (node_num_fanin(node) > MAX_FANIN) {
	    continue;
	}
	array_insert_last(node_t *, cand_node_array, node);
	fanin_nodes = array_alloc(char *, 0);
	array_insert_last(array_t *, fanin_array, fanin_nodes);

	/* Save each fanin of this node and sort them */
	/*--------------------------------------------*/
	foreach_fanin(node, i, fanin) {
	    array_insert_last(char *, fanin_nodes, (char *)fanin);
	}
	array_sort(fanin_nodes, comp_ptr);
    }

    /* Check every mergeable combination of intermediate nodes */
    /*---------------------------------------------------------*/
    for(i = 0, col = 0; i < array_n(cand_node_array); i++) {
	node1 = array_fetch(node_t *, cand_node_array, i);
	fanin1 = array_fetch(array_t *, fanin_array, i);

	for(j = i + 1; j < array_n(cand_node_array); j++) {
	    node2 = array_fetch(node_t *, cand_node_array, j);
	    fanin2 = array_fetch(array_t *, fanin_array, j);

	    count_intsec_union(fanin1, fanin2, &num_intsec, &num_union);
	    if (num_intsec <= MAX_COMMON_FANIN && 
			num_union <= MAX_UNION_FANIN) {
		(void) sm_insert(coeff, i, col);
		(void) sm_insert(coeff, j, col);
		++col;

		/*
		print_fanin(node1);
		print_fanin(node2);
		(void) printf("Node %s and node %s are mergeable\n",
			node_long_name(node1), node_long_name(node2));
		 */
	    }
	}
    }
    if(debug){
	merge_file = ALLOC(char, 100);
	(void)sprintf(merge_file, "/usr/tmp/xln.merge.merge_matrix.XXXXXX");
	(void)mktemp(merge_file);
	mergeFile = fopen(merge_file, "w");
	FREE(merge_file);
	if(mergeFile != NULL) {
	    sm_print(mergeFile, coeff);
	    (void) fclose(mergeFile);

	} else {
	    (void) fprintf(miserr, "No space for mergeFile to be written \n");
	}
    }
	

    /* Solve max cardinality matching problem to get the best merge */
    /*--------------------------------------------------------------*/
    lindo_in =  formulate_Lindo(coeff);
    lindo_out = ALLOC(char, 1000);
    buf       = ALLOC(char, 500);
    (void) sprintf(lindo_out, "/usr/tmp/merge.lindo.out.XXXXXX");
    (void) mktemp(lindo_out);
    (void) sprintf(buf, "lindo < %s > %s", lindo_in, lindo_out);
    (void) system(buf);
    FREE(lindo_in);
    FREE(buf);

    /* Print the results */
    /*-------------------*/
    outputFile = fopen(filename, "w");
    num_match = get_Lindo_result(cand_node_array, coeff, 
		match1_array, match2_array, lindo_out, outputFile);
    FREE(lindo_out);
    if (verbose) (void) printf("Total number of matchings = %d\n\n", num_match);

    (void) fprintf(outputFile, "Merging two CLB's into one CLB\n");
    if (verbose) (void) printf("Merging two CLB's into one CLB\n");
    for(i = 0; i < array_n(match1_array); i++) {
	node1 = array_fetch(node_t *, match1_array, i);
	node2 = array_fetch(node_t *, match2_array, i);
	(void) fprintf(outputFile, "Merge node %s and %s\n", 
			  node_long_name(node1), node_long_name(node2));
	if (verbose) (void) printf("Merge node %s and %s\n", 
			  	    node_long_name(node1), node_long_name(node2));
    }
    (void) fprintf(outputFile, "\n# of CLB's = %d\n", network_num_internal(network) - num_match);
    if (verbose) (void) printf("\n# of CLB's = %d\n", network_num_internal(network) - num_match);
    (void) fclose (outputFile);

    /* Termination */
    /*-------------*/
    array_free(cand_node_array);
    for(j=array_n(fanin_array)-1; j>=0; j--){
	array_free(array_fetch(array_t *, fanin_array, j));
    }
    array_free(fanin_array);
    array_free(match1_array);
    array_free(match2_array);
    sm_free(coeff);

}




/*
 *  Decompose an intermediate nodes in the graph into ULM nodes
 */

void
ACK_decomp(network)

network_t	*network;

{


    lsGen	gen;
    node_t	*node;


    foreach_node(network, gen, node) {
	if (node->type != INTERNAL) {
	    continue;
	}
	/* if (node_num_fanin(node) > MAX_FANIN)  {

	    ACK_decomp_node(network, node);

	}
        */
    }


}



