#include "mis.h"
#include "speed_int.h"
#include "buffer_int.h"

/*
 * Computes the required time at the "pin" input signal of node "fo".
 * This differs from the required time at the node generating the signal
 */
void
sp_buf_req_time(fo, model, pin, reqp, loadp)
node_t *fo;
delay_model_t model;
int pin;
delay_time_t *reqp;	/* RETURN: The required time at the signal */
double *loadp;		/* RETURN: The input cap + wire load cap at input */
{
    pin_phase_t phase;
    double load, auto_route = g_buf_param.auto_route;
    delay_time_t t, req;
    delay_pin_t *pin_delay;

    /*
     * The req time has to be at the signal rather than the gate 
     */
    if (BUFFER(fo)->type == BUF){
	req = BUFFER(fo)->req_time;
	load = BUFFER(fo)->impl.buffer->ip_load;
    } else if (BUFFER(fo)->type == GATE){
	req = BUFFER(fo)->req_time;
	pin_delay = get_pin_delay(fo, pin, model);
	load = pin_delay->load;
    } else{
	if (fo->type == PRIMARY_OUTPUT){
	    req.rise = delay_get_parameter(fo, DELAY_REQUIRED_RISE);
	    req.fall = delay_get_parameter(fo, DELAY_REQUIRED_FALL);
	    if (req.rise == DELAY_VALUE_NOT_GIVEN || req.fall == DELAY_VALUE_NOT_GIVEN){
		req = delay_required_time(fo);
	    }
	    pin_delay = get_pin_delay(fo, pin, model);
	    load = pin_delay->load;
	} else {
	    /* Node in the original network */
	    req = delay_required_time(fo);
	    pin_delay = get_pin_delay(fo, pin, model);
	    phase = pin_delay->phase;
	    t = delay_node_pin(fo, pin, model);
	    sp_compute(phase, &req, t);
	    load = pin_delay->load;
	}
    }
    *reqp = req;
    *loadp = load + auto_route;
}


/*
 * Gets the input capacitance of node. In the case of a gate it will
 * get the input cap of the first pin for now -- Eventually it should
 * get the input cap of most critical input 
 */
double
sp_get_pin_load(node, model)
node_t *node;
delay_model_t model;
{
    double load;
    delay_pin_t *pin_delay;

    if (node->type == PRIMARY_INPUT){
	load = 0;
    } else if (BUFFER(node)->type == BUF){
	load = BUFFER(node)->impl.buffer->ip_load;
    } else {
	pin_delay = get_pin_delay(node, BUFFER(node)->cfi, model);
	load = pin_delay->load;
    }
    return load;
}

/*
 * Returns the drive of the node that fans into nodep. In the case of a
 * gate it will get the drive of the first fanin for now -- Eventually it
 * should get the drive of most critical fanin.  Also returns the phase of
 * selected fanin -- INVERTING || NONINVERTING || NEITHER
 */
delay_time_t
sp_get_input_drive(nodep, model, phase)
node_t *nodep;
delay_model_t model;
pin_phase_t *phase;	/* RETURN: The phase of the selected fanin */
{
    node_t *node;
    delay_time_t drive;
    delay_pin_t *pin_delay;

    if (nodep->type == PRIMARY_INPUT){
	drive.fall = drive.rise = 0.0;
	*phase = PHASE_NONINVERTING;
    } else {
	node = node_get_fanin(nodep, BUFFER(nodep)->cfi);
	if (BUFFER(node)->type == BUF){
	    drive = BUFFER(node)->impl.buffer->drive;
	    *phase = BUFFER(node)->impl.buffer->phase;
	} else {
	    pin_delay = get_pin_delay(node, 0, model);
	    drive = pin_delay->drive;
	    *phase = pin_delay->phase;
	}
    }
    return drive;
}


void
sp_subtract_delay(phase, block, drive, load, req)
pin_phase_t phase;
delay_time_t block, drive, *req;
double load;
{
    delay_time_t delay;

    delay.rise = block.rise + drive.rise * load;
    delay.fall = block.fall + drive.fall * load;

    sp_compute(phase, req, delay);
}

/*
 * Utility routine that will subtract "t" from "req" depending on
 * "phase" -- Used to update required times : "req" is changed in place
 */
void
sp_compute(phase, req, t)
pin_phase_t phase;
delay_time_t *req, t;
{
    switch(phase){
	case PHASE_INVERTING:
	    req->rise -= t.fall;
	    req->fall -= t.rise;
	    break;

	case PHASE_NONINVERTING:
	    req->rise -= t.rise;
	    req->fall -= t.fall;
	    break;

	case PHASE_NEITHER:
	    req->rise -= MAX(t.fall, t.rise);
	    req->fall -= MAX(t.fall, t.rise);
	    break;
	default:
	    break;
    }
}
/*
 * After a redistribution used to update required time data 
 * k is the index of the partition and nodeL is the buffer
 * that was added during the redistribution. Since the required
 * times and caps are known for the fanouts we pass them too.
 */
void
sp_buf_compute_req_time(node, model, req_times, cap_K, k, nodeL)
node_t *node, *nodeL;
delay_model_t model;
delay_time_t *req_times;
double cap_K;
int k;
{
    int i, cfi;
    double load;
    pin_phase_t phase;
    delay_time_t drive, block, t, best;
    delay_pin_t *pin_delay;

    if (node->type == PRIMARY_INPUT || BUFFER(node)->type == GATE){
	cfi = node->type == PRIMARY_INPUT ? 0 : BUFFER(node)->cfi;
	pin_delay = get_pin_delay(node, cfi, model);
	block = pin_delay->block;
	drive = pin_delay->drive;
	phase = pin_delay->phase;
    } else {
	phase = BUFFER(node)->impl.buffer->phase;
	drive = BUFFER(node)->impl.buffer->drive;
	block = BUFFER(node)->impl.buffer->block;
    }

    best.rise = best.fall = POS_LARGE;
    for (i = 0; i < k; i++) {
	best.rise = MIN(best.rise, req_times[i].rise);
	best.fall = MIN(best.fall, req_times[i].fall);
    }
    best.rise = MIN(best.rise, BUFFER(nodeL)->req_time.rise);
    best.fall = MIN(best.fall, BUFFER(nodeL)->req_time.fall);

    load = cap_K + sp_get_pin_load(nodeL, model) + compute_wire_load(node_network(node), 1);
    t.rise = block.rise + drive.rise * load; 
    t.fall = block.fall + drive.fall * load; 
    sp_compute(phase, &best, t);
    BUFFER(node)->req_time = best;
}


delay_time_t
sp_minimum_slack(network)
network_t *network;
{
    lsGen gen;
    node_t *node;
    delay_time_t delay, slack;

    slack.rise = slack.fall = POS_LARGE;

    foreach_primary_input(network, gen, node){
	delay = delay_slack_time(node);
	if (MIN(delay.rise, delay.fall) < MIN(slack.rise, slack.fall)){
	    slack = delay;
	}
    }
    return slack;
}

