#include "mis.h"
#include <math.h>
#include "speed_int.h"
#include "buffer_int.h"

static int sp_compare_buf();
static void sp_insert_buf_in_list();
static void sp_create_noninv_buf_list();

static double sp_inv_load;
static delay_time_t sp_unit_fanout_block ={ 1.0, 1.0};
static delay_time_t sp_unit_fanout_drive ={ 0.2, 0.2};


/*
 * get all the buffers and inverters from the library . If none is 
 * found then returns a dummy buffer with the delay characteristics of
 * the unit fanout model
 */
/* ARGSUSED */
void
buffer_init_globals(lib_present, network, limit, mode, do_decomp, debug)
int lib_present;	/* Generate buffers from the library */
network_t *network;	/* The network we are looking at for buffering */
int limit;		/* Limit for the fanout algorithm */
int mode;		/* Which transformations to carry out */
int do_decomp;		/* Should root node be decomposed */
int debug;		/* Set the level of verbosity */
{
    library_t *lib;

    g_buf_param.auto_route = compute_wire_load(network, 1);
    g_buf_param.mode = mode;
    g_buf_param.do_decomp = do_decomp;
    g_buf_param.debug = debug;
    g_buf_param.glbuftable = st_init_table(st_ptrcmp, st_ptrhash);

    if (lib_present){
	lib = lib_get_library();
	(void)speed_lib_buffers(lib, 2, &(g_buf_param.buffer_list));
	g_buf_param.model = DELAY_MODEL_MAPPED;

    } else {
	(void)speed_lib_buffers(NIL(library_t), 2, &(g_buf_param.buffer_list));
	g_buf_param.model = DELAY_MODEL_UNIT_FANOUT;

    }
    sp_buf_array_from_list();

}

buffer_free_globals()
{
    int i, n = g_buf_param.num_buf;

    /* Delete the buffer structures --- They are in the list and the array */
    for (i = 0; i < n; i++){
	FREE(g_buf_param.buf_array[i]->gate);
	FREE(g_buf_param.buf_array[i]);
    }
    FREE(g_buf_param.buf_array);
    st_free_table(g_buf_param.glbuftable);
    (void)lsDestroy(g_buf_param.buffer_list,(void(*)())0);
}

int
speed_lib_buffers(lib, limit, buffer_list)
library_t *lib;
int limit;
lsList *buffer_list;	/* RETURN: sorted buffer list */
{
    lsGen gen;
    lsGeneric dummy;
    lib_gate_t *gate;
    lib_class_t *class;

    *buffer_list = lsCreate();
    if (lib == NIL(library_t)){
	sp_insert_buf_in_list(*buffer_list, NIL(sp_buffer_t), NIL(lib_gate_t));
	return 1;
    }
    if ((gate = sp_lib_get_inv(lib)) == NIL(lib_gate_t)){
	sp_insert_buf_in_list(*buffer_list, NIL(sp_buffer_t), NIL(lib_gate_t));
	return 1;
    }
    class = lib_gate_class(gate);
    sp_inv_load = limit * ((delay_pin_t *)(gate->delay_info[0]))->load;
    gen = lib_gen_gates(class);
    while (lsNext(gen, &dummy, LS_NH) == LS_OK) {
	gate = (lib_gate_t *)dummy;
	sp_insert_buf_in_list(*buffer_list, NIL(sp_buffer_t), gate);
    }
    (void)lsFinish(gen);

    /* Now we want to get the buffers from the library */
    if ((gate = sp_lib_get_buffer(lib)) != NIL(lib_gate_t)){
	class = lib_gate_class(gate);
	gen = lib_gen_gates(class);
	while (lsNext(gen, &dummy, LS_NH) == LS_OK) {
	    gate = (lib_gate_t *)dummy;
	    sp_insert_buf_in_list(*buffer_list, NIL(sp_buffer_t), gate);
	}
	(void)lsFinish(gen);
    }
    /* Sort the inverters in list -- smallest last */
    (void)lsSort(*buffer_list, sp_compare_buf);
    sp_create_noninv_buf_list(*buffer_list, 2);
    (void)lsSort(*buffer_list, sp_compare_buf);

    if(g_buf_param.debug > 10){
	speed_dump_buffer_list(misout, *buffer_list);
    }
    return 0;
}


void
speed_dump_buffer_list(fp, buffer_list)
FILE *fp;
lsList buffer_list;
{
    lsGen gen;
    char *dummy;
    sp_buffer_t *buffer;
    double d1, d2;

    (void)fprintf(fp,"type        name       area ipcap  blck_r blck_f drve_r drve_f rise   fall\n");
    (void)fprintf(fp,"--------------------------------------------------------------------------\n");
    gen = lsStart(buffer_list);
    while (lsNext(gen, &dummy, LS_NH) == LS_OK) {
	buffer = (sp_buffer_t *)dummy;
	d1 = buffer->block.rise + sp_inv_load * buffer->drive.rise;
	d2 = buffer->block.fall + sp_inv_load * buffer->drive.fall;
	(void)fprintf(fp,"%s    %-14s %5.1f %-6.3f %-6.3f %-6.3f %-6.3f %-6.3f %-6.3f %-6.3f\n",
	(buffer->phase == PHASE_INVERTING? "INV" : "BUF"),
	sp_buffer_name(buffer), buffer->area,buffer->ip_load,
	buffer->block.rise, buffer->block.fall,
	buffer->drive.rise, buffer->drive.fall,
	d1, d2);
    }
    (void)lsFinish(gen);
}

/* Sort the buffers according to size -- hope that the smallest is slowest */
static int
sp_compare_buf(item1, item2)
lsGeneric item1, item2;
{
    sp_buffer_t *buf1, *buf2;

    buf1 = (sp_buffer_t *)item1;
    buf2 = (sp_buffer_t *)item2;

    if (buf1->phase == buf2->phase){
	if (buf2->area < buf1->area) return -1;
	else if (buf2->area > buf1->area) return 1;
	else return 0;
    } else {
	if (buf1->phase == PHASE_INVERTING){
	    return -1;
	} else {
	    return 1;
	}
    }


}
/*
 * When entering the routine the only buffers are single inverters --
 * On exiting the buffer chains have been included 
 * The input list is sorted so that the biggest buffer is at the beginning
 */
static void
sp_create_noninv_buf_list(buffer_list, limit)
lsList buffer_list;
int limit;		/* is the nghd. of the search */
{
    int count;
    lsGen gen, gen1;
    lsList new_list;
    lsHandle handle;
    lsGeneric data, dummy;
    sp_buffer_t *buffer, *buffer1;

    new_list = lsCreate();
    gen = lsStart(buffer_list);
    while (lsNext(gen, &dummy, &handle) == LS_OK) {
	buffer = (sp_buffer_t *)dummy;
	if (buffer->phase == PHASE_INVERTING){
	    sp_insert_buf_in_list(new_list, buffer, buffer->gate[0]);
	    count = 1;
	    gen1 = lsGenHandle(handle, &data, LS_AFTER);
	    while (lsNext(gen1, &dummy, LS_NH) == LS_OK){
		buffer1 = (sp_buffer_t *)dummy;
		if (buffer1->phase == PHASE_INVERTING){
		    sp_insert_buf_in_list(new_list, buffer1, buffer->gate[0]);
		}
		if (++count > limit) break;
	    }
	    (void)lsFinish(gen1);
	}
    }
    (void)lsFinish(gen);

    gen = lsStart(new_list);
    while (lsNext(gen, &dummy, LS_NH) == LS_OK){
	buffer = (sp_buffer_t *)dummy;
	(void)lsNewBegin(buffer_list, (lsGeneric)buffer, LS_NH);
    }
    (void)lsFinish(gen);
    (void)lsDestroy(new_list,(void(*)())0);
}


/*
 * Generates the array of buffers that we need to use  -- 
 * The buf_list is sorted (inverters come earlier) and so is the buf_array
 */
void
sp_buf_array_from_list()
{
    int i = 0;
    lsGen gen;
    lsGeneric dummy;
    sp_buffer_t *buffer;
    lsList buf_list = g_buf_param.buffer_list;

    g_buf_param.num_inv = 0;
    g_buf_param.num_buf = 0;
    g_buf_param.buf_array = ALLOC(sp_buffer_t *, lsLength(buf_list));
    gen = lsStart(buf_list);
    while (lsNext(gen, &dummy, LS_NH) == LS_OK) {
	buffer = (sp_buffer_t *)dummy;
	if (buffer->phase == PHASE_NONINVERTING){
	    g_buf_param.num_buf++;
	} else {
	    g_buf_param.num_inv++;
	}
	g_buf_param.buf_array[i++] = buffer;
    }
    g_buf_param.num_buf += g_buf_param.num_inv;
    (void)lsFinish(gen);
}

/*
 * Sets "gate" as the impl of "node = fanin';"
 * gate must be an inverter in the library 
 */
void
sp_set_inverter(node, fanin, gate)
node_t *node, *fanin;
lib_gate_t *gate;
{
    char *inv_formals[1];
    node_t *inv_actuals[1];

    inv_actuals[0] = fanin;
    inv_formals[0] = lib_gate_pin_name(gate, 0, 1);
    (void)lib_set_gate(node, gate, inv_formals ,inv_actuals, 1);
}

/*
 * Will take the gate implementing "node" and get the extreme gate
 * of that class and fix that as the implementation of "newnode". In the
 * case that "node" has its impl set as a BUF we will set the impl to
 * be the smallest buffer.
 */
/* ARGSUSED */
void
sp_buf_annotate_gate(node, model)
node_t *node;
delay_model_t model;
{
    int cfi;
    lib_gate_t *gate;
    pin_phase_t phase;
    delay_pin_t *pin_delay;
    delay_time_t t, desired, drive;
    sp_buffer_t **buf_array = g_buf_param.buf_array;

    if ((gate = lib_gate_of(node)) == NIL(lib_gate_t)) {
	if (BUFFER(node)->type == NONE){
	    BUFFER(node)->type = BUF;
	    BUFFER(node)->impl.buffer = buf_array[g_buf_param.num_buf -1];
	}
    } else {
	BUFFER(node)->type = GATE;
	BUFFER(node)->impl.gate = gate;
    }
    cfi = sp_buf_get_crit_fanin(node, model);
    pin_delay = get_pin_delay(node, cfi, model);
    phase = pin_delay->phase;
    t = delay_node_pin(node, cfi, model);
    desired = delay_required_time(node);
    sp_compute(phase, &desired, t);
    BUFFER(node)->cfi = cfi;
    BUFFER(node)->req_time = desired;	/* req time at cfi input of node */

    drive = sp_get_input_drive(node, model, &phase);
    BUFFER(node)->prev_dr = drive;
    BUFFER(node)->prev_ph = phase;
    return;
}

lib_gate_t *
buf_get_gate_version(gate, version)
lib_gate_t *gate;
int version;
{
    lsGen gen;
    char *dummy;
    int num_gates;
    lib_gate_t *root_gate;

    num_gates = 0; root_gate = NIL(lib_gate_t);
    gen = lib_gen_gates(lib_gate_class(gate));
    while (lsNext(gen, &dummy, LS_NH) == LS_OK){
	if (num_gates++ == version) {
	    root_gate = (lib_gate_t *)dummy;
	    break;
	}
    }
    (void)lsFinish(gen);

    return root_gate;
}
/*
 * Look at the fanins of node and find the index that is most
 * critical. This index gives us the fanin such that if we decrease
 * the required time at that fanin -- we would most likely decrease the
 * delay thru the entire netw
 */
int 
sp_buf_get_crit_fanin(node, model)
node_t *node;
delay_model_t model;
{
    int i, cfi;
    node_t *fanin;
    double fom, fom_b;
    delay_pin_t *pin_delay;
    pin_phase_t phase;
    delay_time_t arrival, req, delay, act;

    if (node_num_fanin(node) == 1) return 0;

    fom_b = POS_LARGE;
    req = delay_required_time(node);
    foreach_fanin(node, i, fanin){
	/* act is the required time at the i_th fanin signal */
	act = req;
	pin_delay = get_pin_delay(node, i, model);
	phase = pin_delay->phase;
	delay = delay_node_pin(node, i, model);
	sp_compute(phase, &act, delay);

	arrival = delay_arrival_time(fanin);

	/*
	 * figure of merit -- slack value at the signal
	 */
	fom = MIN(act.rise - arrival.rise, act.fall - arrival.fall);
	if (fom < fom_b){
	   fom_b = fom; cfi = i;
	}
    }
    return cfi;
}

/*
 * Replace gate by pref_gate as the implementation of "node" 
 */
void
sp_replace_lib_gate(node, gate, pref_gate)
node_t *node;
lib_gate_t *gate, *pref_gate;
{
    int i, nin;
    char **formals;
    node_t **actuals, *fanin;

    assert(lib_gate_class(gate) == lib_gate_class(pref_gate));
    nin = node_num_fanin(node);
    formals = ALLOC(char *, nin);
    actuals = ALLOC(node_t *, nin);
    foreach_fanin(node, i, fanin){
	formals[i] = lib_gate_pin_name(gate, i, 1);
	actuals[i] = fanin;
    }
    (void)lib_set_gate(node, pref_gate, formals, actuals, nin);
    FREE(formals);
    FREE(actuals);
}


/*
 * Determine the name of the implementation of the node ---
 * Useful only for debugging the buffering routines 
 * Successive calls to this routine use the same storage and
 * so to save the name need to strsav the returned string
 */
char *
sp_name_of_impl(node)
node_t *node;
{
    static char name[BUFSIZ];
    if (node == NIL(node_t)){
	(void)strcpy(name,"--NONE--");
    } else if (node->type == PRIMARY_INPUT){
	(void)strcpy(name,"NODE_PI");
    } else if (BUFFER(node)->type == BUF){
	(void)strcpy(name, sp_buffer_name(BUFFER(node)->impl.buffer));
    } else if (BUFFER(node)->type == GATE){
	(void)strcpy(name, lib_gate_name(BUFFER(node)->impl.gate));
    } else {
	(void)strcpy(name,"-NONE-");
    }
    return name;
}

/*
 * Returns a contrived name for the buffer implementing the node --
 * The name is  of the form "inv1-inv2"  -- uses a static storage too
 */
char *
sp_buffer_name(buffer)
sp_buffer_t *buffer;
{
    int i;
    static char name[BUFSIZ];

    if (buffer == NIL(sp_buffer_t)) (void)strcpy(name,"NONE");
    if (buffer->gate[0] == NIL(lib_gate_t)){ (void)strcpy(name,"UNIT_FAN");
    } else {
	(void)strcpy(name,lib_gate_name(buffer->gate[0]));
	for (i = 1; i < buffer->depth; i++){
	    (void)strcat(name, "-");
	    (void)strcat(name, lib_gate_name(buffer->gate[i]));
	}
    }
    return name;
}


/*
 * Creates a buffer that is buffer "prev" followed by the gate "gate"
 * and inserts the buffer just created into the lsList "buffer_list"
 * Note: "prev" can be a NULL and so can gate
 */
static void
sp_insert_buf_in_list(buffer_list, prev, gate)
lsList buffer_list;
sp_buffer_t *prev;
lib_gate_t *gate;
{
    int i;
    delay_time_t block;
    sp_buffer_t *buffer;
    double load, auto_route = g_buf_param.auto_route;

    buffer = ALLOC(sp_buffer_t,1);
    if (prev == NIL(sp_buffer_t)){
	buffer->depth = 1;
	buffer->gate = ALLOC(lib_gate_t *, 1);
	buffer->gate[0] = gate;
	if (gate != NIL(lib_gate_t)){
	    buffer->area = lib_gate_area(gate);
	    /* Single fanin gates only -- use 0 fanin */
	    buffer->phase = ((delay_pin_t *)(gate->delay_info[0]))->phase;
	    buffer->block = ((delay_pin_t *)(gate->delay_info[0]))->block;
	    buffer->drive = ((delay_pin_t *)(gate->delay_info[0]))->drive;
	    buffer->ip_load = ((delay_pin_t *)(gate->delay_info[0]))->load;
	} else {
	    buffer->area = 1.0;
	    buffer->phase =  PHASE_NONINVERTING;
	    buffer->block = sp_unit_fanout_block;
	    buffer->drive = sp_unit_fanout_drive;
	    buffer->ip_load = 1.0;
	}
    } else {
	buffer->depth = prev->depth + 1;
	buffer->gate = ALLOC(lib_gate_t *,buffer->depth);
	for (i = 0; i < prev->depth; i++){
	    buffer->gate[i] = prev->gate[i];
	}
	buffer->gate[prev->depth] = gate;
	/*
	 * We only chain inverters -- hence dont bother to compute the
	 * phase of the chain.  Assume it is NONINVERTING
	 */
	buffer->phase = PHASE_NONINVERTING;
	buffer->area = prev->area + lib_gate_area(gate);
	load = auto_route + ((delay_pin_t *)(gate->delay_info[0]))->load;
	block = ((delay_pin_t *)(gate->delay_info[0]))->block;
	buffer->block.rise = prev->block.fall + prev->drive.fall * load + block.rise;
	buffer->block.fall = prev->block.rise + prev->drive.rise * load + block.fall;
	buffer->drive = ((delay_pin_t *)(gate->delay_info[0]))->drive;
	buffer->ip_load = prev->ip_load;
    }
    (void)lsNewBegin(buffer_list, (lsGeneric)buffer, LS_NH);
}

/* 
 * Free the storage associated with the buffers in "list" and
 * then destroy the list.   Garbage collection function
 */
void
speed_free_buffer_list(list)
lsList list;
{
    lsGen gen;
    lsGeneric dummy;

    gen = lsStart(list);
    while (lsNext(gen, &dummy, LS_NH) == LS_OK) {
	FREE(((sp_buffer_t *)dummy)->gate);
	FREE(dummy);
    }
    (void)lsDestroy(list,(void(*)())0);
    (void)lsFinish(gen);
}

/*
 * If the node is implemented as a buffer then replace the node by the
 * gates that make up the buffer... Even thought there are more efficient
 * implemetations this particular one is chosen since it retains the 
 * node as the root of the chain (closest to input )
 */
void
sp_implement_buffer_chain(network, node)
network_t *network;
node_t *node;
{
    lsGen gen;
    int n, i, j;
    node_t *temp, *fanin, **fanouts, *fo;
    sp_buffer_t *buffer;

    if (node == NIL(node_t) || node->type == PRIMARY_INPUT) return;
    if (BUFFER(node)->type == BUF){
	buffer = BUFFER(node)->impl.buffer;	
	if (buffer->gate[0] != NIL(lib_gate_t)){
	    fanin = node_get_fanin(node, 0);
	    if (buffer->depth == 1){
		sp_set_inverter(node, fanin, buffer->gate[0]);
		BUFFER(node)->type = GATE;
		BUFFER(node)->cfi = 0;
		BUFFER(node)->impl.gate = buffer->gate[0];
	    } else {
		n = node_num_fanout(node);
		fanouts = ALLOC(node_t *, n);
		i = 0;
		foreach_fanout(node, gen, fo){
		    fanouts[i++] = fo;
		}
		sp_set_inverter(node, fanin, buffer->gate[0]);
		BUFFER(node)->type = GATE;
		BUFFER(node)->cfi = 0;
		BUFFER(node)->impl.gate = buffer->gate[0];
		fanin = node;
		for (j = 1; j < buffer->depth; j++){
		    temp = node_alloc();
		    sp_set_inverter(temp, fanin, buffer->gate[j]);
		    network_add_node(network, temp);
		    fanin = temp;
		    BUFFER(temp)->type = GATE;
		    BUFFER(temp)->cfi = 0;
		    BUFFER(temp)->impl.gate = buffer->gate[j];
		    if (j == buffer->depth - 1){
			/* patch the fanouts of the original node to fanin */
			for (i = 0; i < n; i++){
			    assert(node_patch_fanin(fanouts[i], node, fanin));
			}
		    }
		}
		FREE(fanouts);
	    }
	}
    }
}

/*
 * Add the implementation of the node to be gate
 */
void
buf_add_implementation(node, gate)
node_t *node;
lib_gate_t *gate;
{
    int i, nin;
    char **formals;
    node_t **actuals, *fanin;

    nin = node_num_fanin(node);
    formals = ALLOC(char *, nin);
    actuals = ALLOC(node_t *, nin);
    foreach_fanin(node, i, fanin){
	formals[i] = lib_gate_pin_name(gate, i, 1);
	actuals[i] = fanin;
    }
    (void)lib_set_gate(node, gate, formals, actuals, nin);

    BUFFER(node)->type = GATE;
    BUFFER(node)->impl.gate = gate;
    BUFFER(node)->cfi = 0;

    FREE(formals);
    FREE(actuals);
}

/*
 * Some routines to allow users to set/query the stuff in the BUFFER()
 * field of each node and to query the global defaults.
 */

delay_model_t
buf_get_model()
{
    return (g_buf_param.model);
}

double
buf_get_auto_route()
{
    return (g_buf_param.auto_route);
}

delay_time_t
buf_get_required_time_at_input(node)
node_t *node;
{
    return (BUFFER(node)->req_time);
}

void
buf_set_required_time_at_input(node, req)
node_t *node;
delay_time_t req;
{
    BUFFER(node)->req_time = req;
}

void
buf_set_prev_drive(node, drive)
node_t *node;
delay_time_t drive;
{
    BUFFER(node)->prev_dr = drive;
}

void
buf_set_prev_phase(node, phase)
node_t *node;
pin_phase_t phase;
{
    BUFFER(node)->prev_ph = phase;
}



/*
 * node daemon to allocate the storage with every node 
 */
void
buffer_alloc(node)
node_t *node;
{
    node->BUFFER_SLOT = (char *) ALLOC(sp_node_t, 1);
    BUFFER(node)->type  = NONE;
    BUFFER(node)->cfi = -1;
}

/*
 * node daemon to free the memory associated with the BUFFER_SLOT field 
 */
void
buffer_free(node)
node_t *node;
{
    FREE(node->BUFFER_SLOT);
}
