#include "mis.h"
#include "speed_int.h"
#include "buffer_int.h"
#include <math.h>

int weight_debug; 	      /* debug flags */
int speed_debug;
int sp_add_inv;
int speed_repeat;
int speed_area_reclaim;	      /* If set, reclaim the area -- default */
double speed_thresh;          /* defines the critical threshold */
st_table *speed_plot_table;   /* Used in speed_plot to store plots */
g_buffer_param_t g_buf_param; /* Parameters for the buffering algo */
speed_delay_global_t sp_glb_delay; /* Parameters for the delay tracing */

int
com_speed_up(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int i, c, dist;
    int speed_trace, only_init_decomp, do_script;
    double latest, coeff, thresh;
    delay_model_t model;
    node_t *np;
    array_t *nodevec;

    getopt_reset();
    error_init();

    speed_area_reclaim = TRUE;
    speed_trace = FALSE;
    do_script = TRUE;
    speed_repeat = TRUE;
    only_init_decomp = FALSE;
    weight_debug = FALSE;
    speed_debug = FALSE;
    sp_add_inv = FALSE;
    thresh = DEFAULT_SPEED_THRESH;
    coeff = DEFAULT_SPEED_COEFF;
    dist = DEFAULT_SPEED_DIST;
    model = DELAY_MODEL_UNIT_FANOUT;
    while ((c = getopt(argc, argv, "d:m:w:t:cipTDvr")) != EOF) {
        switch (c) {
        case 'p':
	    sp_add_inv = TRUE;
            break;
        case 'r':
	    speed_area_reclaim = FALSE;
            break;
        case 'c':
            speed_repeat = FALSE;
	    do_script = FALSE;
            break;
        case 'i':
            only_init_decomp = TRUE;
	    do_script = FALSE;
            break;
        case 'T':
            speed_trace = TRUE;
            break;
        case 'v':
            speed_debug = TRUE;
            break;
        case 'D':
            speed_debug = TRUE;
            break;
        case 'w':
            coeff = atof(optarg);
	    if (coeff < 0) coeff = 0;
	    if (coeff > 1) coeff = 1;
            break;
        case 't':
            thresh = atof(optarg);
	    do_script = FALSE;
            break;
        case 'd':
            dist = atoi(optarg);
	    do_script = FALSE;
            break;
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
            	goto speed_usage;
            }
            break;
        default:
            goto speed_usage;
        }
    }

    /* First check if the network is a valid one */
    nodevec = com_get_nodes(*network, argc-optind+1, argv+optind-1);
    if (array_n(nodevec) == 0){
        array_free(nodevec);
	return 0;
    }

    /* If mapped model - make sure library is present */
    if (model == DELAY_MODEL_MAPPED && (lib_get_library() == NIL(library_t))){
	(void) fprintf(misout, "Need to define a library\n");
        array_free(nodevec);
	return 1;
    }
    /* Initialize the globals for the delay routines in this package */
    speed_init_globals(model, 0 /* No library acceleration */);

    /* If a list of nodes is present - decomp only those */
    if (argc - optind != 0) {
        for (i = 0; i < array_n(nodevec); i++) {
            np = array_fetch(node_t *, nodevec, i);
            if (np->type == INTERNAL) {
		speed_up_node(*network, np, coeff, model, 0);
            }
        }
    } else if (do_script){
	if (speed_init_decomp(*network, model, coeff, speed_area_reclaim) ){
	    return 1;
	}
	speed_up_script(network,  model, coeff, speed_trace);
    } else if (only_init_decomp) {
	if (speed_init_decomp(*network, model, coeff, speed_area_reclaim) ){
	    return 1;
	}
    } else if (speed_repeat) {
	if (speed_init_decomp(*network, model, coeff, speed_area_reclaim) ){
	    return 1;
	}
        speed_up_loop(network, thresh, coeff, dist, model, speed_trace);
    } else {
	if (speed_init_decomp(*network, model, coeff, speed_area_reclaim) ){
	    return 1;
	}
        if (delay_trace(*network, model) ){
            (void) delay_latest_output(*network, &latest);
            (void) fprintf(misout,"delay  %5.2f ->  ",latest);
	}
        speed_up_network(*network, thresh, coeff, dist, model);
        if (delay_trace(*network, model) ){
            (void) delay_latest_output(*network, &latest);
            (void) fprintf(misout,"%5.2f\n",latest);
        }
    }

    array_free(nodevec);
    return 0;

speed_usage:
    (void) fprintf(miserr, "usage: speed_up [-ciprvDT] [-d n] [-w n] [-t n.n] [-m model] [node-list]\n");
    (void) fprintf(miserr, "    -c\t\tOnly one pass of speed_up.\n");
    (void) fprintf(miserr, "    -i\t\tOnly do the initial 2_input NAND decomp.\n");
    (void) fprintf(miserr, "    -p\t\tUse explicit inverters (EXPERIMENTAL).\n");
    (void) fprintf(miserr, "    -r\t\tDo not try to reclaim area (EXPERIMENTAL).\n");
    (void) fprintf(miserr, "    -D,-v\t\tDebugging option.\n");
    (void) fprintf(miserr, "    -T\t\tPrint the delays as the speed_up proceeds.\n");
    (void) fprintf(miserr, "    -d\tn\tDistance for collapsing (default = 3).\n");
    (void) fprintf(miserr, "    -w\tn\tRelative weight of area (default = 0).\n");
    (void) fprintf(miserr, "    -t\tn.n\tCritical threshold.\n");
    (void) fprintf(miserr, "    -m\tmodel\tDelay model (default \"unit-fanout\").\n");
    return 1;
}

int
com_buffer_opt(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    lsGen gen;
    node_t *node;
    array_t *nodevec;
    double load, thresh;
    delay_time_t drive;
    st_table *drive_table, *load_table;
    int setting_env, debug, do_decomp, limit, mode, c, trace, single_pass;

    trace = FALSE;
    single_pass = FALSE;
    do_decomp = FALSE;
    debug = FALSE;
    limit = 2;
    mode = -1;
    thresh = 0.5;
    getopt_reset();
    error_init();
    while ((c = getopt(argc, argv, "cdTDf:v:l:t:")) != EOF) {
        switch (c) {
        case 't':
            thresh = atof(optarg);
            break;
	case 'd':
	    do_decomp = TRUE;
	    break;
	case 'c':
	    single_pass = TRUE;
	    break;
	case 'T':
	    trace = TRUE;
	    break;
	case 'f':
	    mode = atoi(optarg);
	    break;
	case 'v':
	    debug = atoi(optarg);
	    break;
	case 'l':
	    limit = atoi(optarg);
	    break;
	case 'D':
	    debug = TRUE;
	    break;
        default:
            goto buffer_opt_usage;
        }
    }

    /* First check if the network is a valid one */
    nodevec = com_get_nodes(*network, argc-optind+1, argv+optind-1);
    if (array_n(nodevec) == 0 ){
        array_free(nodevec);
	return 0;
    }

    /* Check the range of buffering modes (1-7) */
    if (mode < 0){
	/* No value specified -- try all transformations */
	mode = 7;
    } else if (mode < 1 || mode > 7){
	(void)fprintf(misout,"ERROR: Valid range of -f option is 1 to 7\n");
        array_free(nodevec);
	return 1;
    }

    /* Set the global variables -- buffers, flags controlling algorithm */
    if (lib_network_is_mapped(*network)){
	buffer_init_globals(1, *network, limit, mode, do_decomp, debug);
    } else {
	buffer_init_globals(0, *network, limit, mode, do_decomp, debug);
	(void)fprintf(misout,"Network not mapped: Buffering will use the unit-fanout model\n");
    }

    /*
     * Set drive and load of Primary IP/OP respectively to be that of smallest
     * inverter in the library
     */
    load_table = st_init_table(st_ptrcmp, st_ptrhash);
    drive_table = st_init_table(st_ptrcmp, st_ptrhash);

    setting_env = FALSE;
    if (g_buf_param.model == DELAY_MODEL_MAPPED){
	load = (g_buf_param.buf_array[g_buf_param.num_inv-1])->ip_load;
    } else {
	load = 1.0;
    }
    foreach_primary_output(*network, gen, node){
	if (delay_get_parameter(node, DELAY_OUTPUT_LOAD) < 0){
	    setting_env = TRUE;
	    (void)st_insert(load_table, (char *)node, "");
	    delay_set_parameter(node, DELAY_OUTPUT_LOAD, load);
	}
    }
    if (setting_env){
	(void)fprintf(misout,"WARNING: Unspecified primary output loads assumed to be %6.4f\n", load); 
    }

    setting_env = FALSE;
    if (g_buf_param.model == DELAY_MODEL_MAPPED){
	drive = g_buf_param.buf_array[g_buf_param.num_inv-1]->drive;
    } else {
	drive.rise = drive.fall = 0.2;
    }
    foreach_primary_input(*network, gen, node){
	if (delay_get_parameter(node, DELAY_DRIVE_RISE) < 0 ||
		delay_get_parameter(node, DELAY_DRIVE_FALL) < 0){
	    setting_env = TRUE;
	    (void)st_insert(drive_table, (char *)node, "");
	    delay_set_parameter(node, DELAY_DRIVE_RISE, drive.rise);
	    delay_set_parameter(node, DELAY_DRIVE_FALL, drive.fall);
	}
    }
    if (setting_env){
	(void)fprintf(misout,"WARNING: Unspecified primary input drive assumed to be %6.4f\n", drive); 
    }

    /* Annotate the delay data */
    if (!delay_trace(*network, g_buf_param.model)){
	(void)fprintf(misout,"%s\n", error_string());
	return 1;
    }


    if (argc > optind) {
	/* Buffer the output of specified nodes only */
	(void)speed_buffer_array_of_nodes(*network, nodevec);
    } else {
	(void)speed_buffer_network(*network, trace, thresh, single_pass);
    }

    /* Reset the drive and loads of the Primary IP/OP respectively */
    foreach_primary_output(*network, gen, node){
	if (st_is_member(load_table, (char *)node)){
	    delay_set_parameter(node, DELAY_OUTPUT_LOAD, DELAY_NOT_SET);
	}
    }
    foreach_primary_input(*network, gen, node){
	if (st_is_member(drive_table, (char *)node)){
	    delay_set_parameter(node, DELAY_DRIVE_RISE, DELAY_NOT_SET);
	    delay_set_parameter(node, DELAY_DRIVE_FALL, DELAY_NOT_SET);
	}
    }

    st_free_table(drive_table);
    st_free_table(load_table);
    buffer_free_globals();
    array_free(nodevec);

    return 0;

buffer_opt_usage:
    (void)fprintf(misout,"Usage: buffer_opt [-l #] [-c] [-d] [-T] [-D] [-f #] [-v #] [-t #.#] [node_list]\n");
    (void)fprintf(misout,"\t-l \t#\t: Fanout limit (default = 2)\n");
    (void)fprintf(misout,"\t-c \t\t: Buffer only one node in the pass\n");
    (void)fprintf(misout,"\t-d \t\t: Docompose complex gates as part of buffering\n");
    (void)fprintf(misout,"\t-T \t\t: Print the delay as algorithm proceeds\n");
    (void)fprintf(misout,"\t-t \t#.#\t: Threshold for critical paths\n");
    (void)fprintf(misout,"\t-D \t\t: Debugging option\n");
    (void)fprintf(misout,"\t-f \t#\t: Fanout transformations to use (default = 7)\n");
    (void)fprintf(misout,"\t-v \t#\t: Verbosity level (10-20)\n");
    return 1;
}

int
com__reduce_delay(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int flag, c;
    double last, req_delay, percent;
    double init_delay, init_area;
    double final_delay, final_area;
    node_t *po;
    lsGen gen;
    extern void speed_get_stats();
    delay_model_t model;
    library_t *lib;

    percent = 0.0;
    flag = FALSE;
    model = DELAY_MODEL_UNIT_FANOUT;
    getopt_reset();
    error_init();
    while ((c = getopt(argc, argv, "vp:m:")) != EOF) {
        switch (c) {
	case 'v':
	    flag = TRUE;
	    break;
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
            	goto reduce_delay_usage;
            }
            break;
        case 'p':
            percent = atof(optarg);
            break;
        default:
            goto reduce_delay_usage;
        }
    }
    /*
     * Find the current delay and reduce it
     * by percent percent. Report the area and the
     * delay at that stage. Area after mapping .
     */

    if ((lib = lib_get_library()) == NIL(library_t)) {
	(void)fprintf(misout, "Please specify library with read_library ...\n");
	return 1;
    }
    /* Initialize the globals for the delay routines in this package */
    speed_init_globals(model, 0 /* No library acceleration */);

    if (!delay_trace(*network, model)){
	(void)fprintf(misout,"%s\n", error_string());
	return 1;
    }
    (void) delay_latest_output(*network, &last);
    req_delay = last * (1.0 - percent/100.0);

    speed_get_stats(*network, lib, model, &init_delay, &init_area);
    /* Set the desired po required times */
    foreach_primary_output(*network, gen, po){
	delay_set_parameter(po, DELAY_REQUIRED_FALL, req_delay);
	delay_set_parameter(po, DELAY_REQUIRED_FALL, req_delay);
    }

    if (flag) {
    	(void) com_execute(network, "speed_up -T");
    } else {
    	(void) com_execute(network, "speed_up");
    }

    speed_get_stats(*network, lib, model, &final_delay, &final_area);
    (void) fprintf(misout,"\t\tArea  %5.1f -> %-5.1f\n",
	    init_area, final_area);
    (void) fprintf(misout,"\t\tDelay %5.1f -> %-5.1f\n",
	    init_delay, final_delay);

    foreach_primary_output(*network, gen, po){
	delay_set_parameter(po, DELAY_REQUIRED_FALL, DELAY_VALUE_NOT_GIVEN);
	delay_set_parameter(po, DELAY_REQUIRED_FALL, DELAY_VALUE_NOT_GIVEN);
    }

    return 0;
reduce_delay_usage:
    (void) fprintf(miserr, "usage: _reduce_delay [-m model] [-p n.n]\n");
    (void) fprintf(miserr, "    -p\tn.n\tPercent improvement in speed desired.\n");
    (void) fprintf(miserr, "    -m\tmodel\tDelay model. default (\"unit-fanout\").\n");
    return 1;
}


int
com__part_collapse(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    node_t *node;
    double thresh;
    int i, dist, c;
    array_t *nodevec;
    delay_model_t model;

    speed_debug = FALSE;
    dist = DEFAULT_SPEED_DIST;
    thresh = DEFAULT_SPEED_THRESH;
    getopt_reset();
    error_init();
    model = DELAY_MODEL_UNIT_FANOUT;

    while ((c = getopt(argc, argv, "Dm:t:d:")) != EOF) {
        switch (c) {
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
            	goto part_collapse_usage;
            }
            break;
        case 'D':
            speed_debug = TRUE;
            break;
        case 't':
            thresh = atof(optarg);
            break;
        case 'd':
            dist = atoi(optarg);
            break;
        default:
            goto part_collapse_usage;
        }
    }

    /* Initialize the globals for the delay routines in this package */
    speed_init_globals(model, 0 /* No library acceleration */);

    if (argc - optind != 0) {
	if (!delay_trace(*network, model)){
	    (void)fprintf(misout,"%s\n", error_string());
	    return 1;
	}
	set_speed_thresh(*network, thresh);
        nodevec = com_get_nodes(*network, argc-optind+1, argv+optind-1);
        for(i = 0; i < array_n(nodevec); i++){
            node = array_fetch(node_t *, nodevec, i);
            speed_absorb(node, dist);
        }
        array_free(nodevec);
    }
    return 0;

part_collapse_usage:
    (void) fprintf(miserr, "usage: _part_collapse [-m model] [-d n] [-t n.n] list-of-nodes\n");
    (void) fprintf(miserr, "    -d\tn\tDistance for collapsing.\n");
    (void) fprintf(miserr, "    -t\tn.n\tCritical threshold.\n");
    (void) fprintf(miserr, "    -m\tmodel\tDelay model. default (\"unit-fanout\").\n");
    return 1;
}


int
com__print_cutset(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, dist;
    array_t *mincut;
    st_table *table;
    delay_model_t model;
    double thresh, coeff;

    model = DELAY_MODEL_UNIT_FANOUT;
    thresh = DEFAULT_SPEED_THRESH;
    dist = DEFAULT_SPEED_DIST;
    coeff = DEFAULT_SPEED_COEFF;
    speed_debug = TRUE;
    weight_debug = FALSE;

    getopt_reset();
    error_init();
    while ((c = getopt(argc, argv, "w:m:d:t:")) != EOF) {
        switch (c) {
        case 'd':
            dist = atof(optarg);
            break;
        case 'w':
            coeff = atof(optarg);
	    if (coeff < 0) coeff = 0;
	    if (coeff > 1) coeff = 1;
            break;
        case 't':
            thresh = atof(optarg);
            break;
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
            	goto print_cutset_usage;
            }
            break;
        default:
            goto print_cutset_usage;
        }
    }

    /*
     * do a delay trace through the network and
     * set the global variable speed_thresh
     */

    if (!delay_trace(*network, model)){
	(void)fprintf(misout,"%s\n", error_string());
	return 1;
    }
    set_speed_thresh(*network, thresh);
    table = speed_compute_weight(*network, model, coeff, dist, NIL(array_t));

    /* compute the minimum weighted cutset of the critical paths */
    (void) com_execute(network, "_maxflow");
    mincut = cutset(*network, table);
    array_free(mincut);
    (void) com_execute(network, "_maxflow");

    st_free_table(table);
    return 0;

print_cutset_usage:
    (void) fprintf(miserr, "usage: _print_cutset [-t n.n] [-d n] [-w n] [-m model]\n");
    (void) fprintf(miserr, "    -d\tn\tDistance for collapsing.\n");
    (void) fprintf(miserr, "    -w\tn\tRelative weight of area.\n");
    (void) fprintf(miserr, "    -t\tn.n\tCritical threshold.\n");
    (void) fprintf(miserr, "    -m\tmodel\tDelay model. default (\"unit-fanout\").\n");
    return 1;
}


int
com__print_weight(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, dist;
    delay_model_t model;
    double coeff, thresh;
    array_t *nodevec;
    st_table *table;

    dist = DEFAULT_SPEED_DIST;
    thresh = DEFAULT_SPEED_THRESH;
    model = DELAY_MODEL_UNIT_FANOUT;
    coeff = DEFAULT_SPEED_COEFF;
    weight_debug = TRUE;

    getopt_reset();
    error_init();
    while ((c = getopt(argc, argv, "m:d:t:w:")) != EOF) {
        switch (c) {
        case 'd':
            dist = atoi(optarg);
            break;
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
            	goto print_weight_usage;
            }
            break;
        case 'w':
            coeff = atof(optarg);
            break;
        case 't':
            thresh = atof(optarg);
            break;
        default:
            goto print_weight_usage;
        }
    }

    nodevec = com_get_nodes(*network, argc-optind+1, argv+optind-1);

    if (array_n(nodevec) > 0){
	if (!delay_trace(*network, model)){
	    (void)fprintf(misout,"%s\n", error_string());
	    return 1;
	}
	set_speed_thresh(*network, thresh);
	table = speed_compute_weight(*network, model, coeff, dist, nodevec);
	st_free_table (table);
    }

    array_free(nodevec);
    return 0;

print_weight_usage:
    (void) fprintf(miserr, "usage: print_weight [-m model] [-w n.n] [-d n] [-t n.n] list-of-nodes\n");
    (void) fprintf(miserr, "    -d\tn\tDistance for collapsing.\n");
    (void) fprintf(miserr, "    -t\tn.n\tCritical threshold.\n");
    (void) fprintf(miserr, "    -w\tn\tRelative weight of area. \n");
    (void) fprintf(miserr, "    -m\tmodel\tDelay model. default (\"unit-fanout\").\n");
    return 1;
}

int
com__print_last(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c;
    delay_time_t time, best;
    delay_model_t model;
    double cur, max_atime;
    lsGen gen;
    node_t *node;

    max_atime = NEG_LARGE;
    model = DELAY_MODEL_UNIT_FANOUT;

    getopt_reset();
    error_init();
    while ((c = getopt(argc, argv, "m:")) != EOF) {
        switch (c) {
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
            	model = DELAY_MODEL_LIBRARY;
            } else {
            (void) fprintf(misout,"Usage:  _print_latest [-m model]\n");
            return 1;
            }
            break;
        default:
            (void) fprintf(misout,"Usage:  _print_latest [-m model]\n");
            return 1;
        }
    }

    if (delay_trace(*network,model) ){
	foreach_primary_output(*network, gen, node){
		time = delay_arrival_time(node);
		cur =  MAX(time.rise, time.fall);
		if(cur > max_atime){
		    max_atime = cur;
		    best = time;
		}
	}
	(void) fprintf(misout," Latest output arrival time =  %6.2f:%-6.2f\n", best.rise, best.fall);
    } else {
	(void) fprintf(misout,"%s\n", error_string());
    }
    return 0;
}

int
com__speed_delay(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c;
    delay_time_t time;
    delay_model_t model;
    lsGen gen;
    node_t *node;

    model = DELAY_MODEL_UNIT_FANOUT;

    getopt_reset();
    error_init();
    while ((c = getopt(argc, argv, "m:")) != EOF) {
        switch (c) {
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
            	(void) fprintf(misout,"Usage:  _speed_delay [-m model]\n");
            	return 1;
            }
            break;
        default:
            (void) fprintf(misout,"Usage:  _speed_delay [-m model]\n");
            return 1;
        }
    }

    /* Initialize the globals for the delay routines in this package */
    speed_init_globals(model, 0 /* No library acceleration */);

    (void) speed_delay_trace(*network, model);
    foreach_node(*network, gen, node){
        speed_delay_arrival_time(node, &time);
        (void) fprintf(misout," %-10s : arrival=(%-5.2f %5.2f)\n", node_name(node), time.rise, time.fall);
    }
    return 0;
}


int
com_print_level(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    double thresh;
    delay_model_t model;
    int c, crit_only, print_flag, level;

    print_flag = TRUE;
    crit_only = FALSE;
    model = DELAY_MODEL_UNIT_FANOUT;
    thresh = 0.5;

    getopt_reset();
    error_init();
    while ((c = getopt(argc, argv, "m:t:lc")) != EOF) {
        switch (c) {
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
            	(void) fprintf(misout,"Unknown delay model \n");
            	goto print_level_usage;
            }
            break;
        case 't':
            thresh = atof(optarg);
            break;
        case 'l':
	    print_flag = FALSE;
            break;
        case 'c':
	    crit_only = TRUE;
            break;
        default:
	    goto print_level_usage;
        }
    }


    print_flag = (crit_only ? TRUE : print_flag);
    /*
     * Do the delay trace if the critical path needs to be printed 
     */
    if (crit_only){
	if(!delay_trace(*network, model)){
	    (void) fprintf(misout,"%s\n", error_string());
	    return 1;
	}
    }

    level = speed_print_level(*network, thresh, crit_only, print_flag);

    if (!print_flag){
	(void) fprintf(misout, "Total number of levels = %d\n", level);
    }
    return 0;

print_level_usage:
    (void) fprintf(misout,"Usage:  print_level [-m model] [-t thresh] [-l] [-c]\n");
    (void) fprintf(misout, "    -l\t\tPrint only a summary.\n");
    (void) fprintf(misout, "    -c\t\tPrint only critical nodes.\n");
    (void)fprintf(misout,"\t-m \tmodel\t: model for computing delays - default \"unit_fanout\"\n");
    (void)fprintf(misout,"\t-t \t#.#\t: Threshold defining critical paths\n");
    return 1;
	
}

int
com__print_fanout(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    int c, limit, max, print_max_only;
    double thresh;
    delay_time_t delay;
    delay_model_t model;
    node_t *node;
    lsGen gen;

    limit = 5;
    max = -1;
    thresh = 0.5;
    print_max_only = FALSE;
    model = DELAY_MODEL_MAPPED;
    getopt_reset();
    while ((c = getopt(argc, argv, "sn:t:m:")) != EOF) {
        switch (c) {
        case 'm':
            if (strcmp(optarg, "unit") == 0) {
            	model = DELAY_MODEL_UNIT;
            } else if (strcmp(optarg, "unit-fanout") == 0) {
            	model = DELAY_MODEL_UNIT_FANOUT;
            } else if (strcmp(optarg, "mapped") == 0) {
            	model = DELAY_MODEL_MAPPED;
            } else if (strcmp(optarg, "library") == 0) {
		(void)fprintf(misout,"Notice: Using the mapped model instead\n");
            	model = DELAY_MODEL_MAPPED;
            } else {
		goto print_fanout_usage;
            }
            break;
        case 's':
	    print_max_only = TRUE;
            break;
        case 't':
	    thresh = atof(optarg);
            break;
        case 'n':
	    limit = atoi(optarg);
            break;
        default:
            goto print_fanout_usage;
        }
    }

    if (!delay_trace(*network, model)){
	(void)fprintf(misout,"%s\n", error_string());
	return 1;
    }

    if (!print_max_only){
	(void)fprintf(misout, "Nodes with fanout >= %d\n", limit);
	(void)fprintf(misout,"type     node    fo  slack\n");
    }
    foreach_node(*network, gen, node){
        if (node->type != PRIMARY_OUTPUT){
	   delay = delay_slack_time(node);
	   if (!print_max_only){
	       if(node_num_fanout(node) > limit){
		   (void)fprintf(misout,"%s   %-10s %-3d %-5.2f\n", 
		       (node->type == INTERNAL ?"INT" : "PI "),
		       node_name(node), node_num_fanout(node), delay.rise );
	       }
	   } else {
		if (delay.rise < thresh || delay.fall < thresh){
		     max = MAX(max, node_num_fanout(node));
		}
	   }
        }
    }
    if (print_max_only){
	(void)fprintf(misout, "Maximum fanout on crit path = %d\n", max);
    }
    return 0;
print_fanout_usage:
    (void)fprintf(misout,"Usage: print_fanout [-s] [-t #.#] [-n #]\n");
    (void)fprintf(misout,"\t-s \t\t: Print maximum slack on critical path \n");
    (void)fprintf(misout,"\t-m \tmodel\t: Delay model to find critical path \n");
    (void)fprintf(misout,"\t-t \t#.#\t: Threshold of slack defining critical path\n");
    (void)fprintf(misout,"\t-n \t#\t: Print nodes with fanout greater than n\n");
    return 1;
}
/* ARGSUSED */
int
com__is_2ip_and(network, argc, argv)
network_t **network;
int argc;
char **argv;
{
    node_t *node;
    lsGen gen;

    foreach_node(*network, gen, node) {
	if ((node_num_fanin(node) <= 2)  && ((node_function(node) != NODE_OR) &&
		(node_function(node) != NODE_COMPLEX))){
	    continue;
	}
	(void) fprintf(misout,"Network is NOT in 2 input and gates\n");
	return;
    }
    (void) fprintf(misout,"Network is in 2 input and gates\n");
}

init_speed()
{
    speed_debug = FALSE;
    speed_plot_table = st_init_table(strcmp, st_strhash);

    (void) com_add_command("speed_up", com_speed_up, 1);
    (void) com_add_command("buffer_opt", com_buffer_opt, 1);
    (void) com_add_command("_print_fanout", com__print_fanout, 0);
    (void) com_add_command("_part_collapse", com__part_collapse, 1);
    (void) com_add_command("_print_cutset", com__print_cutset, 0);
    (void) com_add_command("_print_weight", com__print_weight, 0);
    (void) com_add_command("print_level", com_print_level, 0);
    (void) com_add_command("_print_latest", com__print_last, 0);
    (void) com_add_command("_speed_delay", com__speed_delay, 0);
    (void) com_add_command("_speed_plot", com__speed_plot, 0);
    (void) com_add_command("_reduce_delay", com__reduce_delay, 1);
    (void) com_add_command("_is_2ip_and", com__is_2ip_and, 0);

    node_register_daemon(DAEMON_ALLOC, buffer_alloc);
    node_register_daemon(DAEMON_FREE, buffer_free);
}

end_speed() 
{
    char *d1;
    char *d2;
    st_generator *gen;
    plot_t *ploter;

    st_foreach_item(speed_plot_table, gen, &d1, &d2) {
	ploter = (plot_t *) d2;
	FREE(d1);
	graphics_kill(ploter);
    }
    st_free_table(speed_plot_table);
}
