#include "mis.h"
#include <math.h>
#include "speed_int.h"
#include "buffer_int.h"

static delay_time_t zerodelay  = {0, 0};
static delay_time_t large_req_time  = {POS_LARGE, POS_LARGE};
static int sp_buffer_weight();
static void sp_already_buffered();


/*
 * exported routine --- needed to set some of the global variables
 * before calling the appropriate routine.
 * Returns 0 if network is not changed, 1 if changed
 */
int
buffer_network(network, do_trace, thresh, single_pass, do_decomp, limit, mode)
network_t *network;
int do_trace;		/* prints the changes in slack as algo proceeds */
double thresh;		/* Determins the critical paths */
int single_pass;	/* Buffer only one node in the pass */
int do_decomp;		/* Decompose large gates into smaller ones */
int limit;		/* Only fanouts > limit are buffered */
int mode;		/* What transformations to allow */
{
    int changed;

    if (lib_network_is_mapped(network)){
	buffer_init_globals(1, network, limit, mode, do_decomp, 0 /* no debug */);
    } else {
	buffer_init_globals(0, network, limit, mode, do_decomp, 0 /* no debug */);
	(void)fprintf(misout,"Network not mapped: Buffering will use the unit-fanout model\n");
    }

    if (!delay_trace(network, g_buf_param.model)){
	(void)fprintf(miserr,"%s\n", error_string());
	return 0;
    }

    changed = speed_buffer_network(network, do_trace, thresh, single_pass);
    changed = (changed == 0 ? 0 : 1);

    buffer_free_globals();

    return changed;
}


/*
 * Buffers the nodes in the network. For now it will select the node
 * on the critical path with the largest fanout -- buffr it -- recompute
 * the delays and keep going till it meets timing constraints or 
 * cannot improve....
 */
int
speed_buffer_network(network, do_trace, thresh, single_pass)
network_t *network;
int do_trace;		/* prints the changes in slack as algo proceeds */
double thresh;		/* Determins the critical paths */
int single_pass;	/* Buffer only one node in the pass */
{
    lsGen gen;
    int changed, i, weight, flag;
    int loop_count, cutset_count, loop_again, required_times_set;
    array_t *mincut;
    st_table *buffered_nodes, *weight_table;
    double dummy;
    node_t *node, *best, *last;
    delay_time_t prev, delay, initial, final;
    delay_model_t model = g_buf_param.model;

    changed = FALSE;
    loop_count = 1;

    (void)delay_trace(network, model);
    /*
     * The flag "requird_times_set" is used to figure out whether to stop
     * when timing constraints are met (if required times at po are specified)
     * or when there is no improvement in delay
     */
    required_times_set = FALSE;
    foreach_primary_output(network, gen, node){
	    required_times_set = (required_times_set ||
		    delay_get_parameter(node, DELAY_REQUIRED_RISE) > 0);
    }

    loop_again = TRUE;
    while (loop_again){		/* There is scope for improvement */
	loop_again = FALSE;
	buffered_nodes = st_init_table(st_ptrcmp, st_ptrhash);
	if (g_buf_param.debug) (void)fprintf(misout,"LOOP NUMBER %d\n", loop_count++);
	/* Initially set the gates to be unbufferd */
	foreach_node(network, gen, node){
	    BUFFER(node)->type  = NONE;
	}
	/* Initial circuit performance */
	if (required_times_set){
	    initial = sp_minimum_slack(network);
	} else {
	    last = delay_latest_output(network, &dummy);
	    initial = delay_arrival_time(last);
	}


	/* Evaluate different cutsets at which to apply buffering */
	flag = TRUE; cutset_count = 1;
	while (flag){
	    flag = FALSE;
	    /*
	     * "prev" is current metric of circuit performance (delay OR slack)
	     */
	    if (required_times_set){
		prev = sp_minimum_slack(network);
	    } else {
		last = delay_latest_output(network, &dummy);
		prev = delay_arrival_time(last);
	    }
	    if (g_buf_param.debug)
		    (void)fprintf(misout,"CUTSET NUMBER %d\n", cutset_count++);

	    set_speed_thresh(network, thresh);
	    weight_table = st_init_table(st_ptrcmp, st_ptrhash);
	    foreach_node(network, gen, node){
		if (node->type == INTERNAL && speed_critical(node)){
		    if (st_is_member(buffered_nodes, (char *)node)){
			/* Node has been buffered earlier */
			weight = MAXWEIGHT * 100;; 
		    } else {
			weight = sp_buffer_weight(node, model);
		    }
		    (void)st_insert(weight_table, (char *)node, (char *)weight);
		}
	    }

	    mincut = cutset(network, weight_table);

	    if (array_n(mincut) > 0){
		for (i = 0; i < array_n(mincut); i++){
		    best = array_fetch(node_t *, mincut, i);
		    if (!st_insert(buffered_nodes, (char *)best, NIL(char))){
			/* node in the cutset has been buffered previously */
			if (speed_buffer_node(network, best, required_times_set)){
			    flag = TRUE;
			    changed = TRUE;
			    sp_already_buffered(buffered_nodes, best);
			    (void)delay_trace(network, model);
			}
			if (!network_check(network)){
			    (void)fprintf(misout,"%s\n", error_string());
			}
		    }
		}
	    }

	    if (single_pass) flag = FALSE;  /* Only one cutset is optimized */

	    st_free_table(weight_table);
	    array_free(mincut);

	    if (required_times_set){
		delay = sp_minimum_slack(network);
	    } else {
		last = delay_latest_output(network, &dummy);
		delay = delay_arrival_time(last);
	    }
	    if (do_trace){
		(void)fprintf(misout,"\t%s (%.2f:%.2f)->(%.2f:%.2f)\n",
			(required_times_set ? "Slack" : "Delay"),
			prev.rise, prev.fall, delay.rise, delay.fall);
	    }
	}
	/*
	 * Completed looking at all cutsets in the current pass : Should we
	 * go on ?? "loop_again" only if the delay/slack has been improved
	 * "delay" is the metric of performance that has been achieved
	 */
	if (required_times_set){
	    final = sp_minimum_slack(network);
	} else {
	    last = delay_latest_output(network, &dummy);
	    final = delay_arrival_time(last);
	}

	if (!single_pass){
	    loop_again = ((fabs(initial.rise-final.rise) > V_SMALL) &&
		    (fabs(initial.fall-final.fall) > V_SMALL));
	}
	st_free_table(buffered_nodes);

    }
    return changed;
}

/* ARGSUSED */
static int
sp_buffer_weight(node, model)
node_t *node;
delay_model_t model;
{
    lsGen gen;
    node_t *fo;
    int fo_num, num_pos, num_neg, num_inv, weight;

    num_pos = num_neg = num_inv = 0;
    foreach_fanout(node, gen, fo){
	if (node_function(fo) == NODE_INV){
	    num_neg += node_num_fanout(fo);
	    num_inv++;
	} else {
	    num_pos++;
	}
    }

    if (num_inv > 1){
	num_neg = 0;
	num_pos += num_inv;
    }

    fo_num = num_neg + num_pos;
    weight = (int)ceil((double)MAXWEIGHT/(double)fo_num);

    return weight;
}

/*
 * Gets the fanout set of node. If there is only one inverter in
 * the fanout set then returns its fanouts too. If more than one inverter
 * then it returns only the immediate fanouts of node
 */
static buf_alg_input_t *
sp_get_fanout_set(node)
node_t *node;
{
    int pin, pin1;
    lsGen gen, gen1;
    node_t *fo, *fo1;
    buf_alg_input_t *alg_input;
    int i=0, number_inv;

    alg_input = ALLOC(buf_alg_input_t, 1);
    alg_input->node = node;
    alg_input->inv_node = NIL(node_t);
    number_inv = alg_input->num_pos = alg_input->num_neg = 0;

    foreach_fanout(node, gen, fo){
	if (node_function(fo) == NODE_INV){
	    alg_input->num_neg += node_num_fanout(fo);
	    alg_input->inv_node = fo;
	    number_inv++;
	} else {
	    (alg_input->num_pos)++;
	}
    }

    if (number_inv > 1){
	alg_input->inv_node = NIL(node_t);
	alg_input->num_neg = 0;
	alg_input->num_pos += number_inv;
    }

    if (number_inv > 1){
	/* get only the positive fanout signals */
	alg_input->fanouts = ALLOC(sp_fanout_t, alg_input->num_pos+number_inv);
    } else {
	alg_input->fanouts = ALLOC(sp_fanout_t, (alg_input->num_pos)+(
		alg_input->num_neg));
    }

    foreach_fanout_pin(node, gen, fo, pin){
	if (node_function(fo) == NODE_INV && number_inv <= 1){
	    foreach_fanout_pin(fo, gen1, fo1, pin1){
		alg_input->fanouts[i].fanout = fo1;
		alg_input->fanouts[i].pin = pin1;
		alg_input->fanouts[i].phase = PHASE_INVERTING;
		i++;
	    }
	} else {
	    alg_input->fanouts[i].fanout = fo;
	    alg_input->fanouts[i].pin = pin;
	    alg_input->fanouts[i].phase = PHASE_NONINVERTING;
	    i++;
	}
    }

    return alg_input;
}

static void
sp_already_buffered(table, node)
st_table *table;
node_t *node;
{
    lsGen gen;
    node_t *fo;
    
    foreach_fanout(node, gen, fo){
	if (BUFFER(fo)->type != NONE){
	    BUFFER(fo)->type = NONE;
	    if (!st_insert(table, (char *)fo, NIL(char))){
		/* Recursively add the nodes that have been added herein */
		sp_already_buffered(table, fo);
	    }
	}
    }
}


/*
 * Optimal buffer selection for the nodes in nodevec based on 
 * the phases and required times of the signals 
 */

int
speed_buffer_array_of_nodes(network, nodevec)
network_t *network;
array_t *nodevec;
{
    lsGen gen;
    node_t *node, *fo;
    int changed;
    int i, do_trace, count;

    count = 0;
    changed = FALSE;	
    do_trace = FALSE;

    for (i = 0; i < array_n(nodevec); i++){
	node = array_fetch(node_t *, nodevec, i);
	if (node->type != PRIMARY_OUTPUT){
	    foreach_fanout(node, gen, fo){
		if (st_is_member(g_buf_param.glbuftable, (char *)fo)){
		    /* Need to update the req time data */
		    do_trace = TRUE;
		    (void)lsFinish(gen);
		    break;
		}
	    }
	    if (do_trace){
		/* Update the delay data via a trace */
		do_trace = FALSE;
		(void)delay_trace(network, g_buf_param.model);
		count++;
		st_free_table(g_buf_param.glbuftable);
		g_buf_param.glbuftable = st_init_table(st_ptrcmp, st_ptrhash);
	    }
	    if (speed_buffer_node(network, node, 0 /*unconstrained */)){
		changed = TRUE;
		(void)st_insert(g_buf_param.glbuftable, (char *)node, NIL(char));
	    }
	}
    }
    if (g_buf_param.debug){
	(void)fprintf(misout,"INFO: %d delay traces were required\n", count);
    }

    return changed;
}

/*
 * Finds the amount by which the the required time at the current node 
 * needs to be increased.
 * The critical path thru node is traversed till a multi-fanout node is
 * encountered and the difference between side input times is found.
 */
delay_time_t 
sp_buf_get_target(node, model)
node_t *node;
delay_model_t model;
{
    lsGen gen;
    int i, pin, more_to_come = TRUE;
    node_t *fin, *fo, *crit_fin, *cur_node;
    double max_arr_time, load;
    delay_time_t t, arr, req_diff;
    delay_time_t cur_req_time, side_req_time;

    /* Initially set the target so as to get zero slack */
    t = delay_slack_time(node);
    req_diff.rise = 0 - t.rise;
    req_diff.fall = 0 - t.fall;

    /* Now look for a multi fanout node along the critical path */
    cur_node = node;
    while (more_to_come){
	more_to_come = FALSE;
	crit_fin = NIL(node_t);
	max_arr_time = -1.0;
	foreach_fanin(cur_node, i, fin){
	    if (speed_critical(fin)){
		/* Choose the one with largest arrival time */
		arr = delay_arrival_time(fin);
		if (MAX(arr.rise, arr.fall) > max_arr_time){
		    crit_fin = fin;
		    max_arr_time = MAX(arr.rise, arr.fall);
		}
	    }
	}

	if (crit_fin == NIL(node_t)){
	    /* The node "node" is not on critical path */
	    return req_diff;
	}

	if (node_num_fanout(crit_fin) > 1){
	    /*
	     * Get the difference between the required time on 
	     * the side outputs and the cur_node
	     */
	    side_req_time = large_req_time;
	    foreach_fanout_pin(crit_fin, gen, fo, pin){
		sp_buf_req_time(fo, model, pin, &t, &load);
		if (fo == cur_node){
		    cur_req_time = t;
		} else {
		    MIN_DELAY(side_req_time, side_req_time, t);
		}
	    }
	    req_diff.rise = side_req_time.rise - cur_req_time.rise;
	    req_diff.fall = side_req_time.fall - cur_req_time.fall;
	    MIN_DELAY(req_diff, req_diff, zerodelay);
	} else {
	    /* This is a fanout free region, continue along the crit path */
	    cur_node = crit_fin;
	    more_to_come = TRUE;
	}
    }
    return req_diff;
}

/*
 * Buffers the fanouts if fanout count is greater than limit 
 * Returns 1 if the fanout structure at the node was changed
 */
int
speed_buffer_node(network, node, constrained)
network_t *network;
node_t *node;
int constrained;	/* == 1 => req_times set at outputs */
{
    buf_alg_input_t *fanout_data;
    delay_time_t req_diff;
    int i, n;
    int level = 0, changed = 0;
    delay_model_t model = g_buf_param.model;

    /*
     * Set the required time at the input of the gate/buffer in the
     * original netw. The setting is done in the BUFFER field of "node"
     */
    sp_buf_annotate_gate(node, model);

    if (constrained) req_diff = sp_buf_get_target(node, model);
    else req_diff = large_req_time;

    if (g_buf_param.debug){
	(void)fprintf(misout,"Buffering %s node %s\n\t (gate = %s, %s-diff = %6.2f:%-6.2f)\n",
	(node->type == INTERNAL ? "INTERNAL" : "PRIMARY_INPUT"),
	node_long_name(node), sp_name_of_impl(node),
	(constrained ? "CONSTRAINED": "UNCONSTRAINED"),
	req_diff.rise, req_diff.fall);
    }

    fanout_data = sp_get_fanout_set(node);
    n = fanout_data->num_pos + fanout_data->num_neg;

    if (n < g_buf_param.limit){
	FREE(fanout_data->fanouts);
	FREE(fanout_data);
	return changed;
    }

    /* Also need to set the appropriate data for the inv_node */
    if (fanout_data->inv_node != NIL(node_t)){
	sp_buf_annotate_gate(fanout_data->inv_node, model);
    }

    for (i = 0; i < n; i++){
	sp_buf_req_time(fanout_data->fanouts[i].fanout, model, 
	    fanout_data->fanouts[i].pin, &(fanout_data->fanouts[i].req),
	    &(fanout_data->fanouts[i].load));
    }
    changed = sp_buffer_recur(network, fanout_data, req_diff, &level);

    FREE(fanout_data->fanouts);
    FREE(fanout_data);

    return changed;
}
