#include "mis.h"
#include "speed_int.h"
#include <math.h>

static void speed_replace();
static int speed_is_fanout_po();
static int speed_resub_alge_node();

bool
speed_critical(np)
node_t *np;
{
    delay_time_t time;

    time = delay_slack_time(np);
    if (time.rise <= speed_thresh || time.fall <= speed_thresh) {
        return TRUE;
    } else {
        return FALSE;
    }
}


static int
speed_is_fanout_po(node)
node_t *node;
{
    lsGen gen;
    node_t *fo;

    foreach_fanout(node, gen, fo){
	if (node_function(fo) == NODE_PO) {
	    (void) lsFinish(gen);
	    return TRUE;
	}
    }
    return FALSE;
}


void
speed_up_network(network, thresh, coeff, dist, model)
network_t *network;
double thresh, coeff;
int dist;
delay_model_t model;
{
    int i;
    st_table *table;
    array_t *mincut;
    lsGen gen, gen1;
    node_t *np, *fo, *fan, *fanin, *fanin_in;


    /* Build up the table of weights */
    set_speed_thresh(network, thresh);
    table = speed_compute_weight(network, model, coeff, dist, NIL(array_t));

    /* compute the minimum weighted cutset of the critical paths */
    mincut = cutset(network, table);

    /*
     * partial_collapse & re-decompose each collapsed node. 
     */ 
    for (i= 0; i < array_n(mincut); i++) {
	np = array_fetch(node_t *, mincut, i);
        speed_absorb(np, dist);
    }
    for (i= 0; i < array_n(mincut); i++) {
        np = array_fetch(node_t *, mincut, i);

        if(np->type == INTERNAL ){
            if(speed_debug){
                (void) fprintf(misout, "\nSPEED DECOMPOSING NODE \t ");
                node_print(misout,np);
            }
            speed_up_node(network, np, coeff, model, 0);
        }
    }

    /* 
     * Due to the collapsing of the inverters, some
     * primary o/p may become buffers. Remove these.
     */
    foreach_primary_output(network, gen, fo){
	fanin = node_get_fanin(fo, 0);
	if (node_function(fanin) == NODE_BUF){
	    fanin_in = node_get_fanin(fanin, 0);
	    assert(node_patch_fanin(fo, fanin, fanin_in));
	    foreach_fanout(fanin, gen1, fan){
		(void) node_collapse(fan, fanin);
	    }
	    if (node_num_fanout(fanin) == 0) speed_network_delete_node(network, fanin);
	} else if (node_function(fanin) == NODE_INV) {
	    foreach_fanout(fanin, gen1, fan){
		(void) node_collapse(fan, fanin);
	    }
	    if (node_num_fanout(fanin) == 0) speed_network_delete_node(network, fanin);
	}
    }

    /*
     *Just a check:  If the routines are good this is not needed
     */
    /*
    if (network_cleanup(network) ) {
	(void) fprintf(misout, " Hey ! cleanup was done \n");
    }
    */

    array_free(mincut);
    st_free_table(table);
}


void
speed_up_node(network, np, coeff, model, delay_flag)
network_t *network;
node_t *np;
double coeff;
delay_model_t model;
int delay_flag;		/* 1=> use user field */
{
    array_t *nodes;
    delay_time_t t;

    /*
     * Check if the node is fit to be decomposed.
     */
    if (np->type != INTERNAL || node_num_literal(np) == 0) {
	speed_single_level_update(np, model, &t);
	return;
    }

    if (node_num_fanin(np) <= 2 && node_num_cube(np) <= 1){
	speed_single_level_update(np, model, &t);
	return;
    }

    nodes = speed_decomp(np, coeff, model, delay_flag);

    if (nodes != NIL(array_t) ){
        speed_replace(network, np, nodes, model);
    } else {
        fail("Speed_decomp returned a null array");
        exit(-1);
    }

    array_free(nodes);
}

int 
speed_init_decomp(network, model, coeff, flag)
network_t *network;
delay_model_t model;
double coeff;
int flag; 	/* If set then do the resubstitution */
{
    node_t *node;
    array_t *nodevec;
    int a_flag, d_flag, i;

    a_flag = speed_area_reclaim;
    d_flag = speed_debug;
    speed_area_reclaim = FALSE;
    speed_debug = FALSE;

    if (sp_add_inv) {
	(void) add_inv_network(network);
    } else {
        (void) network_sweep(network);
    }

    assert(speed_delay_trace(network, model));
    nodevec = network_dfs(network);
    if (speed_debug) {
	(void)fprintf(misout,"List of internal nodes\n");
	for (i = 0; i < array_n(nodevec); i++) {
	    node = array_fetch(node_t *, nodevec, i);
	    node_print(misout, node);
	}
    }
    for (i = 0; i < array_n(nodevec); i++) {
	node = array_fetch(node_t *, nodevec, i);
	if (node->type == INTERNAL){
	    if (d_flag) {
		(void)fprintf(misout,"Initial decomp of ");
		node_print(misout, node);
	    }
	    speed_up_node(network, node, coeff, model, 1);
	}
    }

    /* 
     * During the decomposition into 2-input NAND gates there is a
     * lot of possibility that nodes computing the same logic function
     * exist. Get rid of these by an algebraic resubstitution.
     */

    if (flag) resub_alge_network(network, 0 /* complement not used */);

    if (sp_add_inv) {
	(void) add_inv_network(network);
    } else {
        (void) network_sweep(network);
    }

    speed_debug = d_flag;
    speed_area_reclaim = a_flag;
    array_free(nodevec);
    return 0;
}

static void
speed_replace(network, np, nodes, model)
network_t *network;
node_t *np;
array_t *nodes;
delay_model_t model;
{
    int i;
    char *dummy;
    delay_time_t t;
    st_table *table;
    array_t *inv_array, *added_array, *added_tables;
    node_t *temp, *root_node;

    if (array_n(nodes) > 3){
	inv_array = array_alloc(node_t *, 0);
	added_array = array_alloc(node_t *, 0);
	added_tables = array_alloc(st_table *, 0);
	table = st_init_table(st_ptrcmp, st_ptrhash);

	/*
	 * Now add the nodes in the array. If the inverters
	 * have to be removed, record their occurance and
	 * collapse them later.
	 */
	node_free(array_fetch(node_t *, nodes, 0));
	root_node = array_fetch(node_t *, nodes, 1);
	speed_delay_arrival_time(root_node, &t);

	for (i = array_n(nodes)-1; i > 1; i--) {
	    temp = array_fetch(node_t *, nodes, i);
	    (void)st_insert(table, (char *)temp, "");
	}
	for (i = array_n(nodes)-1; i > 1; i--) {
	    temp = array_fetch(node_t *, nodes, i);
	    if (node_function(temp) != NODE_PI){

		network_add_node(network,temp);
		(void) st_delete(table, (char **)&temp, &dummy);
		if (node_num_fanin(temp) < 2){
		    array_insert_last(node_t *, inv_array, temp);
		} else if (speed_area_reclaim ) {
		/* 
		 * Check if this node is already present -
		 * i.e try an algebraic resubstitution.
		 */
		    array_insert_last(node_t *, added_array, temp);
		    array_insert_last(st_table *, added_tables, st_copy(table));
		}
	    } else {
		node_free(temp);
	    }
	}

	node_replace(np, root_node);
	speed_set_arrival_time(np, t);
	st_free_table(table);

	for (i = 0; i < array_n(added_array); i++) {
	    table = array_fetch(st_table *, added_tables, i);
	    temp = array_fetch(node_t *, added_array, i);
	    if (speed_debug){
		(void)fprintf(misout,"Trying resub for");
		node_print(misout, temp);
	    }
	    if (speed_resub_alge_node(network, temp, table) && speed_debug) {
		(void)fprintf(misout,"After resub");
		node_print(misout, temp);
	    }
	    st_free_table(table );
	}
	array_free(added_array);
	array_free(added_tables);

	/*
	 * Conditional removal of the inverters.
	 */
	if (! sp_add_inv) {
	    for (i = 0; i < array_n(inv_array) ; i++){
		temp = array_fetch(node_t *, inv_array, i);
		speed_delete_single_fanin_node(network, temp);
	    }
	}
	array_free(inv_array);

	if ((node_num_fanin(np) == 1) &&
		(speed_is_fanout_po(np) == FALSE) ){
	    speed_delete_single_fanin_node(network, np);
	}
    } else {
	/*
	 * If the node could be replaced by something this trivial
	 * then one might as well simplify the node itself.
	 * We still need to update the delay times
	 */
	root_node = node_simplify(np, NIL(node_t), NODE_SIM_SIMPCOMP);
	node_replace(np, root_node);
	(void)speed_update_arrival_time(np, model);
    } 
}

/*
 * Substitute node f into other nodes in the network and not in "table".
 * The node "f" might be destroyed if the substitution is made . In that case
 * the routine returns 0, otherwise if nothing happened it returns 1
 */
static int
speed_resub_alge_node(network, f, table)
network_t *network;
node_t *f;
st_table *table;
{
    lsGen gen;
    node_t *np, *fo;
    int i, changed = 1;
    array_t *target, *tl1;

    if (f->type == PRIMARY_INPUT || f->type == PRIMARY_OUTPUT) { 
	return changed;
    }

    if (node_num_literal(f) < 1 ){
	return changed;
    }

    target = array_alloc(node_t *, 0);
    foreach_fanin(f, i, np) {
	tl1 = array_alloc(node_t *, 0);
	foreach_fanout(np, gen, fo) {
	    if (node_num_fanin(fo) <= 2 && (!st_is_member(table, (char *)fo)) ){
		array_insert_last(node_t *, tl1, fo);
	    }
	}
	array_append(target, tl1);
	array_free(tl1);
    }
    array_sort(target, node_compare_id);
    array_uniq(target, node_compare_id, (void (*)()) 0);

    /*
    tl1 = array_alloc(node_t *, 0);
    */
    for(i = 0; i < array_n(target); i++) {
	np = array_fetch(node_t *, target, i);
	if (node_substitute(f, np, 1) ) {
	    /* See if the substituted node can be removed */
	    if (node_num_fanin(f) <= 1) {
		foreach_fanout(f, gen, fo) {
		    (void) node_collapse(fo, f);
		}
		if (node_num_fanout(f) == 0) {
		    if (speed_debug) {
			(void)fprintf(misout,"Substituted out node %s\n",
				node_name(f));
		    }
		    /*
		    tl1 = array_alloc(node_t *, 0);
		    foreach_fanin(np, j, del_node){
			if (del_node != np && node_num_fanout(del_node) == 1){
			    array_insert_last(node_t *, tl1, del_node);
			}
		    }
		    */
		    network_delete_node(network, f);
		    changed = 0;
		    break;
		}
	    }
	}
    }
    /*
    for (j = 0; j < array_n(tl1); j++){
	del_node = array_fetch(node_t *, tl1, j);
	speed_network_delete_node(network, del_node);
    }
    array_free(tl1);
    */
    array_free(target);

    return changed;
}
